/* (C) 2009 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate IP address details.
 *
 * $Id: AddrInfo.cc,v 1.3 2009-06-05 12:47:57 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "AddrInfo.hh"


namespace HTTPTunnel {


// Construction/destruction
AddrInfo::AddrInfo(std::string host)
: m_host(host), m_family(UNKNOWN_FAMILY), m_addrLen(0), m_addr(), m_addrStr("")
{
}

AddrInfo::AddrInfo(const AddrInfo& that)
: m_host(that.m_host), m_family(that.m_family), m_addrLen(that.m_addrLen),
  m_addr(that.m_addr), m_addrStr(that.m_addrStr)
{
}

AddrInfo::~AddrInfo()
{
}


// Assignment
AddrInfo& AddrInfo::operator=(const AddrInfo& that)
{
	// Self assignment
	if(this != &that) {
		m_host = that.m_host;
		m_family = that.m_family;
		m_addrLen = that.m_addrLen;
		m_addr = that.m_addr;
		m_addrStr = that.m_addrStr;
	}

	return *this;
}


/*************/
/* Accessors */
/*************/


// Get hostname as a string
std::string AddrInfo::getHostName(void) const
{
	return m_host;
}

// Get address family
AddrInfo::Family AddrInfo::getFamily(void) const
{
	return m_family;
}

// Get address family as an inet value
int AddrInfo::getInetFamily(void) const
{
	if(m_family == IPV4)
		return AF_INET;

	#if _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE || _POSIX_SOURCE
	else if(m_family == IPV6)
		return AF_INET6;
	#endif

	return AF_UNSPEC;
}

// Get sockaddr len
size_t AddrInfo::getAddrLen(void) const
{
	return m_addrLen;
}

// Get pointer to generic sockaddr
const struct sockaddr* AddrInfo::getAddr(void) const
{
	return &m_addr;
}

// Get IP address as a string
std::string AddrInfo::getAddrStr(void) const
{
	return m_addrStr;
}


/************/
/* Mutators */
/************/


// Set address family
void AddrInfo::setFamily(int family)
{
	if(family == AF_INET)
		m_family = IPV4;

	#if _POSIX_C_SOURCE >= 1 || _XOPEN_SOURCE || _POSIX_SOURCE
	else if(family == AF_INET6)
		m_family = IPV6;
	#endif

	else
		m_family = UNKNOWN_FAMILY;
}

// Set sockaddr details
void AddrInfo::setAddr(size_t addrLen, const struct sockaddr* addr)
{
	m_addrLen = addrLen;
	m_addr = *addr;
}

// Set IP address as a string
void AddrInfo::setAddrStr(std::string addrStr)
{
	m_addrStr = addrStr;
}


}	// namespace

