/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Provide authentication settings and methods.
 *
 * $Id: AuthSettings.cc,v 1.2 2008-06-26 09:47:28 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include <iomanip>
#include <algorithm>
#include "AuthSettings.hh"
#include "Utils.hh"
#include "AuthDigest.hh"
#include "Logger.hh"
#include "Debug.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
AuthSettings::AuthSettings()
: m_authScheme(AUTH_NONE), m_triedSchemes(), m_user(), m_domain(), m_passwd(), m_realm(),
  m_digestNonce(), m_digestQop("auth"), m_digestCNonce(), m_digestCNonceCount(1),
  m_ntlmHandler(), m_reqHeader()
{
}

AuthSettings::AuthSettings(std::string user, std::string domain, std::string passwd)
: m_authScheme(AUTH_UNKNOWN), m_triedSchemes(), m_user(user), m_domain(domain),
  m_passwd(passwd), m_realm(), m_digestNonce(), m_digestQop(), m_digestCNonce(),
  m_digestCNonceCount(1), m_ntlmHandler(), m_reqHeader()
{
}

AuthSettings::AuthSettings(const AuthSettings& that)
: m_authScheme(that.m_authScheme), m_triedSchemes(), m_user(that.m_user),
  m_domain(that.m_domain), m_passwd(that.m_passwd), m_realm(that.m_realm),
  m_digestNonce(that.m_digestNonce), m_digestQop(that.m_digestQop),
  m_digestCNonce(that.m_digestCNonce), m_digestCNonceCount(that.m_digestCNonceCount),
  m_ntlmHandler(), m_reqHeader(that.m_reqHeader)
{
}

AuthSettings::~AuthSettings()
{
}


/*************/
/* Accessors */
/*************/


// Get whether we have already tried (and failed) some authentication scheme
bool AuthSettings::triedAuthScheme(AuthScheme scheme) const
{
	// Have we tried this scheme before?
	std::vector<AuthScheme>::const_iterator ait;
	ait = std::find(m_triedSchemes.begin(), m_triedSchemes.end(), scheme);
	return (ait != m_triedSchemes.end());
}

// Do we have a user/passwd specified?
bool AuthSettings::gotUserPasswd(void) const
{
	return (m_user.size() > 0 && m_passwd.size() > 0);
}

// Do we have a user domain specified?
bool AuthSettings::gotDomain(void) const
{
	return (m_domain.size() > 0);
}

// Get proxy authentication user
std::string AuthSettings::getUser(void) const
{
	return m_user;
}

// Get proxy authentication domain
std::string AuthSettings::getDomain(void) const
{
	return m_domain;
}

// Get proxy authentication passwd
std::string AuthSettings::getPasswd(void) const
{
	return m_passwd;
}

// Get authentication scheme name as a string
std::string AuthSettings::getAuthSchemeStr(AuthScheme scheme)
{
	const char* schemeStr[] = { "NONE", "UNKNOWN", "BASIC", "DIGEST", "NTLM",
		"HOLD" };
	return schemeStr[scheme];
}


/************/
/* Mutators */
/************/


// Set user name & password
void AuthSettings::setUserPasswd(std::string user, std::string passwd)
{
	m_user = user;
	m_passwd = passwd;
}


// Generate a new random nonce
void AuthSettings::newNonce(void)
{
	std::stringstream nonce;
	nonce << std::setw(8) << std::hex << std::setfill('0') << rand();
	m_digestNonce = nonce.str();
}


/*************/
/* Utilities */
/*************/


// Construct tunneling request headers
std::string AuthSettings::getRequestHeader(std::string uri)
{
	std::stringstream res;

	// Using authorization?
	if(m_authScheme == AUTH_HOLD) {
		// Use previously created request header
		res << m_reqHeader;
	}

	else if(m_authScheme == AUTH_BASIC) {
		// Create basic authentication fields
		res << getAuthRequestHeaderID() << ": Basic ";
		res << Utils::Base64Encode(m_user + ":" + m_passwd);
		res << "\r\n";
		m_reqHeader = res.str();

		// Remember trying basic authentication
		finishedAuthScheme(AUTH_BASIC);
	}

	else if(m_authScheme == AUTH_DIGEST) {
		// Get client nonce count
		bool finishedDigest = (m_digestCNonceCount == 1);
		std::stringstream cnoncec;
		cnoncec << std::setw(8) << std::setfill('0') << m_digestCNonceCount++;

		// Calculate digest value
		std::string response = AuthDigest::MD5DigestResponse("MD5", m_user, m_passwd,
			m_realm, m_digestNonce, m_digestCNonce, cnoncec.str(), m_digestQop, uri);

		// Create digest authentication fields
		res << getAuthRequestHeaderID() << ": Digest ";
		res << "username=\"" << m_user << "\"";
		res << ", realm=\"" << m_realm << "\"";
		res << ", nonce=\"" << m_digestNonce << "\"";
		res << ", uri=\"" << uri << "\"";
		res << ", algorithm=\"MD5\"";
		res << ", qop=\"" << m_digestQop << "\"";
		res << ", cnonce=\"" << m_digestCNonce << "\"";
		res << ", nc=" << cnoncec.str();
		res << ", response=\"" << response << "\"";
		res << "\r\n";
		m_reqHeader = res.str();

		// Remember trying digest authentication
		if(finishedDigest) {
			finishedAuthScheme(AUTH_DIGEST);
		}
	}

	else if(m_authScheme == AUTH_NTLM) {
		bool finishedNTLM = false;
		try {
			// Get next set of authentication fields for NTLM
			RawBuff ntlmMsg;
			finishedNTLM = m_ntlmHandler.createNextMsg(ntlmMsg, m_user, m_domain, m_passwd);
			std::string encNtlmMsg;
			unsigned int encLen = ntlmMsg.getLength();
			ntlmMsg.base64encode(encLen, encNtlmMsg);

			// Create NTLM authentication fields
			res << getAuthRequestHeaderID() << ": NTLM ";
			res << encNtlmMsg;
			res << "\r\n";
			m_reqHeader = res.str();
		}
		catch(AuthProcessingError& e) {
			g_pLogger->info("Error creating NTLM request header: " + e.what());
			finishedNTLM = true;
		}

		// Remember trying NTLM authentication
		if(finishedNTLM) {
			finishedAuthScheme(AUTH_NTLM);
		}
	}

	return res.str();
}

// Construct tunneling response headers
std::string AuthSettings::getResponseHeader(std::string uri)
{
	return "";
}


// Update current authentication scheme state
void AuthSettings::updateAuthScheme(AuthScheme scheme,
	const std::vector<std::string>& parms) throw(AuthProcessingError)
{
	#ifdef DEBUG_TUN
	DPRINTV("AuthSettings::updateAuthScheme()");
	std::stringstream dbg;
	dbg << "Scheme: " << getAuthSchemeStr(scheme) << ", Parms:";
	for(unsigned int dpit = 0; dpit < parms.size(); ++dpit) dbg << " " << parms[dpit];
	DPRINTV(dbg.str());
	#endif

	// What authentication scheme?
	if(scheme == AUTH_NONE) {
		// Initialise no authentication
		m_authScheme = AUTH_NONE;
	}

	else if(scheme == AUTH_BASIC) {
		// Get parameters for basic authentication
		std::string realm;
		if(Utils::getParmValue(parms, "realm", realm)) {
			// Initialise basic authentication
			m_authScheme = AUTH_BASIC;
			m_realm = realm;
		}
		else {
			throw AuthProcessingError("Missing basic authentication parameters");
		}
	}

	else if(scheme == AUTH_DIGEST) {
		// Get parameters for digest authentication
		std::string realm, nonce, qop;
		if(Utils::getParmValue(parms, "realm", realm) && Utils::getParmValue(parms,
		"nonce", nonce) && Utils::getParmValue(parms, "qop", qop)) {
			// Initialise digest authentication
			m_authScheme = AUTH_DIGEST;
			m_realm = realm;
			m_digestQop = qop;
			m_digestNonce = nonce;
			m_digestCNonceCount = 1;
			std::stringstream cnonce;
			cnonce << std::setw(8) << std::hex << std::setfill('0') << rand();
			m_digestCNonce = cnonce.str();
		}
		else {
			throw AuthProcessingError("Missing digest authentication parameters");
		}
	}

	else if(scheme == AUTH_NTLM) {
		// Are we starting NTLM?
		if(m_authScheme != AUTH_NTLM) {
			// Initialise NTLM authentication
			m_authScheme = AUTH_NTLM;
		}
		else {
			// Update NTLM authentication
			if(parms.size() == 1) {
				m_ntlmHandler.parseRecvdMsg(parms[0]);
			}
			else {
				throw AuthProcessingError("Wrong number of NTLM authentication parameters");
			}
		}
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Remember trying an authentication scheme
void AuthSettings::finishedAuthScheme(AuthScheme scheme)
{
	// Check we're not remembering a duplicate
	std::vector<AuthScheme>::iterator ait;
	ait = std::find(m_triedSchemes.begin(), m_triedSchemes.end(), scheme);
	if(ait == m_triedSchemes.end()) {
		// Remember trying this scheme
		m_triedSchemes.push_back(scheme);
	}
}


}	// namespace

