/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Provide authentication settings and methods.
 *
 * $Id: AuthSettings.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _AUTH_SETTINGS_HH_
#define _AUTH_SETTINGS_HH_

#include <vector>
#include <string>
#include "Exceptions.hh"
#include "NTLM.hh"

namespace HTTPTunnel {

class AuthSettings {
public:

	/*********/
	/* Types */
	/*********/

	// Authorisation schemes
	enum AuthScheme {
		AUTH_NONE, AUTH_UNKNOWN, AUTH_BASIC, AUTH_DIGEST, AUTH_NTLM,
		AUTH_HOLD
	};


	/***********/
	/* Methods */
	/***********/

	// Construction/destruction
	AuthSettings();
	AuthSettings(std::string user, std::string domain, std::string passwd);
	AuthSettings(const AuthSettings& that);
	virtual ~AuthSettings();

	// Accessors
	bool triedAuthScheme(AuthScheme scheme) const;
	bool gotUserPasswd(void) const;
	bool gotDomain(void) const;
	std::string getUser(void) const;
	std::string getDomain(void) const;
	std::string getPasswd(void) const;
	static std::string getAuthSchemeStr(AuthScheme scheme);

	// Mutators
	void setUserPasswd(std::string user, std::string passwd);
	void newNonce(void);

	// Construct tunneling request/response headers
	virtual std::string getRequestHeader(std::string uri);
	virtual std::string getResponseHeader(std::string uri);

	// Reset state
	virtual void reset(void) = 0;

	// Update authentication scheme state
	void updateAuthScheme(AuthScheme scheme, const std::vector<std::string>& parms)
		throw(AuthProcessingError);


protected:

	/***********/
	/* Methods */
	/***********/

	// Get authorisation request header field identifier
	virtual std::string getAuthRequestHeaderID(void) const = 0;

	// Remember trying an authentication scheme
	void finishedAuthScheme(AuthScheme scheme);


	/********/
	/* Data */
	/********/

	// Current authentication scheme
	AuthScheme						m_authScheme;

	// Tried authentication schemes
	std::vector<AuthScheme>		m_triedSchemes;

	// Proxy auth parameters
	std::string						m_user;
	std::string						m_domain;
	std::string						m_passwd;
	std::string						m_realm;

	// Digest auth parameters
	std::string						m_digestNonce;
	std::string						m_digestQop;
	std::string						m_digestCNonce;
	unsigned int					m_digestCNonceCount;

	// NTLM auth parameters
	NTLM								m_ntlmHandler;

	// Last request header
	std::string						m_reqHeader;


private:

	// Forbid assignment
	AuthSettings& operator=(const AuthSettings& that);

};


}	// namespace

#endif

