/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user, and tracking/controling
 * resource usage.
 *
 * $Id: Authenticator.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include "Authenticator.hh"
#include "Logger.hh"
#include "Utils.hh"
#include "MutexVar.hh"
#include "Tunnel.hh"
#include "Debug.hh"
#include "ServerAuthSettings.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
Authenticator::Authenticator() throw(AuthError)
: m_mtx(), m_users(), m_defaultUserName("DefaultUser"), m_defaultAuthSettings(0)
{
	// Create server authentication settings
	m_defaultAuthSettings = new ServerAuthSettings();

	// Set default username
	m_defaultAuthSettings->setUserPasswd(m_defaultUserName, "");
}

Authenticator::~Authenticator()
{
	MutexVar mv(m_mtx);

	// Free server authentication settings
	delete m_defaultAuthSettings;
	m_defaultAuthSettings = 0;
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Check whether user can receive data
bool Authenticator::getUserCanRecvBytes(std::string name, unsigned int bytes)
{
	bool res = true;

	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		res = user->canRecvBytes(bytes);
	}

	return res;
}


// Get amount of data user can currently send
unsigned int Authenticator::getUserBytesCanSend(std::string name)
{
	unsigned int res = 0;

	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		res = user->getBytesCanSend();
	}

	return res;
}


// Get whether user can create a new tunnel
bool Authenticator::getUserCanCreateTunnel(std::string name)
{
	bool res = true;

	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		res = user->canCreateNewTunnel();
	}

	return res;
}


// Get default authentication settings
ServerAuthSettings* Authenticator::getDefaultAuthSettings(void)
{
	return m_defaultAuthSettings;
}


// Get default username
std::string Authenticator::getDefaultUserName(void)
{
	MutexVar mv(m_mtx);
	return m_defaultAuthSettings->getUser();
}


/************/
/* Mutators */
/************/


// Set whether a user is authenticated
void Authenticator::setUserAuthenticated(std::string name, bool auth,
	Tunnel* tun)
{
	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		// Set whether they are authenticated
		user->setAuthenticated(auth, tun);
	}
}


// Update a users quotas
void Authenticator::setUserQuotas(std::string name, const Quotas& quotas)
{
	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		// Set their quotas
		user->setQuotas(quotas);
	}
}


// Update data user has sent
void Authenticator::setUserSentBytes(std::string name, unsigned int bytes)
{
	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		// Track bytes sent
		user->sentBytes(bytes);
	}
}


/*************/
/* Utilities */
/*************/


// A users tunnel has closed
void Authenticator::userTunnelClosed(std::string name, Tunnel* tun)
{
	DPRINTV("Authenticator::userTunnelClosed() - " + name + ", " + tun->getTidStr());

	MutexVar mv(m_mtx);

	// Does user exist in cache?
	User* user = getUser(name);
	if(user) {
		// Let user know it has one less tunnel in use
		user->removeTunnel(tun);

		// If they're no longer required, uncache user details
		if(!user->keepCached()) {
			freeUser(name);
			DPRINTV("User " + name + " uncached");
		}
	}
}


// Authenticate a new user
void Authenticator::authenticate(std::string userName, Tunnel* tun,
	const std::vector<std::string>& authArgs)
{
	MutexVar mv(m_mtx);

	// Does user exist in cache?
	User* user = getUser(userName);
	if(!user) {
		// Create user and place in cache
		m_users[userName] = createUser(userName, false);
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Lookup user in cached details - caller holds mutex
User* Authenticator::getUser(std::string name)
{
	User* user = 0;
	std::map<std::string, User>::iterator uit = m_users.find(name);
	if(uit != m_users.end()) {
		user = &uit->second;
	}
	#ifdef DEBUG_TUN
	else {
		DPRINTI("Authenticator failed to lookup user " + name);
	}
	#endif

	return user;
}


// Remove user from cache - caller holds mutex
void Authenticator::freeUser(std::string name)
{
	std::map<std::string, User>::iterator uit = m_users.find(name);
	if(uit != m_users.end()) {
		m_users.erase(uit);
	}
	else {
		std::stringstream wrn;
		wrn << "Authenticator failed to lookup user " << name;
		g_pLogger->warn(wrn.str());
	}
}


}	// namespace

