/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user, and tracking/controling
 * resource usage.
 *
 * $Id: Authenticator.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _AUTHENTICATOR_HH_
#define _AUTHENTICATOR_HH_

#include <string>
#include <vector>
#include <map>
#include "Exceptions.hh"
#include "User.hh"
#include "Mutex.hh"

namespace HTTPTunnel {

class Tunnel;
class Quotas;
class ServerAuthSettings;

class Authenticator {
public:
	// Construction/destruction
	Authenticator() throw(AuthError);
	virtual ~Authenticator();


	/*************/
	/* Accessors */
	/*************/

	// Check whether user can receive data
	virtual bool getUserCanRecvBytes(std::string name, unsigned int bytes);

	// Get amount of data user can currently send
	virtual unsigned int getUserBytesCanSend(std::string name);

	// Get whether user can create a new tunnel
	virtual bool getUserCanCreateTunnel(std::string name);

	// Get copy of current user details
	virtual User getUserDetails(std::string name) = 0;

	// Get whether user is authenticated for some tunnel
	virtual bool getUserAuthenticated(std::string name, Tunnel* tun) = 0;

	// Get default authentication settings
	ServerAuthSettings* getDefaultAuthSettings(void);

	// Get default username
	std::string getDefaultUserName(void);


	/************/
	/* Mutators */
	/************/

	// Set whether a user is authenticated
	void setUserAuthenticated(std::string name, bool auth, Tunnel* tun);

	// Update a users quotas
	void setUserQuotas(std::string name, const Quotas& quotas);

	// Update data user has sent
	void setUserSentBytes(std::string name, unsigned int bytes);


	/*************/
	/* Utilities */
	/*************/

	// A users tunnel has closed
	virtual void userTunnelClosed(std::string name, Tunnel* tun);
	
	// Authenticate a new user
	virtual void authenticate(std::string userName, Tunnel* tun,
		const std::vector<std::string>& authArgs);


protected:

	/***********/
	/* Methods */
	/***********/

	// Manage cached user details
	User* getUser(std::string name);
	void freeUser(std::string name);

	// Create a new tunnel user
	virtual User createUser(std::string name, bool keepCached) = 0;


	/********/
	/* Data */
	/********/

	// Mutex
	Mutex									m_mtx;

	// Map of cached users
	std::map<std::string, User>	m_users;

	// Default authentication settings to use
	std::string							m_defaultUserName;
	ServerAuthSettings*				m_defaultAuthSettings;

};


}	// namespace

#endif

