/* (C) 2003 Kevin Manning.
 * (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate function/method callbacks.
 *
 * $Id: Callback.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _CALLBACK_HH_
#define _CALLBACK_HH_

namespace HTTPTunnel {

// Base class interface
class Callback {
public:
	// Construction/destruction
	Callback();
	Callback(const Callback& that);
	virtual ~Callback();

	// Assignment
	virtual Callback& operator=(const Callback& that);

	// Copy ourselves
	virtual Callback* clone(void) = 0;

	// Call the callback
	virtual void operator()(void) = 0;

protected:

private:

};


// 0 argument global function
class ZeroArgGlobalCallback : public Callback {
private:
	// Types
	typedef void (*ZeroArgGlobalCallbackFunc)(void);

public:
	// Construction/destruction
	ZeroArgGlobalCallback(ZeroArgGlobalCallbackFunc cb);
	ZeroArgGlobalCallback(const ZeroArgGlobalCallback& that);
	virtual ~ZeroArgGlobalCallback();

	// Assignment
	virtual ZeroArgGlobalCallback& operator=(const ZeroArgGlobalCallback& that);

	// Copy ourselves
	virtual Callback* clone(void);

	// Call the callback
	virtual void operator()(void);

protected:

	// The function to call
	ZeroArgGlobalCallbackFunc			m_cb;

};


// 1 argument global function
template<class A>
class OneArgGlobalCallback : public Callback {
private:
	// Types
	typedef void (*OneArgGlobalCallbackFunc)(A);

public:
	// Construction/destruction
	OneArgGlobalCallback(OneArgGlobalCallbackFunc cb, A& arg1);
	OneArgGlobalCallback(const OneArgGlobalCallback& that);
	virtual ~OneArgGlobalCallback();

	// Assignment
	virtual OneArgGlobalCallback& operator=(const OneArgGlobalCallback& that);

	// Copy ourselves
	virtual Callback* clone(void);

	// Call the callback
	virtual void operator()(void);

protected:

	// The function to call
	OneArgGlobalCallbackFunc			m_cb;

	// Argument to the callback function
	A											m_arg;

};


// 0 arg class method callback
template<class T>
class ZeroArgClassCallback : public Callback {
private:
	// Types
	typedef void (T::*ZeroArgClassCallbackMethod)(void);

public:
	// Construction/destruction
	ZeroArgClassCallback(ZeroArgClassCallbackMethod cb, T* obj);
	ZeroArgClassCallback(const ZeroArgClassCallback& that);
	virtual ~ZeroArgClassCallback();

	// Assignment
	virtual ZeroArgClassCallback& operator=(const ZeroArgClassCallback& that);

	// Copy ourselves
	virtual Callback* clone(void);

	// Call the callback
	virtual void operator()(void);

protected:

	// The method to call
	ZeroArgClassCallbackMethod				m_cb;

	// Object to callback to
	T*												m_obj;

};


// 1 arg class method callback
template<class T, class A>
class OneArgClassCallback : public Callback {
private:
	// Types
	typedef void (T::*OneArgClassCallbackMethod)(A);

public:
	// Construction/destruction
	OneArgClassCallback(OneArgClassCallbackMethod cb, T* obj, A arg1);
	OneArgClassCallback(const OneArgClassCallback& that);
	virtual ~OneArgClassCallback();

	// Assignment
	OneArgClassCallback& operator=(const OneArgClassCallback& that);

	// Copy ourselves
	virtual Callback* clone(void);

	// Call the callback
	virtual void operator()(void);

protected:

	// The method to call
	OneArgClassCallbackMethod				m_cb;

	// Object to callback to
	T*												m_obj;

	// Argument to the callback method
	A												m_arg1;

};


// 2 arg class method callback
template<class T, class A1, class A2>
class TwoArgClassCallback : public Callback {
private:
	// Types
	typedef void (T::*TwoArgClassCallbackMethod)(A1, A2);

public:
	// Construction/destruction
	TwoArgClassCallback(TwoArgClassCallbackMethod cb, T* obj, A1 arg1, A2 arg2);
	TwoArgClassCallback(const TwoArgClassCallback& that);
	virtual ~TwoArgClassCallback();

	// Assignment
	TwoArgClassCallback& operator=(const TwoArgClassCallback& that);

	// Copy ourselves
	virtual Callback* clone(void);

	// Call the callback
	virtual void operator()(void);

protected:

	// The method to call
	TwoArgClassCallbackMethod				m_cb;

	// Object to callback to
	T*												m_obj;

	// First argument to the callback method
	A1												m_arg1;

	// Second argument to the callback method
	A2												m_arg2;

};


}		// namespace


// Include template methods
#include "CallbackT.cc"

#endif

