/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user at the client, and
 * tracking/controling resource usage.
 *
 * $Id: ClientAuthenticator.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include "ClientAuthenticator.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "Tunnel.hh"
#include "Debug.hh"
#include "Quotas.hh"
#include "ServerAuthSettings.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
ClientAuthenticator::ClientAuthenticator() throw(AuthError)
: Authenticator()
{
	DPRINTV("ClientAuthenticator::ClientAuthenticator()");

	// Create initial user
	std::string initialUser = m_defaultAuthSettings->getUser();
	m_users[initialUser] = createUser(initialUser, true);
}

ClientAuthenticator::~ClientAuthenticator()
{
	DPRINTV("ClientAuthenticator::~ClientAuthenticator()");
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get copy of current user details
User ClientAuthenticator::getUserDetails(std::string name)
{
	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		return *user;
	}

	// Return blank authenticated user details
	User blankUser;
	return blankUser;
}


// Get whether user is authenticated for some tunnel
bool ClientAuthenticator::getUserAuthenticated(std::string name, Tunnel* tun)
{
	// Always authenticated at the client
	return true;
}


/************/
/* Mutators */
/************/


// Set the username & passwd we use to authenticate at the server
void ClientAuthenticator::setUserPasswd(std::string username, std::string passwd)
{
	MutexVar mv(m_mtx);

	// This username becomes our default
	m_defaultUserName = username;
	m_defaultAuthSettings->setUserPasswd(username, passwd);

	// Create this user with default settings
	m_users[username] = createUser(username, false);
}


/***********************/
/* Protected Interface */
/***********************/


// Create a new tunnel user
User ClientAuthenticator::createUser(std::string name, bool keepCached)
{
	// Create user already authenticated, with unlimited quotas
	Quotas quotas;
	User user(true, name, "", quotas, keepCached);
	return user;
}


}	// namespace

