/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate client tunnel configuration settings in a handy blob.
 *
 * $Id: ClientConfigBlob.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "ClientConfigBlob.hh"
#include "Debug.hh"


namespace HTTPTunnel {


// Construction/destruction
ClientConfigBlob::ClientConfigBlob(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, std::string shost, unsigned int sport)
: ConfigBlob(), m_socks(true), m_prot(prot), m_ssl(ssl), m_lhost(lhost), m_lport(lport),
  m_shost(shost), m_sport(sport), m_dhost(""), m_dport(0), m_cliAuth(false),
  m_cliUser(""), m_cliPasswd(""), m_viaProxy(false), m_phost(""), m_pport(0),
  m_proxyAuth(false), m_proxyUser(""), m_proxyPasswd("")
{
}

ClientConfigBlob::ClientConfigBlob(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, std::string shost, unsigned int sport, std::string dhost,
	unsigned int dport)
: ConfigBlob(), m_socks(false), m_prot(prot), m_ssl(ssl), m_lhost(lhost), m_lport(lport),
  m_shost(shost), m_sport(sport), m_dhost(dhost), m_dport(dport), m_cliAuth(false),
  m_cliUser(""), m_cliPasswd(""), m_viaProxy(false), m_phost(""), m_pport(0),
  m_proxyAuth(false), m_proxyUser(""), m_proxyPasswd("")
{
}

ClientConfigBlob::ClientConfigBlob(const ClientConfigBlob& that)
: ConfigBlob(that), m_socks(that.m_socks), m_prot(that.m_prot), m_ssl(that.m_ssl),
  m_lhost(that.m_lhost), m_lport(that.m_lport), m_shost(that.m_shost), m_sport(that.m_sport),
  m_dhost(that.m_dhost), m_dport(that.m_dport), m_cliAuth(that.m_cliAuth),
  m_cliUser(that.m_cliUser), m_cliPasswd(that.m_cliPasswd), m_viaProxy(that.m_viaProxy),
  m_phost(that.m_phost), m_pport(that.m_pport), m_proxyAuth(that.m_proxyAuth),
  m_proxyUser(that.m_proxyUser), m_proxyPasswd(that.m_proxyPasswd)
{
}

ClientConfigBlob::~ClientConfigBlob()
{
}


// Assignment
ClientConfigBlob& ClientConfigBlob::operator=(const ClientConfigBlob& that)
{
	if(this != &that) {
		// Base class
		ConfigBlob::operator=(that);

		// Our members
		m_socks = that.m_socks;
		m_prot = that.m_prot;
		m_ssl = that.m_ssl;
		m_lhost = that.m_lhost;
		m_lport = that.m_lport;
		m_shost = that.m_shost;
		m_sport = that.m_sport;
		m_dhost = that.m_dhost;
		m_dport = that.m_dport;
		m_cliAuth = that.m_cliAuth;
		m_cliUser = that.m_cliUser;
		m_cliPasswd = that.m_cliPasswd;
		m_viaProxy = that.m_viaProxy;
		m_phost = that.m_phost;
		m_pport = that.m_pport;
		m_proxyAuth = that.m_proxyAuth;
		m_proxyUser = that.m_proxyUser;
		m_proxyPasswd = that.m_proxyPasswd;
	}

	return *this;
}


/*************/
/* Accessors */
/*************/


// Get blob type
std::string ClientConfigBlob::getType(void) const
{
	return "CLIENT";
}

// Get blob config as a parsable stream
std::string ClientConfigBlob::getConfig(void) const
{
	std::stringstream cfg;

	/* Format ourselves */

	// Tunnel type
	if(m_socks) {
		// SOCKS
		cfg << "socks" << std::endl;
	}
	else {
		// Fixed
		cfg << "fixed " << m_dhost << " " << m_dport;
		if(m_prot == Enums::UDP)
			cfg << " udp";
		cfg << std::endl;
	}

	// Use SSL?
	if(m_ssl)
		cfg << "ssl" << std::endl;

	// Local server listen address
	cfg << "listen ";
	if(m_lhost.size() > 0)
		cfg << m_lhost << " ";
	cfg << m_lport << std::endl;

	// Tunnel server address
	cfg << "tunsvr " << m_shost << " " << m_sport << std::endl;

	// Client authentication
	if(m_cliAuth)
		cfg << "cliauth " << m_cliUser << " " << m_cliPasswd << std::endl;

	// Proxy settings
	if(m_viaProxy) {
		cfg << "proxy " << m_phost << " " << m_pport << std::endl;
		if(m_proxyAuth)
			cfg << "proxyauth " << m_proxyUser << " " << m_proxyPasswd << std::endl;
	}

	return cfg.str();
}


/************/
/* Mutators */
/************/


// Enable client authentication to the server and specify username & passwd to use
void ClientConfigBlob::setClientAuth(std::string user, std::string passwd)
{
	m_cliAuth = true;
	m_cliUser = user;
	m_cliPasswd = passwd;
}

// Go via a proxy
void ClientConfigBlob::useProxy(std::string host, unsigned int port)
{
	m_viaProxy = true;
	m_phost = host;
	m_pport = port;
	m_proxyAuth = false;
}

// Go via a proxy that requires authentication
void ClientConfigBlob::useProxy(std::string host, unsigned int port, std::string user,
	std::string passwd)
{
	m_viaProxy = true;
	m_phost = host;
	m_pport = port;
	m_proxyAuth = true;
	m_proxyUser = user;
	m_proxyPasswd = passwd;
}


}	// namespace HTTPTunnel

