/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel.
 *
 * $Id: ClientTunneler.cc,v 1.9 2009-05-08 13:44:54 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <vector>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <memory>
#include "ClientTunneler.hh"
#include "Enums.hh"
#include "Debug.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "DummyRawConnection.hh"
#include "HTTPConnection.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ProxySettings.hh"
#include "ServerAuthSettings.hh"
#include "TunnelID.hh"
#include "TunnelerID.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "ClientTunnel.hh"
#include "TestClientTunnel.hh"
#include "MutexVar.hh"
#include "HTTPMsg.hh"
#include "HTTPResponseMsg.hh"
#include "ShutdownTunnelEvent.hh"
#include "HTTPMsgParms.hh"
#include "CheckAckEvent.hh"
#include "PollTimeoutEvent.hh"
#include "SendMsgEvent.hh"
#include "LaunchServerTestEvent.hh"
#include "Utils.hh"
#include "Authenticator.hh"
#include "ConnectionManager.hh"
#include "Server.hh"
#include "Sock.hh"
#include "TCPSockTunClient.hh"
#include "UDPSockTunClient.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ClientTunneler::ClientTunneler(Server* server) throw(TunnelError)
: Tunneler(server), m_tunParms(TestClientTunnel::getDefaultTunParms()),
  m_sockTunClient(0), m_oldTunParmsAge(OLD_TUN_PARMS_AGE)
{
	DPRINTV("ClientTunneler::ClientTunneler()");

	try {
		// Slightly randomise old tunnel parms age
		int otpaRand = ((int)(((float)rand() / RAND_MAX) * OLD_TUN_PARMS_AGE)) / 10;
		m_oldTunParmsAge += otpaRand;

		// Schedule test of connection to server
		MutexVar tmv(m_tunnelsMtx);
		checkScheduleServerTest(0);

		// See if we need a socket tunnel client
		const ClientSideConfig* clientCfg = getClientSideConfig();
		if(clientCfg->isSockTun()) {
			// Get address socket tunnel is listening on
			std::string addr = clientCfg->getListenIP();
			unsigned int port = clientCfg->getListenPort();

			// Create appropriate client for socket tunnel
			if(clientCfg->getProtocol() == Enums::TCP) {
				// TCP socket tunnel
				m_sockTunClient = new TCPSockTunClient(addr, port);
			}
			else {
				// UDP socket tunnel
				m_sockTunClient = new UDPSockTunClient(addr, port);
			}
		}
	}
	catch(TunnelError& e) {
		// Clear up and rethrow
		shutdownSchedulers();
		throw;
	}
}

ClientTunneler::~ClientTunneler()
{
	DPRINTV("ClientTunneler::~ClientTunneler()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();

	// Shutdown any socket tunnel
	if(m_sockTunClient) {
		delete m_sockTunClient;
		m_sockTunClient = 0;
	}
}


/*************/
/* Accessors */
/*************/


// Get reference to current tunneling parameters
ClientTunnelParms ClientTunneler::getTunParms(void)
{
	MutexVar tmv(m_tunnelsMtx);
	return m_tunParms;
}


// Get socket descriptor for client socket tunnel
Sock* ClientTunneler::getSockTunSock(void) throw(TunnelError)
{
	MutexVar tmv(m_tunnelsMtx);

	// Check we have a socket tunnel client
	if(m_sockTunClient == 0)
		throw TunnelError("No socket tunnel client in use");

	// Get socket descriptor from socket tunnel client
	return m_sockTunClient->getSock();
}


/************/
/* Mutators */
/************/


// Enable/disable testing based on current config setting
void ClientTunneler::setServerTesting(void)
{
	DPRINTV("ClientTunneler::setServerTesting()");

	MutexVar tmv(m_tunnelsMtx);

	// Do we have a test tunnel running?
	Tunnel* testTun = getTunFromPred(PredTestTun());
	bool testRunning = testTun != 0;

	// Do we have a test scheduled?
	bool testScheduled = isScheduled<LaunchServerTestEvent>();

	// Are we starting or stopping testing?
	if(g_pConfig->getTesting()) {
		DPRINTV("Enabling testing...");

		// Start testing if it's stopped
		if(!testRunning && !testScheduled) {
			// Schedule test of connection to server
			DPRINTV("Starting testing...");
			checkScheduleServerTest(0);
		}
	}
	else {
		DPRINTV("Disabling testing...");

		// Stop testing if it's started
		if(testScheduled) {
			// Cancel scheduled test
			DPRINTV("Cancelling scheduled test...");
			cancelAllEvents<LaunchServerTestEvent>();
		}
		if(testRunning) {
			// Shutdown runnign test tunnel
			DPRINTV("Aborting running test...");
			scheduleEvent(new ShutdownTunnelEvent(0, this, testTun->getTid(), false));
		}
	}
}


/**********/
/* Events */
/**********/


// Notification of a new socket connection
void ClientTunneler::newConnectionArrived(Sock* insock)
	throw(CommsError)
{
	DPRINTI("ClientTunneler::newConnectionArrived()");

	///////////////////////////////////////////////////////////////
	// In ClientTunneler a new connection always means a new tunnel
	///////////////////////////////////////////////////////////////

	// Do our users quotas allow another tunnel?
	Authenticator* authenticator = m_server->getConfigMutable()->getAuthenticator();
	std::string userName = authenticator->getDefaultUserName();
	bool userQuotaOk = authenticator->getUserCanCreateTunnel(userName);
	if(userQuotaOk) {
		// Create new tunnel config
		TunnelConfig* tcfg = createTunnelConfig();

		// Lock connection details
		MutexVar cmv(m_connectionsMtx);

		// Create raw server connection to handle incoming traffic
		Connection* rawconn = new RawConnection(insock, this, *tcfg->getTunnelID());

		// Create HTTP client connection to handle tunneled traffic
		HTTPConnection* httpconn = new HTTPConnection(this);

		// Save connection details
		addConnection(rawconn);
		addConnection(httpconn);

		// Create tunnel
		Tunnel* tun = new ClientTunnel(tcfg, this);
		tun->addLeftConnection(rawconn);
		tun->addRightConnection(httpconn);

		// Schedule connect out on HTTP side
		const ClientSideConfig* clientCfg = getClientSideConfig();
		m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
			clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL());

		cmv.unlock();

		// Store Tunnel object in our list of active tunnels
		MutexVar tmv(m_tunnelsMtx);
		m_tunnels.push_back(tun);
		tmv.unlock();
	}
	else {
		// Tunnel creation quota exceeded
		std::stringstream err;
		err << "Tunnel creation quota exceeded for user " << userName;
		throw CommsError(err.str());
	}
}


// Notification that a tunnel has insufficient left/right connections to be viable
void ClientTunneler::unviableTunnel(const TunnelID& tid)
{
	DPRINTI("ClientTunneler::unviableTunnel() - " + tid.toString());

	// Lookup tunnel
	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// If tunnel is already closed don't bother checking any further
		if(!tun->isClosed()) {
			// Get details about state of tunnel
			unsigned int leftConns = tun->noLeftConnections();
			unsigned int rightConns = tun->noRightConnections();

			// Check whether our left side warrants preserving
			bool leftPreserve = true;
			if(leftConns == 0) {
				/* In the client, no remaining left connections means we must
				 * abandon this tunnel and close it down.
				 */

				// Is this tunnel in the process of closing?
				if(tun->isClosing()) {
					// Give us a chance to finish closing
					DPRINTI(tid.toString() + " - No lconns, closing: Left preserve yes");
					leftPreserve = true;
				}
				else {
					// No left connections and not closing, so don't preserve
					DPRINTI(tid.toString() + " - No lconns, not closing: Left preserve no");
					leftPreserve = false;
				}
			}

			// Check whether our right side warrants preserving
			bool rightPreserve = true;
			if(rightConns == 0) {
				/* In the client, no remaining right connections means that if
				 * we want to try and preserve this tunnel we should try and
				 * re-establish a connection to the server.
				 */

				// Do we want to try and preserve this tunnel?
				if(!tun->errorLevelExceeded(Tunnel::RIGHT)) {
					// Consider preserving right side
					DPRINTI(tid.toString() + " - No rconns, not erred: Right preserve yes");
					rightPreserve = true;
				}
				else {
					// Don't consider preserving right side
					DPRINTI(tid.toString() + " - No rconns, erred: Right preserve no");
					rightPreserve = false;
				}
			}

			// Both sides agree to preserve?
			if(leftPreserve && rightPreserve) {
				DPRINTI(tid.toString() + " - Trying to preserve");

				// Should we delay preservation of tunnel?
				Tunnel::TunnelState oldTunState = tun->getState();
				unsigned int preserveDelay = 0;
				if(oldTunState != Tunnel::CONNECTED) {
					// Tunnel wasn't connected
					DPRINTV("Tunnel wasn't CONNECTED - delaying preserve connect");
					preserveDelay = NC_PRESERVE_RECONNECT_DELAY;
				}
				else if(!tun->isPersistent()) {
					// Tunnel wasn't persistent
					DPRINTV("Tunnel is non-persistent - delaying preserve connect");
					preserveDelay = NP_PRESERVE_RECONNECT_DELAY;
				}

				// Let tunnel know
				tun->reconnecting();

				// Create replacement HTTP client connection to server
				HTTPConnection* httpconn = new HTTPConnection(this);

				// Save connection details
				addConnection(httpconn);
				tun->addRightConnection(httpconn);

				// Schedule initial connect out HTTP side
				const ClientSideConfig* clientCfg = getClientSideConfig();
				m_connectionMgr->addProxyConnect(httpconn, Time(0,preserveDelay),
					clientCfg->getNextHopIP(), clientCfg->getNextHopPort(),
					Enums::TCP, clientCfg->getSSL());
			}
			else {
				DPRINTI(tid.toString() + " - Not preserving");
				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
			}
		}
		else {
			DPRINTI(tid.toString() + " - Already closed, not preserving further");
		}
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::unviableTunnel() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification of new raw data arrival
void ClientTunneler::rawDataArrived(unsigned int connID, const TunnelID& tid,
	const RawBuff& raw) throw(LookupManagementError)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ClientTunneler::rawDataArrived() - ";
	nfo << raw.getLength() << " bytes on connection " << connID;
	DPRINTV(nfo.str());
	#endif

	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Inform tunnel of raw data arrival
		tun->processRawData(raw);

		// Try sending this data on to the server
		tun->sendTunnelMsg(true, 0);
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::rawDataArrived() - On connection " << connID;
		err << " Unknown tunnel " << tid;
		throw LookupManagementError(err.str());
	}
}


// Notification of a new UDP 'connection'
void ClientTunneler::newUDPConnectionArrived(Sock* insock) throw(CommsError)
{
	////////////////////////////////////////////////////////////////////
	// We currently only support fixed UDP tunnelling, so if we already
	// have a tunnel created we can ignore this 'new connection'.
	////////////////////////////////////////////////////////////////////

	// Do we already have a UDP tunnel to use?
	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);
	if(m_tunnels.size() == 0) {
		DPRINTI("ClientTunneler::newUDPConnectionArrived()");

		// Create new tunnel config
		TunnelConfig* tcfg = createTunnelConfig();

		// Create raw server connection to handle incoming traffic
		Connection* rawconn = new RawConnection(insock, this, *tcfg->getTunnelID());

		// Create HTTP client connection to handle tunneled traffic
		HTTPConnection* httpconn = new HTTPConnection(this);

		// Save connection details
		addConnection(rawconn);
		addConnection(httpconn);

		// Create tunnel
		Tunnel* tun = new ClientTunnel(tcfg, this);
		tun->addLeftConnection(rawconn);
		tun->addRightConnection(httpconn);

		// Schedule connect out on HTTP side
		const ClientSideConfig* clientCfg = getClientSideConfig();
		m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
			clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL());

		// Store Tunnel object in our list of active tunnels
		m_tunnels.push_back(tun);
	}
}


// Check we have received an ack for some sent message
void ClientTunneler::checkAckReceived(void)
{
	std::vector<Tunnel*>::iterator tit;
	MutexVar tmv(m_tunnelsMtx);

	// Check acks for all tunnels
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Let tunnel check we have received any due acks, & resend if not
		Tunnel* tun = *tit;
		if(tun->checkAckTimeout()) {
			tun->sendTunnelMsg(true, 0);
		}
	}

	// Schedule next ack check event
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
}


// Notification of new HTTP message arrival
void ClientTunneler::messageArrived(unsigned int connID, HTTPMsg* msg)
	throw(LookupManagementError)
{
	// Base class
	Tunneler::messageArrived(connID, msg);

	// Lookup connection
	MutexVar cmv(m_connectionsMtx);
	std::vector<Connection*>::iterator cit = getConnection(connID);
	Connection* conn = *cit;

	// Check for HTTP error messages
	bool fatalConnectionError = false;
	const HTTPResponseMsg* respMsg = dynamic_cast<const HTTPResponseMsg*>(msg);
	unsigned short respCode = respMsg->getResponseCode();
	if(respCode == 407 || respCode == 401) {
		// Authentication required
		authenticationFailed(connID, respMsg, respCode);
	}
	else if(respCode == 404) {
		// Probably a proxy error
		responseError(connID, respMsg, respCode);
	}
	else {
		// Standard tunneling msg
		try {
			// Find out which Tunnel this message is from
			MutexVar tmv(m_tunnelsMtx);
			const HTTPMsgParms* parms = msg->getParms();
			TunnelID tid(parms->getTid());
			ClientTunnel* tun = dynamic_cast<ClientTunnel*>(getTunFromPred(PredTunID(&tid)));

			// Have we found the tunnel for this msg?
			if(tun) {
				// Normal response code?
				if(respCode == 200) {
					// We know we are authenticated
					tun->setUserAuthenticated(true);

					// Reset auth settings
					if(tun->getServerAuthSettings() != 0)
						tun->getServerAuthSettings()->reset();
					if(tun->getProxySettings() != 0)
						tun->getProxySettings()->reset();
				}

				// Handle any received tunnel payload
				HTTPConnection* httpConn = dynamic_cast<HTTPConnection*>(conn);
				tun->processTunnelMsg(msg, httpConn);

				// Does this connection report it is closing?
				if(parms->getConnClose() || parms->getProxyConnClose()) {
					#ifdef DEBUG_TUN
					std::stringstream dbg;
					dbg << "Connection id " << connID << " reports it is closing";
					DPRINTV(dbg.str());
					#endif

					// Don't use closing connection any more
					connectionClosing(cit);
				}

				// Check reported tunnel state
				if(parms->getState() == Tunnel::CLOSED) {
					// Notification of tunnel closure, schedule tunnel shutdown
					DPRINTI("Received tunnel shutdown request from server");
					scheduleEvent(new ShutdownTunnelEvent(0, this, tid, true));
				}
				else if(parms->getState() == Tunnel::AUTHREQ) {
					// Send back authentication response
					authenticationRequest(tun);
				}
			}
			else {
				// Unknown tunnel
				std::stringstream wrn;
				wrn << "ClientTunneler::messageArrived() - Unknown tunnel " << tid;
				g_pLogger->warn(wrn.str());
				fatalConnectionError = true;
			}
		}
		catch(InvalidID& e) {
			g_pLogger->error("Error parsing ID: " + e.what());
			fatalConnectionError = true;
		}
	}

	// Did something very serious fail?
	if(fatalConnectionError) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Fatal error processing msg on connection " << connID;
		DPRINTV(dbg.str());
		#endif

		// No point keeping this connection around, kill it now
		destroyConnection(cit);
	}
}


// Notification of some new default tunnel parameters
void ClientTunneler::newTunnelParms(TunnelParms* parms, const TunnelerID& tid)
{
	DPRINTI("ClientTunneler::newTunnelParms() - " + tid.toString());

	// Base class
	Tunneler::newTunnelParms(parms, tid);

	// Save new parms
	MutexVar tmv(m_tunnelsMtx);
	ClientTunnelParms* cparms = dynamic_cast<ClientTunnelParms*>(parms);
	m_tunParms = *cparms;

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "New ClientTunnelParms: Max URL = " << cparms->getURLPayloadSize() << ", ";
	nfo << "Max body = " << cparms->getBodyPayloadSize();
	DPRINTV(nfo.str());
	#endif
}


// Notification of a poll timeout on a tunnel
void ClientTunneler::pollTimeout(const TunnelID& tid)
{
	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Send a poll msg to the server side
		ClientTunnel* cliTun = dynamic_cast<ClientTunnel*>(tun);
		cliTun->pollTunnel();
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::pollTimeout() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification that a tunnel has changed state
void ClientTunneler::tunnelStateChanged(const TunnelID& tid,
	Tunnel::TunnelState state)
{
	DPRINTI("ClientTunneler::tunnelStateChanged() - " + tid.toString());

	// Has tunnel become CONNECTED?
	if(state == Tunnel::CONNECTED) {
		// Lookup tunnel
		MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Do we have a poll scheduled for this tunnel?
			if(!isScheduled<PollTimeoutEvent, TunnelID>(tid)) {
				// Perform first poll of newly connected tunnel
				ClientTunnel* cliTun = dynamic_cast<ClientTunnel*>(tun);
				cliTun->pollTunnel();
			}
		}
		else {
			std::stringstream err;
			err << "ClientTunneler::tunnelStateChanged() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
}


// Notification that we should perform tunnel manangement
void ClientTunneler::performTunnelManagement(void)
{
	// When did we last check tunneling parameters?
	MutexVar tmv(m_tunnelsMtx);
	Time checkTime = Time::now() - m_lastTunParmsCheckTime;
	if(checkTime.getSecs() >= CHECK_TUN_PARMS_PERIOD) {
		m_lastTunParmsCheckTime = Time::now();
		// Do we have completed test tunneling parameters?
		if(m_tunParms.testsCompleted()) {
			// Are our test parameters too old?
			Time parmAge = Time::now() - m_tunParms.getCreateTime();
			if(parmAge.getSecs() >= m_oldTunParmsAge) {
				#ifdef DEBUG_TUN
				std::stringstream dbg;
				dbg << "Removing " << parmAge.getSecs() << " secs old test tunneling parms";
				DPRINTV(dbg.str());
				#endif

				// Reset old parms and schedule a new test
				m_tunParms.resetTestState();
				checkScheduleServerTest(0);
			}
		}
	}

	// Base class
	Tunneler::performTunnelManagement();
}


// Launch test server connection
void ClientTunneler::launchServerTestConnection(void)
{
	DPRINTV("ClientTunneler::launchServerTestConnection()");

	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);

	// Create new test tunnel config
	TunnelConfig* tcfg = createTunnelConfig();

	// Create new test connections
	DummyRawConnection* rawconn = new DummyRawConnection(this, *tcfg->getTunnelID());
	HTTPConnection* httpconn = new HTTPConnection(this);
	addConnection(rawconn);
	addConnection(httpconn);

	// Create test tunnel
	Tunnel* tun = new TestClientTunnel(tcfg, this);
	tun->addLeftConnection(rawconn);
	tun->addRightConnection(httpconn);

	// Schedule connect out to server
	const ClientSideConfig* clientCfg = getClientSideConfig();
	m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
		clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL());

	// Store Tunnel object in our list of active tunnels
	m_tunnels.push_back(tun);
}


/***********************/
/* Protected Interface */
/***********************/


// Get reconnection attempt parameters
void ClientTunneler::getReconnectParms(unsigned int& maxConnectAttempts,
	unsigned int& reconnectDelay, unsigned int failedAttempts)
{
	// Increase reconnect delay with each attempt
	maxConnectAttempts = CLIENT_CONNECT_ATTEMPTS;
	reconnectDelay = failedAttempts * CLIENT_RECONNECT_DELAY;
}


// Check whether we should relaunch a server test - Caller holds mutex
void ClientTunneler::checkScheduleServerTest(Tunnel* tun)
{
	// Is tunnel testing enabled?
	if(!g_pConfig->getTesting())
		return;

	// Got a tunnel to check?
	if(tun) {
		if(tun->isTestTunnel()) {
			// Are these test not completed?
			TestClientTunnel* tcTun = dynamic_cast<TestClientTunnel*>(tun);
			if(!tcTun->testsCompleted()) {
				// Check we don't already have a test launch scheduled
				if(!isScheduled<LaunchServerTestEvent>()) {
					// Schedule relaunch to complete failed testing
					DPRINTV("ClientTunneler::checkScheduleServerTest() - Scheduling relaunch");
					scheduleEvent(new LaunchServerTestEvent(BAD_TEST_LAUNCH_DELAY, this));
				}
			}
		}
	}
	else {
		// Check we don't already have a test tunnel running
		if(getTunFromPred(PredTestTun()) == 0) {
			// Check we don't already have a test launch scheduled
			if(!isScheduled<LaunchServerTestEvent>()) {
				// Launch new test tunnel
				DPRINTV("ClientTunneler::checkScheduleServerTest() - Launching new test");
				scheduleEvent(new LaunchServerTestEvent(0, this));
			}
		}
	}
}


// Create config for tunnel to server
TunnelConfig* ClientTunneler::createTunnelConfig(void)
{
	// Create tunnelID
	ClientSideConfig* clientCfg = getClientSideConfigMutable();
	TunnelID* tid = new TunnelID(clientCfg);

	// Create tunnel config
	return new TunnelConfig(tid, clientCfg->getTunTune(), clientCfg->getProxySettings(),
		clientCfg->getAuthenticator()->getDefaultAuthSettings(), clientCfg->getFixedDestIP(),
		clientCfg->getFixedDestPort());
}


/*********************/
/* Private Interface */
/*********************/


// Respond to an authentication request from the server - Mutexes held by caller
void ClientTunneler::authenticationRequest(Tunnel* tun)
{
	DPRINTI("ClientTunneler::authenticationRequest()");

	// Send back authentication response
	tun->sendTunnelMsg(true, 0);
}


// Notification of an authentication failure
void ClientTunneler::authenticationFailed(unsigned int connID,
	const HTTPResponseMsg* respMsg, unsigned short respCode)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ClientTunneler::authenticationFailed() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	bool proxy = false;

	try {
		// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
		MutexVar cmv(m_connectionsMtx);
		MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		// Lookup carried tunnels (can't rely on having tunnel ID from msg)
		std::vector<Tunnel*> tunnels;
		std::vector<Tunnel*>::iterator tit;
		getAllTunsFromPred(PredTunConnection(conn), tunnels);
		for(tit = tunnels.begin(); tit != tunnels.end(); ++tit) {
			ClientTunnel* tun = dynamic_cast<ClientTunnel*>(*tit);

			try {
				// Reconfigure whatever is necessary for the required auth
				if(respCode == 407) {
					proxy = true;
					setupProxyServerAuthConfig(respMsg->getAuthArgs(), tun->getProxySettings(),
						connID, tun->getTidStr());
				}
				else if(respCode == 401) {
					tun->setUserAuthenticated(false);
					setupProxyServerAuthConfig(respMsg->getAuthArgs(), tun->getServerAuthSettings(),
						connID, tun->getTidStr());
				}

				// Schedule chance to resend messages using authentication
				scheduleEvent(new SendMsgEvent(Time(0,RESEND_AUTH_DELAY), connID, true, tun->getTid(),
					this));
			}
			catch(AuthProcessingError& e) {
				g_pLogger->warn("Authentication error: " + e.what());

				// Fatal tunnel error
				tun->recordConnectionError((proxy? Tunnel::PROXY_AUTH_ERROR : Tunnel::SERVER_AUTH_ERROR));
			}
		}
	}
	catch(TunnelError& e) {
		g_pLogger->warn("Authentication error: " + e.what());

		// Schedule fatal connection error
		scheduleConnectionError(connID, (proxy? Tunnel::PROXY_AUTH_ERROR : Tunnel::SERVER_AUTH_ERROR));
	}
}


// Setup config for required proxy/server authentication
void ClientTunneler::setupProxyServerAuthConfig(const std::map<std::string,
	std::vector<std::string> >& authDetails, AuthSettings* settings, unsigned int connID,
	std::string tidStr) throw(AuthProcessingError)
{
	// Pick authentication scheme
	bool triedScheme = false;
	bool basic = false, digest = false, ntlm = false;
	std::vector<std::string> basicArgs, digestArgs, ntlmArgs;
	std::map<std::string, std::vector<std::string> >::const_iterator ait;
	if((ait = authDetails.find("Basic")) != authDetails.end()) {
		// Can use basic auth
		basic = true;
		basicArgs = ait->second;
	}
	if((ait = authDetails.find("Digest")) != authDetails.end()) {
		// Can use digest auth
		digest = true;
		digestArgs = ait->second;
	}
	if((ait = authDetails.find("NTLM")) != authDetails.end()) {
		// Can use NTLM auth
		ntlm = true;
		ntlmArgs = ait->second;
	}

	// Got nothing to try?
	if(!(basic || digest || ntlm))
		throw AuthProcessingError("No suitable authentication scheme suggested");

	// Setup for basic authentication
	if(!triedScheme && basic) {
		// Check we haven't already tried this
		if(!settings->triedAuthScheme(AuthSettings::AUTH_BASIC)) {
			// Try basic authentication
			DPRINTI("Trying basic authentication");
			try {
				// Setup basic authentication
				settings->updateAuthScheme(AuthSettings::AUTH_BASIC, basicArgs);
				triedScheme = true;
			}
			catch(AuthProcessingError& e) {
				// Need to try something else
				std::stringstream nfo;
				nfo << "Basic authentication error for tunnel " << tidStr << ": " << e.what();
				g_pLogger->info(nfo.str());
			}
		}
		else {
			// Need to try something else
			std::stringstream nfo;
			nfo << "Basic authentication failed for tunnel " << tidStr;
			g_pLogger->info(nfo.str());
		}
	}

	// Setup for digest authentication
	if(!triedScheme && digest) {
		// Check we haven't already tried this
		if(!settings->triedAuthScheme(AuthSettings::AUTH_DIGEST)) {
			// Try digest authentication
			DPRINTI("Trying digest authentication");
			try {
				// Setup digest authentication
				settings->updateAuthScheme(AuthSettings::AUTH_DIGEST, digestArgs);
				triedScheme = true;
			}
			catch(AuthProcessingError& e) {
				// Need to try something else
				std::stringstream nfo;
				nfo << "Digest authentication error for tunnel " << tidStr << ": " << e.what();
				g_pLogger->info(nfo.str());
			}
		}
		else {
			// Need to try something else
			std::stringstream nfo;
			nfo << "Digest authentication failed for tunnel " << tidStr;
			g_pLogger->info(nfo.str());
		}
	}

	// Setup for NTLM authentication
	if(!triedScheme && ntlm) {
		// Check we haven't already tried this
		if(!settings->triedAuthScheme(AuthSettings::AUTH_NTLM)) {
			// Try NTLM authentication
			DPRINTI("Trying NTLM authentication");
			try {
				// Setup NTLM authentication
				settings->updateAuthScheme(AuthSettings::AUTH_NTLM, ntlmArgs);
				triedScheme = true;
			}
			catch(AuthProcessingError& e) {
				// Need to try something else
				std::stringstream nfo;
				nfo << "NTLM authentication error for tunnel " << tidStr << ": " << e.what();
				g_pLogger->info(nfo.str());
			}
		}
		else {
			// Need to try something else
			std::stringstream nfo;
			nfo << "NTLM authentication failed for tunnel " << tidStr;
			g_pLogger->info(nfo.str());
		}
	}

	// Have we got nothing left to try?
	if(!triedScheme)
		throw AuthProcessingError("No suitable authentication scheme left to try");
}


// Handle a msg response error
void ClientTunneler::responseError(unsigned int connID, const HTTPResponseMsg* respMsg,
	unsigned short respCode)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ClientTunneler::responseError() - Connection id " << connID << ", ";
	nfo << "response code: " << respCode;
	DPRINTI(nfo.str());
	#endif

	// Replace this connection with a new one and see if that cures the reponse error
	scheduleConnectionError(connID, Tunnel::GENERAL_COMMS_ERROR);
}


}	// namespace

