/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel.
 *
 * $Id: ClientTunneler.hh,v 1.7 2009-05-08 13:44:54 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _CLIENT_TUNNELER_HH_
#define _CLIENT_TUNNELER_HH_

#include <map>
#include <vector>
#include "Tunneler.hh"
#include "Exceptions.hh"
#include "ClientTunnelParms.hh"

namespace HTTPTunnel {

class Tunnel;
class Connection;
class HTTPResponseMsg;
class AuthSettings;
class ClientSideConfig;
class SockTunClient;
class Sock;


class ClientTunneler : public Tunneler {
public:
   // Construction/destruction
   ClientTunneler(Server* server) throw(TunnelError);
   virtual ~ClientTunneler();


	/*************/
	/* Accessors */
	/*************/

	// Get reference to current tunneling parameters
	ClientTunnelParms getTunParms(void);

	// Get socket descriptor for client socket tunnel
	Sock* getSockTunSock(void) throw(TunnelError);


	/************/
	/* Mutators */
	/************/

	// Enable/disable testing based on current config setting
	virtual void setServerTesting(void);


	/**********/
	/* Events */
	/**********/

	// Notification of a new socket connection
	virtual void newConnectionArrived(Sock* insock) throw(CommsError);

	// Notification that a tunnel has insufficient left/right connections to be viable
	virtual void unviableTunnel(const TunnelID& tid);

	// Notification of new tunneled/raw data arrival
	virtual void rawDataArrived(unsigned int connID, const TunnelID& tid,
		const RawBuff& raw) throw(LookupManagementError);

	// Notification of a new UDP 'connection'
	virtual void newUDPConnectionArrived(Sock* insock) throw(CommsError);

	// Check we have received an ack for some sent message
	virtual void checkAckReceived(void);

	// Notification of new HTTP message arrival
	virtual void messageArrived(unsigned int connID, HTTPMsg* msg)
		throw(LookupManagementError);

	// Notification of some new default tunnel parameters
	virtual void newTunnelParms(TunnelParms* parms, const TunnelerID& tid);

	// Notification of a poll timeout on a tunnel
	virtual void pollTimeout(const TunnelID& tid);

	// Notification that a tunnel has changed state
	virtual void tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state);

	// Notification that we should perform tunnel manangement
	virtual void performTunnelManagement(void);

	// Launch test server connection
	virtual void launchServerTestConnection(void);


protected:

	/*************/
	/* Constants */
	/*************/

	// Number of connect attempts before giving up
	static const unsigned int			CLIENT_CONNECT_ATTEMPTS			= 4;

	// Time between reconnection attempts
	static const unsigned int			CLIENT_RECONNECT_DELAY			= 3;

	// Time to pause (milli-seconds) between preservation reconnects
	static const unsigned int			NC_PRESERVE_RECONNECT_DELAY	= 200;
	static const unsigned int			NP_PRESERVE_RECONNECT_DELAY	= 250;
	static const unsigned int			RESEND_AUTH_DELAY					= 100;

	// Delay (seconds) before relaunching failed test connection
	static const unsigned int			BAD_TEST_LAUNCH_DELAY			= 2;


	/***********/
	/* Methods */
	/***********/

	// Get reconnection attempt parameters
	virtual void getReconnectParms(unsigned int& maxConnectAttempts,
		unsigned int& reconnectDelay, unsigned int failedAttempts);

	// Check whether we should relaunch a server test
	virtual void checkScheduleServerTest(Tunnel* tun);

	// Create config for tunnel to server
	TunnelConfig* createTunnelConfig(void);


	/********/
	/* Data */
	/********/

	// Tunneling parameters to use
	ClientTunnelParms						m_tunParms;

	// Socket tunnel client
	SockTunClient*							m_sockTunClient;


private:

	/***********/
	/* Methods */
	/***********/

	// Respond to an authentication request from the server
	void authenticationRequest(Tunnel* tun);

	// Notification of an authentication failure
	void authenticationFailed(unsigned int connID, const HTTPResponseMsg* respMsg,
		unsigned short respCode);

	// Setup config for required proxy/server authentication
	void setupProxyServerAuthConfig(const std::map<std::string,
		std::vector<std::string> >& authDetails, AuthSettings* settings,
		unsigned int connID, std::string tidStr) throw(AuthProcessingError);

	// Handle a msg response error
	void responseError(unsigned int connID, const HTTPResponseMsg* respMsg,
		unsigned short respCode);


	/********/
	/* Data */
	/********/

	// Age (seconds) at which we retest old tunnel parms
	int										m_oldTunParmsAge;

};


}	// namespace

#endif

