/* (C) 2003 Kevin Manning.
 * (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Wrapper for pthread condition variable.
 *
 * $Id: ConditionVar.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


#include <errno.h>
#include <iostream>
#include "MutexVar.hh"
#include "ConditionVar.hh"

namespace HTTPTunnel {

// Construction/destruction
ConditionVar::ConditionVar() throw(ConditionVarError)
{
   #ifdef DEBUG_THREADS
   std::cerr << "ConditionVar::ConditionVar()" << std::endl;
   #endif

   if(pthread_cond_init(&m_condVar, 0) != 0)
      throw ConditionVarError("pthread_cond_init() failed");
}

ConditionVar::~ConditionVar() throw(ConditionVarError)
{
   #ifdef DEBUG_THREADS
   std::cerr << "ConditionVar::~ConditionVar()" << std::endl;
   #endif

   if(pthread_cond_destroy(&m_condVar) != 0)
      throw ConditionVarError("pthread_cond_destroy() failed");
}


// Wait on an already locked mutex
void ConditionVar::wait(MutexVar& mv) throw(ConditionVarError)
{
   #ifdef DEBUG_THREADS
   std::cerr << "Entering ConditionVar::wait()" << std::endl;
   #endif

   mv.released();
   if(pthread_cond_wait(&m_condVar, mv.getMtx()) != 0)
      throw ConditionVarError("pthread_cond_wait() failed");
   mv.taken();

   #ifdef DEBUG_THREADS
   std::cerr << "Leaving ConditionVar::wait()" << std::endl;
   #endif
}

// Timed wait on an already locked mutex
void ConditionVar::timedwait(MutexVar& mv, const struct timeval& tm)
	throw(ConditionVarError)
{
	#ifdef DEBUG_THREADS
	std::cerr << "Entering ConditionVar::timedwait()" << std::endl;
	#endif

	// Construct timeout
	struct timespec ts;
	ts.tv_sec = tm.tv_sec;
	ts.tv_nsec = tm.tv_usec * 1000;

	mv.released();
	int res = pthread_cond_timedwait(&m_condVar, mv.getMtx(), &ts);
	if(res != 0 && res != ETIMEDOUT)
		throw ConditionVarError("pthread_cond_timedwait() failed");
	mv.taken();

	#ifdef DEBUG_THREADS
	std::cerr << "Leaving ConditionVar::timedwait()" << std::endl;
	#endif
}

// Signal a waiting thread
void ConditionVar::signal(void) throw(ConditionVarError)
{
   #ifdef DEBUG_THREADS
   std::cerr << "ConditionVar::signal()" << std::endl;
   #endif

   if(pthread_cond_signal(&m_condVar) != 0)
      throw ConditionVarError("pthread_cond_signal() failed");
}

// Signal all waiting threads
void ConditionVar::broadcast(void) throw(ConditionVarError)
{
   #ifdef DEBUG_THREADS
   std::cerr << "ConditionVar::broadcast()" << std::endl;
   #endif

   if(pthread_cond_broadcast(&m_condVar) != 0)
      throw ConditionVarError("pthread_cond_broadcast() failed");
}


}	// namespace

