/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * An event to trigger a socket connection to some server.
 *
 * $Id: ConnectEvent.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "ConnectEvent.hh"
#include "Debug.hh"
#include "Exceptions.hh"
#include "Connection.hh"
#include "Logger.hh"
#include "Tunneler.hh"
#include "ConnectResultEvent.hh"
#include "ConnectProxy.hh"
#include "TCPSock.hh"


namespace HTTPTunnel {


// Reference to global logger
extern Logger* g_pLogger;


// Construction/destruction
ConnectEvent::ConnectEvent(const Time& time, Connection* conn,
	ConnectProxy* proxy)
: TimedEvent(time), m_conn(conn), m_proxy(proxy)
{
}

ConnectEvent::~ConnectEvent()
{
}


// Display event details for debugging
void ConnectEvent::details(std::stringstream& dbg)
{
	dbg << "ConnectEvent: ";
	TimedEvent::details(dbg);
}


// Do the connect
void ConnectEvent::execute(void)
{
	// Assume failure by default
	ConnectResultEvent::ConnectResultType result = ConnectResultEvent::GENERAL_ERROR;
	Sock* sock = 0;

	try {
		// Launch the connection
		sock = m_proxy->doConnect();
		result = ConnectResultEvent::SUCCESS;
	}
	catch(AsynchronousConnect& e) {
		// Connect happening asychronously
		g_pLogger->info("Connect pending: " + e.what());
		result = ConnectResultEvent::ASYNC;
	}
	catch(FirewalledConnectError& e) {
		// Connect failed
		g_pLogger->info("Connect failed: " + e.what());
		result = ConnectResultEvent::FIREWALLED;
	}
	catch(ConnectError& e) {
		// Connect failed
		g_pLogger->info("Connect failed: " + e.what());
		result = ConnectResultEvent::GENERAL_ERROR;
	}

	// Do we have a result yet?
	if(result != ConnectResultEvent::ASYNC) {
		try {
			// Let connection know result
			m_conn->connectResult(result, sock);
		}
		catch(CommsError& e) {
			g_pLogger->info("Error handling connect result: " + e.what());
		}
	}
}


/*************/
/* Accessors */
/*************/


// Get ID of connection error is for
unsigned int ConnectEvent::getConnectionID(void) const
{
	return m_conn->getID();
}

// Get host connect failed to
std::string ConnectEvent::getHost(void) const
{
	return m_proxy->getHost();
}

// Get port connect failed to
unsigned int ConnectEvent::getPort(void) const
{
	return m_proxy->getPort();
}


}	// namespace

