/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Perform a blocking connect for a connection in a new thread.
 *
 * $Id: ConnectProxy.cc,v 1.2 2008-11-09 18:50:21 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sstream>
#include "ConnectProxy.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Connection.hh"
#include "MutexVar.hh"
#include "ConnectEvent.hh"
#include "Utils.hh"
#include "TCPSock.hh"
#include "UDPSock.hh"
#include "SSLSock.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ConnectProxy::ConnectProxy(Connection* conn, const Time& time, std::string host,
	unsigned int port, Enums::Protocol prot, bool ssl)
: EventReceiver(), m_mtx(), m_conn(conn), m_host(host), m_port(port),
  m_sockType(prot), m_sockSSL(ssl), m_connecting(false), m_asyncSock(0),
  m_asyncPeriod(0)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "ConnectProxy::ConnectProxy() - For connection " << m_conn->getID();
	DPRINTV(dbg.str());
	#endif

	// Schedule connect event
	scheduleConnect(time);
}

ConnectProxy::~ConnectProxy()
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "ConnectProxy::~ConnectProxy() - For connection " << m_conn->getID();
	DPRINTV(dbg.str());
	#endif

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();

	MutexVar mv(m_mtx);

	// Close asynchronous connecting socket
	if(m_connecting && m_asyncSock != 0) {
		delete m_asyncSock;
		m_asyncSock = 0;
	}
}


/*************/
/* Accessors */
/*************/


// Get address of host we are connecting to
std::string ConnectProxy::getHost(void) const
{
	return m_host;
}

// Get port we are connecting to
unsigned int ConnectProxy::getPort(void) const
{
	return m_port;
}


/*************/
/* Utilities */
/*************/


// Schedule a new connect attempt
void ConnectProxy::scheduleConnect(const Time& time)
{
	scheduleEvent(new ConnectEvent(time, m_conn, this));
}


/**********/
/* Events */
/**********/


// Try to do the connect
Sock* ConnectProxy::doConnect(void) throw(ConnectError)
{
	MutexVar mv(m_mtx);

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ConnectProxy::doConnect() - For connection " << m_conn->getID();
	nfo << " connecting to " << m_host << ":" << m_port;
	DPRINTI(nfo.str());
	#endif

	// Waiting for result of asynchronous connect?
	if(m_connecting) {
		return checkAsyncConnect();
	}
	else {
		return initiateConnect();
	}
}


/*********************/
/* Private Interface */
/*********************/


// Initiate a connect - Caller holds mutex
Sock* ConnectProxy::initiateConnect(void) throw(ConnectError)
{
	DPRINTV("ConnectProxy::initiateConnect()");

	Sock* sock = 0;

	// Let connection prepare for connect attempt
	m_conn->connectAttempt(m_host, m_port, m_sockType, m_sockSSL);

	// Setup socket
	try {
		// What type of socket do we require?
		if(m_sockSSL) {
			// Create SSL socket for connect attempt
			sock = new SSLSock();
		}
		else {
			if(m_sockType == Enums::TCP) {
				// Create TCP socket for connect attempt
				sock = new TCPSock();
			}
			else {
				// Create UDP socket
				sock = new UDPSock();
			}
		}
	}
	catch(SockError& e) {
		throw ConnectError("Failed to initiate connect to server: " + e.what());
	}

	// Do the connect
	try {
		sock->connect(m_host, m_port);
	}
	catch(SockError& e) {
		// Why did it fail to connect?
		if(e.m_errno == EINPROGRESS || e.m_errno == EWOULDBLOCK) {
			// Asynchronous connect
			m_connecting = true;
			m_asyncSock = sock;
			scaleAsyncPeriod(true);
			scheduleConnect(Time(0, m_asyncPeriod));
			throw AsynchronousConnect("Connect will block");
		}
		else {
			// Some other failure
			std::stringstream err;
			err << "Socket connect() error: " << e.m_errno;
			throw ConnectError(err.str());
		}
	}
	catch(CommsError& e) {
		throw ConnectError("Failed to connect to server: " + e.what());
	}

	return sock;
}


// Check whether an asynchronous connect has completed - Caller holds mutex
Sock* ConnectProxy::checkAsyncConnect(void) throw(ConnectError)
{
	DPRINTV("ConnectProxy::checkAsyncConnect()");

	bool resultAvailable = false;
	bool connected = false;
	Sock* sock = 0;

	try {
		// Get connect result
		m_asyncSock->getConnectResult(resultAvailable, connected);
		if(resultAvailable) {
			// Did it succeed or fail?
			if(connected) {
				// Success
				m_connecting = false;
				sock = m_asyncSock;
				m_asyncSock = 0;
			}
			else {
				// Failed
				throw ConnectError("Failed to connect to server");
			}
		}
		else {
			// Connect result not yet available, try again shortly
			scaleAsyncPeriod(false);
			scheduleConnect(Time(0, m_asyncPeriod));
			throw AsynchronousConnect("Connect not yet completed");
		}
	}
	catch(SockError& e) {
		throw ConnectError("Failed to connect to server: " + e.what());
	}

	return sock;
}


// Increase asynchronous connect period
void ConnectProxy::scaleAsyncPeriod(bool reset)
{
	if(reset) {
		m_asyncPeriod = MIN_ASYNC_CHECK_PERIOD;
	}
	else {
		m_asyncPeriod = m_asyncPeriod * 2;
		if(m_asyncPeriod > MAX_ASYNC_CHECK_PERIOD) {
			m_asyncPeriod = MAX_ASYNC_CHECK_PERIOD;
		}
	}
}


}	// namespace

