/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for handling reading/writing for one end of a tunnel
 * connection.
 *
 * $Id: Connection.cc,v 1.5 2009-04-24 15:37:33 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <poll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sstream>
#include <algorithm>
#include "Connection.hh"
#include "Reader.hh"
#include "Writer.hh"
#include "MutexVar.hh"
#include "Tunneler.hh"
#include "Debug.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "ConnectResultEvent.hh"
#include "ShutdownConnectionEvent.hh"
#include "ConnectionStateEvent.hh"
#include "ConnectionManager.hh"
#include "Sock.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Initialise statics
unsigned int Connection::c_idCount = 1;
Mutex Connection::c_idCountMtx;


// Construction/destruction
Connection::Connection(Sock* sock, Tunneler* tunneler)
: m_connMtx(), m_tunneler(tunneler), m_closing(false),
  m_connected(false), m_connecting(false), m_everConnected(false), m_active(true),
  m_errored(false), m_blocked(false), m_firstConnectTime(0), m_inactiveTime(0),
  m_sock(sock), m_reader(0), m_writer(0), m_carriedTunnels(),
  m_connectFailedCount(0), m_destIP(), m_destPort(0), m_destProt(Enums::TCP),
  m_destSSL(false), m_id(0)
{
	// Set connection ID
	m_id = getNextConnectionID();
}

Connection::Connection(Tunneler* tunneler)
: m_connMtx(), m_tunneler(tunneler), m_closing(false),
  m_connected(false), m_connecting(false), m_everConnected(false), m_active(true),
  m_errored(false), m_blocked(false), m_firstConnectTime(0), m_inactiveTime(0),
  m_sock(0), m_reader(0), m_writer(0), m_carriedTunnels(), m_connectFailedCount(0),
  m_destIP(), m_destPort(0), m_destProt(Enums::TCP), m_destSSL(false), m_id(0)
{
	// Set connection ID
	m_id = getNextConnectionID();
}

Connection::~Connection()
{
	// Lock
	MutexVar mv(m_connMtx);

	// Ensure we don't have any remaining tunnels that think they own us
	setInactive();

	// Tidy up reader/writer & close socket
	close();
}


/*************/
/* Accessors */
/*************/


// Are we shutting down?
bool Connection::closing(void) const
{
	return m_closing;
}

// Are we connected?
bool Connection::connected(void) const
{
	return (m_connected && !m_closing && !m_errored);
}

// Are we in the process of conencting?
bool Connection::connecting(void) const
{
	return m_connecting;
}

// Have we ever been connected?
bool Connection::everConnected(void) const
{
	return m_everConnected;
}

// Are we not waiting to be shutdown?
bool Connection::active(void) const
{
	return m_active;
}

// Are we blocked?
bool Connection::blocked(void) const
{
	return m_blocked;
}

// Get time we first connected
time_t Connection::getFirstConnectTime(void) const
{
	return m_firstConnectTime;
}

// Get reference to controlling tunneler
Tunneler* Connection::getTunneler(void)
{
	return m_tunneler;
}

// Get the count of failed connect attempts
unsigned int Connection::getConnectFailedCount(void) const
{
	return m_connectFailedCount;
}

// Get connection ID
unsigned int Connection::getID(void) const
{
	return m_id;
}

// Get destination IP address
std::string Connection::getDestIP(void) const
{
	return m_destIP;
}

// Get destination port
unsigned int Connection::getDestPort(void) const
{
	return m_destPort;
}

// Get destination protocol
Enums::Protocol Connection::getDestProtocol(void) const
{
	return m_destProt;
}

// Get whether SSL is used to destination
bool Connection::getDestSSL(void) const
{
	return m_destSSL;
}

// Get socket
Sock* Connection::getSocket(void)
{
	return m_sock;
}


/************/
/* Mutators */
/************/


// Increment the count of failed connect attempts
void Connection::incConnectFailedCount(void)
{
	MutexVar tmv(m_connMtx);
	m_connectFailedCount++;
}


// We are waiting to be shutdown
void Connection::setInactive(void)
{
	MutexVar tmv(m_connMtx);
	m_active = false;
	m_inactiveTime = Time::now();

	// Let all carried tunnels know they are no longer carried by us
	while(m_carriedTunnels.size() > 0) {
		// Get & remove tunnel from list
		Tunnel* tun = m_carriedTunnels.front();
		m_carriedTunnels.erase(m_carriedTunnels.begin());

		// Let tunnel know
		tun->removeConnection(this);
	}
}


// Indicate we have suffered an error and shouldn't be used anymore
void Connection::setErrored(void)
{
	MutexVar tmv(m_connMtx);
	m_errored = true;

	// Do we have unprocessed data in our reader?
	if(m_reader && m_reader->dataRemaining()) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "Connection " << getID() << " has residual read data after error";
		DPRINTI(nfo.str());
		#endif

		// Let tunneler know we have unprocessed data
		m_tunneler->scheduleConnectionError(getID(), Tunnel::UNPROCESSED_DATA_ERROR);
	}
}

// Set whether we are blocked
void Connection::setBlocked(bool blocked)
{
	MutexVar tmv(m_connMtx);
	m_blocked = blocked;
}


/********************/
/* Public Interface */
/********************/


// Prepare for a connect attempt to be made to the server
void Connection::connectAttempt(std::string host, unsigned int port,
	Enums::Protocol prot, bool ssl) throw(ConnectError)
{
	// Flag we are trying to connect
	MutexVar mv(m_connMtx);
	m_connecting = true;
	m_destIP = host;
	m_destPort = port;
	m_destProt = prot;
	m_destSSL = ssl;

	// Clear up from any previous connection
	close();
}


// Let us know the result of a connection attempt
void Connection::connectResult(ConnectResultEvent::ConnectResultType result,
	Sock* sock) throw(CommsError)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Connection::connectResult() - connection " << getID();
	dbg << ", result: " << ConnectResultEvent::getTypeStr(result);
	DPRINTV(dbg.str());
	#endif

	// Let tunneler know connection attempt is over
	m_tunneler->scheduleEvent(new ConnectResultEvent(m_tunneler, getID(),
		getDestIP(), getDestPort(), getDestProtocol(), getDestSSL(),
		result));

	// No longer trying to connect
	MutexVar mv(m_connMtx);
	m_connecting = false;

	// Save socket result
	m_sock = sock;

	// Success?
	if(result == ConnectResultEvent::SUCCESS) {
		// Create required reader/writer
		createReaderWriter();

		// Let tunneler know we have connected
		m_tunneler->scheduleEvent(new ConnectionStateEvent(m_tunneler, getID(),
			CONNECTED));
	}
}


// Attempt to read from our connected socket
void Connection::read(void) throw(CommsError)
{
	//#ifdef DEBUG_TUN
	//std::stringstream dbg;
	//dbg << "Connection::read() - connection " << m_id;
	//DPRINTV(dbg.str());
	//#endif

	MutexVar mv(m_connMtx);

	// Can't read unless connected & active
	if(!connected())
		throw CommsError("Read attempt while not connected");

	// Instruct reader to do the read
	m_reader->read();
}

// Attempt to write to our connected socket
void Connection::write(MsgParms* parms) throw(CommsError)
{
	MutexVar mv(m_connMtx);

	// Can't write unless connected
	if(!connected())
		throw CommsError("Write attempt while not connected");

	// Insruct writer to do the write
	m_writer->write(parms);
}

// Attempt to flush any buffered data to our connected socket
void Connection::flush(bool retry) throw(CommsError)
{
	MutexVar mv(m_connMtx);

	// Can't write unless connected
	if(!connected())
		throw CommsError("Flush attempt while not connected");

	// Instruct writer to flush
	m_writer->flush(retry);
}

// Send some special out-of-band data
void Connection::writeOutOfBand(RawBuff& outOfBand) throw(CommsError)
{
	DPRINTV("Connection::writeOutOfBand()");

	MutexVar mv(m_connMtx);

	// Can't write unless connected
	if(!connected())
		throw CommsError("Out-of-band write attempt while not connected");

	// Insruct writer to do the write
	m_writer->writeOutOfBand(outOfBand);
}


// Add tunnel from list of carried tunnels
void Connection::addCarriedTunnel(Tunnel* tun)
{
	MutexVar mv(m_connMtx);

	// Check we don't already know
	if(std::find(m_carriedTunnels.begin(), m_carriedTunnels.end(), tun) ==
	m_carriedTunnels.end()) {
		// Record new carried tunnel
		m_carriedTunnels.push_back(tun);
	}
}

// Remove tunnel from list of carried tunnels
void Connection::removeCarriedTunnel(Tunnel* tun)
{
	MutexVar mv(m_connMtx);

	// Find tunnel
	std::vector<Tunnel*>::iterator tit;
	tit = std::find(m_carriedTunnels.begin(), m_carriedTunnels.end(), tun);
	if(tit != m_carriedTunnels.end()) {
		m_carriedTunnels.erase(tit);
	}

	// Do we still carry anything?
	if(m_carriedTunnels.size() == 0) {
		// Schedule our shutdown
		m_tunneler->scheduleEvent(new ShutdownConnectionEvent(m_id, m_tunneler));
	}
}


// Tell carried tunnels about some error
void Connection::tellTunnelsAboutError(Tunnel::TunnelErrorType errType)
{
	// Let all carried tunnels know
	MutexVar tmv(m_connMtx);
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_carriedTunnels.begin(); tit != m_carriedTunnels.end(); ++tit) {
		// Let tunnel know
		(*tit)->recordConnectionError(errType);
	}
}


// Tell tunnels we have connected
void Connection::tellTunnelsAboutConnect(void)
{
	// Let all carried tunnels know
	MutexVar tmv(m_connMtx);
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_carriedTunnels.begin(); tit != m_carriedTunnels.end(); ++tit) {
		(*tit)->connectionConnected(this);
	}
}


// Determine whether this connection can currently be shutdown cleanly
bool Connection::canBeShutdown(void) throw(CommsError)
{
	MutexVar mv(m_connMtx);

	bool actv = active();
	bool conntng = connecting();
	bool queued = false;
	bool delayed = false;

	// Must first be inactive & not trying to connect
	bool res = !actv && !conntng;

	// Check writer
	if(res && m_writer) {
		// Does our writer still have queued data to send?
		queued = !m_writer->flush(false);
		res &= !queued;
	}

	// Leave a small delay after becoming inactive
	if(res) {
		Time delayUntil = m_inactiveTime;
		delayUntil.addInterval(0, DELAY_CLOSE);
		delayed = !delayUntil.elapsed();
		res &= !delayed;
	}

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Connection::canBeShutdown() - connection " << getID();
	dbg << ": " << (res? "true":"false") << " ";
	if(!res) {
		if(actv) dbg << "Actv ";
		if(conntng) dbg << "Cntg ";
		if(queued) dbg << "Qud ";
		if(delayed) dbg << "Dld ";
	}
	DPRINTV(dbg.str());
	#endif

	return res;
}


/***********************/
/* Protected Interface */
/***********************/


// Replace reader
void Connection::replaceReader(Reader* reader)
{
	MutexVar mv(m_connMtx);

	// Free existing reader
	if(m_reader)
		delete m_reader;

	// Save new reader
	m_reader = reader;
}

// Replace writer
void Connection::replaceWriter(Writer* writer)
{
	MutexVar mv(m_connMtx);

	// Free existing writer
	if(m_writer)
		delete m_writer;

	// Save new writer
	m_writer = writer;
}


// We are now connected
void Connection::setConnected(void)
{
	MutexVar mv(m_connMtx);

	// Set connected flags etc.
	m_connected = true;
	m_everConnected = true;
	m_firstConnectTime = time(0);
}


/*********************/
/* Private Interface */
/*********************/


// Tidy up existing socket connection
void Connection::close(void)
{
	MutexVar mv(m_connMtx);

	// Indicate not connected
	m_connected = false;

	// Free any existing reader/writer
	if(m_reader) {
		delete m_reader;
		m_reader = 0;
	}
	if(m_writer) {
		delete m_writer;
		m_writer = 0;
	}

	// Close any existing socket
	if(m_sock) {
		delete m_sock;
		m_sock = 0;
	}
}


// Get next connection ID
unsigned int Connection::getNextConnectionID(void)
{
	unsigned int id = 0;

	MutexVar mv(c_idCountMtx);
	id = c_idCount++;
	if(c_idCount == 0)
		c_idCount = 1;
	mv.unlock();

	return id;
}


}	// namespace

