/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for handling reading/writing for one end of a tunnel
 * connection.
 *
 * $Id: Connection.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _CONNECTION_HH_
#define _CONNECTION_HH_

#include <sys/time.h>
#include <string>
#include <vector>
#include <functional>
#include <time.h>
#include "Mutex.hh"
#include "Exceptions.hh"
#include "Enums.hh"
#include "Tunnel.hh"
#include "ConnectResultEvent.hh"
#include "Time.hh"

namespace HTTPTunnel {

class Reader;
class Writer;
class Tunneler;
class MsgParms;
class Sock;


class Connection {
public:

	/*********/
	/* Types */
	/*********/

	// Connection state changes
	enum ConnectionState { CONNECTED, SOCKSSETUP };


	/***********/
	/* Methods */
	/***********/

	// Construction/destruction
	Connection(Sock* sock, Tunneler* tunneler);
	Connection(Tunneler* tunneler);
	virtual ~Connection();

	// Accessors
	bool closing(void) const;
	bool connected(void) const;
	bool connecting(void) const;
	bool everConnected(void) const;
	bool active(void) const;
	bool blocked(void) const;
	time_t getFirstConnectTime(void) const;
	Tunneler* getTunneler(void);
	unsigned int getConnectFailedCount(void) const;
	unsigned int getID(void) const;
	std::string getDestIP(void) const;
	unsigned int getDestPort(void) const;
	Enums::Protocol getDestProtocol(void) const;
	bool getDestSSL(void) const;
	Sock* getSocket(void);

	// Mutators
	void incConnectFailedCount(void);
	void setInactive(void);
	void setErrored(void);
	void setBlocked(bool blocked);


	/**************/
	/* Predicates */
	/**************/

	// Compare connection first connected times
	struct PredConnectTime :
	public std::binary_function<Connection*, Connection*, bool> {
		bool operator()(const Connection* const& c1, const Connection* const& c2)
		const { return (c1->getFirstConnectTime() < c2->getFirstConnectTime()); }
	};

	// Compare connection IDs
	struct PredID :
	public std::binary_function<Connection*, Connection*, bool> {
		bool operator()(const Connection* const& c1, const Connection* const& c2)
		const { return (c1->getID() < c2->getID()); }
	};

	// Determine whether a connection is connected
	struct PredConnected :
	public std::unary_function<Connection*, bool> {
		bool operator()(const Connection* const& c) const { return c->connected(); }
	};


	/*************/
	/* Utilities */
	/*************/

	// Prepare for a connect attempt to be made to the server
	virtual void connectAttempt(std::string host, unsigned int port,
		Enums::Protocol prot, bool ssl) throw(ConnectError);

	// Let us know the result of a connection attempt
	virtual void connectResult(ConnectResultEvent::ConnectResultType result,
		Sock* sock) throw(CommsError);

	// Attempt to read/write on our connected socket
	virtual void read(void) throw(CommsError);
	virtual void write(MsgParms* parms) throw(CommsError);
	virtual void flush(bool retry) throw(CommsError);

	// Send some special out-of-band data
	void writeOutOfBand(RawBuff& outOfBand) throw(CommsError);

	// Add/remove tunnel from list of carried tunnels
	void addCarriedTunnel(Tunnel* tun);
	void removeCarriedTunnel(Tunnel* tun);

	// Tell carried tunnels about something
	void tellTunnelsAboutError(Tunnel::TunnelErrorType errType);
	void tellTunnelsAboutConnect(void);

	// Determine whether this connection can currently be shutdown cleanly
	bool canBeShutdown(void) throw(CommsError);


protected:

	/***********/
	/* Methods */
	/***********/

	// Construct reader/writer
	virtual void createReaderWriter(void) = 0;

	// Replace reader/writer
	void replaceReader(Reader* reader);
	void replaceWriter(Writer* writer);

	// We are now connected
	void setConnected(void);


	/********/
	/* Data */
	/********/

	// A mutex
	Mutex									m_connMtx;

	// Our controlling tunneler
	Tunneler*							m_tunneler;

	// Connection state flags
	bool									m_closing;
	bool									m_connected;
	bool									m_connecting;
	bool									m_everConnected;
	bool									m_active;
	bool									m_errored;
	bool									m_blocked;

	// Time we first connected
	time_t								m_firstConnectTime;

	// Time we are made inactive
	Time									m_inactiveTime;

	// Socket we read/write with
	Sock*									m_sock;

	// Reader for processing read data
	Reader*								m_reader;

	// Writer for processing and writing data
	Writer*								m_writer;

	// A list of tunnels that we carry
	std::vector<Tunnel*>				m_carriedTunnels;


private:

	/***********/
	/* Methods */
	/***********/

	// Tidy up existing socket connection
	void close(void);

	// Get next connection ID
	static unsigned int getNextConnectionID(void);


	/*************/
	/* Constants */
	/*************/

	// Delay before allowing connection to be closed
	static const time_t				DELAY_CLOSE				= 1000;


	/********/
	/* Data */
	/********/

	// Count of various errors
	unsigned int						m_connectFailedCount;

	// Our destination address
	std::string							m_destIP;
	unsigned int						m_destPort;
	Enums::Protocol					m_destProt;
	bool									m_destSSL;

	// Unique connection ID
	static unsigned int				c_idCount;
	static Mutex						c_idCountMtx;
	unsigned int						m_id;

};


}	// namespace

#endif

