/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Manage all blocking Connection operations without requiring a
 * thread per Connection object.
 *
 * $Id: ConnectionManager.cc,v 1.4 2009-04-24 15:37:33 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <poll.h>
#include <errno.h>
#include <sstream>
#include "ConnectionManager.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Connection.hh"
#include "Tunneler.hh"
#include "Tunnel.hh"
#include "ReadEvent.hh"
#include "ConnectProxy.hh"
#include "Time.hh"
#include "Sock.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ConnectionManager::ConnectionManager()
: EventReceiver(), m_mtx(), m_connections(), m_connectProxies(), m_threadMgr(),
  m_pollThread(0), m_shutDown(false)
{
	DPRINTV("ConnectionManager::ConnectionManager()");

	// Launch connection polling thread
	MutexVar mv(m_mtx);
	ZeroArgClassCallback<ConnectionManager> cb(&ConnectionManager::pollThread, this);
	m_pollThread = m_threadMgr.newThread(cb);
}

ConnectionManager::~ConnectionManager()
{
	DPRINTV("ConnectionManager::~ConnectionManager()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();

	// Take mutexes
	MutexVar mv(m_mtx);

	// Signal we are shutting down
	m_shutDown = true;

	// Shutdown threads
	mv.unlock();
	m_threadMgr.shutdown();
	mv.lock();
	m_pollThread = 0;

	// Remove any remaining managed connections
	while(m_connections.size() > 0) {
		std::map<unsigned int, Connection*>::iterator cit = m_connections.begin();
		delConnection(cit->second);
	}
}


/*************/
/* Utilities */
/*************/


// Add a new connection to the managed list
void ConnectionManager::addConnection(Connection* conn)
{
	MutexVar mv(m_mtx);

	// Get connection ID
	unsigned int connID = conn->getID();

	#ifdef DEBUG_TUN
	// Check we don't already know about this connection
	std::map<unsigned int, Connection*>::iterator cit;
	cit = m_connections.find(connID);
	if(cit != m_connections.end()) {
		std::stringstream err;
		err << "ConnectionManager::addConnection() - ";
		err << "Already managing connection ID " << connID;
		DPRINTE(err.str());
	}
	#endif

	// Add connection to managed list
	m_connections[connID] = conn;
}

// Remove a connection from the managed list
void ConnectionManager::delConnection(Connection* conn)
{
	MutexVar mv(m_mtx);

	// Find connection
	unsigned int connID = conn->getID();
	std::map<unsigned int, Connection*>::iterator cit;
	cit = m_connections.find(connID);
	if(cit != m_connections.end()) {
		// Do we have an active connect proxy for this connection?
		std::map<unsigned int, ConnectProxy*>::iterator cpit;
		cpit = m_connectProxies.find(connID);
		if(cpit != m_connectProxies.end()) {
			// Stop connect proxy
			ConnectProxy* proxy = cpit->second;
			delete proxy;

			// Remove connect proxy from list
			m_connectProxies.erase(cpit);
		}

		// Remove connection from managed list
		m_connections.erase(cit);
	}
	else {
		std::stringstream wrn;
		wrn << "ConnectionManager::delConnection() - ";
		wrn << "Unknown connection ID " << connID;
		g_pLogger->warn(wrn.str());
	}
}


// Setup a proxy to connect out for a connection
void ConnectionManager::addProxyConnect(Connection* conn, const Time& time,
	std::string host, unsigned int port, Enums::Protocol prot, bool ssl)
{
	// Get connection ID
	unsigned int connID = conn->getID();

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "ConnectionManager::addProxyConnect() - For connection " << connID;
	DPRINTV(dbg.str());
	#endif

	MutexVar mv(m_mtx);

	#ifdef DEBUG_TUN
	// Check we actually know about this connection
	std::map<unsigned int, Connection*>::iterator cit;
	cit = m_connections.find(connID);
	if(cit == m_connections.end()) {
		DPRINTE("Connection not found in our managed list");
	}
	#endif

	// Check whether we already have a proxy for this connection
	std::map<unsigned int, ConnectProxy*>::iterator cpit;
	cpit = m_connectProxies.find(connID);
	if(cpit != m_connectProxies.end()) {
		// Reuse this one
		ConnectProxy* proxy = cpit->second;
		proxy->scheduleConnect(time);
	}
	else {
		// Create connect proxy
		ConnectProxy* proxy = new ConnectProxy(conn, time, host, port, prot, ssl);
		m_connectProxies[connID] = proxy;
	}
}

// Clear up after a completed proxy connect
void ConnectionManager::delProxyConnect(Connection* conn)
{
	// Get connection ID
	unsigned int connID = conn->getID();

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "ConnectionManager::delProxyConnect() - For connection " << connID;
	DPRINTV(dbg.str());
	#endif

	MutexVar mv(m_mtx);

	// Lookup proxy
	std::map<unsigned int, ConnectProxy*>::iterator cpit;
	cpit = m_connectProxies.find(connID);
	ConnectProxy* proxy = cpit->second;

	#ifdef DEBUG_TUN
	// Check we actually know about this connection
	std::map<unsigned int, Connection*>::iterator cit;
	cit = m_connections.find(connID);
	if(cit == m_connections.end()) {
		DPRINTE("Connection not found in our managed list");
	}
	// Check we know about proxy for this connection
	if(cpit == m_connectProxies.end()) {
		DPRINTE("Unknown proxy for this connection");
	}
	#endif

	// Clear up proxy
	m_connectProxies.erase(cpit);
	delete proxy;
}


// Display some audit stats
void ConnectionManager::audit(void)
{
	DPRINTV("ConnectionManager::audit()");

	std::stringstream nfo;
	MutexVar mv(m_mtx);

	// Number of managed connections
	nfo.str("");
	nfo << "Number of managed connections: " << m_connections.size();
	DPRINTV(nfo.str());

	// Number of active connect proxies
	nfo.str("");
	nfo << "Number of active connect proxies: " << m_connectProxies.size();
	DPRINTV(nfo.str());
}


/**********/
/* Events */
/**********/


// Try reading from our managed connections
void ConnectionManager::readConnections(void)
{
	MutexVar mv(m_mtx);

	// How many connected & active Connections should we poll for?
	std::map<unsigned int, Connection*>::iterator cit;
	for(cit = m_connections.begin(); cit != m_connections.end(); ++cit) {
		// Worth checking this connection?
		Connection* conn = cit->second;
		if(conn->connected() && conn->active() && !conn->blocked()) {
			try {
				// Get this connections socket
				Sock* sock = conn->getSocket();

				// Read everything we can from this socket
				bool ready = false, pollError = false, pollHup = false;
				do {
					try {
						// Poll socket
						if(sock) {
							ready = sock->pollRead(0, pollError, pollHup);
						}
					}
					catch(SockError& e) {
						std::stringstream err;
						err << "Error polling socket for connection id " << conn->getID();
						throw CommsError(err.str());
					}

					if(ready) {
						// Descriptor ready for reading
						try {
							conn->read();
						}
						catch(NotReadyError& e) {
							// Socket wasn't actually ready to read
							g_pLogger->info("Socket not really ready for reading: " + e.what());
							ready = false;
						}

						// Was there also an error condition?
						if(pollError) {
							if(pollHup) {
								std::stringstream err;
								err << "Poll returned HUP for socket read for connection id " << conn->getID();
								throw ConnectionClosed(err.str());
							}
							else {
								std::stringstream err;
								err << "Poll returned error for socket read for connection id " << conn->getID();
								throw CommsError(err.str());
							}
						}
					}
				} while(ready);
			}
			catch(BadResponseCodeError& e) {
				// Something returned an unexpected HTTP error response code
				g_pLogger->info("Bad HTTP response code: " + e.what());
				conn->setErrored();
				if(!conn->closing()) {
					// Schedule error reporting event
					Tunneler* tunneler = conn->getTunneler();
					tunneler->scheduleConnectionError(conn->getID(), Tunnel::BAD_RESPONSE_ERROR);
				}
			}
			catch(ConnectionClosed& e) {
				g_pLogger->info("Connection closed: " + e.what());
				conn->setErrored();
				if(!conn->closing() && conn->active()) {
					// Schedule error reporting event
					Tunneler* tunneler = conn->getTunneler();
					tunneler->scheduleConnectionError(conn->getID(), Tunnel::CONNECTION_CLOSED_ERROR);
				}
			}
			catch(CommsError& e) {
				// A terminal error while reading
				g_pLogger->info("Comms error: " + e.what());
				conn->setErrored();
				if(!conn->closing() && conn->active()) {
					// Schedule error reporting event on connections controlling tunneler
					Tunneler* tunneler = conn->getTunneler();
					tunneler->scheduleConnectionError(conn->getID(), Tunnel::GENERAL_COMMS_ERROR);
				}
			}
		}
	}
}


/*********************/
/* Private Interface */
/*********************/


// Connection polling thread
void ConnectionManager::pollThread(void)
{
	MutexVar mv(m_mtx);

	while(!m_shutDown) {
		// Look for connections with active sockets
		if(m_connections.size() > 0 && !isScheduled<ReadEvent>()) {
			std::vector<int> sockFDs;
			std::map<unsigned int, Connection*>::iterator cit;
			for(cit = m_connections.begin(); cit != m_connections.end(); ++cit) {
				Connection* conn = cit->second;
				if(conn->connected() && conn->active() && !conn->blocked()) {
					Sock* sock = conn->getSocket();
					if(sock)
						sockFDs.push_back(sock->getFD());
				}
			}

			// Build list of socket descriptors to poll
			unsigned int noSocks = sockFDs.size();
			if(noSocks > 0) {
				struct pollfd pollStruct[noSocks];
				for(unsigned int i = 0; i < noSocks; ++i) {
					pollStruct[i].fd = sockFDs[i];;
					pollStruct[i].events = POLLIN;
					pollStruct[i].revents = 0;
				}

				// Poll sockets
				int readyfds = poll(pollStruct, noSocks, 0);
				if(readyfds > 0) {
					// Events pending
					scheduleEvent(new ReadEvent(0, this));
				}
				else if(readyfds < 0) {
					// Poll error
					g_pLogger->error("ConnectionManager::pollThread() - Poll error");
				}
			}
		}

		// Sleep
		Time sleepTime(0, POLL_PERIOD);
		m_pollThread->sleep(mv, sleepTime.getTimeVal());
	}

	DPRINTV("ConnectionManager::pollThread() - Returning");
}


}	// namespace

