/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * An object capable of receiving and executing events.
 *
 * $Id: EventReceiverT.cc,v 1.2 2008-11-22 20:44:00 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "Event.hh"
#include "TimedEvent.hh"


namespace HTTPTunnel {


/********************/
/* Public Interface */
/********************/


// Reschedule an event
template<class T, class I>
void EventReceiver::rescheduleEvent(I identity, const Time& time)
{
	MutexVar emv(m_eventsMtx);

	// Look for matching event
	std::vector<Event*>::iterator eit = findEvent<T, I>(identity);
	if(eit != m_events.end()) {
		// Cancel this event
		Event* event = *eit;
		m_events.erase(eit);

		// Reschedule it for new time
		event->setRunTime(time);
		scheduleEvent(event);
	}
}


// Cancel events of some type
template<class T, class I>
void EventReceiver::cancelEvents(I identity)
{
	bool keepLooking = false;

	// Look through plain event queue
	MutexVar emv(m_eventsMtx);
	do {
		keepLooking = false;

		// Look for plain event
		std::vector<Event*>::iterator eit = findEvent<T, I>(identity);
		if(eit != m_events.end()) {
			// Cancel this event
			Event* event = *eit;
			m_events.erase(eit);
			delete event;

			// Look for another matching event
			keepLooking = true;
		}
	} while(keepLooking);
}


// Cancel all events of some type
template<class T>
void EventReceiver::cancelAllEvents(void)
{
	bool keepLooking = false;

	// Look through plain event queue
	MutexVar emv(m_eventsMtx);
	do {
		keepLooking = false;

		// Look for plain event
		std::vector<Event*>::iterator eit = findEvent<T>();
		if(eit != m_events.end()) {
			// Cancel this event
			Event* event = *eit;
			m_events.erase(eit);
			delete event;

			// Look for another matching event
			keepLooking = true;
		}
	} while(keepLooking);
}


// Cancel and return an event of some type
template<class T, class I>
T* EventReceiver::cancelAndGetEvent(I identity)
{
	T* event = 0;

	// Look through plain event queue
	MutexVar emv(m_eventsMtx);
	std::vector<Event*>::iterator eit = findEvent<T, I>(identity);
	if(eit != m_events.end()) {
		// Cancel and return this event
		event = dynamic_cast<T*>(*eit);
		m_events.erase(eit);
	}

	return event;
}


// Return whether some event is currently scheduled
template<class T>
bool EventReceiver::isScheduled(void)
{
	bool res = false;

	// Iterate over all queued events
	MutexVar emv(m_eventsMtx);
	std::vector<Event*>::iterator eit;
	for(eit = m_events.begin(); eit != m_events.end(); ++eit) {
		// Event of the right type?
		T* event = dynamic_cast<T*>(*eit);
		if(event) {
			res = true;
			break;
		}
	}

	return res;
}

// Return whether some identified event is currently scheduled
template<class T, class I>
bool EventReceiver::isScheduled(I identity)
{
	// Look for plain event
	MutexVar emv(m_eventsMtx);
	bool res = (findEvent<T, I>(identity) != m_events.end());
	return res;
}



/*********************/
/* Private Interface */
/*********************/


// Find event of a particular type matching some identity - Caller holds mutex
template<class T, class I>
std::vector<Event*>::iterator EventReceiver::findEvent(I identity)
{
	// Iterate over all queued events
	std::vector<Event*>::iterator eit;
	for(eit = m_events.begin(); eit != m_events.end(); ++eit) {
		// Event of the right type?
		T* event = dynamic_cast<T*>(*eit);
		if(event) {
			// Test identity
			if(event->identify(identity))
				break;
		}
	}

	return eit;
}

// Find any event of a particular type - Caller holds mutex
template<class T>
std::vector<Event*>::iterator EventReceiver::findEvent(void)
{
	// Iterate over all queued events
	std::vector<Event*>::iterator eit;
	for(eit = m_events.begin(); eit != m_events.end(); ++eit) {
		// Event of the right type?
		T* event = dynamic_cast<T*>(*eit);
		if(event) {
			break;
		}
	}

	return eit;
}


}	// namespace

