/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Exceptions thrown by the application.
 *
 * $Id: Exceptions.hh,v 1.4 2009-02-13 18:14:13 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _EXCEPTIONS_HH_
#define _EXCEPTIONS_HH_

#include <exception>
#include <vector>
#include <string>

namespace HTTPTunnel {


/**************/
/* Base class */
/**************/

class TunnelError : public std::exception {
public:
	TunnelError(std::string e) : err(e) {}
	virtual ~TunnelError() throw() {}
	virtual std::string what(void) { return err; }
protected:
	std::string err;
};


/************************/
/* Configuration errors */
/************************/

class ConfigError : public TunnelError {
public: ConfigError(std::string e) : TunnelError(e) {}
};


/*************************/
/* Communications errors */
/*************************/

class CommsError : public TunnelError {
public: CommsError(std::string e) : TunnelError(e) {}
};

// Server error
class ServerError : public CommsError {
public: ServerError(std::string e) : CommsError(e) {}
};

// General data reading error
class ReadError : public CommsError {
public: ReadError(std::string e) : CommsError(e) {}
};

// Connection closed
class ConnectionClosed : public ReadError {
public: ConnectionClosed(std::string e) : ReadError(e) {}
};

// Received some error response code
class BadResponseCodeError : public ReadError {
public: BadResponseCodeError(std::string e, unsigned short respCode)
	: ReadError(e), m_respCode(respCode) {}
	unsigned short m_respCode;
};

// General data writing error
class WriteError : public CommsError {
public: WriteError(std::string e) : CommsError(e) {}
};

// Connect failed error
class ConnectError : public CommsError {
public: ConnectError(std::string e) : CommsError(e) {}
};

// Firewalled connect failed error
class FirewalledConnectError : public ConnectError {
public: FirewalledConnectError(std::string e) : ConnectError(e) {}
};

// Asynchronous connect result
class AsynchronousConnect : public ConnectError {
public: AsynchronousConnect(std::string e) : ConnectError(e) {}
};

// Resending error
class ResendError : public CommsError {
public: ResendError(std::string e) : CommsError(e) {}
};

// Not ready for read/write error
class NotReadyError : public CommsError {
public: NotReadyError(std::string e) : CommsError(e) {}
};


/*****************/
/* Socket errors */
/*****************/

class SockError : public TunnelError {
public: SockError(std::string e, int en) : TunnelError(e), m_errno(en) {}
	int m_errno;
};


/*********************/
/* Processing errors */
/*********************/

class ProcessingError : public TunnelError {
public: ProcessingError(std::string e) : TunnelError(e) {}
};

// Invalid ID
class InvalidID : public ProcessingError {
public: InvalidID(std::string e) : ProcessingError(e) {}
};

// Invalid tunnel ID string
class InvalidTunnelID : public InvalidID {
public: InvalidTunnelID(std::string e) : InvalidID(e) {}
};

// Invalid tunneler ID string
class InvalidTunnelerID : public InvalidID {
public: InvalidTunnelerID(std::string e) : InvalidID(e) {}
};

// No data for the requested operation
class NoDataError : public ProcessingError {
public: NoDataError(std::string e) : ProcessingError(e) {}
};

// Authentication processing
class AuthProcessingError : public ProcessingError {
public: AuthProcessingError(std::string e) : ProcessingError(e) {}
};


/******************/
/* Parsing errors */
/******************/

class ParsingError : public TunnelError {
public: ParsingError(std::string e) : TunnelError(e) {}
};

// Error constructing HTTP message from raw data stream
class HTTPMsgParsingError : public ParsingError {
public:
	HTTPMsgParsingError(std::string e, unsigned int line)
	: ParsingError(e), m_line(line) {}
	unsigned int m_line;
};

// Wrong message type parsing error
class HTTPMsgWrongMsgType : public HTTPMsgParsingError {
public:
	HTTPMsgWrongMsgType(std::string e, unsigned int line)
	: HTTPMsgParsingError(e, line) {}
};

// Error in message payload content
class ContentParsingError : public HTTPMsgParsingError {
public:
	ContentParsingError(std::string e, unsigned int line)
	: HTTPMsgParsingError(e, line) {}
};

// Not enough data in buffer to complete parsing
class NotEnoughDataParsingError : public ContentParsingError {
public:
	NotEnoughDataParsingError(std::string e, unsigned int reqd, unsigned int recvd,
		unsigned int line)
	: ContentParsingError(e, line), m_reqd(reqd), m_recvd(recvd) {}
	unsigned int m_reqd, m_recvd;
};

// Reported encoded content length doesn't equal received encoded length
class BadContentLengthParsingError : public ContentParsingError {
public:
	BadContentLengthParsingError(std::string e, unsigned int line)
	: ContentParsingError(e, line) {}
};


/*********************/
/* Management errors */
/*********************/

class ManagementError : public TunnelError {
public: ManagementError(std::string e) : TunnelError(e) {}
};

// Failed to find some managed resource
class LookupManagementError : public ManagementError {
public: LookupManagementError(std::string e) : ManagementError(e) {}
};


/*****************/
/* Memory errors */
/*****************/

class MemoryError : public TunnelError {
public: MemoryError(std::string e) : TunnelError(e) {}
};


/****************/
/* Event errors */
/****************/

class EventError : public TunnelError {
public: EventError(std::string e) : TunnelError(e) {}
};


/********************************/
/* Server authentication errors */
/********************************/

class AuthError : public TunnelError {
public: AuthError(std::string e) : TunnelError(e) {}
};

// Error in accessing an authentication database
class DBServerAuthError : public AuthError {
public: DBServerAuthError(std::string e) : AuthError(e) {}
};


/******************/
/* General errors */
/******************/

class GeneralError : public TunnelError {
public: GeneralError(std::string e) : TunnelError(e) {}
};

// Host address info lookup error
class AddrLookupError : public GeneralError {
public: AddrLookupError(std::string e) : GeneralError(e) {}
};


/*****************/
/* Thread errors */
/*****************/

class ThreadError : public TunnelError {
public: ThreadError(std::string e) : TunnelError(e) {}
};

// Thread not found
class ThreadNotFound : public ThreadError {
public: ThreadNotFound(std::string e) : ThreadError(e) {}
};

// Mutex error
class MutexError : public ThreadError {
public: MutexError(std::string e) : ThreadError(e) {}
};

// Condition variable error
class ConditionVarError : public ThreadError {
public: ConditionVarError(std::string e) : ThreadError(e) {}
};

// Thread specific data error
class ThreadDataError : public ThreadError {
public: ThreadDataError(std::string e) : ThreadError(e) {}
};


}		// namespace HTTPTunnel

#endif

