/* (C) 2008 Michael Weiser.
 * (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Authenticate a user at the server against a MySQL database.
 *
 * $Id: FileServerAuthenticator.cc,v 1.2 2008-11-14 17:59:19 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "FileServerAuthenticator.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "Tunnel.hh"
#include "ServerAuthSettings.hh"
#include "Tunneler.hh"
#include "ServerSideConfig.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
FileServerAuthenticator::FileServerAuthenticator() throw(AuthError)
: ServerAuthenticator()
{
	DPRINTV("FileServerAuthenticator::FileServerAuthenticator()");

	// Configure to use digest authentication
	getDefaultAuthSettings()->setServerSideAuth(AuthSettings::AUTH_DIGEST, g_pConfig->getRealm());

	// Create initial user
	std::string initialUser = m_defaultAuthSettings->getUser();
	m_users[initialUser] = createUser(initialUser, true);
}

FileServerAuthenticator::~FileServerAuthenticator()
{
	DPRINTV("FileServerAuthenticator::~FileServerAuthenticator()");
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get copy of current user details
User FileServerAuthenticator::getUserDetails(std::string name)
{
	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		return *user;
	}

	// Return blank unauthenticated user details
	User blankUser;
	blankUser.setAuthenticated(false, 0);
	return blankUser;
}


/*************/
/* Utilities */
/*************/


// Authenticate a new user
void FileServerAuthenticator::authenticate(std::string userName, Tunnel* tun,
	const std::vector<std::string>& authArgs)
{
	DPRINTV("FileServerAuthenticator::authenticate()");

	MutexVar mv(m_mtx);

	// Don't bother for default user name
	if(userName != m_defaultUserName) {
		// Let base class cache this user name
		ServerAuthenticator::authenticate(userName, tun, authArgs);

		// Check user authentication
		ServerSideConfig *ssc = tun->getTunneler()->getServerSideConfigMutable();
		std::string passwd = ssc->getPasswd(userName);
		ServerAuthSettings* settings = tun->getServerAuthSettings();
		if(checkDigestAuthResponse(userName, passwd, settings, authArgs)) {
			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "User '" << userName << "' authenticated against config file";
			DPRINTV(nfo.str());
			#endif

			User* userDetails = getUser(userName);
			Quotas q(ssc->getBandwidth(userName), ssc->getNoTunnels(userName));
			userDetails->setAll(userName, passwd, true, q, tun);
		}
		else {
			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "User '" << userName << "' failed authentication against config file";
			DPRINTV(nfo.str());
			#endif
		}
	}
}


}	// namespace

