/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Apply firewalling rules for the tunnel server.
 *
 * $Id: Firewall.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include "Firewall.hh"
#include "Utils.hh"
#include "Logger.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


Firewall::Firewall(std::string ips, std::string ports)
: m_mtx(), m_denyIPs(), m_denyWildcardIPs(), m_denyPorts()
{
	// Add initial blocked IPs/Ports
	if(ips.size() > 0)
		denyIPs(ips);
	if(ports.size() > 0)
		denyPorts(ports);
}

Firewall::~Firewall()
{
}


/********************/
/* Public Interface */
/********************/


// Is a requested tunnel destination allowed?
bool Firewall::isAllowed(std::string dest)
{
	bool allowed = true;

	// Split address into IP and port
	std::vector<std::string> parms;
	Utils::SplitString(dest, ":", parms);
	if(parms.size() == 2) {
		std::string ip = parms[0];
		unsigned int port = atoi(parms[1].c_str());
		allowed = isAllowed(ip, port);
	}
	else {
		// Deny badly formatted destination
		allowed = false;
	}

	return allowed;
}

bool Firewall::isAllowed(std::string ip, unsigned int port)
{
	bool allowed = true;

	// Port banned?
	if(portBanned(port)) {
		allowed = false;
		std::stringstream nfo;
		nfo << "Destination port " << port << " in denied list";
		g_pLogger->info(nfo.str());
	}
	else if(ipBanned(ip)) {
		allowed = false;
		std::stringstream nfo;
		nfo << "Destination IP address " << ip << " in denied list";
		g_pLogger->info(nfo.str());
	}

	return allowed;
}


// Add new banned IPs
void Firewall::denyIPs(std::string dest)
{
	// Split list
	std::vector<std::string> ips;
	Utils::SplitString(dest, ",", ips);

	MutexVar mv(m_mtx);

	for(unsigned int i = 0; i < ips.size(); ++i) {
		// Basic or wildcarded address?
		std::string ip = ips[i];
		if(ip[ip.size() - 1] == '*') {
			// Trim *
			ip = ip.substr(0, ip.size() - 1);

			// Ignore duplicates
			std::vector<std::string>::iterator ipit;
			ipit = std::find(m_denyWildcardIPs.begin(), m_denyWildcardIPs.end(), ip);
			if(ipit == m_denyWildcardIPs.end())
				m_denyWildcardIPs.push_back(ip);
		}
		else {
			// Resolve IP address
			ip = resolve(ip);
			if(ip.size() > 0) {
				// Ignore duplicates
				std::list<std::string>::iterator ipit;
				ipit = std::find(m_denyIPs.begin(), m_denyIPs.end(), ip);
				if(ipit == m_denyIPs.end())
					m_denyIPs.push_back(ip);
			}
		}
	}

	// Maintain sorted list
	m_denyIPs.sort();
}

// Add new banned ports
void Firewall::denyPorts(std::string dest)
{
	// Split list
	std::vector<std::string> ports;
	Utils::SplitString(dest, ",", ports);

	MutexVar mv(m_mtx);

	for(unsigned int i = 0; i < ports.size(); ++i) {
		unsigned int port = atoi(ports[i].c_str());
		if(port > 0 && port <= 65535) {
			// Ignore duplicates
			std::list<unsigned int>::iterator pit;
			pit = std::find(m_denyPorts.begin(), m_denyPorts.end(), port);
			if(pit == m_denyPorts.end())
				m_denyPorts.push_back(port);
		}
	}

	// Maintain sorted list
	m_denyPorts.sort();
}


// Serialize to a file
void Firewall::serialize(std::ofstream& file)
{
	MutexVar mv(m_mtx);

	// Firewalled IPs
	if(m_denyIPs.size() > 0 || m_denyWildcardIPs.size() > 0) {
		file << "denyips ";
		bool first = true;
		std::list<std::string>::iterator ipit;
		for(ipit = m_denyIPs.begin(); ipit != m_denyIPs.end(); ++ipit) {
			if(!first)
				file << ",";
			else
				first = true;
			file << *ipit;
		}
		std::vector<std::string>::iterator wcit;
		for(wcit = m_denyWildcardIPs.begin(); wcit != m_denyWildcardIPs.end(); ++wcit) {
			if(!first)
				file << ",";
			else
				first = true;
			file << *wcit << "*";
		}
		file << std::endl;
	}

	// Firewalled ports
	if(m_denyPorts.size() > 0) {
		file << "denyports ";
		bool first = true;
		std::list<unsigned int>::iterator pit;
		for(pit = m_denyPorts.begin(); pit != m_denyPorts.end(); ++pit) {
			if(!first)
				file << ",";
			else
				first = true;
			file << *pit;
		}
		file << std::endl;
	}
}


/*********************/
/* Private Interface */
/*********************/


// Is an IP address banned
bool Firewall::ipBanned(std::string ip)
{
	MutexVar mv(m_mtx);

	bool banned = false;

	// Resolve to numeric address
	ip = resolve(ip);
	if(ip.size() > 0) {
		// First check simple list
		if(std::binary_search(m_denyIPs.begin(), m_denyIPs.end(), ip)) {
			banned = true;
		}
		else {
			// Check wildcard list
			std::vector<std::string>::iterator ipit;
			for(ipit = m_denyWildcardIPs.begin(); ipit != m_denyWildcardIPs.end(); ++ipit) {
				std::string wcip = *ipit;
				ip = ip.substr(0, wcip.size());
				if(ip == wcip)
					banned = true;
			}
		}
	}
	else {
		banned = true;
	}

	return banned;
}


// Is a port banned
bool Firewall::portBanned(unsigned int port)
{
	MutexVar mv(m_mtx);
	return std::binary_search(m_denyPorts.begin(), m_denyPorts.end(), port);
}


// Resolve an address to a string IP address
std::string Firewall::resolve(std::string address)
{
	std::string ip = Utils::resolveIP(address);
	if(ip.size() == 0) {
		g_pLogger->warn("Firewall failed to resolve address " + address);
	}
	return ip;
}


}	// namespace

