/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration.
 *
 * $Id: GlobalConfig.cc,v 1.9 2009-03-20 11:52:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <unistd.h>
#include <inttypes.h>
#include "Debug.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"
#include "SockTunClientConfigBlob.hh"
#include "SockTunServerConfigBlob.hh"
#include "ProxySettings.hh"
#include "ClientServer.hh"
#include "ServerServer.hh"
#include "MutexVar.hh"
#include "Tunneler.hh"


namespace HTTPTunnel {


// Construction/destruction
GlobalConfig::GlobalConfig(std::string procName)
: m_mtx(), m_procName(procName), m_daemon(true), m_testing(false), m_hostName(),
  m_trueHostName(), m_debugLevel(ERROR), m_pid(getpid()), m_pidFile(),
  m_realm("FireDrill"), m_servers(), m_serverConfigs(), m_dataLogStem("/tmp/Lin-"),
  m_bigEndian(false), m_version()
{
	// Create pseudo unique local hostname
	std::stringstream hostnamestr;
	char hostname[OUR_HOST_NAME_MAX + 1];
	if(gethostname(hostname, sizeof(hostname)) == 0) {
		m_trueHostName = hostname;
		hostnamestr << hostname;
	}

	// Try to ensure pseudeo-uniqueness by appending some random number
	srand(time(0));
	int randNum = (int)(((float)rand() / RAND_MAX) * 65535);
	hostnamestr << std::hex << std::setw(4) << std::setfill('0') << randNum;
	m_hostName = hostnamestr.str();

	// Determine endianness of platform
	union { char array[4]; uint32_t chars; } testUnion;
	char c = 'a';
	for(unsigned int x = 0; x < 4; x++) {
		testUnion.array[x] = c++;
	}
	if(testUnion.chars == 0x61626364) {
		m_bigEndian = true;
	}

	// Create version identifier
	std::stringstream ver;
	ver << LIBVERSION << " " << PLATNAME;
	m_version = ver.str();
}

GlobalConfig::~GlobalConfig()
{
	MutexVar mv(m_mtx);

	// Stop local servers
	destroyLocalServers();
}


/*************/
/* Utilities */
/*************/


// Serialize to a config file
void GlobalConfig::dumpToFile(std::string name) throw(ConfigError)
{
	MutexVar mv(m_mtx);

	// Open file for writing
	std::ofstream file(name.c_str());
	if(file) {
		/* Header */

		file << "##########################" << std::endl;
		file << "## " << m_procName << " config" << std::endl;
		file << "## Version " << m_version << std::endl;
		file << "##########################" << std::endl;

		/* Global options */

		file << std::endl;
		file << "#################" << std::endl;
		file << "## Global options" << std::endl;
		file << "#################" << std::endl;
		file << std::endl;
		file << "[GLOBAL]" << std::endl;

		// Daemon?
		if(!getDaemon()) {
			file << "nodaemon" << std::endl;
		}

		// Testing?
		if(getTesting()) {
			file << "dotest" << std::endl;
		}

		// Debug level
		file << "debug " << getDebug() << std::endl;

		// PID file
		std::string pidfile = getPIDFile();
		if(pidfile.size() > 0) {
			file << "pidfile " << pidfile << std::endl;
		}

		/* Servers config */

		file << std::endl;

		// Get config for all servers
		std::vector<SideConfig*>::const_iterator cit;
		for(cit = m_serverConfigs.begin(); cit != m_serverConfigs.end(); ++cit) {
			// Serialize config from this server
			(*cit)->serialize(file);
		}
	}
	else {
		std::stringstream err;
		err << "Failed to open config file " << name << " for writing";
		throw ConfigError(err.str());
	}
}


/**********************/
/* Accessors/Mutators */
/**********************/


// Configure as a server
#ifdef AUTH_FILE
void GlobalConfig::configureServer(bool ssl, bool sockTun, std::string cert,
	std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth, std::map<std::string, std::string> &passwds,
	std::map<std::string, unsigned int> &tunnelnos, std::map<std::string,
	unsigned int> &bandwidths) throw(AuthError)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ServerSideConfig(ssl, sockTun, cert, certKey, lhost,
		lport, firewallIPs, firewallPorts, auth, passwds, tunnelnos, bandwidths));
}
#else
void GlobalConfig::configureServer(bool ssl, bool sockTun, std::string cert,
	std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth) throw(AuthError)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ServerSideConfig(ssl, sockTun, cert, certKey, lhost,
		lport, firewallIPs, firewallPorts, auth));
}
#endif


// Reconfigure a server
#ifdef AUTH_FILE
void GlobalConfig::reconfigureServer(unsigned int no, bool ssl, bool sockTun,
	std::string cert, std::string certKey, std::string lhost, unsigned int lport,
	std::string firewallIPs, std::string firewallPorts, bool auth,
	std::map<std::string, std::string> &passwds, std::map<std::string, unsigned int> &tunnelnos,
	std::map<std::string, unsigned int> &bandwidths) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ServerSideConfig(ssl,
		sockTun, cert, certKey, lhost, lport, firewallIPs, firewallPorts, auth, passwds,
		tunnelnos, bandwidths));

	// Start replacement server
	createNamedLocalServer(no);
}
#else
void GlobalConfig::reconfigureServer(unsigned int no, bool ssl, bool sockTun,
	std::string cert, std::string certKey, std::string lhost, unsigned int lport,
	std::string firewallIPs, std::string firewallPorts, bool auth) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ServerSideConfig(ssl,
		sockTun, cert, certKey, lhost, lport, firewallIPs, firewallPorts, auth));

	// Start replacement server
	createNamedLocalServer(no);
}
#endif


#ifdef AUTH_FILE
// Configure default server
unsigned int GlobalConfig::configureDefaultServer(void) throw(AuthError)
{
	std::map<std::string, std::string>	  passwds;
	std::map<std::string, unsigned int>	 tunnelnos, bandwidths;
	MutexVar mv(m_mtx);

	// Pick some sensible defaults
	configureServer(false, false, "", "", "", 8080, "", "", false, passwds,
		tunnelnos, bandwidths);

	// Return ID for new config
	return getLastConfigID();
}
#else
// Configure default server
unsigned int GlobalConfig::configureDefaultServer(void) throw(AuthError)
{
	MutexVar mv(m_mtx);

	// Pick some sensible defaults
	configureServer(false, false, "", "", "", 8080, "", "", false);

	// Return ID for new config
	return getLastConfigID();
}
#endif


// Configure as a client
void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, bool sockTun,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
	std::string tune)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, sockTun, lhost, lport,
		shost, sport, suser, spasswd, dhost, dport, tune));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, bool sockTun,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
	std::string tune, std::string phost, unsigned int pport, std::string puser,
	std::string ppasswd)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, sockTun, lhost, lport,
		shost, sport, suser, spasswd, dhost, dport, tune, phost, pport, puser, ppasswd));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, std::string shost, unsigned int sport, std::string suser,
	std::string spasswd, std::string tune)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, lhost, lport,
		shost, sport, suser, spasswd, tune));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, std::string shost, unsigned int sport, std::string suser,
	std::string spasswd, std::string tune, std::string phost, unsigned int pport,
	std::string puser, std::string ppasswd)
{
	MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, lhost, lport,
		shost, sport, suser, spasswd, tune, phost, pport, puser, ppasswd));
}

// Reconfigure a client
void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	bool sockTun, std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
	std::string tune) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, sockTun, lhost, lport, shost, sport, suser, spasswd, dhost, dport, tune));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	bool sockTun, std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
	std::string tune, std::string phost, unsigned int pport, std::string puser,
	std::string ppasswd) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, sockTun, lhost, lport, shost, sport, suser, spasswd, dhost, dport, tune,
		phost, pport, puser, ppasswd));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string tune) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, lhost, lport, shost, sport, suser, spasswd, tune));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string tune, std::string phost,
	unsigned int pport, std::string puser, std::string ppasswd) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, lhost, lport, shost, sport, suser, spasswd, tune, phost, pport, puser,
		ppasswd));

	// Start replacement server
	createNamedLocalServer(no);
}

// Configure default client
unsigned int GlobalConfig::configureDefaultClient(void)
{
	MutexVar mv(m_mtx);

	// Pick some sensible defaults
	configureClient(Enums::TCP, false, "", 1080, "localhost", 8080, "", "", "default");

	// Return ID for new config
	return getLastConfigID();
}


// Setup any additional config required for client tunnelers
void GlobalConfig::additionalClientConfig(const TunnelerID& tid,
	const ClientConfigBlob& blob) throw(TunnelError)
{
	// No additional configuration currently required for clients
}

// Setup any additional config required for server tunnelers
void GlobalConfig::additionalServerConfig(const TunnelerID& tid,
	const ServerConfigBlob& blob) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Lookup current config for this server
	unsigned int no = getConfigIDFromTunnelerID(tid);
	ServerSideConfig* svrCfg = dynamic_cast<ServerSideConfig*>(m_serverConfigs[no]);
	if(svrCfg) {
		// Is this a socket tunnel?
		if(svrCfg->isSockTun()) {
			// Get socket tunnel config blob
			const SockTunServerConfigBlob* sockTunBlob;
			sockTunBlob = dynamic_cast<const SockTunServerConfigBlob*>(&blob);
			if(sockTunBlob) {
				// Set user callback
				svrCfg->setUserCallback(sockTunBlob->getUserCallback());
			}
			else {
				std::stringstream err;
				err << "Error setting additional config for server tunneler id " << tid;
				err << ": Config says it's a socket tunnel, but blob is of wrong type";
				throw TunnelError(err.str());
			}
		}
	}
	else {
		std::stringstream err;
		err << "Error setting additional config for server tunneler id " << tid;
		err << ": It is not a server";
		throw TunnelError(err.str());
	}
}


// Get ID of last configured client/server tunneler
TunnelerID GlobalConfig::getLastTunnelerID(void) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Return tunneler ID for most recent config
	unsigned int idx = getLastConfigID();
	SideConfig* cfg = m_serverConfigs[idx];
	return cfg->getTunnelerID();
}


// Get version identifier
std::string GlobalConfig::getVersionStr(void) const
{
	return m_version;
}

// Get local host name
std::string GlobalConfig::getUniqueHostName(void) const
{
	return m_hostName;
}
std::string GlobalConfig::getHostName(void) const
{
	return m_trueHostName;
}

// Get/set daemon flag
void GlobalConfig::setDaemon(bool flag)
{
	m_daemon = flag;
}
bool GlobalConfig::getDaemon(void) const
{
	return m_daemon;
}

// Get/set debug level
void GlobalConfig::setDebug(unsigned int debug)
{
	m_debugLevel = debug;
}
unsigned int GlobalConfig::getDebug(void) const
{
	return m_debugLevel;
}

// Get/set tunnel testing flag
void GlobalConfig::setTesting(bool flag, bool tellTunnelers)
{
	m_testing = flag;

	// Do we need to let active tunnelers know?
	if(tellTunnelers)
		activateTesting();
}
bool GlobalConfig::getTesting(void) const
{
	return m_testing;
}

// Get our process ID
unsigned int GlobalConfig::getPid(void) const
{
	return m_pid;
}

// Get/set PID file name
void GlobalConfig::setPIDFile(std::string pidfile)
{
	m_pidFile = pidfile;
}
std::string GlobalConfig::getPIDFile(void) const
{
	return m_pidFile;
}


// Get/set the default realm
std::string GlobalConfig::getRealm(void) const
{
	return m_realm;
}
void GlobalConfig::setRealm(std::string realm)
{
	m_realm = realm;
}


// Get/set data log file name stem
std::string GlobalConfig::getDataLogFileStem(void) const
{
	return m_dataLogStem;
}
void GlobalConfig::setDataLogFileStem(std::string stem)
{
	m_dataLogStem = stem;
}


// Create local servers
void GlobalConfig::createLocalServers(void) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Create server for each set of config
	for(unsigned int cit = 0; cit < m_serverConfigs.size(); ++cit)
		createNamedLocalServer(cit);
}

// Create named local server
void GlobalConfig::createNamedLocalServer(unsigned int no) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Get config for named server
	SideConfig* cfg = m_serverConfigs[no];

	// Create server
	if(cfg->isClient()) {
		// Create client side server
		ClientSideConfig* ccfg = dynamic_cast<ClientSideConfig*>(cfg);
		m_servers.insert(m_servers.begin() + no, new ClientServer(ccfg));
	}
	else {
		// Create server side server
		ServerSideConfig* scfg = dynamic_cast<ServerSideConfig*>(cfg);
		m_servers.insert(m_servers.begin() + no, new ServerServer(scfg));
	}
}

// Create named local server
void GlobalConfig::createNamedLocalServer(const TunnelerID& tid) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Get index for named server
	unsigned int no = getConfigIDFromTunnelerID(tid);

	// Create named server
	createNamedLocalServer(no);
}


// Destroy local servers
void GlobalConfig::destroyLocalServers(void)
{
	MutexVar mv(m_mtx);

	// Stop local servers
	while(m_servers.size() > 0)
		destroyNamedLocalServer(0);
}

// Destroy specified local server/config
void GlobalConfig::destroyNamedLocalServer(unsigned int no) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Check we have at least a config with this id
	if(m_serverConfigs.size() <= no) {
		std::stringstream err;
		err << "No local server config with id " << no << " found to destroy";
		throw TunnelError(err.str());
	}

	// Remove server config
	std::vector<SideConfig*>::iterator cit = m_serverConfigs.begin() + no;
	SideConfig* cfg = *cit;
	m_serverConfigs.erase(cit);

	// Got a corresponding server for this config?
	if(m_servers.size() > no) {
		std::vector<Server*>::iterator sit = m_servers.begin() + no;
		Server* svr = *sit;
		if(svr->getConfig() == cfg) {
			// Clear up server
			delete svr;
			m_servers.erase(sit);
		}
	}

	// Delete server config
	delete cfg;
}

// Destroy specified local server/config
void GlobalConfig::destroyNamedLocalServer(const TunnelerID& tid) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Get index for named server
	unsigned int no = getConfigIDFromTunnelerID(tid);

	// Destroy named server
	destroyNamedLocalServer(no);
}



// Initiate clean shutdown of all local server tunnels
void GlobalConfig::shutdownNamedLocalServerTunnels(unsigned int no) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Check we have at least a server with this id
	if(m_servers.size() <= no) {
		std::stringstream err;
		err << "No local server with id " << no << " found to shutdown";
		throw TunnelError(err.str());
	}
	else {
		// Get server
		std::vector<Server*>::iterator sit = m_servers.begin() + no;
		Server* svr = *sit;

		// Initiate shutdown on this servers tunneler
		Tunneler* tunneler = svr->getTunneler();
		if(tunneler) {
			tunneler->initiateAllTunnelsShutdown();
		}
	}
}

// Initiate clean shutdown of all local server tunnels
void GlobalConfig::shutdownNamedLocalServerTunnels(const TunnelerID& tid) throw(TunnelError)
{
	MutexVar mv(m_mtx);

	// Get index for named server
	unsigned int no = getConfigIDFromTunnelerID(tid);

	// Destroy named server
	shutdownNamedLocalServerTunnels(no);
}


// Have we got at least one client or server configured?
bool GlobalConfig::confCliSvr(void)
{
	return (m_serverConfigs.size() > 0);
}


// Are we running on a big-endian platform?
bool GlobalConfig::bigEndian(void)
{
	return m_bigEndian;
}


// Get socket descriptor for a client socket tunnel
Sock* GlobalConfig::getCliSockTunSock(const TunnelerID& tid) throw(TunnelError)
{
	Sock* sock = 0;

	MutexVar mv(m_mtx);

	try {
		// Lookup client tunneler
		unsigned int no = getConfigIDFromTunnelerID(tid);
		ClientServer* cliSvr = getClientServer(no);

		// Get socket descriptor
		sock = cliSvr->getSockTunSock();
	}
	catch(TunnelError& e) {
		std::stringstream err;
		err << "Error retrieving socket for client socket tunneler id " << tid;
		err << ": " << e.what();
		throw TunnelError(err.str());
	}

	return sock;
}


// Get stats for all tunnels carried by the specified tunneler
void GlobalConfig::getTunnelerTunnelStats(const TunnelerID& tid,
	std::vector<TunnelStats>& stats) throw(TunnelError)
{
	// Clear results
	stats.clear();

	// Lookup tunneler server
	MutexVar mv(m_mtx);
	unsigned int no = getConfigIDFromTunnelerID(tid);
	Server* svr = m_servers[no];

	// Get stats from server
	svr->getTunnelStats(stats);
}


/*********************/
/* Private Interface */
/*********************/


// Get ID of last client/server config - Caller holds mutex
unsigned int GlobalConfig::getLastConfigID(void) const throw(TunnelError)
{
	// Most recent config will be last index in list
	if(m_serverConfigs.size() > 0)
		return (m_serverConfigs.size() - 1);

	// No configured clients/servers
	throw TunnelError("No currently configured clients/servers to get ID for");
}


// Get ID of client/server config from tunneler ID - Caller holds mutex
unsigned int GlobalConfig::getConfigIDFromTunnelerID(const TunnelerID& tid)
	throw(TunnelError)
{
	// Look for matching tunneler ID in list of configs
	bool foundIt = false;
	unsigned int idx;
	for(idx = 0; idx < m_serverConfigs.size(); ++idx) {
		SideConfig* cfg = m_serverConfigs[idx];
		TunnelerID thatTid = cfg->getTunnelerID();
		if(tid == thatTid) {
			foundIt = true;
			break;
		}
	}

	// Not found?
	if(!foundIt) {
		std::stringstream err;
		err << "No currently configured tunneler found with ID " << tid;
		throw TunnelError(err.str());
	}

	// Return index of matching config
	return idx;
}


// Lookup client tunneler with the specified index - Caller holds mutex
ClientServer* GlobalConfig::getClientServer(unsigned int idx) throw(TunnelError)
{
	// Check we have a server with this id
	if(m_servers.size() <= idx) {
		std::stringstream err;
		err << "No client tunneler found with index " << idx;
		throw TunnelError(err.str());
	}

	// Check this is actually a client server/tunneler
	ClientServer* cliSvr = dynamic_cast<ClientServer*>(m_servers[idx]);
	if(cliSvr == 0) {
		std::stringstream err;
		err << "Tunneler at index " << idx << " is not a client tunneler";
		throw TunnelError(err.str());
	}

	return cliSvr;
}

// Lookup server tunneler with the specified index - Caller holds mutex
ServerServer* GlobalConfig::getServerServer(unsigned int idx) throw(TunnelError)
{
	// Check we have a server with this id
	if(m_servers.size() <= idx) {
		std::stringstream err;
		err << "No server tunneler found with index " << idx;
		throw TunnelError(err.str());
	}

	// Check this is actually a client server/tunneler
	ServerServer* svrSvr = dynamic_cast<ServerServer*>(m_servers[idx]);
	if(svrSvr == 0) {
		std::stringstream err;
		err << "Tunneler at index " << idx << " is not a server tunneler";
		throw TunnelError(err.str());
	}

	return svrSvr;
}


// Tell all client tunnelers about a change to testing
void GlobalConfig::activateTesting(void)
{
	MutexVar mv(m_mtx);

	// Iterate over tunnelers
	for(unsigned int idx = 0; idx < m_servers.size(); ++idx) {
		Server* svr = m_servers[idx];

		// Start/stop testing on this tunnel
		svr->getTunneler()->setServerTesting();
	}
}


}	// namespace HTTPTunnel

