/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration.
 *
 * $Id: GlobalConfig.hh,v 1.6 2009-03-20 11:52:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _GLOBAL_CONFIG_HH_
#define _GLOBAL_CONFIG_HH_

#include <string>
#include <vector>
#ifdef AUTH_FILE
#include <map>
#endif
#include <sys/types.h>
#include "Enums.hh"
#include "Exceptions.hh"
#include "Mutex.hh"
#include "TunnelerID.hh"
#include "TunnelStats.hh"

namespace HTTPTunnel {

class Authenticator;
class Server;
class ClientServer;
class ServerServer;
class SideConfig;
class ClientConfigBlob;
class ServerConfigBlob;
class Sock;

class GlobalConfig {
public:

	/*********************/
	/* Types & Constants */
	/*********************/

	// Maximum size of local hostname
	static const unsigned int		OUR_HOST_NAME_MAX		= 64;


   // Construction/destruction
   GlobalConfig(std::string procName);
   ~GlobalConfig();


	/*************/
	/* Utilities */
	/*************/

	// Serialize to a config file
	void dumpToFile(std::string name) throw(ConfigError);


	/**********************/
	/* Accessors/mutators */
	/**********************/

	// Client/server settings
	#ifdef AUTH_FILE
	void configureServer(bool ssl, bool sockTun, std::string cert, std::string certKey,
		std::string lhost, unsigned int lport, std::string firewallIPs,
		std::string firewallPorts, bool auth, std::map<std::string, std::string> &passwds,
		std::map<std::string, unsigned int> &tunnelnos,
		std::map<std::string, unsigned int> &bandwidths) throw(AuthError);
	#else
	void configureServer(bool ssl, bool sockTun, std::string cert, std::string certKey,
		std::string lhost, unsigned int lport, std::string firewallIPs,
		std::string firewallPorts, bool auth) throw(AuthError);
	#endif

	#ifdef AUTH_FILE
	void reconfigureServer(unsigned int no, bool ssl, bool sockTun, std::string cert,
		std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
		std::string firewallPorts, bool auth, std::map<std::string, std::string> &passwds,
		std::map<std::string, unsigned int> &tunnelnos,
		std::map<std::string, unsigned int> &bandwidths) throw(TunnelError);
	#else
	void reconfigureServer(unsigned int no, bool ssl, bool sockTun, std::string cert,
		std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
		std::string firewallPorts, bool auth) throw(TunnelError);
	#endif

	unsigned int configureDefaultServer(void) throw(AuthError);

	void configureClient(Enums::Protocol prot, bool ssl, bool sockTun, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string dhost, unsigned int dport, std::string tune);
	void configureClient(Enums::Protocol prot, bool ssl, bool sockTun, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string dhost, unsigned int dport, std::string tune,
		std::string phost, unsigned int pport, std::string puser, std::string ppasswd);
	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string tune);
	void configureClient(Enums::Protocol prot, bool ssl, std::string lhost,
		unsigned int lport, std::string shost, unsigned int sport, std::string suser,
		std::string spasswd, std::string tune, std::string phost, unsigned int pport,
		std::string puser, std::string ppasswd);

	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl, bool sockTun, 
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
		std::string tune) throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl, bool sockTun,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
		std::string tune, std::string phost, unsigned int pport, std::string puser,
		std::string ppasswd) throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string tune) throw(TunnelError);
	void reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
		std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
		std::string suser, std::string spasswd, std::string tune, std::string phost,
		unsigned int pport, std::string puser, std::string ppasswd) throw(TunnelError);

	unsigned int configureDefaultClient(void);


	// Setup any additional config required for socket tunnels
	void additionalClientConfig(const TunnelerID& tid, const ClientConfigBlob& blob)
		throw(TunnelError);
	void additionalServerConfig(const TunnelerID& tid, const ServerConfigBlob& blob)
		throw(TunnelError);

	// Get ID of last configured client/server tunneler
	TunnelerID getLastTunnelerID(void) throw(TunnelError);

	// Get version identifier
	std::string getVersionStr(void) const;

	// Get pseudo unique local hostname
	std::string getUniqueHostName(void) const;
	std::string getHostName(void) const;

	// Get/set daemon flag
	void setDaemon(bool flag);
	bool getDaemon(void) const;

	// Get/set debug level
	void setDebug(unsigned int debug);
	unsigned int getDebug(void) const;

	// Get/set tunnel testing flag
	void setTesting(bool flag, bool tellTunnelers);
	bool getTesting(void) const;

	// Process values
	unsigned int getPid(void) const;
	void setPIDFile(std::string);
	std::string getPIDFile(void) const;

	// Default realm
	std::string getRealm(void) const;
	void setRealm(std::string realm);

	// Data log file stem
	std::string getDataLogFileStem(void) const;
	void setDataLogFileStem(std::string stem);

	// Create/destroy local servers
	void createLocalServers(void) throw(TunnelError);
	void createNamedLocalServer(unsigned int no) throw(TunnelError);
	void createNamedLocalServer(const TunnelerID& tid) throw(TunnelError);
	void destroyLocalServers(void);
	void destroyNamedLocalServer(unsigned int no) throw(TunnelError);
	void destroyNamedLocalServer(const TunnelerID& tid) throw(TunnelError);

	// Initiate clean shutdown of all local server tunnels
	void shutdownNamedLocalServerTunnels(unsigned int no) throw(TunnelError);
	void shutdownNamedLocalServerTunnels(const TunnelerID& tid) throw(TunnelError);

	// Have we got at least one client or server configured?
	bool confCliSvr(void);

	// Are we running on a big-endian platform?
	bool bigEndian(void);

	// Get socket descriptor for a client socket tunnel
	Sock* getCliSockTunSock(const TunnelerID& tid) throw(TunnelError);

	// Get stats for all tunnels carried by the specified tunneler
	void getTunnelerTunnelStats(const TunnelerID& tid, std::vector<TunnelStats>& stats)
		throw(TunnelError);


private:

	/***********/
	/* Methods */
	/***********/

	// Get ID of last client/server config
	unsigned int getLastConfigID(void) const throw(TunnelError);

	// Get ID of client/server config from tunneler ID
	unsigned int getConfigIDFromTunnelerID(const TunnelerID& tid) throw(TunnelError);

	// Lookup client/server tunneler with the specified index
	ClientServer* getClientServer(unsigned int idx) throw(TunnelError);
	ServerServer* getServerServer(unsigned int idx) throw(TunnelError);

	// Tell all client tunnelers about a change to testing
	void activateTesting(void);


	/********/
	/* Data */
	/********/

	// Mutex
	Mutex								m_mtx;

	// Process name to use
	std::string						m_procName;

	// Should we daemonise?
	bool								m_daemon;

	// Should we do tunnel testing?
	bool								m_testing;

	// Pseudo unique local host identifer
	std::string						m_hostName;
	std::string						m_trueHostName;

	// Debug level
	unsigned int					m_debugLevel;

	// Our process ID
	pid_t								m_pid;
	std::string						m_pidFile;

	// Default realm
	std::string						m_realm;

	// Local servers and their config
	std::vector<Server*>			m_servers;
	std::vector<SideConfig*>	m_serverConfigs;

	// Data log file stem
	std::string						m_dataLogStem;

	// Big-endian platform flag
	bool								m_bigEndian;

	// Version identifier
	std::string						m_version;

};


}	// namespace HTTPTunnel

#endif

