/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP GET request message.
 *
 * $Id: HTTPGetMsg.cc,v 1.3 2008-07-11 11:22:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <cstring>
#include <sstream>
#include <iostream>
#include "HTTPGetMsg.hh"
#include "Exceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "ClientSideConfig.hh"
#include "HTTPMsgParms.hh"
#include "Payload.hh"
#include "Utils.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
HTTPGetMsg::HTTPGetMsg()
: HTTPRequestMsg(), m_uuData(), m_repUULen(0), m_uriEncFields()
{
}

HTTPGetMsg::HTTPGetMsg(const HTTPMsgParms* parms, std::string uri,
	const ClientSideConfig* cfg)
: HTTPRequestMsg(parms, uri, cfg), m_uuData(), m_repUULen(0), m_uriEncFields()
{
	// Construct headers
	unsigned int payloadLen = m_parms->getPayload()->getPayloadSize();
	if(payloadLen > 0) {
		// Encode payload data into header
		std::stringstream ourhdr;
		packPayload(m_parms->getPayload()->getPayload(), ourhdr);
		m_ourHeader += ourhdr.str();
	}

	// CONNECT specific fields
	if(m_parms->getState() == Tunnel::CONNECT) {
		// Send tunnel destination details
		m_ourHeader += "HT-Des: " + parms->getTunDest() + "\r\n";

		// Send tunnel tuning
		m_ourHeader += "HT-Tun: " + parms->getTunTune() + "\r\n";

		// Send client version
		m_ourHeader += "HT-Cvr: " + parms->getCliVer() + "\r\n";

		// Also include all our header fields as a test of any proxy filtering
		m_ourHeader += "HT-Dat: \"\"\r\n";
		m_ourHeader += "HT-Dln: 0\r\n";
		m_ourHeader += "HT-Spl: 0\r\n";
	}

	// Construct request line
	std::stringstream req;
	req << "GET " << getRequestURI() << " HTTP/" << m_parms->getVer();
	m_requestLine = req.str();
}

HTTPGetMsg::HTTPGetMsg(const HTTPGetMsg& that)
: HTTPRequestMsg(that), m_uuData(that.m_uuData), m_repUULen(that.m_repUULen),
  m_uriEncFields(that.m_uriEncFields)
{
}

HTTPGetMsg::~HTTPGetMsg()
{
}

// Construction from a raw read data stream
HTTPGetMsg::HTTPGetMsg(RawBuff& readdata) throw(HTTPMsgParsingError)
: HTTPRequestMsg(), m_uuData(), m_repUULen(0)
{
	/* Look for basic components of GET message header -
		GET		- message identifier
		/X.htm	- URI
		HTTP/x.y	- protocol version
	*/

	const unsigned char *getStart, *getEnd;
	if(!readdata.findString("GET", getStart, getEnd))
		throw HTTPMsgWrongMsgType("GET message identifier not found", m_parsingLine);

	// Extract URI
	const unsigned char *uriStart1, *uriStart2, *uriEnd;
	if(readdata.findString("/", uriStart1, uriEnd)) {
		if(readdata.findString(".htm", uriStart2, uriEnd)) {
			// Get URI
			m_uri = makeString(uriStart1, uriEnd);
			m_parms->setURI(m_uri);
		}
		else
			throw HTTPMsgParsingError("End of GET URI not found", m_parsingLine);
	}
	else
		throw HTTPMsgParsingError("Start of GET URI not found", m_parsingLine);

	// Extract protocol version
	const unsigned char *httpStart, *httpEnd;
	if(!readdata.findString("HTTP/", httpStart, httpEnd))
		throw HTTPMsgParsingError("GET protocol id not found", m_parsingLine);

	const unsigned char *nlStart, *nlEnd;
	if(!findNewLine(readdata, nlStart, nlEnd))
		throw HTTPMsgParsingError("Expected newline not found", m_parsingLine);
	m_parms->setVer(makeString(httpEnd, nlStart));

	// Copy entire request line
	m_requestLine = makeString(getStart, nlStart);

	// Are we passing tunneling data via URI?
	bool dataInURI = false;
	RawBuff reqLineBuff(m_requestLine);
	if(reqLineBuff.findString("?", uriStart1, uriEnd)) {
		if(reqLineBuff.findString(" ", uriStart1, uriStart2, uriEnd)) {
			if(uriEnd - 1 > uriStart1 + 1) {
				// Get encoded tunneling fields
				m_uriEncFields = makeString(uriStart1 + 1, uriEnd - 1);
				dataInURI = true;
			}
			else {
				throw HTTPMsgParsingError("Too little encoded URI data", m_parsingLine);
			}
		}
		else {
			throw HTTPMsgParsingError("End of GET encoded URI data not found",
				m_parsingLine);
		}
	}

	// Discard everything from the first line in the data stream
	readdata.erase(nlEnd);
	m_parsingLine++;

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "HTTPGetMsg::HTTPGetMsg() - Got URI: " << m_uri;
	nfo << ", protocol: " << m_parms->getVer();
	DPRINTV(nfo.str());
	#endif

	// Let base classes look for their headers
	bool eoh;
	std::string hdrline;
	parseHdrs(readdata, eoh, hdrline);

	// Expect final newline to terminate non-content part of message
	if(!findNewLine(readdata, nlStart, nlEnd))
		throw HTTPMsgParsingError("GET final newline terminator not found", m_parsingLine);
	readdata.erase(nlEnd);
	m_parsingLine++;

	// Did we see data passed in the URI?
	if(dataInURI) {
		bool parseURI = true;

		// Was this the first CONNECT msg of a session?
		Tunnel::TunnelState state = m_parms->getState();
		if(state == Tunnel::UNKNOWNSTATE || state == Tunnel::CONNECT) {
			// Remember whether we need to use header tunneling all the time
			parseURI = !allReqHeaders();
			m_parms->setUseHeaderFields(!parseURI);
		}

		if(parseURI) {
			// Try to parse tunneling fields from coded URI
			std::string decoded = Utils::URLDecode(m_uriEncFields) + "\r\n";
			RawBuff decodedHeaders(decoded);
			parseHdrs(decodedHeaders, eoh, hdrline);
		}
	}

	// Now deal with any content
	extractContent(readdata);
}

// Assignment
HTTPGetMsg& HTTPGetMsg::operator=(const HTTPGetMsg& that)
{
	HTTPRequestMsg::operator=(that);
	m_uuData = that.m_uuData;
	m_repUULen = that.m_repUULen;
	m_uriEncFields = that.m_uriEncFields;
	return *this;
}


/***********************/
/* Protected interface */
/***********************/


// Parse and extract any message content
void HTTPGetMsg::extractContent(RawBuff& readdata) throw(ParsingError)
{
	DPRINTV("HTTPGetMsg::extractContent()");

	// Are we expecting any content data?
	if(m_repUULen > 0) {
		#ifdef DEBUG_TUN
		std::stringstream str;
		str << "HTTPGetMsg::extractContent() - Expecting " << m_repUULen;
		str << " bytes of UUEncoded content";
		DPRINTV(str.str());
		#endif

		if(m_uuData.size() == m_repUULen) {
			// UUDecode data into our payload buffer
			try {
				m_parms->getPayloadMutable()->getPayloadMutable().uudecode(m_uuData);
			}
			catch(MemoryError& e) {
				// Rethrow
				throw ParsingError("Error decoding UUEncoded data: " + e.what());
			}
		}
		else {
			#ifdef DEBUG_TUN
			std::stringstream str;
			str << "HTTPGetMsg::extractContent() - Wrong UUEncoded data length: ";
			str << "Reported " << m_repUULen << ", but got " << m_uuData.size();
			DPRINTV(str.str());
			#endif

			throw BadContentLengthParsingError("Incorrect encoded content length",
				m_parsingLine);
		}
	}
}


// Attempt to parse headers from a raw byte-stream
void HTTPGetMsg::parseHdrs(RawBuff& data, bool& eoh, std::string& hdrline)
	throw(HTTPMsgParsingError)
{
	/* Look for any GET headers - everything else is handled elsehere */

	bool ok = true;
	eoh = false;

	// Parse as much as we can
	while(ok && !eoh) {
		// First let base class have a go
		HTTPRequestMsg::parseHdrs(data, eoh, hdrline);
		if(!eoh) {
			/* Base class didn't like a header, let us try */

			// Is it one of ours?
			if(strncmp(hdrline.c_str(), "HT-Dat:", strlen("HT-Dat:")) == 0) {
				// Some UUEncoded data
				std::string::size_type quotestart, quoteend;
				quotestart = hdrline.find('\"');
				quoteend = hdrline.rfind('\"');
				if(quotestart != std::string::npos && quoteend != std::string::npos) {
					std::string packed(hdrline, quotestart, (quoteend - quotestart) + 1);
					unpackPayload(packed);
					m_seenUUData = true;
					m_lastTunId = "HT-Dat: ";
				}
				else {
					throw HTTPMsgParsingError("Failed to parse start/end of UUEncoded data"
						" from get msg", m_parsingLine);
				}
			}
			else if(strncmp(hdrline.c_str(), "HT-Dln:", strlen("HT-Dln:")) == 0) {
				// Length of uuencoded data
				if(sscanf(hdrline.c_str(), "HT-Dln: %d", &m_repUULen) != 1) {
					throw HTTPMsgParsingError("Failed to parse UUEncoded length"
						" from get msg", m_parsingLine);
				}
				m_seenUULen = true;
				m_lastTunId = "HT-Dln: ";
			}
			else if(strncmp(hdrline.c_str(), "HT-Des:", strlen("HT-Des:")) == 0) {
				// Get address of destination to tunnel to
				std::string::size_type i = hdrline.find(' ');
				if(i != std::string::npos) {
					std::string dest(hdrline, i + 1);
					m_parms->setTunDest(dest);
					m_seenDest = true;
					m_lastTunId = "HT-Des: ";
				}
				else {
					throw HTTPMsgParsingError("Failed to parse tunnel destination"
						" from get msg", m_parsingLine);
				}
			}
			else if(strncmp(hdrline.c_str(), "HT-Tun:", strlen("HT-Tun:")) == 0) {
				// Get tunnel tuning
				std::string::size_type i = hdrline.find(' ');
				if(i != std::string::npos) {
					std::string tune(hdrline, i + 1);
					m_parms->setTunTune(tune);
					m_seenTune = true;
					m_lastTunId = "HT-Tun: ";
				}
				else {
					throw HTTPMsgParsingError("Failed to parse tunnel tuning"
						" from get msg", m_parsingLine);
				}
			}
		}
	}
}


// Encode payload data into a header string
void HTTPGetMsg::packPayload(const RawBuff& payload, std::stringstream& str)
{
	// Get uuencoded data
	std::string uue;
	unsigned int length = payload.getLength();
	payload.uuencode(length, uue);

	// If we're packing into header fields, limit the maximum line length
	bool limitLineLength = m_parms->getUseHeaderFields();

	// Write out uuencoded data as header lines
	std::string::size_type bytesToGo = uue.size();
	std::string::size_type pos = 0;
	do {
		str << "HT-Dat: \"";

		if(!limitLineLength || (bytesToGo < MAX_UU_CHARS_LINE)) {
			// Not enough left for a full line
			str << uue.substr(pos);
			bytesToGo = 0;
		}
		else {
			// Full line
			str << uue.substr(pos, MAX_UU_CHARS_LINE);
			bytesToGo -= MAX_UU_CHARS_LINE;
			pos += MAX_UU_CHARS_LINE;
		}

		str << "\"\r\n";
	} while(bytesToGo > 0);

	// Add encoded data length field
	str << "HT-Dln: " << uue.size() << "\r\n";
}

// Extract packed payload data from a header string
void HTTPGetMsg::unpackPayload(std::string& header)
{
	std::string::size_type pos = 0;

	// Iterate over encoded data
	while(pos != std::string::npos) {
		// Skip starting "
		pos++;

		// Find terminating "
		std::string::size_type term = header.find('\"', pos);

		// Extract encoded string
		m_uuData += header.substr(pos, term - pos);

		// Look for possible next starting "
		pos = header.find('\"', term + 1);
	}
}


}	// namespace HTTPTunnel

