/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP message.
 *
 * $Id: HTTPMsg.cc,v 1.5 2009-06-18 11:51:28 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <string>
#include <cstring>
#include <sstream>
#include <vector>
#include <map>
#include <iostream>
#include "HTTPMsg.hh"
#include "Exceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Payload.hh"
#include "RawBuff.hh"
#include "Tunnel.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;
extern Logger* g_pLogger;


// Construction/destruction
HTTPMsg::HTTPMsg()
: m_parms(0), m_generalHeader(), m_reqRespHeader(), m_entityHeader(), m_ourHeader(),
  m_parsingLine(0), m_seenTID(false), m_seenSeqNo(false), m_seenConn(false),
  m_seenDest(false), m_seenUUData(false), m_seenUULen(false), m_seenAck(false),
  m_seenQuotas(false), m_seenCliVer(false), m_seenSplit(false), m_seenTest(false),
  m_seenTune(false), m_seenNotif(false), m_authArgs(), m_lastTunId(""),
  m_lastTunIdSplit(false)
{
	// Construct default msg parameters
	m_parms = new HTTPMsgParms();
}

HTTPMsg::HTTPMsg(const HTTPMsgParms* parms)
: m_parms(0), m_generalHeader(), m_reqRespHeader(), m_entityHeader(), m_ourHeader(),
  m_parsingLine(0), m_seenTID(false), m_seenSeqNo(false), m_seenConn(false),
  m_seenDest(false), m_seenUUData(false), m_seenUULen(false), m_seenAck(false),
  m_seenQuotas(false), m_seenCliVer(false), m_seenSplit(false), m_seenTest(false),
  m_seenTune(false), m_seenNotif(false), m_authArgs(), m_lastTunId(""),
  m_lastTunIdSplit(false)
{
	// Take our own copy of paramters
	m_parms = dynamic_cast<HTTPMsgParms*>(parms->clone());

	// Construct headers
	std::stringstream genhdr, ourhdr;
	genhdr << "Cache-Control: no-cache\r\n";
	ourhdr << "HT-Tid: " << m_parms->getTid() << "\r\n";
	ourhdr << "HT-Seq: " << m_parms->getPayload()->getSeqNo() << "\r\n";

	// Include acknowledgments
	if(m_parms->getAck().size() > 0)
		ourhdr << "HT-Ack: " << m_parms->getAck() << "\r\n";

	// Split payload indication
	if(m_parms->getPayload()->getSplit())
		ourhdr << "HT-Spl: 1\r\n";

	// Mark connection state as appropriate
	Tunnel::TunnelState state = m_parms->getState();
	switch(state) {
		case(Tunnel::CONNECT):
		case(Tunnel::AUTHREQ):
		case(Tunnel::RECONNECT):
		case(Tunnel::CLOSING):
		case(Tunnel::CONNECTED): {
			genhdr << "Connection: keep-alive\r\n";
			ourhdr << "HT-Cst: " << (unsigned int)(state) << "\r\n";
			break;
		}
		case(Tunnel::CLOSED): {
			genhdr << "Connection: close\r\n";
			ourhdr << "HT-Cst: " << (unsigned int)(state) << "\r\n";
			break;
		}
		case(Tunnel::UNKNOWNSTATE): {
			g_pLogger->error("Tunnel state is UNKNOWN in HTTPMsg()");
			break;
		}
	}

	// Mark tunnel testing state
	Tunnel::TunnelTestState testState = m_parms->getTestState();
	if(testState != Tunnel::NO_TESTING || state == Tunnel::CONNECT) {
		ourhdr << "HT-Tst: " << (unsigned int)(testState) << "\r\n";
	}

	// Format any notifications
	std::vector<std::string> notifs = m_parms->getNotifs();
	for(std::vector<std::string>::iterator nit = notifs.begin(); nit != notifs.end(); ++nit) {
		ourhdr << "HT-Ntf: " << *nit << "\r\n";
	}

	// Tidy up
	m_generalHeader = genhdr.str();
	m_ourHeader = ourhdr.str();
}

HTTPMsg::HTTPMsg(const HTTPMsg& that)
: m_parms(0), m_generalHeader(that.m_generalHeader), m_reqRespHeader(that.m_reqRespHeader),
  m_entityHeader(that.m_entityHeader), m_ourHeader(that.m_ourHeader),
  m_parsingLine(that.m_parsingLine), m_seenTID(that.m_seenTID),
  m_seenSeqNo(that.m_seenSeqNo), m_seenConn(that.m_seenConn),
  m_seenDest(that.m_seenDest), m_seenUUData(that.m_seenUUData),
  m_seenUULen(that.m_seenUULen), m_seenAck(that.m_seenAck),
  m_seenQuotas(that.m_seenQuotas), m_seenCliVer(that.m_seenCliVer),
  m_seenSplit(that.m_seenSplit), m_seenTest(that.m_seenTest),
  m_seenTune(that.m_seenTune), m_seenNotif(that.m_seenNotif),
  m_authArgs(that.m_authArgs), m_lastTunId(that.m_lastTunId),
  m_lastTunIdSplit(that.m_lastTunIdSplit)
{
	// Copy parameters
	if(that.m_parms)
		m_parms = dynamic_cast<HTTPMsgParms*>(that.m_parms->clone());
}

HTTPMsg::~HTTPMsg()
{
	// Free up parameters
	if(m_parms) {
		delete m_parms;
		m_parms = 0;
	}
}

// Assignment
HTTPMsg& HTTPMsg::operator=(const HTTPMsg& that)
{
	// Self assignment?
	if(this != &that) {
		// Copy parameters
		if(m_parms) {
			delete m_parms;
			m_parms = 0;
		}
		if(that.m_parms)
			m_parms = dynamic_cast<HTTPMsgParms*>(that.m_parms->clone());

		// Other members
		m_generalHeader = that.m_generalHeader;
		m_reqRespHeader = that.m_reqRespHeader;
		m_entityHeader = that.m_entityHeader;
		m_ourHeader = that.m_ourHeader;
		m_parsingLine = that.m_parsingLine;
		m_seenTID = that.m_seenTID;
		m_seenSeqNo = that.m_seenSeqNo;
		m_seenConn = that.m_seenConn;
		m_seenDest = that.m_seenDest;
		m_seenUUData = that.m_seenUUData;
		m_seenUULen = that.m_seenUULen;
		m_seenAck = that.m_seenAck;
		m_seenQuotas = that.m_seenQuotas;
		m_seenCliVer = that.m_seenCliVer;
		m_seenSplit = that.m_seenSplit;
		m_seenTest = that.m_seenTest;
		m_seenTune = that.m_seenTune;
		m_seenNotif = that.m_seenNotif;
		m_authArgs = that.m_authArgs;
		m_lastTunId = that.m_lastTunId;
		m_lastTunIdSplit = that.m_lastTunIdSplit;
	}

	return *this;
}


/********************/
/* Public Interface */
/********************/


// Write ourselves out as a stream of bytes
void HTTPMsg::output(RawBuff& dest)
{
	// Somewhere to build up complete message
	std::stringstream output;

	// Get message start line
	std::string start = getStartLine();
	// Build up complete header
	std::string header = getHeader();

	#ifdef DEBUG_TUN
	std::stringstream str1;
	str1 << "HTTPMsg::output() - Sending ";
	str1 << m_parms->getPayload()->getPayload().getLength() << " bytes of payload";
	DPRINTV(str1.str());
	#endif

	// Build up complete message
	output << start << "\r\n";

	if(header.length() > 0)
		output << header;			// Each header line includes it's own CRLF

	output << "\r\n";

	// Append header part of msg to destination
	std::string fullHeader = output.str();
	dest.append((const unsigned char*)fullHeader.c_str(), fullHeader.size());
}


// Get message parameters
const HTTPMsgParms* HTTPMsg::getParms(void) const
{
	return m_parms;
}
HTTPMsgParms* HTTPMsg::getParmsMutable(void)
{
	return m_parms;
}

// Get authentication paramaters
const std::map<std::string, std::vector<std::string> >& HTTPMsg::getAuthArgs(void) const
{
	return m_authArgs;
}


// Minimum headers passed as expected?
bool HTTPMsg::allReqHeaders(void) const
{
	return (m_seenTID && m_seenSeqNo && m_seenConn && m_seenUUData &&
		m_seenUULen && m_seenAck && m_seenCliVer && m_seenSplit &&
		m_seenTest && m_seenTune);
}
bool HTTPMsg::allRespHeaders(void) const
{
	return (m_seenTID && m_seenSeqNo && m_seenConn && m_seenAck);
}


/***********************/
/* Protected interface */
/***********************/


// Attempt to parse headers from a raw byte-stream
void HTTPMsg::parseHdrs(RawBuff& data, bool& eoh, std::string& hdrline)
	throw(HTTPMsgParsingError)
{
	bool ok = true;
	bool thisSplit = false;
	eoh = false;

	/* Look for any general headers - everything else is handled in other classes */
	const unsigned char *start, *end;

	// Parse as much as we can
	while(ok) {

		// We need a complete line to parse
		if(findNewLine(data, start, end, thisSplit)) {
			// Is there anything before the newline?
			 if(start != data.begin()) {

				// Extract this header field
				hdrline = makeString(data.begin(), start);
				if(m_lastTunIdSplit) {
					// This is actually a split line
					hdrline = m_lastTunId + hdrline;
				}

				// Clear out parse buff
				data.erase(end);
				if(thisSplit)
					data.eraseWS();
				m_lastTunIdSplit = thisSplit;
				m_parsingLine++;

				// Is it one of ours?
				if(strncmp(hdrline.c_str(), "HT-Seq:", strlen("HT-Seq:")) == 0) {
					// Get msg sequence number
					unsigned int seqNo;
					if(sscanf(hdrline.c_str(), "HT-Seq: %d", &seqNo) != 1) {
						throw HTTPMsgParsingError("Failed to parse sequence number from msg",
							m_parsingLine);
					}
					m_parms->getPayloadMutable()->setSeqNo(seqNo);
					m_seenSeqNo = true;
					m_lastTunId = "HT-Seq: ";

					#ifdef DEBUG_TUN
					std::stringstream dbg;
					dbg << "HTTPMsg::parseHdrs() - Got sequence no: " << seqNo;
					DPRINTV(dbg.str());
					#endif
				}
				else if(strncmp(hdrline.c_str(), "HT-Ack:", strlen("HT-Ack:")) == 0) {
					// Get acknowledgment
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string ack(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got ack: " + ack);
						m_parms->setAck(ack);
						m_seenAck = true;
						m_lastTunId = "HT-Ack: ";
					}
					else {
						throw HTTPMsgParsingError("Failed to parse ack from msg", m_parsingLine);
					}
				}
				else if(strncmp(hdrline.c_str(), "HT-Tid:", strlen("HT-Tid:")) == 0) {
					// Get reported tunnel ID
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string tid(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got TID: " + tid);
						m_parms->setTid(tid);
						m_seenTID = true;
						m_lastTunId = "HT-Tid: ";
					}
					else {
						throw HTTPMsgParsingError("Failed to parse TID from msg", m_parsingLine);
					}
				}
				else if(strncmp(hdrline.c_str(), "HT-Cst:", strlen("HT-Cst:")) == 0) {
					// Closing?
					unsigned int state;
					if(sscanf(hdrline.c_str(), "HT-Cst: %d", &state) != 1) {
						throw HTTPMsgParsingError("Failed to parse connection state from msg",
							m_parsingLine);
					}
					m_parms->setState((Tunnel::TunnelState)(state));
					m_seenConn = true;
					m_lastTunId = "HT-Cst: ";
				}
				else if(strncmp(hdrline.c_str(), "HT-Cvr:", strlen("HT-Cvr:")) == 0) {
					// Get client version ID
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string text(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got client version: " + text);
						m_parms->setCliVer(text);
						m_seenCliVer = true;
						m_lastTunId = "HT-Cvr: ";
					}
					else {
						throw HTTPMsgParsingError("Failed to parse client version from msg",
							m_parsingLine);
					}
				}
				else if(strncmp(hdrline.c_str(), "HT-Spl:", strlen("HT-Spl:")) == 0) {
					// Get payload split indicator
					unsigned int split;
					if(sscanf(hdrline.c_str(), "HT-Spl: %d", &split) != 1) {
						throw HTTPMsgParsingError("Failed to parse split indicator from msg",
							m_parsingLine);
					}
					m_parms->getPayloadMutable()->setSplit(split != 0);
					m_seenSplit = true;
					m_lastTunId = "HT-Spl: ";
				}
				else if(strncmp(hdrline.c_str(), "HT-Tst:", strlen("HT-Tst:")) == 0) {
					// Get tunnel test state
					unsigned int state;
					if(sscanf(hdrline.c_str(), "HT-Tst: %d", &state) != 1) {
						throw HTTPMsgParsingError("Failed to parse test state from msg", m_parsingLine);
					}
					m_parms->setTestState((Tunnel::TunnelTestState)(state));
					m_seenTest = true;
					m_lastTunId = "HT-Tst: ";

					#ifdef DEBUG_TUN
					std::stringstream dbg;
					dbg << "HTTPMsg::parseHdrs() - Got test state: " << state;
					DPRINTV(dbg.str());
					#endif
				}
				else if(strncmp(hdrline.c_str(), "HT-Ntf:", strlen("HT-Ntf:")) == 0) {
					// Get notification msg
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string text(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got notification: " + text);
						m_parms->addNotif(text);
						m_seenNotif = true;
						m_lastTunId = "HT-Ntf: ";
					}
					else {
						throw HTTPMsgParsingError("Failed to parse notification from msg",
							m_parsingLine);
					}
				}

				// Is it a standard HTTP header field of interest?
				else if(strncmp(hdrline.c_str(), "Connection:", strlen("Connection:")) == 0) {
					// Get HTTP connection option
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string text(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got HTTP connection option: " + text);
						m_parms->setConnOption(text);
					}
					else {
						throw HTTPMsgParsingError("Failed to parse HTTP connection option from msg",
							m_parsingLine);
					}
				}
				else if(strncmp(hdrline.c_str(), "Proxy-Connection:", strlen("Proxy-Connection:")) == 0) {
					// Get proxy HTTP connection option
					std::string::size_type i = hdrline.find(' ');
					if(i != std::string::npos) {
						std::string text(hdrline, i + 1);
						DPRINTV("HTTPMsg::parseHdrs() - Got proxy HTTP connection option: " + text);
						m_parms->setProxyConnOption(text);
					}
					else {
						throw HTTPMsgParsingError("Failed to parse proxy HTTP connection option from msg",
							m_parsingLine);
					}
				}

				else {
					// Unknown header
					ok = false;
				}
			}
			else {
				// Newline found at start of line, must be end of headers
				eoh = true;
				break;
			}
		}
		else {
			// Header line terminating new line not found
			throw HTTPMsgParsingError("Msg header terminating newline not found",
				m_parsingLine);
		}
	}
}


// Find the next newline in some raw data
bool HTTPMsg::findNewLine(RawBuff& buff, const unsigned char*& start,
	const unsigned char*& end) const
{
	// Try looking for "\r\n" or "\n"
	return (buff.findString("\r\n", start, end) || buff.findString("\n", start, end));
}

// Find the next newline in some raw data, also checking for split lines
bool HTTPMsg::findNewLine(RawBuff& buff, const unsigned char*& start,
	const unsigned char*& end, bool& split) const
{
	split = false;

	// Try looking for "\r\n"
	bool res = buff.findString("\r\n", start, end);
	if(!res) {
		// Try looking for just "\n"
		res = buff.findString("\n", start, end);
		if(!res) {
			// Try looking for lines joined with ","
			res = buff.findString(",", start, end);
			if(res)
				split = true;
		}
	}

	return res;
}


// Construct the longest string we can from a raw data buffer
std::string HTTPMsg::makeString(const unsigned char* start,
	const unsigned char* end) const
{
	std::string::size_type maxSize = end - start;

	// Does the raw buffer contain a null terminator?
	if(memchr(start, 0, maxSize) != 0) {
		return std::string((const char*)start);
	}

	return std::string((const char*)start, maxSize);
}


// Get header string
std::string HTTPMsg::getHeader(void) const
{
	std::string header = m_generalHeader + m_reqRespHeader + m_entityHeader;
	if(m_parms->getUseHeaderFields())
		header += m_ourHeader;
	return header;
}


}	// namespace HTTPTunnel

