/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP POST request message.
 *
 * $Id: HTTPPostMsg.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <sstream>
#include <iostream>
#include "HTTPPostMsg.hh"
#include "Exceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "ClientSideConfig.hh"
#include "HTTPMsgParms.hh"
#include "Payload.hh"


namespace HTTPTunnel {


// Reference to the logger
extern Logger* g_pLogger;


// Construction/destruction
HTTPPostMsg::HTTPPostMsg()
: HTTPRequestMsg()
{
	// Construct headers
	std::stringstream enthdr;
	enthdr << "Content-Length: 0\r\n";
	m_entityHeader += enthdr.str();
}

HTTPPostMsg::HTTPPostMsg(const HTTPMsgParms* parms, std::string uri,
	const ClientSideConfig* cfg)
: HTTPRequestMsg(parms, uri, cfg)
{
	// Construct headers
	std::stringstream enthdr;
	enthdr << "Content-Length: " << m_parms->getPayload()->getPayload().getLength() << "\r\n";
	m_entityHeader += enthdr.str();

	// Construct request line
	std::stringstream req;
	req << "POST " << getRequestURI() << " HTTP/" << m_parms->getVer();
	m_requestLine = req.str();
}

HTTPPostMsg::HTTPPostMsg(const HTTPPostMsg& that)
: HTTPRequestMsg(that)
{
}

HTTPPostMsg::~HTTPPostMsg()
{
}


// Construction from a raw read data stream
HTTPPostMsg::HTTPPostMsg(RawBuff& readdata) throw(HTTPMsgParsingError)
: HTTPRequestMsg()
{
	/* Look for basic components of POST message header -
		POST		- message identifier
		/X.htm	- URI
		HTTP/x.y	- protocol version
	*/

	const unsigned char *postStart, *postEnd;
	if(!readdata.findString("POST", postStart, postEnd))
		throw HTTPMsgWrongMsgType("POST message identifier not found");

	// Extract URI
	const unsigned char *uriStart1, *uriStart2, *uriEnd;
	if(readdata.findString("/", uriStart1, uriEnd)) {
		if(readdata.findString(".htm", uriStart2, uriEnd)) {
			// Get URI
			m_uri = makeString(uriStart1, uriEnd);
		}
		else
			throw HTTPMsgParsingError("POST end of URI not found", m_parsingLine);
	}
	else
		throw HTTPMsgParsingError("POST start of URI not found", m_parsingLine);

	// Extract protocol version
	const unsigned char *httpStart, *httpEnd;
	if(!readdata.findString("HTTP/", httpStart, httpEnd))
		throw HTTPMsgParsingError("POST protocol id not found", m_parsingLine);

	const unsigned char *nlStart, *nlEnd;
	if(!findNewLine(readdata, nlStart, nlEnd))
		throw HTTPMsgParsingError("Expected newline not found", m_parsingLine);
	m_parms->setVer(makeString(httpEnd, nlStart));

	// Copy entire request line
	m_requestLine = makeString(getStart, nlStart);

	// Discard everything in the raw data stream from the first line
	readdata.erase(nlEnd);
	m_parsingLine++;

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "HTTPPostMsg::HTTPPostMsg() - Got URI: " << m_uri;
	nfo << ", protocol: " << m_parms->getVer();
	DPRINTV(nfo.str());
	#endif

	// Let base classes look for their headers
	bool eoh;
	std::string hdrline;
	parseHdrs(readdata, eoh, hdrline);

	// Expect final newline to terminate non-content part of message
	if(!findNewLine(readdata, nlStart, nlEnd))
		throw HTTPMsgParsingError("POST final newline terminator not found", m_parsingLine);
	readdata.erase(nlEnd);
	m_parsingLine++;

	// Now look far any content
	extractContent(readdata);
}

// Assignment
HTTPPostMsg& HTTPPostMsg::operator=(const HTTPPostMsg& that)
{
	HTTPRequestMsg::operator=(that);
	return *this;
}


/********************/
/* Public Interface */
/********************/


// Write ourselves out as a stream of bytes
void HTTPPostMsg::output(RawBuff& dest)
{
	// Base class stuff
	HTTPRequestMsg::output(dest);

	// Write out payload data
	dest.append(m_parms->getPayload()->getPayload());
}


/***********************/
/* Protected interface */
/***********************/


// Parse and extract any message content
void HTTPPostMsg::extractContent(RawBuff& readdata) throw(ParsingError)
{
	DPRINTV("HTTPPostMsg::extractContent");

	// Are we expecting any content data?
	if(m_repConLen > 0) {
		#ifdef DEBUG_TUN
		std::stringstream str;
		str << "HTTPPostMsg::extractContent() - Expecting " << m_repConLen;
		str << " bytes of content";
		DPRINTV(str.str());
		#endif

		if(readdata.getLength() >= m_repConLen) {
			// Copy content data to end of message
			m_parms->getPayloadMutable()->getPayloadMutable().append(readdata, m_repConLen);

			// Remove extracted data from buffer
			readdata.erase(m_repConLen);
		}
		else {
			#ifdef DEBUG_TUN
			std::stringstream str;
			str << "HTTPPostMsg::extractContent() - Only " << readdata.getLength();
			str << " bytes left in buffer";
			DPRINTV(str.str());
			#endif

			throw NotEnoughContentParsingError("Not enough content data");
		}
	}
}


// Attempt to parse headers from a raw byte-stream
void HTTPPostMsg::parseHdrs(RawBuff& data, bool& eoh, std::string& hdrline)
	throw(HTTPMsgParsingError)
{
	/* Look for any POST headers - everything else is handled elsehere */

	bool ok = true;
	eoh = false;

	// Parse as much as we can
	while(ok && !eoh) {
		// Let base class have a go
		HTTPRequestMsg::parseHdrs(data, eoh, hdrline);
		if(!eoh) {
			// Base class didn't like a header and we have no headers, ignore it
		}
	}
}


}	// namespace HTTPTunnel

