/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * An HTTP data reader.
 *
 * $Id: HTTPReader.cc,v 1.2 2008-07-25 12:14:53 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "HTTPReader.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "Connection.hh"
#include "HTTPGetMsg.hh"
#include "HTTPResponseMsg.hh"
#include "TunnelID.hh"
#include "Tunneler.hh"
#include "AutoPtr.hh"
#include "MsgArrivedEvent.hh"
#include "SideConfig.hh"
#include "Server.hh"


namespace HTTPTunnel {


#ifdef DEBUG_LOG_DATA
// Reference to log file
extern std::ofstream* g_pPayloadReadLogFile;
extern unsigned int g_PayloadReadChars;
#endif

// Global references
extern Logger* g_pLogger;


// Construction/destruction
HTTPReader::HTTPReader(Connection* conn)
: Reader(conn), m_buff(), m_parseState(UNDEFINED), m_msgErrorCount(0),
  m_lastRecvdData(0), m_nextReqdData(0), m_lastErrLine(0)
{
	DPRINTV("HTTPReader::HTTPReader()");

	// Set message parse state
	if(m_conn->getTunneler()->getServer()->getConfig()->isClient())
		m_parseState = RECEIVE;
	else
		m_parseState = CONNECT;
}

HTTPReader::~HTTPReader()
{
	DPRINTV("HTTPReader::~HTTPReader()");

	#ifdef DEBUG_TUN
	if(m_buff.getLength() > 0) {
		DPRINTW("HTTPReader::~HTTPReader() - Error! Residual read data");
	}
	#endif
}


/*************/
/* Accessors */
/*************/


// Do we have unprocessed data remaining?
bool HTTPReader::dataRemaining(void) const
{
	return (m_buff.getLength() > 0);
}


/***********************/
/* Protected Interface */
/***********************/


// Process some newly read raw data - Caller holds mutex
void HTTPReader::processData(const unsigned char* data, unsigned int length)
	throw(ReadError)
{
	// Append data to parse buffer
	m_buff.append(data, length);

	#ifdef DEBUG_TUN
	std::stringstream err1;
	err1 << "HTTPReader::processData() - Before parse, buff contains ";
	err1 << m_buff.getLength() << " bytes";
	DPRINTV(err1.str());
	#endif

	// Do we have an indication of how much data we're going to need to
	// contruct this HTTP msg?
	if(m_nextReqdData > 0) {
		if(m_buff.getLength() <= m_nextReqdData) {
			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Not enough data to attempt parse, need at least " << m_nextReqdData;
			DPRINTV(dbg.str());
			#endif

			return;
		}
	}

	// Somewhere to copy raw data buffer while attempting HTTP msg construction
	RawBuff parseBuffCopy(m_buff);

	try {
		// Get tunneler reference
		Tunneler* tunneler = m_conn->getTunneler();

		// Try to parse and process as many messages as we can from our current data
		while(parseBuffCopy.getLength() > 0) {

			// Ptr to created message
			AutoPtr<HTTPMsg> msg;

			// What message do we expect next?
			switch(m_parseState) {
				case(CONNECT): {
					// Try to parse & create a GET msg
					msg = new HTTPGetMsg(parseBuffCopy);
					DPRINTV("HTTPReader::processData() - Processed GET connect");

					// Now we just send
					m_parseState = SEND;

					break;
				}

				case(SEND): {
					// Try to parse & create a GET msg
					msg = new HTTPGetMsg(parseBuffCopy);
					DPRINTV("HTTPReader::processData() - Processed GET");
					break;
				}

				case(RECEIVE): {
					// Try to parse & create a response msg
					HTTPResponseMsg* respMsg = new HTTPResponseMsg(parseBuffCopy);
					msg = respMsg;
					DPRINTV("HTTPReader::processData() - Processed RESPONSE msg");

					// Check response code
					checkMsgResponseCode(respMsg);

					break;
				}

				default: {
					throw ReadError("Unhandled parse state in HTTPReader");
				}
			}

			/************************************/
			/* Message type agnostic processing */
			/************************************/

			#ifdef DEBUG_LOG_DATA
			// Log read payload data
			Payload* payload = msg->getParmsMutable()->getPayloadMutable();
			logData(payload->getPayload(), g_pPayloadReadLogFile, g_PayloadReadChars);
			#endif

			// Schedule message arrival event for tunneler
			msg.freeOnDelete(false);
			Event* evnt = new MsgArrivedEvent(m_conn->getID(), tunneler, msg.getMutable());
			tunneler->scheduleEvent(evnt);

			// No error in processing, so purge processed data
			m_buff = parseBuffCopy;
			m_msgErrorCount = 0;
			m_lastRecvdData = 0;
			m_nextReqdData = 0;
		}
	}
	catch(NotEnoughDataParsingError& e) {
		#ifdef DEBUG_TUN
		std::stringstream err;
		err << "Not enough data to create HTTP msg: " << e.what();
		DPRINTV(err.str());
		#endif

		// Are we progressing towards having enough data?
		if(!(e.m_recvd > m_lastRecvdData)) {
			// We don't seem to be progressing, treat as a serious error
			if(msgError(PARSING_ERROR_MAJOR)) {
				throw ReadError("Too many errors in HTTP data");
			}
		}
		m_lastRecvdData = e.m_recvd;
		m_nextReqdData = e.m_reqd;
	}
	catch(BadContentLengthParsingError& e) {
		#ifdef DEBUG_TUN
		std::stringstream err;
		err << "Bad encoded data length in HTTP msg: " << e.what();
		DPRINTV(err.str());
		#endif

		// Error in communications
		throw ReadError("Error processing HTTP data");
	}
	catch(HTTPMsgParsingError& e) {
		#ifdef DEBUG_TUN
		std::stringstream err;
		err << "Parsing error for HTTP msg on line " << e.m_line;
		err << ": " << e.what();
		DPRINTV(err.str());
		#endif

		// Are we progressing through msg?
		if(!(e.m_line > m_lastErrLine)) {
			// We don't seem to be progressing, treat as a serious error
			if(msgError(PARSING_ERROR_MAJOR)) {
				throw ReadError("Too many errors in HTTP data");
			}
		}
		m_lastErrLine = e.m_line;
	}
	catch(ParsingError& e) {
		#ifdef DEBUG_TUN
		std::stringstream err;
		err << "General error creating HTTP msg: " << e.what();
		DPRINTV(err.str());
		#endif

		// Error in communications
		throw ReadError("Error processing HTTP data");
	}
	catch(InvalidID& e) {
		std::stringstream err;
		err << "Error constructing ID: " << e.what();
		g_pLogger->error(err.str());

		// Re-throw
		throw ReadError("Error processing read ID");
	}

	#ifdef DEBUG_TUN
	std::stringstream err2;
	err2 << "HTTPReader::processData() - After parse, buff contains ";
	err2 << m_buff.getLength() << " bytes";
	DPRINTV(err2.str());
	#endif
}


// Check a read message's HTTP response code
void HTTPReader::checkMsgResponseCode(HTTPResponseMsg* msg) throw(ReadError)
{
	bool ok = false;
	unsigned short respCode = msg->getResponseCode();

	// Everything OK?
	if(respCode == 200) {
		// Normal response
		ok = true;
	}
	else if(respCode == 401 || respCode == 407) {
		// Authentication required is not necessarily an error
		ok = true;
	}
	else if(respCode == 404) {
		// Probably a proxy connection error, not necessarily an error
		ok = true;
	}

	// Throw exception for bad response codes
	if(!ok) {
		#ifdef DEBUG_TUN
		if(respCode == 500) {
			// Server error response
			DPRINTV("Got HTTP error response 500 from server");
		}
		#endif

		// Tell caller
		std::stringstream err;
		err << "Got HTTP response code " << respCode;
		throw BadResponseCodeError(err.str(), respCode);
	}
}


// Deal with an error in HTTP msg processing
bool HTTPReader::msgError(unsigned int inc)
{
	// Increment error count
	m_msgErrorCount += inc;

	// Should we give up?
	if(m_msgErrorCount >= MAX_PROCESSING_ERRORS)
		return true;
	return false;
}


}	// namespace

