/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP request message.
 *
 * $Id: HTTPRequestMsg.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <cstring>
#include <sstream>
#include <iostream>
#include "HTTPRequestMsg.hh"
#include "Exceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "ClientSideConfig.hh"
#include "ProxySettings.hh"
#include "ServerAuthSettings.hh"
#include "HTTPMsgParms.hh"
#include "Utils.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
HTTPRequestMsg::HTTPRequestMsg()
: HTTPMsg(), m_uri(), m_requestLine(), m_cfg(0)
{
}

HTTPRequestMsg::HTTPRequestMsg(const HTTPMsgParms* parms, std::string uri,
	const ClientSideConfig* cfg)
: HTTPMsg(parms), m_uri(uri), m_requestLine(), m_cfg(cfg)
{
	// Construct headers
	std::stringstream rrhdr;

	// Host address
	rrhdr << "Host: " << m_cfg->getResolvedServerIP();
	if(m_cfg->getServerPort() != 80)
		rrhdr << ":" << m_cfg->getServerPort();
	rrhdr << "\r\n";

	// Via proxy?
	if(m_cfg->viaProxy()) {
		ProxySettings* pxySettings = m_parms->getProxySettings();
		if(pxySettings)
			rrhdr << m_parms->getProxySettings()->getRequestHeader(uri);
	}

	// Server authentication
	if(!m_parms->getUserAuthenticated()) {
		rrhdr << m_parms->getServerAuthSettings()->getRequestHeader(uri);
	}

	m_reqRespHeader += rrhdr.str();
}

HTTPRequestMsg::HTTPRequestMsg(const HTTPRequestMsg& that)
: HTTPMsg(that), m_uri(that.m_uri), m_requestLine(that.m_requestLine)
{
}

HTTPRequestMsg::~HTTPRequestMsg()
{
}

// Assignment
HTTPRequestMsg& HTTPRequestMsg::operator=(const HTTPRequestMsg& that)
{
	HTTPMsg::operator=(that);
	m_uri = that.m_uri;
	m_requestLine = that.m_requestLine;
	return *this;
}


/********************/
/* Public Interface */
/********************/


/***********************/
/* Protected interface */
/***********************/


// Get start line string
std::string HTTPRequestMsg::getStartLine(void) const
{
	// Base class has no start line, so just return ours
	return m_requestLine;
}


// Attempt to parse headers from a raw byte-stream
void HTTPRequestMsg::parseHdrs(RawBuff& data, bool& eoh, std::string& hdrline)
	throw(HTTPMsgParsingError)
{
	bool ok = true;
	eoh = false;

	// Parse as much as we can
	while(ok && !eoh) {
		// First let base class have a go
		HTTPMsg::parseHdrs(data, eoh, hdrline);
		if(!eoh) {
			/* Base class didn't like a header, let us try */

			// Is it one of ours?
			if(strncmp(hdrline.c_str(), "Authorization:", strlen("Authorization:")) == 0) {
				// Response from client to authenticate itself
				std::string::size_type scmBeg = hdrline.find(' ');
				if(scmBeg != std::string::npos) {
					// Get authentication scheme used
					std::string authLine(hdrline, scmBeg + 1);
					std::string::size_type scmEnd = authLine.find(' ');
					std::string schemeName = authLine.substr(0, scmEnd);

					// Get authentication arguments
					std::vector<std::string> argList;
					if(scmEnd != std::string::npos) {
						std::string schemeArgs(authLine, scmEnd + 1);
						Utils::SplitString(schemeArgs, " ", argList);

						// Strip redundant commas from the end of any parms
						for(unsigned int i = 0; i < argList.size(); ++i) {
							std::string arg = argList[i];
							if(arg.size() > 0) {
								if(arg[arg.size() - 1] == ',')
									arg = arg.substr(0, arg.size() - 1);
								argList[i] = arg;
							}
						}
					}

					// Save parsed authentication scheme details
					m_authArgs[schemeName] = argList;
				}
				else {
					throw HTTPMsgParsingError("Failed to parse server auth from response",
						m_parsingLine);
				}
			}

			else {
				// Unknown header
				ok = false;
			}
		}
	}
}


// Get URL to include in request
std::string HTTPRequestMsg::getRequestURI(void) const
{
	std::stringstream uri;

	// Are we using a proxy?
	if(m_cfg && m_cfg->viaProxy())
		uri << "http://" << m_cfg->getServerIP() << ":" << m_cfg->getServerPort();

	// Append base uri
	uri << m_uri;

	// Are we using our header fields?
	if(!m_parms->getUseHeaderFields() || m_parms->getState() == Tunnel::CONNECT) {
		// Need to encode tunneling header fields in URI
		uri << "?" << Utils::URLEncode(m_ourHeader);
	}

	return uri.str();
}


}	// namespace HTTPTunnel

