/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP response message.
 *
 * $Id: HTTPResponseMsg.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <cstring>
#include <sstream>
#include <iostream>
#include "HTTPResponseMsg.hh"
#include "Exceptions.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "ServerSideConfig.hh"
#include "HTTPMsgParms.hh"
#include "Payload.hh"
#include "Utils.hh"
#include "ServerAuthSettings.hh"
#include "Quotas.hh"


namespace HTTPTunnel {


// Reference to the logger
extern Logger* g_pLogger;


// Construction/destruction
HTTPResponseMsg::HTTPResponseMsg()
: HTTPMsg(), m_responseCode(200), m_responsePhrase("OK"), m_responseLine(),
  m_repContentLen(0)
{
	// Construct headers
	std::stringstream enthdr;
	enthdr << "Content-Length: 0\r\n";
	m_entityHeader += enthdr.str();
}

HTTPResponseMsg::HTTPResponseMsg(const HTTPMsgParms* parms, ServerSideConfig* cfg)
: HTTPMsg(parms), m_responseCode(parms->getRespCode()),
  m_responsePhrase(parms->getRespPhrase()), m_responseLine(), m_repContentLen(0)
{
	// Construct response line
	std::stringstream resp;
	resp << "HTTP/" << m_parms->getVer() << " " << m_responseCode;
	resp << " " << m_responsePhrase;
	m_responseLine = resp.str();

	// Server authentication fields
	if(!m_parms->getUserAuthenticated()) {
		ServerAuthSettings* authSettings = m_parms->getServerAuthSettings();
		if(authSettings)
			m_reqRespHeader += authSettings->getResponseHeader(m_parms->getURI());
	}

	// AuthReq fields
	if(m_parms->getState() == Tunnel::AUTHREQ) {
		// Quotas
		std::stringstream quotastr;
		quotastr << "HT-Quo: ";
		quotastr << m_parms->getUser().getQuotas() << "\r\n";
		m_ourHeader += quotastr.str();
	}

	// Not using headers for tunneling?
	if(!m_parms->getUseHeaderFields()) {
		// Include tunneling headers in payload
		RawBuff ourHeaders(m_ourHeader);
		ourHeaders.append(0);
		m_parms->getPayloadMutable()->getPayloadMutable().prepend(ourHeaders);
	}

	// Construct headers
	std::stringstream enthdr;
	enthdr << "Content-Length: " << m_parms->getPayload()->getPayload().getLength() << "\r\n";
	m_entityHeader += enthdr.str();
}

HTTPResponseMsg::HTTPResponseMsg(const HTTPResponseMsg& that)
: HTTPMsg(that), m_responseCode(that.m_responseCode),
  m_responsePhrase(that.m_responsePhrase), m_responseLine(that.m_responseLine),
  m_repContentLen(that.m_repContentLen)
{
}

HTTPResponseMsg::~HTTPResponseMsg()
{
}

// Construction from a raw read data stream
HTTPResponseMsg::HTTPResponseMsg(RawBuff& readdata) throw(HTTPMsgParsingError)
: HTTPMsg()
{
	const unsigned char *start, *end;

	/* Look for basic components of response message header -
		HTTP/x.y - protocol version
		int		- response code
		string	- response phrase
	*/

	if(!readdata.findString("HTTP/", start, end)) {
		throw HTTPMsgParsingError("Response protocol id not found", m_parsingLine);
	}

	const unsigned char* verStart = end;
	if(!readdata.findString(" ", start, end)) {
		throw HTTPMsgParsingError("Space after response protocol id not found",
			m_parsingLine);
	}

	// Extract protocol version
	m_parms->setVer(makeString(verStart, start));

	const unsigned char *nlStart, *nlEnd;
	if(!findNewLine(readdata, nlStart, nlEnd))
		throw HTTPMsgParsingError("Expected newline not found", m_parsingLine);

	// Extract response code and phrase
	std::string codePhraseStr = makeString(end, readdata.end());
	int code;
	char phrase[readdata.getLength()];
	if(sscanf(codePhraseStr.c_str(), "%d %s", &code, phrase) != 2)
		throw HTTPMsgParsingError("Response code/phrase not found", m_parsingLine);

	// Copy entire response line, and then we've finished with it
	m_responseLine = makeString(start, nlEnd);
	m_responseCode = code;
	m_responsePhrase = phrase;
	readdata.erase(nlEnd);
	m_parsingLine++;

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "HTTPResponseMsg::HTTPResponseMsg() - Got response code: " << m_responseCode;
	nfo << ", protocol: " << m_parms->getVer();
	DPRINTV(nfo.str());
	#endif

	// Look for headers
	bool eoh;
	std::string hdrline;
	parseHdrs(readdata, eoh, hdrline);

	// Expect final newline to terminate non-content part of message
	if(!findNewLine(readdata, start, end)) {
		throw HTTPMsgParsingError("Response final newline terminator not found",
			m_parsingLine);
	}
	readdata.erase(end);
	m_parsingLine++;

	// Now look for content
	extractContent(readdata);

	// Did we see our headers?
	if(!allRespHeaders()) {
		// Normal response?
		if(m_responseCode == 200) {
			// Remember we seem to be using non-header based tunneling
			m_parms->setUseHeaderFields(false);

			// Look for end of headers in payload marker
			const unsigned char* headersEnd = 0;
			RawBuff& payload = m_parms->getPayloadMutable()->getPayloadMutable();
			if(payload.findChar(0, headersEnd)) {
				// Extract headers from payload & try to parse them
				unsigned int headersLength = headersEnd - payload.begin();
				RawBuff payloadHeaders(payload, headersLength);
				payloadHeaders.append("\r\n", 2);
				parseHdrs(payloadHeaders, eoh, hdrline);

				// Remove headers from proper payload
				payload.erase(headersLength + 1);
			}
			else {
				throw HTTPMsgParsingError("End of headers in payload not found",
					m_parsingLine);
			}
		}
	}
}

// Assignment
HTTPResponseMsg& HTTPResponseMsg::operator=(const HTTPResponseMsg& that)
{
	HTTPMsg::operator=(that);
	m_responseCode = that.m_responseCode;
	m_responsePhrase = that.m_responsePhrase;
	m_responseLine = that.m_responseLine;
	m_repContentLen = that.m_repContentLen;
	return *this;
}


/********************/
/* Public Interface */
/********************/



// Write ourselves out as a stream of bytes
void HTTPResponseMsg::output(RawBuff& dest)
{
	// Base class stuff
	HTTPMsg::output(dest);

	// Write out main payload data
	dest.append(m_parms->getPayload()->getPayload());
}


// Get our response code
unsigned short HTTPResponseMsg::getResponseCode(void) const
{
	return m_responseCode;
}


/***********************/
/* Protected interface */
/***********************/


// Parse and extract any message content
void HTTPResponseMsg::extractContent(RawBuff& readdata) throw(ParsingError)
{
	DPRINTV("HTTPResponseMsg::extractContent");

	// Are we expecting any content data?
	if(m_repContentLen > 0) {
		#ifdef DEBUG_TUN
		std::stringstream str;
		str << "HTTPResponseMsg::extractContent() - Expecting " << m_repContentLen;
		str << " bytes of content";
		DPRINTV(str.str());
		#endif

		unsigned int recvdLen = readdata.getLength();
		if(recvdLen >= m_repContentLen) {
			// Copy content data to end of message
			m_parms->getPayloadMutable()->getPayloadMutable().append(readdata, m_repContentLen);

			// Remove extracted data from buffer
			readdata.erase(m_repContentLen);
		}
		else {
			#ifdef DEBUG_TUN
			std::stringstream str;
			str << "HTTPResponseMsg::extractContent() - Only " << recvdLen;
			str << " bytes left in buffer";
			DPRINTV(str.str());
			#endif

			throw NotEnoughDataParsingError("Not enough content", m_repContentLen, recvdLen,
				m_parsingLine);
		}
	}
}


// Get start line string
std::string HTTPResponseMsg::getStartLine(void) const
{
	// Base class has no start line, so just return ours
	return m_responseLine;
}


// Attempt to parse headers from a raw byte-stream
void HTTPResponseMsg::parseHdrs(RawBuff& data, bool& eoh, std::string& hdrline)
	throw(HTTPMsgParsingError)
{
	bool ok = true;
	eoh = false;

	// Parse as much as we can
	while(ok && !eoh) {
		// First let base class have a go
		HTTPMsg::parseHdrs(data, eoh, hdrline);
		if(!eoh) {
			/* Base class didn't like a header, let us try */
					
			// Is it one of ours?
			if(strncmp(hdrline.c_str(), "Proxy-Authenticate:", strlen("Proxy-Authenticate:")) == 0 ||
			strncmp(hdrline.c_str(), "WWW-Authenticate:", strlen("WWW-Authenticate:")) == 0) {
				// Request from proxy/server to authenticate ourselves
				std::string::size_type scmBeg = hdrline.find(' ');
				if(scmBeg != std::string::npos) {
					// Get suggested authentication scheme
					std::string authLine(hdrline, scmBeg + 1);
					std::string::size_type scmEnd = authLine.find(' ');
					std::string schemeName = authLine.substr(0, scmEnd);

					// Get authentication arguments
					std::vector<std::string> argList;
					if(scmEnd != std::string::npos) {
						std::string schemeArgs(authLine, scmEnd + 1);
						Utils::SplitString(schemeArgs, " ", argList);

						// Strip redundant commas from the end of any parms
						for(unsigned int i = 0; i < argList.size(); ++i) {
							std::string arg = argList[i];
							if(arg.size() > 0) {
								if(arg[arg.size() - 1] == ',')
									arg = arg.substr(0, arg.size() - 1);
								argList[i] = arg;
							}
						}
					}

					// Save parsed authentication scheme details
					m_authArgs[schemeName] = argList;
				}
				else {
					throw HTTPMsgParsingError("Failed to parse proxy/server auth from response",
						m_parsingLine);
				}
			}
			else if(strncmp(hdrline.c_str(), "HT-Quo:", strlen("HT-Quo:")) == 0) {
				// Tunnel quotas
				std::string::size_type i = hdrline.find(' ');
				if(i != std::string::npos) {
					std::string quotastr(hdrline, i + 1);
					DPRINTV("HTTPResponseMsg::parseHdrs() - Got quotas: " + quotastr);
					Quotas quotas(quotastr);
					m_parms->setQuotas(quotas);
					m_seenQuotas = true;
					m_lastTunId = "HT-Quo: ";
				}
				else {
					throw HTTPMsgParsingError("Failed to parse quotas from msg", m_parsingLine);
				}
			}
			else if(strncmp(hdrline.c_str(), "Content-Length:", strlen("Content-Length:")) == 0) {
				// Get reported content length
				if(sscanf(hdrline.c_str(), "Content-Length: %d", &m_repContentLen) != 1) {
					throw HTTPMsgParsingError("Failed to parse content length from msg",
						m_parsingLine);
				}
			}
		}
	}
}


}	// namespace HTTPTunnel

