/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * HTTP Tunneler instance. Top level library interface.
 *
 * $Id: HTTPTunneler.hh,v 1.6 2009-03-20 13:25:35 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _HTTP_TUNNELER_HH_
#define _HTTP_TUNNELER_HH_

#include <string>
#include "Mutex.hh"
#include "Exceptions.hh"
#include "GlobalConfigBlob.hh"
#include "ClientConfigBlob.hh"
#include "SockTunClientConfigBlob.hh"
#include "ServerConfigBlob.hh"
#include "SockTunServerConfigBlob.hh"
#include "TunnelerID.hh"
#include "SockTunCallback.hh"
#include "ClientSockTunHandle.hh"
#include "ServerSockTunHandle.hh"

namespace HTTPTunnel {

class HTTPTunneler {
public:

	/**********************************/
	/* HTTPTunneler instance handling */
	/**********************************/

	// Get/Release singleton instance
	static HTTPTunneler* Get(void) throw(TunnelError);
	static void Release(void) throw(TunnelError);

	// Set name to use when logging
	static void SetName(std::string name);

	// Get version details
	static std::string GetLibVersion(void);
	static std::string GetLibBuildName(void);


	/*******************************/
	/* Control the HTTP tunnelling */
	/*******************************/

	// Start/Stop the background HTTP tunnelling
	void Start(void) throw(TunnelError);
	void Stop(void) throw(TunnelError);


	/*********************************/
	/* Configure the HTTP tunnelling */
	/*********************************/

	// Configure from a file
	void ConfigureFromFile(std::string fileName) throw(TunnelError);
	void ReconfigureFromFile(std::string fileName) throw(TunnelError);

	// Regenerate the named config file with defaults
	void RegenerateCfgFile(std::string fileName) throw(TunnelError);

	// Configure directly
	void ConfigureGlobal(const GlobalConfigBlob& blob) throw(TunnelError);
	TunnelerID ConfigureClientTunneler(const ClientConfigBlob& blob) throw(TunnelError);
	TunnelerID ConfigureServerTunneler(const ServerConfigBlob& blob) throw(TunnelError);

	// Initiate a clean shutdown on a configured tunneler
	void ShutdownTunneler(const TunnelerID& tid, bool block = false) throw(TunnelError);

	// Remove configured tunneler directly
	void RemoveTunneler(const TunnelerID& tid) throw(TunnelError);


	/********************************/
	/* Socket based HTTP tunnelling */
	/********************************/

	// Launch a HTTP tunneling server socket
	ServerSockTunHandle LaunchServerSocket(const SockTunServerConfigBlob& blob)
		throw(TunnelError);

	// Launch a HTTP tunneling client socket
	ClientSockTunHandle LaunchClientSocket(const SockTunClientConfigBlob& blob)
		throw(TunnelError);


	/***********************************/
	/* Query the HTTP tunnelling state */
	/***********************************/

	// Is the specified tunneler currently active?
	bool IsTunnelerActive(const TunnelerID& tid) throw(TunnelError);


private:

	/***********/
	/* Methods */
	/***********/

	// Construction/destruction
	HTTPTunneler() throw(TunnelError);
	~HTTPTunneler() throw(TunnelError);

	// Perform one-off initialisation/shutdown
	void initialise(void) throw(TunnelError);
	void shutdown(void) throw(TunnelError);

	// PID file management
	void createPIDFile(void) throw(TunnelError);
	void removePIDFile(void);

	// Daemonise
	void daemonise(void) throw(TunnelError);

	#ifdef DEBUG_LOG_DATA
	// Open data debugging logs
	void openDebugLogs(void);
	#endif


	/********/
	/* Data */
	/********/

	// Mutex
	Mutex								m_mtx;

	// Name of any PID file
	std::string						m_pidFileName;

	// Flag to indicate whether we have started running
	bool								m_running;

	// Flag to indicate if we have daemonised
	bool								m_daemonised;

	// Singleton instance
	static HTTPTunneler*			c_tunneler;
	static Mutex					c_mtx;

	// Name to use when logging
	static std::string			c_name;

};


}	// namespace

#endif

