/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * An HTTP data writer.
 *
 * $Id: HTTPWriter.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "HTTPWriter.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Exceptions.hh"
#include "Connection.hh"
#include "HTTPGetMsg.hh"
#include "HTTPResponseMsg.hh"
#include "Tunneler.hh"
#include "TunnelID.hh"
#include "HTTPMsgParms.hh"
#include "MutexVar.hh"
#include "Server.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"


namespace HTTPTunnel {


// Reference to log file
#ifdef DEBUG_LOG_DATA
extern std::ofstream* g_pPayloadWriteLogFile;
extern unsigned int g_PayloadWriteChars;
#endif

// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Initialise static data
std::vector<unsigned short> HTTPWriter::c_requestStr;
Mutex HTTPWriter::c_requestStrMtx;


// Construction/destruction
HTTPWriter::HTTPWriter(Connection* conn)
: Writer(conn), m_msgState(UNDEFINED)
{
	DPRINTV("HTTPWriter::HTTPWriter()");

	// Set message creation state
	if(m_conn->getTunneler()->getServer()->getConfig()->isClient())
		m_msgState = INIT_CONNECT;
	else
		m_msgState = RESP_CONNECT;
}

HTTPWriter::~HTTPWriter()
{
	DPRINTV("HTTPWriter::~HTTPWriter()");
}


/********************/
/* Public Interface */
/********************/


/***********************/
/* Protected Interface */
/***********************/


// Attempt to process some data to send out
void HTTPWriter::processData(MsgParms* parms, RawBuff& out) throw(WriteError)
{
	#ifdef DEBUG_LOG_DATA
	// Log data
	logData(parms->getPayload()->getPayload(), g_pPayloadWriteLogFile, g_PayloadWriteChars);
	#endif

	// Check we have HTTP msg parameters
	HTTPMsgParms* httpParms = dynamic_cast<HTTPMsgParms*>(parms);
	if(httpParms) {
		// Create HTTP msg
		HTTPMsg* msg = createMsg(httpParms);
		if(msg) {
			// Write message out as a byte stream to destination buffer
			msg->output(out);

			// Finished with message
			delete msg;
		}
	}
	else {
		throw WriteError("HTTPWriter given wrong type of MsgParms");
	}
}


// Construct appropriate HTTP message to encapsulate some data
HTTPMsg* HTTPWriter::createMsg(HTTPMsgParms* parms) throw(WriteError)
{
	// Construct HTTP message of appropriate type containing our pending data
	HTTPMsg* msg = 0;

	if(m_msgState == INIT_CONNECT) {
		// Include our version ID in msg to server
		parms->setCliVer(g_pConfig->getVersionStr());

		// Create get connect msg
		DPRINTV("HTTPWriter::createMsg() - CONNECTING with GET");
		msg = new HTTPGetMsg(parms, createRequestURL(),
			m_conn->getTunneler()->getClientSideConfig());
		m_msgState = SENDING;
	}
	else if(m_msgState == RESP_CONNECT) {
		// Create response msg
		DPRINTV("HTTPWriter::createMsg() - RESPONDING to connect");
		msg = new HTTPResponseMsg(parms,
			m_conn->getTunneler()->getServerSideConfigMutable());
		m_msgState = RECEIVING;
	}
	else if(m_msgState == SENDING) {
		// Create get msg
		DPRINTV("HTTPWriter::createMsg() - SENDING with GET");
		msg = new HTTPGetMsg(parms, createRequestURL(),
			m_conn->getTunneler()->getClientSideConfig());
	}
	else if(m_msgState == RECEIVING) {
		// Create response msg
		DPRINTV("HTTPWriter::createMsg() - RECEIVING with 200 OK");
		msg = new HTTPResponseMsg(parms,
			m_conn->getTunneler()->getServerSideConfigMutable());
	}
	else {
		DPRINTV("HTTPWriter::createMsg() - Error: Unknown state");
		throw WriteError("Attempt to get pending data while in unknown state");
	}

	return msg;
}

// Create a new URL for some request
std::string HTTPWriter::createRequestURL(void)
{
	static const char alphabet[] = { 'a','b','c','d','e','f','g','h','i',
		'j', 'k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
		'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
		'R','S','T','U','V','W','X','Y','Z','0','1','2','3','4','5','6','7',
		'8','9' };

	// Increment file string
	MutexVar mv(c_requestStrMtx);
	bool carry;
	unsigned int col = 0;
	do {
		carry = false;

		if(c_requestStr.size() <= col)
			c_requestStr.push_back(0);

		unsigned short val = c_requestStr[col] + 1;
		if(val > sizeof(alphabet)) {
			c_requestStr[col] = 1;
			carry = true;
			col++;
		}
		else {
			c_requestStr[col] = val;
		}
	} while(carry);

	// Build complete URL
	std::stringstream url;
	url << "/";
	for(col = 0; col < c_requestStr.size(); ++col)
		url << alphabet[c_requestStr[col] - 1];
	url << ".htm";

	return url.str();
}


}	// namespace

