/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Perform logging to either stderr or syslog.
 *
 * $Id: Logger.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sys/time.h>
#include <syslog.h>
#include <string>
#include <iostream>
#include <sstream>
#include <time.h>
#include "Logger.hh"
#include "MutexVar.hh"
#include "GlobalConfig.hh"
#include "Debug.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;



/****************************/
/* Construction/destruction */
/****************************/

/* If constructed with syslog=true we will log all messages via
   syslog. If syslog=false we will simply log all messages to
   stderr. */

Logger::Logger(std::string pname, bool time, bool syslog)
: m_procName(pname), m_time(time), m_syslog(syslog), m_mtx()
{
	// Open the syslog channel
	openlog(m_procName.c_str(), 0, 0);
}

Logger::~Logger()
{
	// Close syslog channel
	closelog();
}


/*******************/
/* Logging methods */
/*******************/

// Log an error
void Logger::error(std::string msg)
{
	// Check debug level
	if(getDebugLevel() >= ERROR) {
		// Tag message as an error
		std::string logMsg = "ERROR: " + msg;

		if(m_syslog) {
			syslog(ERR_FACILITY | LOG_ERR, logMsg.c_str());
		}
		else {
			MutexVar mv(m_mtx);
			// Not logging to syslog, send to stderr instead
			if(m_time)
				std::cerr << getTimeStr() << ": ";
			std::cerr << logMsg << std::endl;
		}
	}
}

// Log a warning
void Logger::warn(std::string msg)
{
	// Check debug level
	if(getDebugLevel() >= WARN) {
		// Tag message as a warning
		std::string logMsg = "WARN: " + msg;

		if(m_syslog) {
			syslog(ERR_FACILITY | LOG_WARNING, logMsg.c_str());
		}
		else {
			MutexVar mv(m_mtx);
			// Not logging to syslog, send to stderr instead
			if(m_time)
				std::cerr << getTimeStr() << ": ";
			std::cerr << logMsg << std::endl;
		}
	}
}

// Log an informational message
void Logger::info(std::string msg)
{
	// Check debug level
	if(getDebugLevel() >= INFO) {
		// Tag message as an informational
		std::string logMsg = "INFO: " + msg;

		if(m_syslog) {
			syslog(ERR_FACILITY | LOG_NOTICE, logMsg.c_str());
		}
		else {
			MutexVar mv(m_mtx);
			// Not logging to syslog, send to stderr instead
			if(m_time)
				std::cerr << getTimeStr() << ": ";
			std::cerr << logMsg << std::endl;
		}
	}
}


/*******************/
/* Tracing methods */
/*******************/

// Log a new trace message
void Logger::trace(std::string msg, unsigned int level)
{
	// Check debug level
	if(getDebugLevel() >= level) {
		if(m_syslog) {
			// Send message to syslog
			syslog(TRACE_FACILITY | LOG_NOTICE, msg.c_str());
		}

		// Also send tracing to stderr
		MutexVar mv(m_mtx);
		if(m_time)
			std::cerr << getTimeStr() << ": ";
		std::cerr << msg << std::endl;
	}
}


/*********************/
/* Private Interface */
/*********************/


// Get current debug level
unsigned int Logger::getDebugLevel(void) const
{
	#ifdef DEBUG_TUN
	unsigned int level = VERBOSE;
	#else
	unsigned int level = QUIET;
	#endif

	if(g_pConfig)
		level = g_pConfig->getDebug();

	return level;
}


// Get time as a tring for including in logging
std::string Logger::getTimeStr(void) const
{
	std::stringstream str;
	str << time(0);
	return str.str();
}


}		// namespace HTTPTunnel


