/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Handle an NTLM authentication session.
 *
 * $Id: NTLM.cc,v 1.3 2008-06-26 10:05:32 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstring>
#include <strings.h>
#include <cstdlib>
#include "NTLM.hh"
#include "Utils.hh"
#include "Logger.hh"
#include "GlobalConfig.hh"
#include "Debug.hh"
#include "AuthDigest.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
NTLM::NTLM()
: m_state(NTLM_NEG), m_dataBlock(), m_chalDomain(), m_chalFlags(0), m_chalChallenge(),
  m_chalContext(), m_chalTargets()
{
}

NTLM::~NTLM()
{
}


/*************/
/* Utilities */
/*************/


// Create next NTLM message
bool NTLM::createNextMsg(RawBuff& buff, const std::string& user, const std::string& domain,
	const std::string& passwd) throw(AuthProcessingError)
{
	DPRINTV("NTLM::createNextMsg()");

	bool finished = false;

	// Clear buffers to start
	buff.clear();
	m_dataBlock.clear();

	// What type of msg is next?
	switch(m_state) {
		case(NTLM_NEG): {
			// Create negotiate msg
			createNegotiateMsg(buff, domain);
			m_state = NTLM_CHAL;
			break;
		}
		case(NTLM_CHAL): {
			throw AuthProcessingError("Can't create unsupported NTLM challenge msg");
			break;
		}
		case(NTLM_AUTH): {
			// Create auth msg
			createAuthMsg(buff, user, domain, passwd);
			finished = true;
			break;
		}
		default: {
			throw AuthProcessingError("Can't create unknown NTLM msg type");
			break;
		}
	}

	return finished;
}


// Parse received NTLM message
void NTLM::parseRecvdMsg(const std::string& msg) throw(AuthProcessingError)
{
	DPRINTV("NTLM::parseRecvdMsg()");

	// Base64 decode msg
	RawBuff rawMsg = Utils::Base64Decode(msg);
	RawBuff origMsg = rawMsg;

	// NTLMSSP signature
	std::string sig = decOem(rawMsg, strlen("NTLMSSP"));
	rawMsg.erase(1);
	if(sig != "NTLMSSP")
		throw AuthProcessingError("Missing NTLMSSP signature in NTLM msg");
	DPRINTV("Parsed NTLM signature");

	// Msg type
	unsigned int msgType = decLong(rawMsg);
	if(msgType != (unsigned int)NTLM_CHAL) {
		std::stringstream err;
		err << "Error parsing unknown NTLM msg type " << msgType;
		throw AuthProcessingError(err.str());
	}
	#ifdef DEBUG_TUN
	std::stringstream nfo1;
	nfo1 << "Parsed NTLM msg type: " << msgType;
	DPRINTV(nfo1.str());
	#endif

	// Target domain
	SecurityBuff domainBuff = decSecBuff(rawMsg, origMsg);

	// Flags
	m_chalFlags = decLong(rawMsg);
	#ifdef DEBUG_TUN
	std::stringstream nfo2;
	nfo2 << "Parsed NTLM CHALLENGE flags: " << std::hex << std::setfill('0') << m_chalFlags;
	DPRINTV(nfo2.str());
	#endif

	// Decode target domain
	if(m_chalFlags & FLAG_UNICODE)
		m_chalDomain = decUnicode(domainBuff);
	else
		m_chalDomain = decOem(domainBuff);
	DPRINTV("Parsed NTLM CHALLENGE target domain: " + m_chalDomain);

	// Challenge
	if(rawMsg.getLength() < 8)
		throw AuthProcessingError("Missing challenge in NTLM msg");
	m_chalChallenge = RawBuff(rawMsg.begin(), 8);
	rawMsg.erase(8);
	DPRINTV("Parsed NTLM CHALLENGE challenge");

	// Context
	if(rawMsg.getLength() < 8)
		throw AuthProcessingError("Missing context in NTLM msg");
	m_chalContext = RawBuff(rawMsg.begin(), 8);
	rawMsg.erase(8);
	DPRINTV("Parsed NTLM CHALLENGE context");

	// Target address list
	m_chalTargets.clear();
	SecurityBuff addrListBuff = decSecBuff(rawMsg, origMsg);
	RawBuff addrList = addrListBuff.getData();
	NTLMTarget::TargetType tgtType = NTLMTarget::TGT_NONE;
	do {
		// Type & length
		tgtType = (NTLMTarget::TargetType)decShort(addrList);
		unsigned int tgtLen = decShort(addrList);
		if(tgtType != NTLMTarget::TGT_NONE) {
			// Name
			std::string tgtName = decUnicode(addrList, tgtLen);

			// Add target to list
			m_chalTargets.push_back(NTLMTarget(tgtType, tgtName));
		}
	} while(tgtType != NTLMTarget::TGT_NONE);
	#ifdef DEBUG_TUN
	DPRINTV("Parsed NTLM CHALLENGE targets:");
	std::vector<NTLMTarget>::iterator tit;
	for(tit = m_chalTargets.begin(); tit != m_chalTargets.end(); ++tit) {
		std::stringstream nfo3;
		nfo3 << "  Type: " << tit->m_type << ", Name: " << tit->m_name;
		DPRINTV(nfo3.str());
	}
	#endif

	// Update state
	m_state = NTLM_AUTH;
}


// Reset our state
void NTLM::reset(void)
{
	// Reset/clear data
	m_state = NTLM_NEG;
	m_dataBlock.clear();
	m_chalDomain = "";
	m_chalFlags = 0;
	m_chalChallenge.clear();
	m_chalContext.clear();
	m_chalTargets.clear();
}


/*********************/
/* Private Interface */
/*********************/


// Create NTLM negotiate message
void NTLM::createNegotiateMsg(RawBuff& buff, const std::string& domain) throw(AuthProcessingError)
{
	DPRINTV("NTLM::createNegotiateMsg()");

	try {
		// NTLMSSP signature
		buff.append("NTLMSSP");
		buff.append(0x00);

		// Msg type
		buff.append(encLong(NTLM_NEG));

		// Flags
		buff.append(encLong(0x00088207));

		// Workstation domain & name
		std::string hostname = g_pConfig->getHostName();
		DPRINTV("Domain name: " + domain);
		DPRINTV("Workstation name: " + hostname);
		RawBuff wsDomainBuff = encOem(domain);
		RawBuff wsNameBuff = encOem(hostname);
		encSecBuff(wsDomainBuff, buff, 32);
		encSecBuff(wsNameBuff, buff, 32);

		// Data block
		buff.append(m_dataBlock);
	}
	catch(TunnelError& e) {
		throw AuthProcessingError("Error creating NTLM negotiate msg: " + e.what());
	}
}


// Create NTLM auth message
void NTLM::createAuthMsg(RawBuff& buff, const std::string& user, const std::string& domain,
	const std::string& passwd) throw(AuthProcessingError)
{
	DPRINTV("NTLM::createAuthMsg()");

	try {
		// NTLMSSP signature
		buff.append("NTLMSSP");
		buff.append(0x00);

		// Msg type
		buff.append(encLong(NTLM_AUTH));

		// Create LM and NTLM response fields
		unsigned char lmResponse[24];
		unsigned char ntlmResponse[24];
		if(m_chalFlags & FLAG_NTLM2) {
			// Use NTLM2
			DPRINTV("Creating NTLM2 session response");
			getNTLM2SessionResponse(passwd, m_chalChallenge.begin(), lmResponse, ntlmResponse);
		}
		else {
			// Plain NTLM
			DPRINTV("Creating NTLM response");
			getLMResponse(passwd, m_chalChallenge.begin(), lmResponse);
			getNTLMResponse(passwd, m_chalChallenge.begin(), ntlmResponse);
		}

		#ifdef DEBUG_TUN
		DPRINTV("Challenge: " + Utils::ToHex(m_chalChallenge.begin(), 8));
		DPRINTV("LM Response: " + Utils::ToHex(lmResponse, 24));
		DPRINTV("NTLM Response: " + Utils::ToHex(ntlmResponse, 24));
		#endif

		// LM response
		RawBuff lmBuff(lmResponse, sizeof(lmResponse));
		encSecBuff(lmBuff, buff, 64);

		// NTLM response
		RawBuff ntlmBuff(ntlmResponse, sizeof(ntlmResponse));
		encSecBuff(ntlmBuff, buff, 64);

		// Domain name
		DPRINTV("Domain name: " + domain);
		if(m_chalFlags & FLAG_UNICODE) {
			RawBuff domainName = encUnicode(domain);
			encSecBuff(domainName, buff, 64);
		}
		else {
			RawBuff domainName = encOem(domain);
			encSecBuff(domainName, buff, 64);
		}

		// User name
		DPRINTV("User name: " + user);
		if(m_chalFlags & FLAG_UNICODE) {
			RawBuff userName = encUnicode(user);
			encSecBuff(userName, buff, 64);
		}
		else {
			RawBuff userName = encOem(user);
			encSecBuff(userName, buff, 64);
		}

		// Host name
		std::string host = g_pConfig->getHostName();
		DPRINTV("Host name: " + host);
		if(m_chalFlags & FLAG_UNICODE) {
			RawBuff hostName = encUnicode(host);
			encSecBuff(hostName, buff, 64);
		}
		else {
			RawBuff hostName = encOem(host);
			encSecBuff(hostName, buff, 64);
		}

		// Session key
		buff.append(encLong(0));
		buff.append(encLong(0));

		// Flags
		buff.append(encLong(m_chalFlags));

		// Data block
		buff.append(m_dataBlock);
	}
	catch(TunnelError& e) {
		throw AuthProcessingError("Error creating NTLM auth msg: " + e.what());
	}
}


// Get response for NTLM LM authentication
void NTLM::getLMResponse(const std::string& passwd, const unsigned char* challenge,
	unsigned char* response)
{
	// Create LM hash
	unsigned char hash[16];
	createLMHash(passwd, hash);

	// Create response
	createLMResponse(hash, challenge, response);
}

// Get response for NTLM authentication
void NTLM::getNTLMResponse(const std::string& passwd, const unsigned char* challenge,
	unsigned char* response)
{
	// Create NTLM hash
	unsigned char hash[16];
	createNTLMHash(passwd, hash);

	// Create response
	createLMResponse(hash, challenge, response);
}

// Get response for NTLM2 session authentication
void NTLM::getNTLM2SessionResponse(const std::string& passwd, const unsigned char* challenge,
	unsigned char* clientNonce, unsigned char* response)
{
	// Create client nonce
	createNTLMClientNonce(clientNonce);

	// Create NTLM2 session hash
	unsigned char sessionNonce[16];
	memcpy(sessionNonce, challenge, 8);
	memcpy(sessionNonce + 8, clientNonce, 8);
	unsigned char sessionHash[16];
	AuthDigest::MD5Digest(sessionNonce, 16, sessionHash);

	// Create NTLM hash
	unsigned char hash[16];
	createNTLMHash(passwd, hash);

	// Create response
	createLMResponse(hash, sessionHash, response);
}

// Create response for NTLM LM authentication
void NTLM::createLMResponse(const unsigned char* hash, const unsigned char* challenge,
   unsigned char* response)
{
   // Pad hash to 21 bytes
   unsigned char padHash[21];
   memset(padHash, 0, 21);
   memcpy(padHash, hash, 16);

   // Create DES keys from padded hash
   unsigned char loKey[8], midKey[8], hiKey[8];
   Utils::CreateDESKey(padHash, loKey);
   Utils::CreateDESKey(padHash + 7, midKey);
   Utils::CreateDESKey(padHash + 14, hiKey);

   // Encrypt challenge with DES keys
   unsigned char loResp[8], midResp[8], hiResp[8];
   Utils::DESCrypt(challenge, loKey, loResp, true, g_pConfig->bigEndian());
   Utils::DESCrypt(challenge, midKey, midResp, true, g_pConfig->bigEndian());
   Utils::DESCrypt(challenge, hiKey, hiResp, true, g_pConfig->bigEndian());

   // Construct final response
   memcpy(response, loResp, 8);
   memcpy(response + 8, midResp, 8);
   memcpy(response + 16, hiResp, 8);
}

// Create the hash for NTLM LM authentication
void NTLM::createLMHash(const std::string& passwd, unsigned char* hash)
{
   // Convert passwd to 14 bytes uppercase
   unsigned char keyBytes[14];
   memset(keyBytes, 0, 14);
   unsigned int length = passwd.size();
   if(length > 14)
      length = 14;
   for(unsigned int i = 0; i < length; ++i)
      keyBytes[i] = toupper(passwd[i]);

   // Create DES keys from passwd
   unsigned char loKey[8], hiKey[8];
   Utils::CreateDESKey(keyBytes, loKey);
   Utils::CreateDESKey(keyBytes + 7, hiKey);

   // Encrypt constant with DES keys
   unsigned char loHash[8], hiHash[8];
   Utils::DESCrypt("KGS!@#$%", loKey, loHash, true, g_pConfig->bigEndian());
   Utils::DESCrypt("KGS!@#$%", hiKey, hiHash, true, g_pConfig->bigEndian());

   // Create final combined hash
   memcpy(hash, loHash, 8);
   memcpy(hash + 8, hiHash, 8);
}

// Create the hash for NTLM authentication
void NTLM::createNTLMHash(const std::string& passwd, unsigned char* hash)
{
   // Convert passwd to Unicode
   unsigned int length = passwd.size();
   unsigned char keyBytes[length * 2];
   memset(keyBytes, 0, sizeof(keyBytes));
   for(unsigned int i = 0; i < length; ++i) {
      keyBytes[i*2] = passwd[i];
      keyBytes[i*2 + 1] = 0;
   }

   // Generate MD4 hash of passwd
   AuthDigest::MD4Digest(keyBytes, sizeof(keyBytes), hash);
}

// Create random 8 byte (padded to 24 byte) client nonce
void NTLM::createNTLMClientNonce(unsigned char* nonce)
{
   // 8 random bytes
   for(int i = 0; i < 8; ++i)
      nonce[i] = (unsigned char)(((float)rand() / RAND_MAX) * 255);

   // Pad to 24 bytes
   bzero(nonce + 8, 24 - 8);
}


/********************/
/* Encoding Methods */
/********************/


// Encode a value as a little-endian unsigned short
RawBuff NTLM::encShort(unsigned int val) const throw(MemoryError)
{
	RawBuff ushort;

	// Convert 2 bytes
	unsigned char b1 = (val & 0xFF);
	unsigned char b2 = ((val & 0xFF00) >> 8);
	ushort.append(b1);
	ushort.append(b2);

	return ushort;
}

// Encode a value as a little-endian unsigned long
RawBuff NTLM::encLong(unsigned int val) const throw(MemoryError)
{
	RawBuff ulong;

	// Convert 4 bytes
	unsigned char b1 = (val & 0xFF);
	unsigned char b2 = ((val & 0xFF00) >> 8);
	unsigned char b3 = ((val & 0xFF0000) >> 16);
	unsigned char b4 = ((val & 0xFF000000) >> 24);
	ulong.append(b1);
	ulong.append(b2);
	ulong.append(b3);
	ulong.append(b4);

	return ulong;
}

// Encode a string as unicode
RawBuff NTLM::encUnicode(std::string str) const throw(MemoryError)
{
	RawBuff ucode;

	// Convert each char from original string
	for(unsigned int i = 0; i < str.size(); ++i) {
		ucode.append(str[i]);
		ucode.append(0x00);
	}

	return ucode;
}

// Encode a string as OEM
RawBuff NTLM::encOem(std::string str) const throw(MemoryError)
{
	RawBuff oem;

	// Convert each char from original string
	for(unsigned int i = 0; i < str.size(); ++i) {
		char ch = toupper(str[i]);
		oem.append(ch);
	}

	return oem;
}

// Append to a buffer in security buffer format
void NTLM::encSecBuff(const RawBuff& data, RawBuff& buff, unsigned int constLen)
	throw(MemoryError)
{
	// Append data to data block
	unsigned int dataLen = data.getLength();
	unsigned int dataStart = constLen + m_dataBlock.getLength();
	m_dataBlock.append(data);

	// Append security buffer details
	buff.append(encShort(dataLen));
	buff.append(encShort(dataLen));
	buff.append(encLong(dataStart));
}


/********************/
/* Decoding Methods */
/********************/


// Decode a little-endian unsigned short
unsigned int NTLM::decShort(RawBuff& buff) const throw(AuthProcessingError)
{
	unsigned int val = 0;

	// Length check
	if(buff.getLength() >= 2) {
		// Convert 2 bytes
		const unsigned char* ptr = buff.begin();
		unsigned int b1 = (unsigned int)(*ptr++);
		unsigned int b2 = (unsigned int)(*ptr++);
		val = b1 | (b2 << 8);

		// Clear out buffer
		buff.erase(2);
	}
	else {
		throw AuthProcessingError("Insufficient NTLM data for decoding short");
	}

	return val;
}

// Decode a little-endian unsigned long
unsigned int NTLM::decLong(RawBuff& buff) const throw(AuthProcessingError)
{
	unsigned int val = 0;

	// Length check
	if(buff.getLength() >= 4) {
		// Convert 4 bytes
		const unsigned char* ptr = buff.begin();
		unsigned int b1 = (unsigned int)(*ptr++);
		unsigned int b2 = (unsigned int)(*ptr++);
		unsigned int b3 = (unsigned int)(*ptr++);
		unsigned int b4 = (unsigned int)(*ptr++);
		val = b1 | (b2 << 8) | (b3 << 16) | (b4 << 24);

		// Clear out buffer
		buff.erase(4);
	}
	else {
		throw AuthProcessingError("Insufficient NTLM data for decoding long");
	}

	return val;
}

// Decode a unicode string
std::string NTLM::decUnicode(RawBuff& buff, unsigned int len) const
	throw(AuthProcessingError)
{
	std::stringstream val;

	// Length check
	if(buff.getLength() >= len) {
		// Should be even number of bytes
		if(len % 2 == 0) {
			// Decode each char
			const unsigned char* ptr = buff.begin();
			for(unsigned int i = 0; i < len / 2; ++i) {
				val << (char)(*ptr);
				ptr += 2;
			}

			// Clear out buffer
			buff.erase(len);
		}
		else {
			throw AuthProcessingError("Odd length for decoding unicode");
		}
	}
	else {
		throw AuthProcessingError("Insufficient NTLM data for decoding unicode");
	}

	return val.str();
}

// Decode a unicode string from a security buffer
std::string NTLM::decUnicode(const SecurityBuff& buff) const throw(AuthProcessingError)
{
	RawBuff rawBuff = buff.getData();
	return decUnicode(rawBuff, rawBuff.getLength());
}

// Decode an OEM string
std::string NTLM::decOem(RawBuff& buff, unsigned int len) const
	throw(AuthProcessingError)
{
	std::stringstream val;

	// Length check
	if(buff.getLength() >= len) {
		// Decode each char
		const unsigned char* ptr = buff.begin();
		for(unsigned int i = 0; i < len; ++i) {
			val << (char)(*ptr++);
		}

		// Clear out buffer
		buff.erase(len);
	}
	else {
		throw AuthProcessingError("Insufficient NTLM data for decoding OEM");
	}

	return val.str();
}

// Decode an OEM string from a security buffer
std::string NTLM::decOem(const SecurityBuff& buff) const throw(AuthProcessingError)
{
	RawBuff rawBuff = buff.getData();
	return decOem(rawBuff, rawBuff.getLength());
}

// Decode a security buffer
SecurityBuff NTLM::decSecBuff(RawBuff& buff, const RawBuff& origMsg) const
	throw(AuthProcessingError)
{
	unsigned int dataLen = 0;
	unsigned int buffLen = 0;
	unsigned int buffStart = 0;
	const unsigned char* data = origMsg.begin();

	// Check minimum length
	if(buff.getLength() >= 8) {
		// Decode security buffer parms
		dataLen = decShort(buff);
		buffLen = decShort(buff);
		buffStart = decLong(buff);

		// Get pointer to actual data
		if(origMsg.getLength() >= (buffStart + buffLen)) {
			data += buffStart;
		}
		else {
			throw AuthProcessingError("Insufficient NTLM data for decoding secbuff");
		}
	}
	else {
		throw AuthProcessingError("Insufficient NTLM data for decoding secbuff");
	}

	return SecurityBuff(dataLen, buffLen, buffStart, data);
}


}	// namespace

