/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Handle an NTLM authentication session.
 *
 * $Id: NTLM.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _NTLM_HH_
#define _NTLM_HH_

#include <string>
#include <vector>
#include "Exceptions.hh"
#include "RawBuff.hh"
#include "SecurityBuff.hh"

namespace HTTPTunnel {

class NTLM {
public:
	// Construction/destruction
	NTLM();
	~NTLM();

	// Create next NTLM message
	bool createNextMsg(RawBuff& buff, const std::string& user, const std::string& domain,
		const std::string& passwd) throw(AuthProcessingError);

	// Parse received NTLM message
	void parseRecvdMsg(const std::string& msg) throw(AuthProcessingError);

	// Reset our state
	void reset(void);


private:

	/*********/
	/* Types */
	/*********/

	// NTLM states
	enum NTLMState { NTLM_NEG = 1, NTLM_CHAL, NTLM_AUTH };

	// Target
	struct NTLMTarget {
		enum TargetType { TGT_NONE = 0, TGT_SVR, TGT_DMN, TGT_FQDNS, TGT_DNS };
		NTLMTarget(TargetType ttype, std::string name) : m_type(ttype), m_name(name) {}
		~NTLMTarget() {}

		TargetType		m_type;
		std::string		m_name;
	};


	/*************/
	/* Constants */
	/*************/

	// Flags
	static const unsigned int				FLAG_UNICODE			= 0x00000001;
	static const unsigned int				FLAG_OEM					= 0x00000002;
	static const unsigned int				FLAG_NTLM2				= 0x00080000;


	/***********/
	/* Methods */
	/***********/

	// Forbid assignment
	NTLM(const NTLM& that);
	NTLM& operator=(const NTLM& that);

	// NTLM message creation
	void createNegotiateMsg(RawBuff& buff, const std::string& domain) throw(AuthProcessingError);
	void createAuthMsg(RawBuff& buff, const std::string& user, const std::string& domain,
		const std::string& passwd) throw(AuthProcessingError);
	void getLMResponse(const std::string& passwd, const unsigned char* challenge,
		unsigned char* response);
	void getNTLMResponse(const std::string& passwd, const unsigned char* challenge,
		unsigned char* response);
	void getNTLM2SessionResponse(const std::string& passwd, const unsigned char* challenge,
		unsigned char* clientNonce, unsigned char* response);
	void createLMResponse(const unsigned char* hash, const unsigned char* challenge,
		unsigned char* response);
	void createLMHash(const std::string& passwd, unsigned char* hash);
	void createNTLMHash(const std::string& passwd, unsigned char* hash);
	void createNTLMClientNonce(unsigned char* nonce);

	// Buffer encoding
	RawBuff encShort(unsigned int val) const throw(MemoryError);
	RawBuff encLong(unsigned int val) const throw(MemoryError);
	RawBuff encUnicode(std::string str) const throw(MemoryError);
	RawBuff encOem(std::string str) const throw(MemoryError);
	void encSecBuff(const RawBuff& data, RawBuff& buff, unsigned int constLen)
		throw(MemoryError);

	// Buffer decoding
	unsigned int decShort(RawBuff& buff) const throw(AuthProcessingError);
	unsigned int decLong(RawBuff& buff) const throw(AuthProcessingError);
	std::string decUnicode(RawBuff& buff, unsigned int len) const throw(AuthProcessingError);
	std::string decUnicode(const SecurityBuff& buff) const throw(AuthProcessingError);
	std::string decOem(RawBuff& buff, unsigned int len) const throw(AuthProcessingError);
	std::string decOem(const SecurityBuff& buff) const throw(AuthProcessingError);
	SecurityBuff decSecBuff(RawBuff& buff, const RawBuff& origMsg) const throw(AuthProcessingError);


	/********/
	/* Data */
	/********/

	// Current state
	NTLMState						m_state;

	// Data block under construction
	RawBuff							m_dataBlock;

	// Challenge msg fields
	std::string						m_chalDomain;
	unsigned int					m_chalFlags;
	RawBuff							m_chalChallenge;
	RawBuff							m_chalContext;
	std::vector<NTLMTarget>		m_chalTargets;

};


}	// namespace

#endif

