/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Parse options from either the command line or a config file.
 *
 * $Id: OptionsParser.cc,v 1.8 2009-02-27 11:56:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>
#include <iostream>
#include "OptionsParser.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"
#include "Utils.hh"
#include "Firewall.hh"
#include "ConfigBlob.hh"
#include "GlobalConfigBlob.hh"


namespace HTTPTunnel {


// The global config
extern GlobalConfig* g_pConfig;


// Construction/destruction
#ifdef AUTH_FILE
OptionsParser::OptionsParser()
: m_localIP(), m_localPort(0), m_sslCert(), m_sslCertKey(), m_cliFixed(false),
  m_cliSOCKS(false), m_cliProxy(false), m_cliSocketTun(false), m_cliProtocol(Enums::TCP),
  m_cliServerIP(), m_cliServerPort(0), m_cliSSL(false), m_cliProxyIP(), m_cliProxyPort(0),
  m_cliProxyUser(), m_cliProxyPasswd(), m_cliDestIP(), m_cliDestPort(0), m_cliAuthUser(),
  m_cliAuthPasswd(), m_cliTune(), m_svrFirewallIPs(), m_svrFirewallPorts(), m_svrAuth(false),
  m_svrSSL(false), m_svrSocketTun(false), m_passwds(), m_tunnelnos(), m_bandwidths()
{
	// Initialise parsing state
	sectionFinished(false, false, false);
}
#else
OptionsParser::OptionsParser()
: m_localIP(), m_localPort(0), m_sslCert(), m_sslCertKey(), m_cliFixed(false),
  m_cliSOCKS(false), m_cliProxy(false), m_cliSocketTun(false), m_cliProtocol(Enums::TCP),
  m_cliServerIP(), m_cliServerPort(0), m_cliSSL(false), m_cliProxyIP(), m_cliProxyPort(0),
  m_cliProxyUser(), m_cliProxyPasswd(), m_cliDestIP(), m_cliDestPort(0), m_cliAuthUser(),
  m_cliAuthPasswd(), m_cliTune(), m_svrFirewallIPs(), m_svrFirewallPorts(), m_svrAuth(false),
  m_svrSSL(false), m_svrSocketTun(false)
{
	// Initialise parsing state
	sectionFinished(false, false, false);
}
#endif

OptionsParser::~OptionsParser()
{
}


// Parse the contents of a config file
void OptionsParser::parseCfgFile(const char* name) throw(ConfigError)
{
	// Open file
	std::ifstream file(name);
	if(file) {
		int lineNo = 0;
		char readBuff[1024];
		bool inGlobal = false, inClient = false, inServer = false;

		// Process line at a time
		while(file.getline(readBuff, sizeof(readBuff), '\n')) {
			lineNo++;
			std::string line(readBuff);

			if(line[0] == '#' || line.length() <= 1) {
				// Comment line or blank line
				continue;
			}

			try {
				// Start of new section?
				if(line[0] == '[') {
					// What type of section?
					if(line == "[GLOBAL]") {
						// Old section possibly finished, new one starting
						sectionFinished(inGlobal, inClient, inServer);
						inGlobal = true, inClient = false, inServer = false;
					}
					else if(line == "[CLIENT]") {
						// Old section possibly finished, new one starting
						sectionFinished(inGlobal, inClient, inServer);
						inGlobal = false, inClient = true, inServer = false;
					}
					else if(line == "[SERVER]") {
						// Old section possibly finished, new one starting
						sectionFinished(inGlobal, inClient, inServer);
						inGlobal = false, inClient = false, inServer = true;
					}
				}
				else {
					// Line within existing section
					if(inGlobal)
						parseGlobal(line);
					else if(inClient)
						parseClient(line);
					else if(inServer)
						parseServer(line);
				}
			}
			catch(ConfigError& e) {
				// Rethrow
				std::stringstream err;
				err << "Error in config file at line " << lineNo << ": " << e.what();
				throw ConfigError(err.str());
			}
		}

		// Old section possibly finished
		sectionFinished(inGlobal, inClient, inServer);
	}
	else {
		std::stringstream err;
		err << "Failed to open config file " << name;
		throw ConfigError(err.str());
	}
}


// Parse some config blob
void OptionsParser::parseBlob(const ConfigBlob& blob) throw(ConfigError)
{
	// Get details from blob
	std::string blobType = blob.getType();
	std::string blobCfg = blob.getConfig();

	// What type of config are we parsing?
	bool global = (blobType == "GLOBAL");
	bool client = (blobType == "CLIENT");
	bool server = (blobType == "SERVER");

	// Parse config line at a time
	std::stringstream cfgStream(blobCfg);
	char readBuff[1024];
	while(cfgStream.getline(readBuff, sizeof(readBuff), '\n')) {
		std::string line(readBuff);

		try {
			if(global)
				parseGlobal(line);
			else if(client)
				parseClient(line);
			else if(server)
				parseServer(line);
		}
		catch(ConfigError& e) {
			// Rethrow
			std::stringstream err;
			err << "Error in " << blobType << " config blob: " << e.what();
			throw ConfigError(err.str());
		}
	}

	// Finished parsing
	sectionFinished(global, client, server);
}


// Create default config file
void OptionsParser::createDefaultConfigFile(const char* name) throw(ConfigError)
{
	// Open file
	std::ofstream file(name);
	if(file) {
		// Write sensible default config
		file << "############################" << std::endl;
		file << "## Default HTTPTunnel config" << std::endl;
		file << "############################" << std::endl;
		file << std::endl;
		file << "#################" << std::endl;
		file << "## Global options" << std::endl;
		file << "#################" << std::endl;
		file << std::endl;
		file << "[GLOBAL]" << std::endl;
		file << "debug 2" << std::endl;
		file << std::endl;
		file << std::endl;
		file << "#######################" << std::endl;
		file << "## Client configuration" << std::endl;
		file << "#######################" << std::endl;
		file << std::endl;
		file << "#[CLIENT]" << std::endl;
		file << "#socks" << std::endl;
		file << "#listen 1080" << std::endl;
		file << "#tunsvr myserver.localnet" << std::endl;
		file << std::endl;
		file << std::endl;
		file << "#######################" << std::endl;
		file << "## Server configuration" << std::endl;
		file << "#######################" << std::endl;
		file << std::endl;
		file << "#[Server]" << std::endl;
		file << "#listen 80" << std::endl;

		// Close file
		file.close();
	}
	else {
		std::stringstream err;
		err << "Failed to open config file for writing " << name;
		throw ConfigError(err.str());
	}
}


/*********************/
/* Private Interface */
/*********************/


// Check and get numeric value
unsigned int OptionsParser::getNumeric(const std::string& from, std::string err) const
	throw(ConfigError)
{
	if(isNumeric(from)) {
		return atoi(from.c_str());
	}
	else {
		throw ConfigError("Error in parameters for " + err);
	}
}


// Check a string is numeric
bool OptionsParser::isNumeric(const std::string& str) const
{
	bool res = true;

	for(std::string::size_type i = 0; i < str.size(); ++i) {
		if(!isdigit(str[i])) {
			res = false;
			break;
		}
	}

	return res;
}


// Parse global section line
void OptionsParser::parseGlobal(std::string line) throw(ConfigError)
{
	bool ok = false;

	// Split
	std::vector<std::string> parms;
	Utils::SplitString(line, " ", parms);

	// At least one parm?
	if(parms.size() >= 1) {
		// At least 2 parms?
		if(parms.size() >= 2) {
			if(parms[0] == "debug") {
				// Set debug level
				ok = true;
				unsigned int debug = getNumeric(parms[1], "debug");
				g_pConfig->setDebug(debug);
			}
			else if(parms[0] == "pidfile") {
				// Set PID file name
				ok = true;
				g_pConfig->setPIDFile(parms[1]);
			}
			else if(parms[0] == "datafilestem") {
				// Data log file name stem
				ok = true;
				g_pConfig->setDataLogFileStem(parms[1]);
			}
		}
		else {
			// 1 Arg
			if(parms[0] == "nodaemon") {
				// Set daemon flag
				ok = true;
				g_pConfig->setDaemon(false);
			}
			else if(parms[0] == "dotest") {
				// Set tunnel testing flag
				ok = true;
				g_pConfig->setTesting(true, false);
			}
		}
	}

	// Recognised ok?
	if(!ok)
		throw ConfigError("Unknown GLOBAL config option: " + line);
}

// Parse client section line
void OptionsParser::parseClient(std::string line) throw(ConfigError)
{
	bool ok = false;

	// Split
	std::vector<std::string> parms;
	Utils::SplitString(line, " ", parms);

	// At least one parm?
	if(parms.size() >= 1) {
		// At least 2 parms?
		if(parms.size() >= 2) {
			// At least 3 parms?
			if(parms.size() >= 3) {
				// At least 4 parms?
				if(parms.size() >= 4) {
					// 4 args
					if(parms[0] == "fixed") {
						// Address of fixed tunnel destination
						ok = true;
						m_cliFixed = true;
						m_cliSOCKS = false;
						m_cliDestIP = parms[1];
						m_cliDestPort = getNumeric(parms[2], "fixed");
						if(parms[3] == "udp")
							m_cliProtocol = Enums::UDP;
						else
							m_cliProtocol = Enums::TCP;
					}
				}
				else {
					// 3 args
					if(parms[0] == "tunsvr") {
						// Address of server end of tunnel
						ok = true;
						m_cliServerIP = parms[1];
						m_cliServerPort = getNumeric(parms[2], "tunsvr");
					}
					else if(parms[0] == "fixed") {
						// Address of fixed tunnel destination
						ok = true;
						m_cliFixed = true;
						m_cliSOCKS = false;
						m_cliDestIP = parms[1];
						m_cliDestPort = getNumeric(parms[2], "fixed");
						m_cliProtocol = Enums::TCP;
					}
					else if(parms[0] == "proxy") {
						// Address of proxy to use
						ok = true;
						m_cliProxy = true;
						m_cliProxyIP = parms[1];
						m_cliProxyPort = getNumeric(parms[2], "proxy");
					}
					else if(parms[0] == "proxyauth") {
						// Authorisation to use at proxy
						ok = true;
						m_cliProxyUser = parms[1];
						m_cliProxyPasswd = parms[2];
					}
					else if(parms[0] == "cliauth") {
						// Client authentication user/passwd
						ok = true;
						m_cliAuthUser = parms[1];
						m_cliAuthPasswd = parms[2];
					}
					else if(parms[0] == "listen") {
						// Local listening address & port
						ok = true;
						m_localIP = parms[1];
						m_localPort = getNumeric(parms[2], "listen");
					}
				}
			}
			else {
				// 2 args
				if(parms[0] == "listen") {
					// Local listening port
					ok = true;
					m_localPort = getNumeric(parms[1], "listen");
				}
				else if(parms[0] == "tunsvr") {
					// Address of server end of tunnel
					ok = true;
					m_cliServerIP = parms[1];
					m_cliServerPort = 80;
				}
				else if(parms[0] == "tune") {
					// Tune tunnelling behaviour
					if(parms[1] == "bursty" || parms[1] == "realtime" || parms[1] == "default") {
						ok = true;
						m_cliTune = parms[1];
					}
				}
			}
		}
		else {
			// 1 arg
			if(parms[0] == "socks") {
				// SOCKS client
				ok = true;
				m_cliSOCKS = true;
				m_cliFixed = false;
			}
			else if(parms[0] == "ssl") {
				// Use https
				ok = true;
				m_cliSSL = true;
			}
			else if(parms[0] == "socktun") {
				// Socket tunneler
				ok = true;
				m_cliSocketTun = true;
			}
		}
	}

	// Recognised ok?
	if(!ok)
		throw ConfigError("Unknown CLIENT config option: " + line);
}

// Parse server section line
void OptionsParser::parseServer(std::string line) throw(ConfigError)
{
	bool ok = false;

	// Split
	std::vector<std::string> parms;
	Utils::SplitString(line, " ", parms);

	// At least one parm?
	if(parms.size() >= 1) {
		// At least 2 parms?
		if(parms.size() >= 2) {
			// At least 3 parms?
			if(parms.size() >= 3) {
			   // At least 4 parms?
				if(parms.size() >= 4) {
			   	// At least 5 parms?
            	if(parms.size() >= 5) {
						// 5 parms
						#ifdef AUTH_FILE
						if(parms[0] == "user") {
							ok = true;
							m_passwds[parms[1]] = parms[2];
							m_tunnelnos[parms[1]] = getNumeric(parms[3], "user");
							m_bandwidths[parms[1]] = getNumeric(parms[4], "user");
						}
						#endif
					}
				}
				else {
					// 3 args
					if(parms[0] == "listen") {
						// Local listening address & port
						ok = true;
						m_localIP = parms[1];
						m_localPort = getNumeric(parms[2], "listen");
					}
					else if(parms[0] == "ssl") {
						// Use https
						ok = true;
						m_svrSSL = true;
						m_sslCert = parms[1];
						m_sslCertKey = parms[2];
					}
				}
			}
			else {
				// 2 args
				if(parms[0] == "listen") {
					// Local listening port
					ok = true;
					m_localPort = getNumeric(parms[1], "listen");
				}
				else if(parms[0] == "denyips") {
					// Firewalled IPs
					ok = true;
					m_svrFirewallIPs = parms[1];
				}
				else if(parms[0] == "denyports") {
					// Firewalled ports
					ok = true;
					m_svrFirewallPorts = parms[1];
				}
			}
		}
		else {
			// 1 arg
			if(parms[0] == "svrauth") {
				// Use authentication at the server
				ok = true;
				m_svrAuth = true;
			}
			else if(parms[0] == "socktun") {
				// Socket tunneler
				ok = true;
				m_svrSocketTun = true;
			}
		}
	}

	// Recognised ok?
	if(!ok)
		throw ConfigError("Unknown SERVER config option: " + line);
}


// Use parsed options for a section to create runtime config
void OptionsParser::sectionFinished(bool global, bool client, bool server)
	throw(ConfigError)
{
	// What section has finished?
	if(global) {
		// Nothing for now
	}
	else if(client) {
		// Construct client config
		if(m_cliSOCKS && m_cliProxy) {
			// SOCKS client, via proxy
			if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
			m_cliProxyIP != "" && m_cliProxyPort != 0) {
				g_pConfig->configureClient(m_cliProtocol, m_cliSSL, m_localIP,
					m_localPort, m_cliServerIP, m_cliServerPort, m_cliAuthUser,
					m_cliAuthPasswd, m_cliTune, m_cliProxyIP, m_cliProxyPort,
					m_cliProxyUser, m_cliProxyPasswd);
			}
			else {
				throw ConfigError("Incomplete CLIENT SOCKS (via proxy) config");
			}
		}
		else if(m_cliSOCKS && !m_cliProxy) {
			// SOCKS client, no proxy
			if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0) {
				g_pConfig->configureClient(m_cliProtocol, m_cliSSL, m_localIP,
					m_localPort, m_cliServerIP, m_cliServerPort, m_cliAuthUser,
					m_cliAuthPasswd, m_cliTune);
			}
			else {
				throw ConfigError("Incomplete CLIENT SOCKS config");
			}
		}
		else if(m_cliFixed && m_cliProxy) {
			// Direct client, via proxy
			if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
			gotValidDest() && m_cliProxyIP != "" && m_cliProxyPort != 0) {
				g_pConfig->configureClient(m_cliProtocol, m_cliSSL, m_cliSocketTun,
					m_localIP, m_localPort, m_cliServerIP, m_cliServerPort, m_cliAuthUser,
					m_cliAuthPasswd, m_cliDestIP, m_cliDestPort, m_cliTune, m_cliProxyIP,
					m_cliProxyPort, m_cliProxyUser, m_cliProxyPasswd);
			}
			else {
				throw ConfigError("Incomplete CLIENT fixed (via proxy) config");
			}
		}
		else if(m_cliFixed && !m_cliProxy) {
			// Direct client, no proxy
			if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
			gotValidDest()) {
				g_pConfig->configureClient(m_cliProtocol, m_cliSSL, m_cliSocketTun,
					m_localIP, m_localPort, m_cliServerIP, m_cliServerPort, m_cliAuthUser,
					m_cliAuthPasswd, m_cliDestIP, m_cliDestPort, m_cliTune);
			}
			else {
				throw ConfigError("Incomplete CLIENT fixed config");
			}
		}
		else {
			throw ConfigError("You must specify the CLIENT type as one of socks/fixed");
		}
	}
	else if(server) {
		// Construct server config
		if(m_localPort != 0) {
			#ifdef AUTH_FILE
			g_pConfig->configureServer(m_svrSSL, m_svrSocketTun, m_sslCert, m_sslCertKey,
				m_localIP, m_localPort, m_svrFirewallIPs, m_svrFirewallPorts, m_svrAuth,
				m_passwds, m_tunnelnos, m_bandwidths);
			#else
			g_pConfig->configureServer(m_svrSSL, m_svrSocketTun, m_sslCert, m_sslCertKey,
				m_localIP, m_localPort, m_svrFirewallIPs, m_svrFirewallPorts, m_svrAuth);
			#endif
		}
		else {
			throw ConfigError("Incomplete SERVER config");
		}
	}

	// Reset all parsing variables
	m_localIP = "";
	m_localPort = 0;
	m_sslCert = "";
	m_sslCertKey = "";
	m_cliFixed = false;
	m_cliSOCKS = false;
	m_cliProxy = false;
	m_cliSocketTun = false;
	m_cliProtocol = Enums::TCP;
	m_cliServerIP = "";
	m_cliServerPort = 0;
	m_cliSSL = false;
	m_cliProxyIP = "";
	m_cliProxyPort = 0;
	m_cliProxyUser = "";
	m_cliProxyPasswd = "";
	m_cliDestIP = "";
	m_cliDestPort = 0;
	m_cliAuthUser = "";
	m_cliAuthPasswd = "";
	m_cliTune = "";
	m_svrFirewallIPs = "";
	m_svrFirewallPorts = "";
	m_svrAuth = false;
	m_svrSSL = false;
	m_svrSocketTun = false;

	#ifdef AUTH_FILE
	m_passwds.clear();
	m_tunnelnos.clear();
	m_bandwidths.clear();
	#endif
}


// Check we have a valid local listening address
bool OptionsParser::gotValidLocal(void) const
{
	// Must either have a valid port, or the address is the special case "-"
	return (m_localPort != 0 || m_localIP == "-");
}

// Check we have a valid destination address
bool OptionsParser::gotValidDest(void) const
{
	// Must have valid IP and port, or the address is the special case "-"
	return ((m_cliDestIP != "" && m_cliDestPort != 0) || m_cliDestIP == "-");
}


}	// namespace

