/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a raw buffer of bytes.
 *
 * $Id: RawBuff.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include "RawBuff.hh"
#include "Logger.hh"
#include "Exceptions.hh"
#include "Utils.hh"


namespace HTTPTunnel {


// Reference to the logger
extern Logger* g_pLogger;


// Construction/destruction
RawBuff::RawBuff()
: m_data(0), m_length(0)
{
}

RawBuff::RawBuff(unsigned int length, bool zero) throw(MemoryError)
: m_data(0), m_length(0)
{
	if(length) {
		// Allocate buffer of requested size
		m_data = (unsigned char*)malloc(length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");
		m_length = length;

		// If required, zero buffer contents
		if(zero)
			memset(m_data, 0, m_length);
	}
}

RawBuff::RawBuff(const unsigned char* data, unsigned int length) throw(MemoryError)
: m_data(0), m_length(0)
{
	if(length) {
		// Copy data
		m_data = (unsigned char*)malloc(length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = length;
		memcpy(m_data, data, m_length);
	}
}

RawBuff::RawBuff(std::string& data) throw(MemoryError)
: m_data(0), m_length(0)
{
	const char* rawdata = data.c_str();
	unsigned int length = data.size();

	if(length) {
		// Copy data
		m_data = (unsigned char*)malloc(length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = length;
		memcpy(m_data, rawdata, m_length);
	}
}

RawBuff::RawBuff(const std::vector<unsigned char>& data) throw(MemoryError)
: m_data(0), m_length(0)
{
	if(data.size()) {
		// Copy data
		m_data = (unsigned char*)malloc(data.size());
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = data.size();
		std::copy(data.begin(), data.end(), m_data);
	}
}

RawBuff::RawBuff(const RawBuff& that) throw(MemoryError)
: m_data(0), m_length(0)
{
	if(that.m_length) {
		// Copy data
		m_data = (unsigned char*)malloc(that.m_length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = that.m_length;
		memcpy(m_data, that.m_data, m_length);
	}
}

RawBuff::RawBuff(const RawBuff& that, unsigned int length) throw(MemoryError)
: m_data(0), m_length(0)
{
	// Range check
	if(length > that.m_length)
		length = that.m_length;

	if(length) {
		// Copy data
		m_data = (unsigned char*)malloc(length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = length;
		memcpy(m_data, that.m_data, m_length);
	}
}

RawBuff::~RawBuff()
{
	// Empty ourselves
	clear();
}


// Assignment operators
RawBuff& RawBuff::operator=(const RawBuff& that) throw(MemoryError)
{
	// Self assignment?
	if(this != &that) {
		// Free any existing data
		clear();

		// Copy data
		m_data = (unsigned char*)malloc(that.m_length);
		if(m_data == 0)
			throw MemoryError("RawBuff malloc failed");

		m_length = that.m_length;
		memcpy(m_data, that.m_data, m_length);
	}

	return *this;
}

RawBuff& RawBuff::operator=(const std::vector<unsigned char>& that) throw(MemoryError)
{
	// Free any existing data
	clear();

	// Copy data
	m_data = (unsigned char*)malloc(that.size());
	if(m_data == 0)
		throw MemoryError("RawBuff malloc failed");

	m_length = that.size();
	std::copy(that.begin(), that.end(), m_data);

	return *this;
}


// Equality operator
bool RawBuff::operator==(const RawBuff& that) const
{
	bool equal = false;

	bool lengthEqual = (m_length == that.m_length);
	if(lengthEqual) {
		equal = (memcmp(m_data, that.m_data, m_length) == 0);
	}

	return equal;
}


/*************/
/* Accessors */
/*************/
 

// Get length of data
unsigned int RawBuff::getLength(void) const
{
	return m_length;
}


// Get pointer to start of data
const unsigned char* RawBuff::begin(void) const
{
	return m_data;
}

// Get pointer to 1 beyond end of data
const unsigned char* RawBuff::end(void) const
{
	return m_data + m_length;
}


// Append ourselves to a vector
void RawBuff::appendToVec(std::vector<unsigned char>& vec)
{
	// Append read data to our buffer
	vec.reserve(vec.size() + m_length);
	std::back_insert_iterator<std::vector<unsigned char> > bit(vec);
	std::copy(m_data, m_data + m_length, bit);
}


/************/
/* Mutators */
/************/


// Empty ourselves
void RawBuff::clear(void)
{
	// Free data
	if(m_data)
		free(m_data);

	m_data = 0;
	m_length = 0;
}

// Erase some number of bytes from the start of our data
void RawBuff::erase(unsigned int length) throw(MemoryError)
{
	// Do nothing?
	if(length == 0)
		return;

	// Erase everything?
	if(length >= m_length) {
		clear();
	}
	else {
		// Calculate what data to keep
		unsigned char* keepDataPtr = m_data + length;
		unsigned int keepLength = m_length - length;

		// Replace old data with new
		replace(keepDataPtr, keepLength);
	}
}

// Erase everything from the start of our data to some position
void RawBuff::erase(const unsigned char* end) throw(MemoryError)
{
	// Check pointer is within our data buffer
	if(end < begin() || end > this->end())
		throw MemoryError("RawBuff::erase() - Pointer outside buffer range");

	// Erase everything?
	if(end == this->end()) {
		clear();
	}
	else {
		// Calculate what data to keep
		const unsigned char* keepDataPtr = end;
		unsigned int keepLength = m_length - (end - m_data);

		// Replace old data with new
		replace(keepDataPtr, keepLength);
	}
}


// Erase whitespace from our start
void RawBuff::eraseWS(void) throw(MemoryError)
{
	// Calculate number of whitespace chars are at our start
	unsigned int wsCount = 0;
	while(wsCount < m_length) {
		char c = m_data[wsCount];
		if(c != ' ' && c != '\t')
			break;
		wsCount++;
	}

	// Erase whitespace
	if(wsCount > 0)
		erase(wsCount);
}


// Append some data
void RawBuff::append(const void* data, unsigned int length)
	throw(MemoryError)
{
	const unsigned char* ucdata = (const unsigned char*)(data);

	if(length) {
		// Realloc extra buffer space
		m_data = (unsigned char*)realloc(m_data, m_length + length);
		if(m_data == 0) {
			m_length = 0;
			throw MemoryError("RawBuff realloc failed");
		}

		// Append extra data
		memcpy(m_data + m_length, ucdata, length);
		m_length = m_length + length;
	}
}

void RawBuff::append(std::string data) throw(MemoryError)
{
	// Append string data
	append(data.c_str(), data.size());
}

void RawBuff::append(const std::vector<unsigned char>& data) throw(MemoryError)
{
	// Append all data
	append(data, data.size());
}

void RawBuff::append(const std::vector<unsigned char>& data, unsigned int length)
	throw(MemoryError)
{
	// Range check
	if(length > data.size())
		length = data.size();

	if(length) {
		// Realloc extra buffer space
		m_data = (unsigned char*)realloc(m_data, m_length + length);
		if(m_data == 0) {
			m_length = 0;
			throw MemoryError("RawBuff realloc failed");
		}

		// Append extra data
		unsigned char* pos = m_data + m_length;
		std::copy(data.begin(), data.begin() + length, pos);
		m_length = m_length + length;
	}
}

void RawBuff::append(const RawBuff& data) throw(MemoryError)
{
	// Append all data
	append(data, data.getLength());
}

void RawBuff::append(const RawBuff& data, unsigned int length) throw(MemoryError)
{
	// Range check
	if(length > data.m_length)
		length = data.m_length;

	if(length) {
		// Realloc extra buffer space
		m_data = (unsigned char*)realloc(m_data, m_length + length);
		if(m_data == 0) {
			m_length = 0;
			throw MemoryError("RawBuff realloc failed");
		}

		// Append extra data
		memcpy(m_data + m_length, data.begin(), length);
		m_length = m_length + length;
	}
}

void RawBuff::append(unsigned char data) throw(MemoryError)
{
	// Realloc extra buffer space
	m_data = (unsigned char*)realloc(m_data, m_length + 1);
	if(m_data == 0) {
		m_length = 0;
		throw MemoryError("RawBuff realloc failed");
	}

	// Append extra data
	m_data[m_length] = data;
	m_length++;
}


// Prepend some data
void RawBuff::prepend(const RawBuff& data) throw(MemoryError)
{
	// Prepend all data
	prepend(data, data.getLength());
}

void RawBuff::prepend(const RawBuff& data, unsigned int length) throw(MemoryError)
{
	// Range check
	if(length > data.m_length)
		length = data.m_length;

	if(length) {
		// Realloc extra buffer space
		m_data = (unsigned char*)realloc(m_data, m_length + length);
		if(m_data == 0) {
			m_length = 0;
			throw MemoryError("RawBuff realloc failed");
		}

		// Shuffle along exising data
		memmove(m_data + length, m_data, m_length);

		// Prepend extra data
		memcpy(m_data, data.begin(), length);
		m_length = m_length + length;
	}
}


/*************/
/* Utilities */
/*************/


// Look for some string within ourselves
bool RawBuff::findString(const std::string& str, const unsigned char*& start,
	const unsigned char*& end) const
{
	return findString(str, m_data, start, end);
}

// Look for some string within ourselves, beginning at some offset
bool RawBuff::findString(const std::string& str, const unsigned char* begin,
	const unsigned char*& start, const unsigned char*& end) const
{
	// Check for null
	if(m_data == 0)
		return false;

	// Range check beginning
	int startChar = begin - m_data;
	if(startChar < 0)
		startChar = 0;
	else if((unsigned int)startChar >= m_length)
		startChar = m_length - 1;

	bool res = false;
	unsigned int bytesLeft;
	unsigned int index = (unsigned int)startChar;
	unsigned int strLen = str.size();
	unsigned char* buffEnd = m_data + m_length;
	int firstChar = str[0];

	// Keep going while there is room in our buffer for the string
	while((bytesLeft = m_length - index) >= strLen) {
		// Find first character of string
		void* buffPtr = m_data + index;
		unsigned char* firstCharPos = (unsigned char*)memchr(buffPtr, firstChar,
			bytesLeft);

		if(firstCharPos) {
			// Room for complete string?
			if(firstCharPos + strLen <= buffEnd) {
				// Look for complete string
				if(memcmp(firstCharPos, str.c_str(), strLen) == 0) {
					// Found it
					res = true;
					start = firstCharPos;
					end = start + strLen;
					break;
				}
				else {
					// Keep looking from here
					index = (firstCharPos - m_data) + 1;
				}
			}
			else {
				// String not found
				break;
			}
		}
		else {
			// String not found
			break;
		}
	}

	return res;
}


// Look for first occurance of some character
bool RawBuff::findChar(unsigned char ch, const unsigned char*& pos) const
{
	// Check for null
	if(m_data == 0)
		return false;

	pos = (unsigned char*)memchr(m_data, ch, m_length);
	return (pos != 0);
}


// Uuencode some of our data
void RawBuff::uuencode(unsigned int& length, std::string& res) const
{
	res = Utils::ModBase64Encode(toString(length));
	length = res.size();
}

// Uudecode some string into ourselves
void RawBuff::uudecode(std::string& uuencoded) throw(MemoryError)
{
	*this = Utils::ModBase64Decode(uuencoded);
}


// Base64 encode some of our data
void RawBuff::base64encode(unsigned int& length, std::string& res) const
{
	res = Utils::Base64Encode(toString(length));
	length = res.size();
}


// Write our contents as a string
std::string RawBuff::toString(void) const
{
	if(m_data == 0 || m_length == 0)
		return "";
	return std::string((char*)m_data, (char*)(m_data + m_length));
}

// Write some of our contents as a string
std::string RawBuff::toString(unsigned int& length) const
{
	length = (length <= m_length)? length : m_length;
	if(m_data == 0 || length == 0)
		return "";
	return std::string((char*)m_data, (char*)(m_data + length));
}


/*********************/
/* Private Interface */
/*********************/


// Replace our data with something new
void RawBuff::replace(const void* data, unsigned int length)
	throw(MemoryError)
{
	// Copy new data
	unsigned char* newData = (unsigned char*)malloc(length);
	if(newData == 0)
		throw MemoryError("RawBuff malloc failed");
	memcpy(newData, data, length);

	// Free old data
	clear();

	// Point to new data
	m_data = newData;
	m_length = length;
}


}	// namespace

