/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a raw buffer of bytes.
 *
 * $Id: RawBuff.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _RAW_BUFF_HH_
#define _RAW_BUFF_HH_

#include <vector>
#include "Exceptions.hh"


namespace HTTPTunnel {


class RawBuff {
public:
	// Construction/destruction
	RawBuff();
	RawBuff(unsigned int length, bool zero) throw(MemoryError);
	RawBuff(const unsigned char* data, unsigned int length) throw(MemoryError);
	RawBuff(std::string& data) throw(MemoryError);
	RawBuff(const std::vector<unsigned char>& data) throw(MemoryError);
	RawBuff(const RawBuff& that) throw(MemoryError);
	RawBuff(const RawBuff& that, unsigned int length) throw(MemoryError);
	~RawBuff();

	// Assignment operators
	RawBuff& operator=(const RawBuff& that) throw(MemoryError);
	RawBuff& operator=(const std::vector<unsigned char>& that) throw(MemoryError);

	// Equality operator
	bool operator==(const RawBuff& that) const;


	/*************/
	/* Accessors */
	/*************/

	// Get length of data
	unsigned int getLength(void) const;

	// Get pointer to start of data
	const unsigned char* begin(void) const;
	// Get pointer to 1 beyond end of data
	const unsigned char* end(void) const;

	// Append ourselves to a vector
	void appendToVec(std::vector<unsigned char>& vec);


	/************/
	/* Mutators */
	/************/

	// Empty ourselves
	void clear(void);

	// Erase some number of bytes from the start of our data
	void erase(unsigned int length) throw(MemoryError);
	// Erase everything from the start of our data to some position
	void erase(const unsigned char* end) throw(MemoryError);

	// Erase whitespace from our start
	void eraseWS(void) throw(MemoryError);

	// Append some data
	void append(const void* data, unsigned int length) throw(MemoryError);
	void append(std::string data) throw(MemoryError);
	void append(const std::vector<unsigned char>& data) throw(MemoryError);
	void append(const std::vector<unsigned char>& data, unsigned int length)
		throw(MemoryError);
	void append(const RawBuff& data) throw(MemoryError);
	void append(const RawBuff& data, unsigned int length) throw(MemoryError);
	void append(unsigned char data) throw(MemoryError);

	// Prepend some data
	void prepend(const RawBuff& data) throw(MemoryError);
	void prepend(const RawBuff& data, unsigned int length) throw(MemoryError);


	/*************/
	/* Utilities */
	/*************/

	// Look for some string within ourselves
	bool findString(const std::string& str, const unsigned char*& start,
		const unsigned char*& end) const;
	bool findString(const std::string& str, const unsigned char* begin,
		const unsigned char*& start, const unsigned char*& end) const;

	// Look for first occurance of some character
	bool findChar(unsigned char ch, const unsigned char*& pos) const;

	// Uuencode/Uudecode some of our data
	void uuencode(unsigned int& length, std::string& res) const;
	void uudecode(std::string& uuencoded) throw(MemoryError);

	// Base64 encode some of our data
	void base64encode(unsigned int& length, std::string& res) const;

	// Write our contents as a string
	std::string toString(void) const;
	std::string toString(unsigned int& length) const;


private:

	/***********/
	/* Methods */
	/***********/

	// Replace our data with something new
	void replace(const void* data, unsigned int length) throw(MemoryError);


	/********/
	/* Data */
	/********/

	// Pointer to our raw data
	unsigned char*						m_data;

	// Length of data
	unsigned int						m_length;

};


}	// namespace

#endif

