/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of a raw tunnel
 * connection.
 *
 * $Id: RawConnection.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "RawConnection.hh"
#include "Tunneler.hh"
#include "RawReader.hh"
#include "RawWriter.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "GlobalConfig.hh"
#include "ServerSideConfig.hh"
#include "Firewall.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;


// Construction/destruction
RawConnection::RawConnection(Sock* sock, Tunneler* tunneler, const TunnelID& tid)
: Connection(sock, tunneler), m_tid(tid)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "RawConnection::RawConnection(server) - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// Construct reader/writer
	createReaderWriter();
}

RawConnection::RawConnection(Tunneler* tunneler, const TunnelID& tid)
: Connection(tunneler), m_tid(tid)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "RawConnection::RawConnection(client) - connection " << getID();
	DPRINTV(dbg.str());
	#endif
}

RawConnection::~RawConnection()
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "RawConnection::~RawConnection() - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// We are closing down
	MutexVar cmv(m_connMtx);
	m_closing = true;
}


/********************/
/* Public Interface */
/********************/


// Prepare for a connect attempt to be made to the server
void RawConnection::connectAttempt(std::string host, unsigned int port,
	Enums::Protocol prot, bool ssl) throw(ConnectError)
{
	// Check to see if connection is allowed
	if(!m_tunneler->getServerSideConfigMutable()->getFirewall()->isAllowed(host, port)) {
		#ifdef DEBUG_TUN
		std::stringstream wrn;
		wrn << "RawConnection::connectAttempt() - Refused connect to ";
		wrn << "host " << host << " port " << port;
		g_pLogger->warn(wrn.str());
		#endif

		throw FirewalledConnectError("Connect destination is firewalled");
	}

	// Let base class take it from here
	Connection::connectAttempt(host, port, prot, ssl);
}


/***********************/
/* Protected Interface */
/***********************/


// Construct reader/writer
void RawConnection::createReaderWriter(void)
{
	// Create raw reader & writer
	MutexVar mv(m_connMtx);
	m_reader = new RawReader(this, m_tid);
	m_writer = new RawWriter(this);

	// Only truly connected once reader & writer are constructed
	setConnected();
}


} // namespace


