/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for data readers.
 *
 * $Id: Reader.cc,v 1.3 2008-07-25 12:14:53 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef DEBUG_LOG_DATA
#include <iomanip>
#endif

#include <sstream>
#include <cstring>
#include <cerrno>
#include "Reader.hh"
#include "Debug.hh"
#include "Mutex.hh"
#include "MutexVar.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "Connection.hh"
#include "Tunneler.hh"
#include "Sock.hh"


namespace HTTPTunnel {


#ifdef DEBUG_LOG_DATA
// Reference to log file mutex
extern Mutex g_logMtx;
#endif

// Reference to the logger
extern Logger* g_pLogger;


// Construction/destruction
Reader::Reader(Connection* conn)
: m_mtx(), m_conn(conn), m_readbuff()
{
}

Reader::~Reader()
{
}


/********************/
/* Public Interface */
/********************/


// Read and process some data from our socket
void Reader::read(void) throw(CommsError)
{
	MutexVar mv(m_mtx);

	try {
		// Read data from socket
		Sock* sock = m_conn->getSocket();
		int noread = sock->read(m_readbuff, sizeof(m_readbuff));

		// Try to process what we currently have buffered
		processData(m_readbuff, noread);
	}
	catch(SockError& e) {
		std::stringstream dbg;
		dbg << "Error reading from socket for connection " << m_conn->getID();
		dbg << ": " << e.m_errno << " " << strerror(e.m_errno);
		throw ReadError(dbg.str());
	}
}


/*************/
/* Accessors */
/*************/


// Do we have unprocessed data remaining?
bool Reader::dataRemaining(void) const
{
	return false;
}


/***********************/
/* Protected Interface */
/***********************/


#ifdef DEBUG_LOG_DATA
// Log some newly read data
void Reader::logData(const RawBuff& data, std::ofstream* file,
	unsigned int& lineChar)
{
	// Lock global log file
	MutexVar mv(g_logMtx);

	// Get pointer to data
	const unsigned char* dataPtr = data.begin();

	// Log bytes
	for(unsigned int i = 0; i < data.getLength(); ++i) {
		*file << std::hex << std::setw(2) << std::setfill('0');
		*file << (unsigned int)dataPtr[i];

		// Space or newline?
		if((++lineChar % 20) == 0) {
			lineChar = 0;
			*file << std::endl;
		}
		else
			*file << " ";
	}

	// Flush
	*file << std::flush;
}
#endif


}	// namespace

