/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel acting as a
 * SOCKS proxy.
 *
 * $Id: SOCKSClientTunneler.cc,v 1.4 2009-04-24 15:37:33 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include <vector>
#include "SOCKSClientTunneler.hh"
#include "Exceptions.hh"
#include "Enums.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ProxySettings.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "ClientTunnel.hh"
#include "SOCKSConnection.hh"
#include "HTTPConnection.hh"
#include "ShutdownTunnelEvent.hh"
#include "Authenticator.hh"
#include "ConnectionManager.hh"
#include "Server.hh"
#include "Sock.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
SOCKSClientTunneler::SOCKSClientTunneler(Server* server) throw(TunnelError)
: ClientTunneler(server)
{
	DPRINTV("SOCKSClientTunneler::SOCKSClientTunneler()");
}

SOCKSClientTunneler::~SOCKSClientTunneler()
{
	DPRINTV("SOCKSClientTunneler::~SOCKSClientTunneler()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();
}


/**********/
/* Events */
/**********/


// Notification of a new socket connection
void SOCKSClientTunneler::newConnectionArrived(Sock* insock)
	throw(CommsError)
{
	DPRINTI("SOCKSClientTunneler::newConnectionArrived()");

	/////////////////////////////////////////////////////////////////////
	// In SOCKSClientTunneler a new connection always means a new tunnel,
	// but we don't want to create that tunnel until we have completed
	// The SOCKS negotiation.
	/////////////////////////////////////////////////////////////////////

	// Do our users quotas allow another tunnel?
	Authenticator* authenticator = m_server->getConfigMutable()->getAuthenticator();
	std::string userName = authenticator->getDefaultUserName();
	bool userQuotaOk = authenticator->getUserCanCreateTunnel(userName);
	if(userQuotaOk) {
		// Create SOCKS connection to handle incoming data
		MutexVar cmv(m_connectionsMtx);
		Connection* rawconn = new SOCKSConnection(insock, this);
		addConnection(rawconn);
	}
	else {
		// Tunnel creation quota exceeded
		std::stringstream err;
		err << "Tunnel creation quota exceeded for user " << userName;
		throw CommsError(err.str());
	}
}


// Notification that a connection has changed state
void SOCKSClientTunneler::connectionStateChange(unsigned int connID,
	Connection::ConnectionState state)
{
	// Let base class know
	ClientTunneler::connectionStateChange(connID, state);

	// Something for us?
	if(state == Connection::SOCKSSETUP) {

		///////////////////////////////////////////////////////////////////
		// In SOCKSClientTunneler this means that the SOCKS connection
		// has completed negotiation.
		///////////////////////////////////////////////////////////////////

		try {
			// Lookup connection
			MutexVar cmv(m_connectionsMtx);
			std::vector<Connection*>::iterator cit = getConnection(connID);
			SOCKSConnection* socksconn = dynamic_cast<SOCKSConnection*>(*cit);

			// Create HTTP client connection to handle tunneled traffic
			HTTPConnection* httpconn = new HTTPConnection(this);
			addConnection(httpconn);

			// Create new tunnel config
			TunnelConfig* tcfg = createTunnelConfig(socksconn);

			// Let SOCKS connection know about new tunnels ID
			socksconn->setTunnelID(*tcfg->getTunnelID());

			// Create tunnel
			Tunnel* tun = new ClientTunnel(tcfg, this);
			tun->addLeftConnection(socksconn);
			tun->addRightConnection(httpconn);

			// Schedule connect out on HTTP side
			const ClientSideConfig* clientCfg = getClientSideConfig();
			m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
				clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL());

			cmv.unlock();

			// Store Tunnel object in our list of active tunnels
			MutexVar tmv(m_tunnelsMtx);
			m_tunnels.push_back(tun);
			tmv.unlock();
		}
		catch(LookupManagementError& e) {
			g_pLogger->warn("Error looking up connection: " + e.what());
		}
	}
}


// Notification that a tunnel has changed state
void SOCKSClientTunneler::tunnelStateChanged(const TunnelID& tid,
	Tunnel::TunnelState state)
{
	DPRINTI("SOCKSClientTunneler::tunnelStateChanged() - " + tid.toString());

	// Let base class know
	ClientTunneler::tunnelStateChanged(tid, state);

	// Lookup tunnel
	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Check this isn't a test tunnel
		if(!tun->isTestTunnel()) {
			// Lookup SOCKS connection
			std::vector<Tunnel::Selector*> sel;
			sel.push_back(new Tunnel::SelLeftConn());
			SOCKSConnection* socksconn = dynamic_cast<SOCKSConnection*>(tun->getConnFromSel(sel));
			if(socksconn) {
				try {
					// Let connection know
					if(state == Tunnel::CONNECTED) {
						socksconn->tunnelConnected(true);
					}
				}
				catch(CommsError& e) {
					// Log error
					g_pLogger->error("Error processing tunnel state change notification: " +
						e.what());

					// Report error on this connection
					if(!socksconn->closing()) {
						scheduleConnectionError(socksconn->getID(), Tunnel::GENERAL_COMMS_ERROR);
					}
				}
			}
			else {
				// Abandon this tunnel
				DPRINTI("Failed to lookup SOCKS connection on tunnel");
				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
			}
		}
	}
	else {
		std::stringstream wrn;
		wrn << "State change notification for unknown tunnel " << tid;
		g_pLogger->warn(wrn.str());
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Create config for tunnel to server
TunnelConfig* SOCKSClientTunneler::createTunnelConfig(SOCKSConnection* inConn)
{
	// Create tunnelID
	ClientSideConfig* clientCfg = getClientSideConfigMutable();
	TunnelID* tid = new TunnelID(clientCfg);

	// Get destination IP & port from SOCKS connection
	std::string destIP = inConn->getSOCKSDestIP();
	unsigned int destPort = inConn->getSOCKSDestPort();

	// Create tunnel config
	return new TunnelConfig(tid, clientCfg->getTunTune(), clientCfg->getProxySettings(),
		clientCfg->getAuthenticator()->getDefaultAuthSettings(), destIP, destPort);
}


}	// namespace


