/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of a raw tunnel
 * SOCKS connection.
 *
 * $Id: SOCKSConnection.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "SOCKSConnection.hh"
#include "Tunneler.hh"
#include "GlobalConfig.hh"
#include "Debug.hh"
#include "SOCKSReader.hh"
#include "RawWriter.hh"
#include "TunnelID.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;


// Construction/destruction
SOCKSConnection::SOCKSConnection(Sock* sock, Tunneler* tunneler)
: Connection(sock, tunneler)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "SOCKSConnection::SOCKSConnection(server) - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// Construct reader/writer
	createReaderWriter();
}

SOCKSConnection::~SOCKSConnection()
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "SOCKSConnection::~SOCKSConnection() - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// We are closing down
	MutexVar cmv(m_connMtx);
	m_closing = true;
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get the IP address of the SOCKS destination
std::string SOCKSConnection::getSOCKSDestIP(void) const
{
	SOCKSReader* reader = dynamic_cast<SOCKSReader*>(m_reader);
	return reader->getDestIP();
}

// Get the port of the SOCKS destination
unsigned int SOCKSConnection::getSOCKSDestPort(void) const
{
	SOCKSReader* reader = dynamic_cast<SOCKSReader*>(m_reader);
	return reader->getDestPort();
}


/************/
/* Mutators */
/************/


// Tell our SOCKS reader what tunnel ID they are working for
void SOCKSConnection::setTunnelID(const TunnelID& tid)
{
	MutexVar mv(m_connMtx);
	SOCKSReader* reader = dynamic_cast<SOCKSReader*>(m_reader);
	reader->setTunnelID(tid);
}

// Notification that the tunnel we carry has been established
void SOCKSConnection::tunnelConnected(bool tunOk) throw(CommsError)
{
	MutexVar mv(m_connMtx);
	SOCKSReader* reader = dynamic_cast<SOCKSReader*>(m_reader);
	reader->tunnelConnected(tunOk);
}


/***********************/
/* Protected Interface */
/***********************/


// Construct reader/writer
void SOCKSConnection::createReaderWriter(void)
{
	// Create raw reader & writer
	MutexVar mv(m_connMtx);
	m_reader = new SOCKSReader(this);
	m_writer = new RawWriter(this);

	// Only truly connected once reader & writer are constructed
	setConnected();
}


}	// namespace

