/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A SOCKS data reader.
 *
 * $Id: SOCKSReader.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include <iomanip>
#include "SOCKSReader.hh"
#include "Connection.hh"
#include "TunnelID.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "Tunneler.hh"
#include "ConnectionStateEvent.hh"


namespace HTTPTunnel {


// Construction/destruction
SOCKSReader::SOCKSReader(Connection* conn)
: RawReader(conn), m_state(SOCKS_VERSION), m_buff(), m_version(0),
  m_destIP(), m_destPort(0), m_atype(0x01)
{
	DPRINTV("SOCKSReader::SOCKSReader()");
}

SOCKSReader::~SOCKSReader()
{
	DPRINTV("SOCKSReader::~SOCKSReader()");
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get the IP address of the SOCKS destination
std::string SOCKSReader::getDestIP(void) const
{
	return m_destIP;
}

// Get the port of the SOCKS destination
unsigned int SOCKSReader::getDestPort(void) const
{
	return m_destPort;
}


/*************/
/* Utilities */
/*************/


// Notification of whether our tunnel connected ok
void SOCKSReader::tunnelConnected(bool tunOk) throw(CommsError)
{
	// Are we waiting for this notification?
	if(m_state == SOCKS_REQUEST_RSP) {
		socksRequestRsp(tunOk);
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Process some newly read SOCKS data
void SOCKSReader::processData(const unsigned char* data, unsigned int length)
	throw(ReadError)
{
	DPRINTV("SOCKSReader::processData()");

	// What SOCKS protocol stage are we in?
	switch(m_state) {
		case(SOCKS_VERSION): {
			socksVersion(data, length);
			break;
		}
		case(SOCKS_REQUEST_RCV): {
			socksRequestRcv(data, length);
			break;
		}
		case(SOCKS_REQUEST_RSP): {
			throw ReadError("SOCKS protocol error: Read in unexpected state");
			break;
		}
		case(SOCKS_CONNECTED): {
			// Let base class handle it
			RawReader::processData(data, length);
			break;
		}
		default: {
			throw ReadError("Unknown SOCKS protocol state");
		}
	}
}


/*********************/
/* Private Interface */
/*********************/


// Parse & respond to a SOCKS version message
void SOCKSReader::socksVersion(const unsigned char* data, unsigned int length)
	throw(ReadError)
{
	DPRINTV("SOCKSReader::socksVersion()");

	// Append read data to our buffer
	MutexVar mv(m_mtx);
	m_buff.append(data, length);

	// SOCKS version message is at least 2 bytes
	if(m_buff.getLength() >= 2) {
		const unsigned char* buff = m_buff.begin();
		m_version = *buff++;

		#ifdef DEBUG_TUN
		std::stringstream nfo1;
		nfo1 << "SOCKSReader::socksVersion() - Version " << (int)m_version;
		DPRINTI(nfo1.str());
		#endif

		if(m_version == 5) {
			// SOCKS 5
			unsigned char nmethods = *buff++;
			unsigned int expectedMsgLen = 2 + nmethods;

			// Got enough data for entire message?
			if(m_buff.getLength() == expectedMsgLen) {
				bool ok = false;

				for(unsigned int i = 0; i < nmethods; ++i) {
					// Get suggested method
					unsigned char method = *buff++;

					#ifdef DEBUG_TUN
					std::stringstream nfo1;
					nfo1 << "SOCKSReader::socksVersion() - Method " << (int)method;
					DPRINTI(nfo1.str());
					#endif

					// We currently only support "no authentication"
					if(method == 0) {
						// Build and send SOCKS response msg
						unsigned char response[] = { 0x05, 0x00 };
						socksSendMsg(response, sizeof(response));

						// We're all done here
						ok = true;
						break;
					}
				}

				// Responded ok?
				if(ok) {
					// Clear out message buffer
					m_buff.clear();

					// Advance state
					m_state = SOCKS_REQUEST_RCV;
				}
				else {
					// No suitable method suggested
					unsigned char response[] = { 0x05, 0xFF };
					socksSendMsg(response, sizeof(response));

					// Give up on this connection
					throw ReadError("SOCKS protocol error: Bad SOCKS5 method received");
				}
			}
			else {
				// Too much data?
				if(m_buff.getLength() > expectedMsgLen) {
					// SOCKS protocol error, give up
					throw ReadError("SOCKS protocol error: Bad SOCKS5 request length");
				}
			}
		}
		else if(m_version == 4) {
			// SOCKS4
			unsigned char command = *buff++;

			#ifdef DEBUG_TUN
			std::stringstream nfo1;
			nfo1 << "SOCKSReader::socksVersion() - Command " << (int)command;
			DPRINTI(nfo1.str());
			#endif

			// Got enough data for entire message?
			if(m_buff.getLength() >= 8) {
				bool ok = false;

				// Only currently support Connect
				if(command == 1) {
					// Get port
					unsigned int port = *buff++;
					port <<= 8;
					port |= *buff++;

					// IPv4 - 4 bytes
					std::stringstream address;
					for(int i = 0; i < 4-1; ++i)
						address << (int)(*buff++) << ".";
					address << (int)(*buff++);

					// Save address & port
					m_destIP = address.str();
					m_destPort = port;

					#ifdef DEBUG_TUN
					std::stringstream nfo;
					nfo << "SOCKS4 request address " << m_destIP;
					nfo << ", port " << m_destPort;
					DPRINTV(nfo.str());
					#endif

					// Valid IP address?
					if(m_destIP.substr(0, 6) == "0.0.0.") {
						// Skip over any username
						while(*buff != 0)
							buff++;
						buff++;

						// Look for SOCKS4a domain name
						address.str("");
						while(*buff != 0)
							address << *buff++;
						m_destIP = address.str();

						#ifdef DEBUG_TUN
						std::stringstream nfo2;
						nfo2 << "SOCKS4a request address " << m_destIP;
						nfo2 << ", port " << m_destPort;
						DPRINTV(nfo2.str());
						#endif
					}

					ok = true;
				}

				if(ok) {
					// Clear out message buffer
					m_buff.clear();

					// Advance state
					m_state = SOCKS_REQUEST_RSP;

					// Let tunneler try to establish tunnel to destination
					Tunneler* tunneler = m_conn->getTunneler();
					tunneler->scheduleEvent(new ConnectionStateEvent(tunneler,
						m_conn->getID(), Connection::SOCKSSETUP));
				}
				else {
					// No suitable command
					std::vector<unsigned char> response;
					response.push_back(0x00);
					response.push_back(0x5b);
					for(int i = 0; i < 2+4; i++)
						response.push_back(0x00);
					socksSendMsg(response);

					// Give up on this connection
					throw ReadError("SOCKS protocol error: Bad SOCKS4 command received");
				}
			}
			else {
				// SOCKS protocol error, give up
				throw ReadError("SOCKS protocol error: Bad SOCKS4 request length");
			}
		}
	}
}


// Parse & respond to a SOCKS5 request message
void SOCKSReader::socksRequestRcv(const unsigned char* data, unsigned int length)
	throw(ReadError)
{
	DPRINTV("SOCKSReader::socksRequestRcv()");

	// Append read data to our buffer
	MutexVar mv(m_mtx);
	m_buff.append(data, length);

	// Got minimal header?
	if(m_buff.getLength() >= 4) {
		const unsigned char* buff = m_buff.begin();
		#ifdef DEBUG_TUN
		unsigned char version = *buff++;
		#else
		buff++;
		#endif
		unsigned char cmd = *buff++;
		buff++;
		m_atype = *buff++;

		#ifdef DEBUG_TUN
		std::stringstream nfo1;
		nfo1 << "SOCKSReader::socksRequest() - Version " << (int)version << ", ";
		nfo1 << "Cmd " << (int)cmd << ", AType " << int(m_atype);
		DPRINTI(nfo1.str());
		#endif

		// Supported command?
		if(cmd != 0x01) {
			// Currently only support TCP connect
			throw ReadError("SOCKS protocol error: Unsupported request command");
		}

		// Get address
		std::stringstream address;
		switch(m_atype) {
			case(0x01): {
				// IPv4 - 4 bytes
				for(int i = 0; i < 4-1; ++i)
					address << (int)(*buff++) << ".";
				address << (int)(*buff++);
				break;
			}
			case(0x03): {
				// FQDN
				unsigned char len = *buff++;
				for(int i = 0; i < len; ++i)
					address << (char)(*buff++);
				break;
			}
			case(0x04): {
				// IPv6 - 16 bytes
				address << std::hex << std::setw(2) << std::setfill('0');
				for(int i = 0; i < 16-1; ++i)
					address << (int)(*buff++) << ":";
				address << (int)(*buff++);
				break;
			}
			default: {
				throw ReadError("SOCKS protocol error: Bad address type in request");
			}
		}

		// Get port
		unsigned int port = *buff++;
		port <<= 8;
		port |= *buff;

		// Save address & port
		m_destIP = address.str();
		m_destPort = port;

		#ifdef DEBUG_TUN
		std::stringstream nfo2;
		nfo2 << "SOCKS5 request address " << m_destIP << ", port " << m_destPort;
		DPRINTV(nfo2.str());
		#endif

		// Finished with msg buff
		m_buff.clear();

		// Advance state
		m_state = SOCKS_REQUEST_RSP;

		// Let tunneler try to establish tunnel to destination
		Tunneler* tunneler = m_conn->getTunneler();
		tunneler->scheduleEvent(new ConnectionStateEvent(tunneler,
			m_conn->getID(), Connection::SOCKSSETUP));
	}
}


// Send back a SOCKS connect response message
void SOCKSReader::socksRequestRsp(bool ok) throw(ReadError)
{
	DPRINTV("SOCKSReader::socksRequestRsp()");

	// Advance state
	MutexVar mv(m_mtx);
	m_state = SOCKS_CONNECTED;

	// SOCKS response msg
	std::vector<unsigned char> response;

	if(m_version == 5) {
		// Basic header fields
		response.push_back(0x05);
		if(ok)
			response.push_back(0x00);
		else
			response.push_back(0x01);
		response.push_back(0x00);
		response.push_back(0x01);

		// IPv4 address/Port
		for(int i = 0; i < 4+2; i++)
			response.push_back(0x00);
	}
	else if(m_version == 4) {
		// Basic header fields
		response.push_back(0x00);
		if(ok)
			response.push_back(0x5a);
		else
			response.push_back(0x5b);

		// Port/IPv4 address
		for(int i = 0; i < 2+4; i++)
			response.push_back(0x00);
	}

	// Send response out of band
	socksSendMsg(response);
}


// Send a SOCKS message - Caller holds mutex
void SOCKSReader::socksSendMsg(unsigned char* msg, size_t len) throw(ReadError)
{
	try {
		// Build and send SOCKS msg
		RawBuff oob(msg, len);
		m_conn->writeOutOfBand(oob);
	}
	catch(TunnelError& e) {
		// Rethrow
		throw ReadError("Failed to send SOCKS msg: " + e.what());
	}
}

// Send a SOCKS message - Caller holds mutex
void SOCKSReader::socksSendMsg(std::vector<unsigned char>& msg) throw(ReadError)
{
	try {
		// Build and send SOCKS msg
		RawBuff oob(msg);
		m_conn->writeOutOfBand(oob);
	}
	catch(TunnelError& e) {
		// Rethrow
		throw ReadError("Failed to send SOCKS msg: " + e.what());
	}
}


}	// namespace

