/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an SSL socket.
 *
 * $Id: SSLSock.hh,v 1.4 2008-11-21 17:59:39 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _SSL_SOCK_HH_
#define _SSL_SOCK_HH_

#ifdef USE_SSL

#include <pthread.h>
#include <bio.h>
#include <ssl.h>
#include "Sock.hh"

namespace HTTPTunnel {

class SSLSock : public Sock {
public:
	// Construction/destruction
	SSLSock(bool blocking = false) throw(SockError);
	SSLSock(std::string cert, std::string key, bool blocking = false) throw(SockError);
	SSLSock(int fd, std::string cliAddr, BIO* sbio, SSL* ssl) throw(SockError);
	virtual ~SSLSock();

	// Utilities
	virtual void newConnection(Tunneler* tunneler) throw(SockError);
	virtual void listen(void) throw(SockError);
	virtual Sock* accept(void) throw(SockError);
	virtual void connect(std::string addr, unsigned int port) throw(SockError, CommsError);
	virtual int read(unsigned char* buff, size_t buffSize) throw(SockError, CommsError);
	virtual int write(const unsigned char* buff, size_t buffSize) throw(SockError, CommsError);
	virtual void getConnectResult(bool& resultAvailable, bool& connected) throw(SockError);
	virtual void bind(std::string addr, unsigned int& port) throw(SockError);

	virtual bool pollRead(int timeout, bool& err, bool& hup) throw(SockError);
	virtual bool pollWrite(int timeout, bool& err) throw(SockError);

	// Initialise/close SSL library
	static void initSSL(void) throw(CommsError);
	static void shutdownSSL(void) throw(CommsError);


protected:

	/***********/
	/* Methods */
	/***********/

	// Close SSL session
	virtual void close(void) throw(SockError);

	// Setup blocking/nonblocking mode
	virtual void blocknonblock(void) throw(SockError);


private:

	/*************/
	/* Constants */
	/*************/

	// Maximum number of connect retries to attempt
	static const unsigned int		MAX_CONNECT_RETRY		= 20;


	/***********/
	/* Methods */
	/***********/

	// Forbid copying
	SSLSock(const SSLSock& that);
	SSLSock& operator=(const SSLSock& that);

	// Tidy up after a failed accept
	void failedAccept(BIO* sbio);


	/********/
	/* Data */
	/********/

	// Flag to indicate whether SSL library has been initialised/shutdown yet
	static bool						c_initSSL;
	static bool						c_shutdownSSL;

	// OpenSSL data
	SSL_CTX*							m_ctx;
	BIO*								m_abio;
	BIO*								m_sbio;
	SSL*								m_ssl;

	// Certificate/key files
	std::string						m_cert;
	std::string						m_certKey;

	// Count of number of times we've retried a connect
	unsigned int					m_connectRetry;

};


// OpenSSL callbacks
void pthreadsLockingCallback(int mode, int type, const char *file, int line);
unsigned long pthreadsThreadId(void);


}	// namespace

#else		// Not using SSL
#include "TCPSock.hh"
namespace HTTPTunnel {
class SSLSock : public TCPSock {
public:
	SSLSock() throw(SockError) {}
	SSLSock(std::string cert, std::string key) throw(SockError) {}
	static void initSSL(void) throw(CommsError) {}
	static void shutdownSSL(void) throw(CommsError) {}
};
}  // namespace
#endif	// Using SSL

#endif

