/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A TCP/IP socket server to listen for incoming connections.
 *
 * $Id: Server.cc,v 1.4 2008-11-22 20:44:00 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <errno.h>
#include "Server.hh"
#include "Debug.hh"
#include "Enums.hh"
#include "SideConfig.hh"
#include "ServerSideConfig.hh"
#include "Callback.hh"
#include "Exceptions.hh"
#include "Tunneler.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "TCPSock.hh"
#include "UDPSock.hh"
#include "SSLSock.hh"
#include "Time.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
Server::Server(SideConfig* serverConfig) throw(TunnelError)
: m_thread(0), m_shutDown(false), m_serverConfig(serverConfig), m_sock(0),
  m_tunneler(0), m_serverThreadStartMtx(), m_serverThreadStartCond(),
  m_serverThreadContMtx(), m_serverThreadContCond(), m_threadMgr(), m_mtx()
{
}

Server::~Server()
{
	// Take mutexes
	MutexVar mv(m_mtx);

	// Signal we are no longer running
	m_shutDown = true;

	// Shutdown threads
	mv.unlock();
	m_threadMgr.shutdown();
	mv.lock();
	m_thread = 0;

	// Finished with sockets
	if(m_sock) {
		delete m_sock;
		m_sock = 0;
	}

	// Delete tunneler
	delete m_tunneler;
	m_tunneler = 0;
}


/*************/
/* Accessors */
/*************/


// Get config settings
const SideConfig* Server::getConfig(void) const
{
	return m_serverConfig;
}

SideConfig* Server::getConfigMutable(void)
{
	return m_serverConfig;
}


// Get tunneler
Tunneler* Server::getTunneler(void)
{
	return m_tunneler;
}


// Get tunnel stats for our tunnelers carried tunnels
void Server::getTunnelStats(std::vector<TunnelStats>& stats)
{
	MutexVar mv(m_mtx);

	// Get IDs for active tunnels
	std::vector<TunnelID> tids;
	m_tunneler->getActiveTunnelIDs(tids);
	for(std::vector<TunnelID>::iterator tit = tids.begin(); tit != tids.end(); ++tit) {
		// Get & copy stats for this tunnel
		TunnelStats* ts = m_tunneler->getTunnelStats(*tit);
		stats.push_back(*ts);

		// Free stats
		delete ts;
	}
}


/***********************/
/* Protected interface */
/***********************/


// Wait for arriving connections (never returns)
void Server::startServer(void)
{
	// Notify launching thread we have started
	MutexVar stcMtx(m_serverThreadContMtx);
	MutexVar stsMtx(m_serverThreadStartMtx);
	m_serverThreadStartCond.signal();
	stsMtx.unlock();

	// Wait for permission to start accepting connections
	m_serverThreadContCond.wait(stcMtx);
	stcMtx.unlock();
	DPRINTI("Server::startServer() - Waiting for connection");

	// Take mutex
	MutexVar mv(m_mtx);

	while(!m_shutDown) {
		try {
			// Poll socket until we think it's worth trying to accept()
			bool pollError, pollHup;
			if(m_sock->pollRead(0, pollError, pollHup)) {
				// Socket connection pending
				m_sock->newConnection(m_tunneler);
			}
			else {
				// Poll timeout, wait a while before retrying
				Time pollTime(0, POLL_TIMEOUT);
				m_thread->sleep(mv, pollTime.getTimeVal());
			}
		}
		catch(SockError& e) {
			if(e.m_errno == EINTR) {
				// Interrupted system call
				continue;
			}
			else {
				// General error, just carry on
				continue;
			}
		}
	}

	DPRINTI("Server closing down");
}


// Configure our server socket
void Server::configureSocket(bool ssl) throw(TunnelError)
{
	// Create listening socket
	if(ssl) {
		// Use SSL
		if(m_serverConfig->isServer()) {
			// Need both SSL certificate & private key at server
			ServerSideConfig* scfg = dynamic_cast<ServerSideConfig*>(m_serverConfig);
			m_sock = new SSLSock(scfg->getCertificate(), scfg->getCertificateKey());
		}
		else {
			// Don't require anything at client
			m_sock = new SSLSock();
		}
	}
	else {
		// Unencrypted comms
		if(m_serverConfig->getProtocol() == Enums::TCP) {
			m_sock = new TCPSock();
		}
		else {
			m_sock = new UDPSock();
		}
	}

	// Get address & port to bind server on
	std::string listenIP = m_serverConfig->getListenIP();
	unsigned int listenPort = m_serverConfig->getListenPort();
	if(listenIP == "-") {
		// Pick our own port to bind to on loopback
		listenIP = "127.0.0.1";
		listenPort = 0;
	}

	// Bind to requested address
	m_sock->bind(listenIP, listenPort);
	m_sock->listen();

	// Update listening address & port incase they've changed
	m_serverConfig->setListenIP(listenIP);
	m_serverConfig->setListenPort(listenPort);

	// Launch server
	ZeroArgClassCallback<Server> cb(&Server::startServer, this);
	m_thread = m_threadMgr.newThread(cb);
}


}	// namespace

