/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate settings for server authentication.
 *
 * $Id: ServerAuthSettings.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <sstream>
#include <iomanip>
#include "ServerAuthSettings.hh"
#include "Logger.hh"
#include "Debug.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
ServerAuthSettings::ServerAuthSettings()
: AuthSettings()
{
}

ServerAuthSettings::ServerAuthSettings(const ServerAuthSettings& that)
: AuthSettings(that)
{
}

ServerAuthSettings::~ServerAuthSettings()
{
}


/************/
/* Mutators */
/************/


// Setup server side authentication settings
void ServerAuthSettings::setServerSideAuth(AuthScheme scheme, std::string realm)
	throw(AuthProcessingError)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "ServerAuthSettings::setServerSideAuth() - ";
	dbg << "Scheme: " << getAuthSchemeStr(scheme) << ", Realm: " << realm;
	DPRINTV(dbg.str());
	#endif

	m_authScheme = scheme;
	m_realm = realm;

	// What authentication scheme?
	if(m_authScheme == AUTH_DIGEST) {
		// Create initial digest CNonce
		m_digestCNonceCount = 1;
		std::stringstream cnonce;
		cnonce << std::setw(8) << std::hex << std::setfill('0') << rand();
		m_digestCNonce = cnonce.str();
	}
	else {
		// Unsupported authentication scheme at the server
		throw AuthProcessingError("Unsupported server authentication scheme");
	}
}


/*************/
/* Utilities */
/*************/


// Construct tunneling server response header
std::string ServerAuthSettings::getResponseHeader(std::string uri)
{
	std::stringstream res;

	// Using authorization?
	if(m_authScheme == AUTH_BASIC) {
		// Create basic authentication fields
		res << "WWW-Authenticate: Basic realm=\"" << m_realm << "\"\r\n";

		// Remember trying basic authentication
		finishedAuthScheme(AUTH_BASIC);
	}
	else if(m_authScheme == AUTH_DIGEST) {
		// Create digest authentication fields
		res << "WWW-Authenticate: Digest realm=\"" << m_realm << "\", ";
		res << "nonce=\"" << m_digestNonce << "\", ";
		res << "qop=\"" << m_digestQop << "\", ";
		res << "stale=false";
		res << "\r\n";

		// Remember trying digest authentication
		finishedAuthScheme(AUTH_DIGEST);
	}

	return res.str();
}


// Clear and reset any exisiting authentication scheme details
void ServerAuthSettings::reset(void)
{
	// Reset current scheme
	m_authScheme = AUTH_NONE;

	// Clear tried schemes
	m_triedSchemes.clear();
}


/***********************/
/* Protected Interface */
/***********************/


// Get authorisation request header field identifier
std::string ServerAuthSettings::getAuthRequestHeaderID(void) const
{
	return "Authorization";
}


}	// namespace

