/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user at the server, and
 * tracking/controling resource usage.
 *
 * $Id: ServerAuthenticator.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include "ServerAuthenticator.hh"
#include "Logger.hh"
#include "Utils.hh"
#include "MutexVar.hh"
#include "Tunnel.hh"
#include "Debug.hh"
#include "ServerAuthSettings.hh"
#include "AuthDigest.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
ServerAuthenticator::ServerAuthenticator() throw(AuthError)
: Authenticator()
{
}

ServerAuthenticator::~ServerAuthenticator()
{
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get whether user is authenticated for some tunnel
bool ServerAuthenticator::getUserAuthenticated(std::string name, Tunnel* tun)
{
	bool res = true;

	MutexVar mv(m_mtx);

	// Lookup user in cached details
	User* user = getUser(name);
	if(user) {
		// Is user authenticated for this tunnel?
		res = user->getAuthenticated(tun);
	}

	return res;
}


/*************/
/* Utilities */
/*************/


// Log details about a new tunnel connection
unsigned int ServerAuthenticator::logNewConnection(ServerTunnel* tun)
{
	// Do nothing by default
	return 0;
}


// Log details about a closing tunnel
void ServerAuthenticator::logCloseTunnel(ServerTunnel* tun)
{
	// Do nothing by default
}


/***********************/
/* Protected Interface */
/***********************/


// Create a new tunnel user
User ServerAuthenticator::createUser(std::string name, bool keepCached)
{
	// Create user un-authenticated
	Quotas quotas;
	User user(false, name, "", quotas, keepCached);
	return user;
}


// Check a digest authentication response
bool ServerAuthenticator::checkDigestAuthResponse(std::string& user,
	std::string& passwd, ServerAuthSettings* settings,
	const std::vector<std::string>& authArgs)
{
	bool res = false;

	// Get required authentication parameters
	std::string realm, nonce, uri, alg, qop, cnonce, nc, response;
	if(Utils::getParmValue(authArgs, "realm", realm) &&
	Utils::getParmValue(authArgs, "nonce", nonce) &&
	Utils::getParmValue(authArgs, "uri", uri) &&
	Utils::getParmValue(authArgs, "algorithm", alg) &&
	Utils::getParmValue(authArgs, "qop", qop) &&
	Utils::getParmValue(authArgs, "cnonce", cnonce) &&
	Utils::getParmValue(authArgs, "nc", nc) &&
	Utils::getParmValue(authArgs, "response", response)) {
		// Calculate expected response & compare
		std::string calcresponse = AuthDigest::MD5DigestResponse(alg, user,
			passwd, realm, nonce, cnonce, nc, qop, uri);
		res = (response == calcresponse);
	}

	return res;
}


}	// namespace

