/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration for
 * a server tunneler.
 *
 * $Id: ServerSideConfig.cc,v 1.4 2008-11-09 18:50:21 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "ServerSideConfig.hh"
#include "Enums.hh"
#include "ServerAuthSettings.hh"
#include "GlobalConfig.hh"
#include "Firewall.hh"
#include "SockTunCallback.hh"

#ifdef AUTH_DB
#include "DBServerAuthenticator.hh"
#endif
#ifdef AUTH_FILE
#include "FileServerAuthenticator.hh"
#endif
#include "DefaultServerAuthenticator.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;


// Construction/destruction
#ifdef AUTH_FILE
ServerSideConfig::ServerSideConfig(bool ssl, bool sockTun, std::string cert,
	std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth, std::map<std::string, std::string> &passwds,
	std::map<std::string, unsigned int> &tunnelnos,
	std::map<std::string, unsigned int> &bandwidths) throw(AuthError)
: SideConfig(Enums::TCP, ssl, sockTun, lhost, lport), m_useAuth(auth), m_firewall(0),
  m_cert(cert), m_certKey(certKey), m_userCallback(0), m_passwds(passwds),
  m_tunnelnos(tunnelnos), m_bandwidths(bandwidths)
{
	// Create firewall
	m_firewall = new Firewall(firewallIPs, firewallPorts);
}
#else
ServerSideConfig::ServerSideConfig(bool ssl, bool sockTun, std::string cert,
	std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth) throw(AuthError)
: SideConfig(Enums::TCP, ssl, sockTun, lhost, lport), m_useAuth(auth), m_firewall(0),
  m_cert(cert), m_certKey(certKey), m_userCallback(0)
{
	// Create firewall
	m_firewall = new Firewall(firewallIPs, firewallPorts);
}
#endif

ServerSideConfig::~ServerSideConfig()
{
	// Destroy firewall
	if(m_firewall) {
		delete m_firewall;
		m_firewall = 0;
	}

	// Free user callback
	if(m_userCallback) {
		delete m_userCallback;
		m_userCallback = 0;
	}
}


/*************/
/* Accessors */
/*************/


// Server config?
bool ServerSideConfig::isServer(void) const
{
	return true;
}

// Get firewall
Firewall* ServerSideConfig::getFirewall(void)
{
	return m_firewall;
}

// Get SSL certificate
std::string ServerSideConfig::getCertificate(void) const
{
	return m_cert;
}

// Get SSL private key
std::string ServerSideConfig::getCertificateKey(void) const
{
	return m_certKey;
}

// Get user callback
const SockTunCallback* ServerSideConfig::getUserCallback(void) const
{
	return m_userCallback;
}

#ifdef AUTH_FILE
std::string ServerSideConfig::getPasswd(std::string user)
{
	return m_passwds[user];
}

unsigned int ServerSideConfig::getNoTunnels(std::string user)
{
	return m_tunnelnos[user];
}

unsigned int ServerSideConfig::getBandwidth(std::string user)
{
	return m_bandwidths[user];
}
#endif


/************/
/* Mutators */
/************/


// Turn authentication on/off
void ServerSideConfig::setAuthenticate(bool auth) throw(AuthError)
{
	m_useAuth = auth;
}

// Set user callback
void ServerSideConfig::setUserCallback(const SockTunCallback* cb)
{
	m_userCallback = cb->clone();
}


/********************/
/* Public Interface */
/********************/


// Create authenticator
void ServerSideConfig::createAuthenticator(void) throw(AuthError)
{
	// Create new one
	if(m_useAuth) {
		#ifdef AUTH_DB
		// Create database authenticator
		m_authenticator = new DBServerAuthenticator();

		#elif defined(AUTH_FILE)
		// Create database authenticator
		m_authenticator = new FileServerAuthenticator();

		#else
		// Create default authenticator
		m_authenticator = new DefaultServerAuthenticator();

		#endif
	}
	else {
		// Create default authenticator
		m_authenticator = new DefaultServerAuthenticator();
	}
}


// Serialize to a file
void ServerSideConfig::serialize(std::ofstream& file)
{
	// Server header
	file << "#######################" << std::endl;
	file << "## Server configuration" << std::endl;
	file << "#######################" << std::endl;
	file << std::endl;
	file << "[SERVER]" << std::endl;

	// Local listening address
	file << "listen ";
	if(getListenIP().size() > 0) {
		file << getListenIP() << " ";
	}
	file << getListenPort() << std::endl;

	// Server authentication
	if(m_useAuth) {
		file << "svrauth" << std::endl;
	}

	// SSL certificate & key
	if(m_ssl) {
		file << "ssl " << getCertificate() << " " << getCertificateKey();
		file << std::endl;
	}

	// Firewall settings
	m_firewall->serialize(file);

	file << std::endl;
}


}	// namespace

