/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Server side Tunnel.
 *
 * $Id: ServerTunnel.cc,v 1.3 2008-07-11 11:22:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <sstream>
#include "ServerTunnel.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "GlobalConfig.hh"
#include "Authenticator.hh"
#include "ServerAuthenticator.hh"
#include "Utils.hh"
#include "HTTPMsg.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "HTTPConnection.hh"
#include "ServerTunneler.hh"
#include "AuthSuccessEvent.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ServerTunnel::ServerTunnel(TunnelConfig* tcfg, ServerTunneler* tunneler)
: Tunnel(tcfg, tunneler, LEFT, AUTHREQ), m_logKey(0), m_authFailureCount(0),
  m_cliVerStr()
{
	DPRINTI("ServerTunnel::ServerTunnel()");
}

ServerTunnel::~ServerTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "ServerTunnel::~ServerTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif

	MutexVar mv(m_mtx);

	// Tell user they own one less tunnel
	ServerAuthenticator* authenticator = getServerAuthenticator();
	authenticator->userTunnelClosed(m_user, this);

	// Log tunnel closed
	authenticator->logCloseTunnel(this);
}


/*************/
/* Accessors */
/*************/


// Get source addresses of connections for this tunnel
std::string ServerTunnel::getSrcConnectionAddresses(void)
{
	std::stringstream res;

	MutexVar mv(m_mtx);

	// Get addresses for left side connections
	std::vector<Connection*>::iterator cit;
	bool comma = false;
	for(cit = m_leftConn.begin(); cit != m_leftConn.end(); ++cit) {
		// Should only be HTTPConnections in server side tunnel
		HTTPConnection* conn = dynamic_cast<HTTPConnection*>(*cit);
		if(conn) {
			// Do we have a client address?
			std::string cliAddr = conn->getClientAddr();
			if(cliAddr.size() > 0) {
				if(comma)
					res << ",";
				else
					comma = true;
				res << cliAddr;
			}
		}
	}

	return res.str();
}

// Get destination addresses of connections for this tunnel
std::string ServerTunnel::getDstConnectionAddresses(void)
{
	std::stringstream res;

	MutexVar mv(m_mtx);

	// Get addresses for right side connections
	std::vector<Connection*>::iterator cit;
	bool comma = false;
	for(cit = m_rightConn.begin(); cit != m_rightConn.end(); ++cit) {
		// Should only be RawConnections in server side tunnel
		RawConnection* conn = dynamic_cast<RawConnection*>(*cit);
		if(conn) {
			// Do we have a destination address?
			std::string destAddr = conn->getDestIP();
			unsigned int destPort = conn->getDestPort();
			if(destPort > 0) {
				if(comma)
					res << ",";
				else
					comma = true;
				res << destAddr << ":" << destPort;
			}
		}
	}

	return res.str();
}


// Get logging key
unsigned int ServerTunnel::getLogKey(void) const
{
	return m_logKey;
}

// Get client version identifier
std::string ServerTunnel::getClientVersionStr(void) const
{
	return m_cliVerStr;
}


/************/
/* Mutators */
/************/


// A connection that carries us has connected
void ServerTunnel::connectionConnected(Connection* conn)
{
	MutexVar mv(m_mtx);

	// Base class
	Tunnel::connectionConnected(conn);

	// Maybe log this new connection
	if(m_tunnelState == CONNECTED) {
		ServerAuthenticator* authenticator = getServerAuthenticator();
		m_logKey = authenticator->logNewConnection(this);
	}
}


/********************/
/* Public Interface */
/********************/


// Process some newly arrived tunnel msg from one of our connections
void ServerTunnel::processTunnelMsg(HTTPMsg* msg, HTTPConnection* httpConn)
{
	DPRINTV("ServerTunnel::processTunnelMsg() - " + m_tid.toString());

	MutexVar mv(m_mtx);

	// Do we need to try and authenticate client?
	if(!userAuthenticated()) {
		DPRINTV("Client not authenticated");
		authenticateClient(msg->getAuthArgs());
	}

	// Is the client authenticated now?
	if(userAuthenticated()) {
		// Base class
		Tunnel::processTunnelMsg(msg, httpConn);
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Create complete HTTP msg parameters for sending down tunnel - Caller hold mutex
HTTPMsgParms* ServerTunnel::makeTunnelMsg(const HTTPMsgParms* recvd, bool retry)
	throw(MemoryError)
{
	// Get basic parameters from base class
	HTTPMsgParms* parms = Tunnel::makeTunnelMsg(recvd, retry);

	// Do we need to authenticate the client?
	if(!userAuthenticated()) {
		parms->setRespCodePhrase(401, "Unauthorized");
	}

	// Is the client version ok?
	if(recvd->getState() == CONNECT) {
		m_cliVerStr = recvd->getCliVer();

		// Split version fields
		std::vector<std::string> verFields;
		Utils::SplitString(m_cliVerStr, " ", verFields);
		std::string cliVerNum = verFields[0];

		// Split version number
		std::vector<std::string> verNumFields;
		Utils::SplitString(cliVerNum, ".", verNumFields);
		if(verNumFields.size() == 3) {
			unsigned int cliMajor = atoi(verNumFields[0].c_str());
			unsigned int cliMinor = atoi(verNumFields[1].c_str());
			unsigned int cliPoint = atoi(verNumFields[2].c_str());

			// Check version fields
			bool verOk = false;
			if(cliMajor > MIN_CLIVER_MAJOR) {
				verOk = true;
			}
			else if(cliMajor == MIN_CLIVER_MAJOR) {
				if(cliMinor > MIN_CLIVER_MINOR) {
					verOk = true;
				}
				else if(cliMinor == MIN_CLIVER_MINOR) {
					if(cliPoint >= MIN_CLIVER_POINT) {
						verOk = true;
					}
				}
			}

			if(!verOk) {
				// Client too old, so reject them
				std::stringstream nfo;
				nfo << "Client version " << cliVerNum << " is too old, rejecting them";
				g_pLogger->info(nfo.str());

				// Tell client they are forbidden
				parms->setRespCodePhrase(403, "Forbidden");
			}
		}
		else {
			// Reject client
			std::stringstream wrn;
			wrn << "Bad format for client version identifier: " << m_cliVerStr;
			g_pLogger->warn(wrn.str());

			// Tell client they are forbidden
			parms->setRespCodePhrase(403, "Forbidden");
		}
	}

	return parms;
}


// Successfully sent a new message down the tunnel - Caller hold mutex
void ServerTunnel::sentNewTunnelMsg(HTTPConnection* conn, HTTPMsgParms* parms)
{
	// Only remember if the sent msg is part of a normal authenticated tunnel session
	bool remember = userAuthenticated();
	Tunnel::sentNewTunnelMsg(conn, parms, remember);
}


// Determine whether a multi-ack implies we should resend a msg - Caller holds mutex
bool ServerTunnel::checkForMultiAckResendRequest(SentMsgParms* sentMsgParms)
{
	bool ackCountExceeded = false;

	// Update ack count
	sentMsgParms->m_ackCount++;

	/* The client may send several msgs over a persistant connection with
	 * the same ack'd sequence number. So don't consider this a resend request
	 * unless they've exceeded the maximum possible unacked count.
	 */

	if(sentMsgParms->m_ackCount >= m_maxUnackedMsgs) {
		DPRINTV("ServerTunnel::checkForMultiAckResendRequest() - Multi-Ack resend request");
		ackCountExceeded = true;
		sentMsgParms->m_ackCount = 0;
	}

	return ackCountExceeded;
}


// Can we send a new tunnel msg? - Caller holds mutex
bool ServerTunnel::clearToSendNewMsg(HTTPConnection* conn)
{
	bool sendNew = false;

	try {
		// Check for unacked management msgs and consider resending them early
		checkUnackedMgmtMsgs();

		// New message or resend of old message?
		sendNew = !resendUnackedMsgs(conn);
	}
	catch(ResendError& e) {
		// Resending error
		g_pLogger->warn("Error resending msg: " + e.what());
		recordConnectionError(MSG_RESEND_ERROR);
	}

	return sendNew;
}


// Check whether any connections need to be blocked/unblocked - Caller holds mutex
void ServerTunnel::checkBlockedConnections(void)
{
	// Check raw connections
	std::vector<Tunnel::Selector*> filt;
	filt.push_back(new SelRightConn());
	std::vector<Connection*> conns;
	getConnsFromSel(filt, conns);

	// Do we want to block or unblock?
	bool unblock = false;
	bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
	if(!block) {
		// Only unblock if we're well below full
		unsigned int buffFill = (unsigned int)(((float)m_rawDataQueueSize /
			(float)MAX_RAW_DATA_FILL) * 100);
		unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
	}

	// Check each connection
	std::vector<Connection*>::iterator cit;
	for(cit = conns.begin(); cit != conns.end(); ++cit) {
		// Get connection
		Connection* conn = *cit;

		// Need to change blocked state?
		if(block)
			conn->setBlocked(true);
		else if(unblock)
			conn->setBlocked(false);
	}
}


// Advance our internal state machine - Caller holds mutex
void ServerTunnel::advanceState(const HTTPMsgParms* parms)
{
	// Respond to the reported state from this msg
	switch(parms->getState()) {
		case(CONNECT): {
			// Now we need to authenticate
			setState(AUTHREQ);
			break;
		}
		case(RECONNECT): {
			// Reconnected
			setState(CONNECTED);
			break;
		}
		case(CLOSING): {
			// Closing
			setState(CLOSING);
			break;
		}
		case(CLOSED): {
			// Closed
			setState(CLOSED);
			break;
		}
		case(CONNECTED): {
			// Stay connected
			setState(CONNECTED);
			break;
		}
		case(AUTHREQ):
		case(UNKNOWNSTATE):
		default: {
			std::stringstream err;
			err << "Unexpected ServerTunnel reported msg state: " << parms->getState();
			g_pLogger->error(err.str());
			break;
		}
	}
}

// The tunnel state has changed - Caller holds mutex
void ServerTunnel::stateChanged(TunnelState oldState)
{
	// Base class
	Tunnel::stateChanged(oldState);

	// Log if we've become CONNECTED
	if(m_tunnelState == CONNECTED) {
		ServerAuthenticator* authenticator = getServerAuthenticator();
		m_logKey = authenticator->logNewConnection(this);
	}
}


// Get maximum allowable payload size - Caller holds mutex
unsigned int ServerTunnel::getMaxPayloadSize(bool useHeader)
{
	// Get payload size from server tunnel parameters
	ServerTunneler* svrTunneler = dynamic_cast<ServerTunneler*>(m_tunneler);
	ServerTunnelParms tunParms = svrTunneler->getTunParms(m_tid.getTunnelerID());
	return tunParms.getMaxPayloadSize(useHeader);
}


// Get server side authenticator
ServerAuthenticator* ServerTunnel::getServerAuthenticator(void)
{
	// Get base authenticator
	Authenticator* auth = getAuthenticator();

	// Convert to ServerAuthenticator and return
	ServerAuthenticator* svrAuth = dynamic_cast<ServerAuthenticator*>(auth);
	return svrAuth;
}


/*********************/
/* Private Interface */
/*********************/


// Try to authenticate a client - Caller holds mutex
void ServerTunnel::authenticateClient(const std::map<std::string,
	std::vector<std::string> >& authDetails)
{
	// Get authentication arguments for a suitable scheme
	std::vector<std::string> authArgs;
	if(authDetails.size() > 0)
		authArgs = authDetails.begin()->second;

	// Get supplied user name
	std::string user;
	if(Utils::getParmValue(authArgs, "username", user)) {
		// Use supplied user name
		m_user = user;
	}

	// First check quotas for number of tunnels
	Authenticator* authenticator = getAuthenticator();
	if(authenticator->getUserCanCreateTunnel(m_user) || isTestTunnel()) {
		// Try to authenticate
		authenticator->authenticate(m_user, this, authArgs);
		if(userAuthenticated()) {
			DPRINTV("Client authentication OK");
			m_authFailureCount = 0;

			// Let controlling tunneler know
			m_tunneler->scheduleEvent(new AuthSuccessEvent(m_tunneler, m_tid));
		}
		else {
			DPRINTV("Client authentication failed");
			m_authFailureCount++;

			// Give up if the client keeps failing to authenticate
			if(m_authFailureCount >= MAX_AUTH_FAILURES)
				recordConnectionError(CLIENT_AUTH_FAILED_ERROR);
		}
	}
	else {
		// User trying to exceed tunnel number quota
		std::stringstream nfo;
		nfo << "User " << m_user << " exceeding tunnel count quota, ";
		nfo << "not allowing them to authenticate";
		g_pLogger->info(nfo.str());
		recordConnectionError(FATAL_QUOTAS_ERROR);
	}
}


// Is this tunnel user authenticated? - Caller holds mutex
bool ServerTunnel::userAuthenticated(void)
{
	Authenticator* authenticator = getAuthenticator();
	return authenticator->getUserAuthenticated(m_user, this);
}


}	// namespace

