/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for server end of tunnel.
 *
 * $Id: ServerTunneler.cc,v 1.7 2009-03-20 11:52:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "Tunneler.hh"
#include "ServerTunneler.hh"
#include "Debug.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "DummyRawConnection.hh"
#include "HTTPConnection.hh"
#include "GlobalConfig.hh"
#include "ServerSideConfig.hh"
#include "Logger.hh"
#include "ServerTunnel.hh"
#include "TestServerTunnel.hh"
#include "HTTPMsg.hh"
#include "HTTPRequestMsg.hh"
#include "ShutdownTunnelEvent.hh"
#include "CheckAckEvent.hh"
#include "HTTPMsgParms.hh"
#include "ErrorMsgParms.hh"
#include "Authenticator.hh"
#include "ConnectionManager.hh"
#include "Sock.hh"
#include "TCPSockTunServer.hh"
#include "UDPSockTunServer.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ServerTunneler::ServerTunneler(Server* server) throw(TunnelError)
: Tunneler(server), m_tunParms(), m_tcpSockTunServer(0), m_udpSockTunServer(0)
{
	DPRINTV("ServerTunneler::ServerTunneler()");

	try {
		// Create authenticator
		ServerSideConfig* serverCfg = getServerSideConfigMutable();
		serverCfg->createAuthenticator();

		// See if we need a socket tunnel server
		if(serverCfg->isSockTun()) {
			m_tcpSockTunServer = new TCPSockTunServer(serverCfg->getUserCallback());
			m_udpSockTunServer = new UDPSockTunServer(serverCfg->getUserCallback());
		}
	}
	catch(TunnelError& e) {
		// Clear up and rethrow
		shutdownSchedulers();
		throw;
	}
}

ServerTunneler::~ServerTunneler()
{
	DPRINTV("ServerTunneler::~ServerTunneler()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();

	// Shutdown any socket tunnel servers
	if(m_tcpSockTunServer) {
		delete m_tcpSockTunServer;
		m_tcpSockTunServer = 0;
	}
	if(m_udpSockTunServer) {
		delete m_udpSockTunServer;
		m_udpSockTunServer = 0;
	}
}


/*************/
/* Accessors */
/*************/


// Get reference to current tunneling parameters
ServerTunnelParms ServerTunneler::getTunParms(const TunnelerID& tunnelerID)
{
	// Assume default parms unless we find better
	ServerTunnelParms parms = TestServerTunnel::getDefaultTunParms();

	// Lookup parms for the specified client tunneler ID
	MutexVar tmv(m_tunnelsMtx);
	std::map<TunnelerID, ServerTunnelParms>::iterator pit = m_tunParms.find(tunnelerID);
	if(pit != m_tunParms.end()) {
		// Got parameters for specified tunneler ID
		parms = pit->second;
	}

	return parms;
}


/**********/
/* Events */
/**********/


// Notification of a new socket connection
void ServerTunneler::newConnectionArrived(Sock* insock)
	throw(CommsError)
{
	DPRINTI("ServerTunneler::newConnectionArrived()");

	//////////////////////////////////////////////////////////////////
	// In ServerTunneler we can't really do much until we are told
	// whether this is a connection from a new tunnel, or a new pipe
	// within an existing tunnel.
	//////////////////////////////////////////////////////////////////

	// Create HTTP connection to handle incoming traffic
	MutexVar cmv(m_connectionsMtx);
	HTTPConnection* httpconn = new HTTPConnection(insock, this);
	addConnection(httpconn);
}


// Notification that a tunnel has insufficient left/right connections to be viable
void ServerTunneler::unviableTunnel(const TunnelID& tid)
{
	DPRINTI("ServerTunneler::unviableTunnel() - " + tid.toString());

	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// If tunnel is already closed don't bother checking any further
		if(!tun->isClosed()) {
			// Get details about state of tunnel
			unsigned int leftConns = tun->noLeftConnections();
			unsigned int rightConns = tun->noRightConnections();

			// Check whether our left side warrants preserving
			bool leftPreserve = true;
			if(leftConns == 0) {
				/* In the server, no remaining left connections means that if
				 * we want to try and preserve this tunnel we should wait for
				 * a re-connection from the client.
				 */

				// Do we want to try and preserve this tunnel?
				if(!tun->errorLevelExceeded(Tunnel::LEFT)) {
					// Consider preserving left side
					DPRINTI(tid.toString() + " - No lconns, not erred: Left preserve yes");
					leftPreserve = true;
				}
				else {
					// Don't consider preserving left side
					DPRINTI(tid.toString() + " - No lconns, erred: Left preserve no");
					leftPreserve = false;
				}
			}

			// Check whether our right side warrants preserving
			bool rightPreserve = true;
			if(rightConns == 0) {
				if(!tun->everConnected(Tunnel::RIGHT)) {
					// Never been connected right, so might want to let it try
					if(!tun->errorLevelExceeded()) {
						// Haven't failed to connect right, so let it try
						DPRINTI(tid.toString() + " - No rconns, not erred: Right preserve yes");
						rightPreserve = true;
					}
					else {
						// Maybe failed to connect right, don't try anymore
						DPRINTI(tid.toString() + " - No rconns, erred: Right preserve no");
						rightPreserve = false;
					}
				}
				else {
					// Is this tunnel in the process of closing?
					if(tun->isClosing()) {
						// Give us a chance to finish closing
						DPRINTI(tid.toString() + " - No rconns, closing: Right preserve yes");
						rightPreserve = true;
					}
					else {
						// Have been connected right and not closing, so don't preserve 
						DPRINTI(tid.toString() + " - No rconns, not closing: Right preserve no");
						rightPreserve = false;
					}
				}
			}

			// Both sides agree to preserve?
			if(leftPreserve && rightPreserve) {
				DPRINTI(tid.toString() + " - Trying to preserve");
				scheduleEvent(new ShutdownTunnelEvent(TUNNEL_RECONNECT_TIMEOUT,
					this, tid, false));
			}
			else {
				DPRINTI(tid.toString() + " - Not preserving");
				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
			}
		}
		else {
			DPRINTI(tid.toString() + " - Already closed, not preserving further");
		}
	}
	else {
		std::stringstream err;
		err << "ServerTunneler::unviableTunnel() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification of new raw data arrival
void ServerTunneler::rawDataArrived(unsigned int connID, const TunnelID& tid,
	const RawBuff& raw) throw(LookupManagementError)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ServerTunneler::rawDataArrived() - ";
	nfo << raw.getLength() << " bytes on connection " << connID;
	DPRINTV(nfo.str());
	#endif

	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Inform tunnel of raw data arrival
		tun->processRawData(raw);
	}
	else {
		std::stringstream err;
		err << "ServerTunneler::rawDataArrived() - On connection " << connID;
		err << " Unknown tunnel " << tid;
		throw LookupManagementError(err.str());
	}
}


// Check we have received an ack for some sent message
void ServerTunneler::checkAckReceived(void)
{
	std::vector<Tunnel*>::iterator tit;
	MutexVar tmv(m_tunnelsMtx);

	// Check acks for all tunnels
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Let tunnel check we have received any due acks
		Tunnel* tun = *tit;
		tun->checkAckTimeout();
	}

	// Schedule next ack check event
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
}


// Notification of new HTTP message arrival
void ServerTunneler::messageArrived(unsigned int connID, HTTPMsg* msg)
	throw(LookupManagementError)
{
	// Base class
	Tunneler::messageArrived(connID, msg);

	// Lookup connection
	MutexVar cmv(m_connectionsMtx);
	std::vector<Connection*>::iterator cit = getConnection(connID);
	Connection* conn = *cit;
	HTTPConnection* httpConn = dynamic_cast<HTTPConnection*>(conn);

	// Process msg arrival
	bool fatalConnectionError = false;
	try {
		// Find out which Tunnel this message is from
		MutexVar tmv(m_tunnelsMtx);
		const HTTPMsgParms* parms = msg->getParms();
		Tunnel::TunnelState state = parms->getState();
		TunnelID tid(parms->getTid());
		Tunnel* tun = getTunFromPred(PredTunID(&tid));

		// Do we already know about the tunnel this msg claims it is from?
		if(tun) {
			// Inform tunnel of payload arrival
			tun->processTunnelMsg(msg, httpConn);

			// What state does this message report the tunnel as?
			if(state == Tunnel::CONNECT) {
				// A new pipe connection within this tunnel
				DPRINTI("Got new pipe connection within tunnel");
				newTunnelPipeConnectionRequest(tun, conn);
			}
			// Is there a notification of tunnel closure?
			else if(state == Tunnel::CLOSED) {
				// Schedule tunnel shutdown
				DPRINTI("Received tunnel shutdown request from client");
				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, true));
			}
			else if(state == Tunnel::CLOSING) {
				// Treat like a normal connected tunnel for now
				DPRINTV("Peer reports tunnel as CLOSING");
			}

			// Check for reconnection events
			if(state == Tunnel::RECONNECT) {
				// Reconnection for existing tunnel
				DPRINTI("Got explicit reconnection request from client");
				tunnelReconnectionRequest(tun, conn, state);
			}
			else {
				// Check for new connection from proxy
				if(!tun->ownsConnection(conn)) {
					// Treat like a reconnection
					DPRINTI("Got reconnection request, probably from proxy");
					tunnelReconnectionRequest(tun, conn, state);
				}
			}

			// Send back any queued data/acks to the client
			tun->respondToTunnelMsg(conn, parms);
		}

		else {
			// Unknown tunnel ID; what state does this message report the tunnel as?
			if(state == Tunnel::CONNECT) {
				// Connection for a new tunnel
				tun = newTunnelConnectionRequest(conn, new TunnelID(tid), parms);
				if(tun) {
					// Inform tunnel of payload arrival, & send back any queued data/acks
					tun->processTunnelMsg(msg, httpConn);
					tun->respondToTunnelMsg(conn, parms);
				}
				else {
					// Failed to create tunnel
					fatalConnectionError = true;
				}
			}
			else {
				std::stringstream wrn;
				wrn << "Unexpected state " << state << " in msg for unknown tunnel " << tid.toString();
				g_pLogger->warn(wrn.str());
				fatalConnectionError = true;
			}
		}

		// Does this connection report it is closing?
		if(!fatalConnectionError && (parms->getConnClose() || parms->getProxyConnClose())) {
			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Connection id " << connID << " reports it is closing";
			DPRINTV(dbg.str());
			#endif

			// Don't use closing connection any more
			connectionClosing(cit);
		}
	}
	catch(InvalidID& e) {
		g_pLogger->error("Error parsing ID: " + e.what());
		fatalConnectionError = true;
	}

	// Did something very serious fail?
	if(fatalConnectionError) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Fatal error processing msg on connection " << connID;
		DPRINTV(dbg.str());
		#endif

		// No point keeping this connection around, kill it now
		sendFatalErrorResponse(conn, msg);
		cit = getConnection(connID);
		destroyConnection(cit);
	}
}


// Notification of some new default tunnel parameters
void ServerTunneler::newTunnelParms(TunnelParms* parms, const TunnelerID& tid)
{
	DPRINTI("ServerTunneler::newTunnelParms() - " + tid.toString());

	// Base class
	Tunneler::newTunnelParms(parms, tid);

	// Are these brand new parms or an update to existing parms?
	MutexVar tmv(m_tunnelsMtx);
	ServerTunnelParms* sparms = dynamic_cast<ServerTunnelParms*>(parms);
	std::map<TunnelerID, ServerTunnelParms>::iterator pit = m_tunParms.find(tid);
	if(pit != m_tunParms.end()) {
		// We already have parms for this client tunneler
		pit->second = *sparms;

		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "New ServerTunnelParms for existing tunneler " << tid;
		nfo << " : Max = " << sparms->getMaxPayloadSize(true);
		DPRINTV(nfo.str());
		#endif
	}
	else {
		// Save parms from new client tunneler
		std::pair<TunnelerID, ServerTunnelParms> newParms(tid, *sparms);
		m_tunParms.insert(newParms);

		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "ServerTunnelParms for new tunneler " << tid;
		nfo << " : Max = " << sparms->getMaxPayloadSize(true);
		DPRINTV(nfo.str());
		#endif
	}
}


// Notification that authentication has succeeded for a tunnel
void ServerTunneler::authenticationSuccess(const TunnelID& tid)
{
	DPRINTI("ServerTunneler::authenticationSuccess() - " + tid.toString());

	// Lookup tunnel
	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		/* Now we have authenticated the connecting client, setup raw connection
		 * out to tunnel destination.
		 */

		// How we handle raw connection side depends on whether this is a test tunnel
		if(tun->isTestTunnel()) {
			// Create a dummy raw connection that never actually does anything
			RawConnection* rawconn = new DummyRawConnection(this, tid);
			addConnection(rawconn);
			tun->addRightConnection(rawconn);
		}
		else {
			// Create proper raw connection
			RawConnection* rawconn = new RawConnection(this, tid);
			addConnection(rawconn);
			tun->addRightConnection(rawconn);

			// Address of server to connect to
			const TunnelConfig* tcfg = tun->getConfig();
			std::string svrIP = tcfg->getDestIP();
			unsigned int svrPort = tcfg->getDestPort();
			Enums::Protocol svrProtocol = tcfg->getDestProtocol();

			// Are we for a socket tunnel?
			const ServerSideConfig* serverCfg = getServerSideConfig();
			if(serverCfg->isSockTun()) {
				// What type of socket tunnel do we need to use?
				if(svrProtocol == Enums::TCP) {
					// Get address of TCP socket tunnel server to connect to
					svrIP = m_tcpSockTunServer->getListenIP();
					svrPort = m_tcpSockTunServer->getListenPort();
				}
				else {
					// Get address of UDP socket tunnel server to connect to
					svrIP = m_udpSockTunServer->getListenIP();
					svrPort = m_udpSockTunServer->getListenPort();
				}

				#ifdef DEBUG_TUN
				std::stringstream dbg;
				dbg << "ServerTunneler connecting to local ";
				if(svrProtocol == Enums::TCP) dbg << "TCP"; else dbg << "UDP";
				dbg << " SockTunServer at " << svrIP << ":" << svrPort;
				DPRINTV(dbg.str());
				#endif
			}

			// Schedule initial connect out raw side
			m_connectionMgr->addProxyConnect(rawconn, 0, svrIP, svrPort, svrProtocol, false);
		}
	}
	else {
		std::stringstream err;
		err << "ServerTunneler::authenticationSuccess() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification that we should perform tunnel manangement
void ServerTunneler::performTunnelManagement(void)
{
	// When did we last check tunneling parameters?
	MutexVar tmv(m_tunnelsMtx);
	Time checkTime = Time::now() - m_lastTunParmsCheckTime;
	if(checkTime.getSecs() >= CHECK_TUN_PARMS_PERIOD) {
		m_lastTunParmsCheckTime = Time::now();

		#ifdef DEBUG_TUN
		if(m_tunParms.size() > 0) {
			std::stringstream dbg1;
			dbg1 << "ServerTunneler::performTunnelManagement() - Managing ";
			dbg1 << m_tunParms.size() << " test tunneling parms";
			DPRINTV(dbg1.str());
		}
		#endif

		// Check our store of tunneling parameters for old details
		std::map<TunnelerID, ServerTunnelParms>::iterator pit;
		bool foundOne = false;
		do {
			foundOne = false;
			for(pit = m_tunParms.begin(); pit != m_tunParms.end(); ++pit) {
				// Are these parms old enough to consider removing?
				Time parmAge = Time::now() - pit->second.getCreateTime();
				if(parmAge.getSecs() >= EXPIRE_OLD_TUN_PARMS_AGE) {
					// Do we have an active test tunnel trying to update them?
					if(getTunFromPred(PredTestTunID(&pit->first)) == 0) {
						#ifdef DEBUG_TUN
						std::stringstream dbg2;
						dbg2 << "Removing " << parmAge.getSecs() << " secs old test ";
						dbg2 << "tunneling parms for " << pit->first;
						DPRINTV(dbg2.str());
						#endif

						// Remove old parms
						m_tunParms.erase(pit);
						foundOne = true;
						break;
					}
				}
			}
		} while(foundOne);
	}

	// Base class
	Tunneler::performTunnelManagement();
}


/***********************/
/* Protected Interface */
/***********************/


// Get reconnection attempt parameters
void ServerTunneler::getReconnectParms(unsigned int& maxConnectAttempts,
	unsigned int& reconnectDelay, unsigned int failedAttempts)
{
	// Keep reconnect delay constant at the server
	maxConnectAttempts = SERVER_CONNECT_ATTEMPTS;
	reconnectDelay = SERVER_RECONNECT_DELAY;
}


/*********************/
/* Private Interface */
/*********************/


// Create a new carried tunnel in response to a received message - Caller holds mutexes
Tunnel* ServerTunneler::newTunnelConnectionRequest(Connection* conn, TunnelID* tid,
	const HTTPMsgParms* parms)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ServerTunneler::newTunnelConnectionRequest() - " << *tid;
	DPRINTI(nfo.str());
	#endif

	// Create tunnel config
	ServerSideConfig* serverCfg = getServerSideConfigMutable();
	TunnelConfig* tcfg = new TunnelConfig(tid, parms->getTunTune(), parms->getTunDest(),
		serverCfg->getAuthenticator()->getDefaultAuthSettings());

	// Create tunnel
	ServerTunnel* tun = 0;
	if(parms->getTestState() != Tunnel::NO_TESTING) {
		// Request for new test tunnel
		checkNewTestTunnelConnection(tid->getTunnelerID());
		tun = new TestServerTunnel(tcfg, this);
	}
	else {
		// Plain tunnel
		tun = new ServerTunnel(tcfg, this);
	}

	// Add incoming connection
	tun->addLeftConnection(conn);

	// Store tunnel object in our list of active tunnels
	m_tunnels.push_back(tun);

	return tun;
}

// Reconfigure an existing tunnel in response to a reconnection message
void ServerTunneler::tunnelReconnectionRequest(Tunnel* tun, Connection* conn,
	Tunnel::TunnelState reportedState)
{
	DPRINTI("ServerTunneler::tunnelReconnectionRequest() - " + tun->getTidStr());

	// Let tunnel know it has a new incoming connection
	tun->addLeftConnection(conn);

	// Is it worth keeping this tunnel?
	if(reportedState != Tunnel::CLOSED) {
		// Cancel any pending timeout
		cancelEvents<ShutdownTunnelEvent, TunnelID>(tun->getTid());
	}
}

// Reconfigure an existing tunnel in response to a new pipe connection
void ServerTunneler::newTunnelPipeConnectionRequest(Tunnel* tun, Connection* conn)
{
	DPRINTI("ServerTunneler::newTunnelPipeConnectionRequest() - " + tun->getTidStr());

	// Check we don't already know about this connection
	if(!tun->ownsConnection(conn)) {
		// Let tunnel know it has a new connection to manage
		tun->addLeftConnection(conn);
	}
}


// Send back a tunnel error response to a client
void ServerTunneler::sendFatalErrorResponse(Connection* conn, const HTTPMsg* msg)
{
	DPRINTV("ServerTunneler::sendFatalErrorResponse()");

	ErrorMsgParms* respParms = 0;

	try {
		// Create suitable msg parameters for response
		const HTTPMsgParms* recvd = msg->getParms();
		respParms = new ErrorMsgParms(recvd);

		// Send error response back to client
		conn->write(respParms);
	}
	catch(TunnelError& e) {
		g_pLogger->warn("Failed to send error response msg: " + e.what());
	}

	// Free up response message parameters
	if(respParms)
		delete respParms;
}


// Check if a new test tunnel connection should clear out any old results - Caller holds mutex
void ServerTunneler::checkNewTestTunnelConnection(const TunnelerID& tunnelerID)
{
	// Do we already have test parameters for the connecting tunneler?
	std::map<TunnelerID, ServerTunnelParms>::iterator pit = m_tunParms.find(tunnelerID);
	if(pit != m_tunParms.end()) {
		// If the existing test parms are completed, let this new test replace them
		if(pit->second.testsCompleted()) {
			// Reset parms test state
			DPRINTV("Retesting tunneling parms for " + tunnelerID.toString());
			pit->second.resetTestState();
		}
		else {
			// Do we have a test already running for the connecting tunneler?
			Tunnel* oldTun = getTunFromPred(PredTestTunID(&tunnelerID));
			if(oldTun != 0) {
				DPRINTV("Got replacement tunneling test connection for " + tunnelerID.toString());

				// Shutdown existing test tunnel and let this one replace it
				scheduleEvent(new ShutdownTunnelEvent(0, this, oldTun->getTid(), true));
			}
		}
	}
}


}	// namespace

