/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for server end of tunnel.
 *
 * $Id: ServerTunneler.hh,v 1.4 2009-03-20 11:52:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _SERVER_TUNNELER_HH_
#define _Server_TUNNELER_HH_

#include <map>
#include <vector>
#include "Tunneler.hh"
#include "Exceptions.hh"
#include "ServerTunnelParms.hh"
#include "TunnelerID.hh"

namespace HTTPTunnel {

class Tunnel;
class Connection;
class HTTPMsg;
class ServerSideConfig;
class TCPSockTunServer;
class UDPSockTunServer;


class ServerTunneler : public Tunneler {
public:
   // Construction/destruction
   ServerTunneler(Server* server) throw(TunnelError);
   virtual ~ServerTunneler();


	/*************/
	/* Accessors */
	/*************/

	// Get reference to current tunneling parameters
	ServerTunnelParms getTunParms(const TunnelerID& tunnelerID);


	/**********/
	/* Events */
	/**********/

	// Notification of a new socket connection
	virtual void newConnectionArrived(Sock* insock) throw(CommsError);

	// Notification that a tunnel has insufficient left/right connections to be viable
	virtual void unviableTunnel(const TunnelID& tid);

	// Notification of new tunneled/raw data arrival
	virtual void rawDataArrived(unsigned int connID, const TunnelID& tid,
		const RawBuff& raw) throw(LookupManagementError);

	// Check we have received an ack for some sent message
	virtual void checkAckReceived(void);

	// Notification of new HTTP message arrival
	virtual void messageArrived(unsigned int connID, HTTPMsg* msg)
		throw(LookupManagementError);

	// Notification of some new default tunnel parameters
	virtual void newTunnelParms(TunnelParms* parms, const TunnelerID& tid);

	// Notification that authentication has succeeded for a tunnel
	virtual void authenticationSuccess(const TunnelID& tid);

	// Notification that we should perform tunnel manangement
	virtual void performTunnelManagement(void);


protected:

	/*************/
	/* Constants */
	/*************/

	// Number of connect attempts before giving up
	static const unsigned int			SERVER_CONNECT_ATTEMPTS		= 3;

	// Time between reconnection attempts
	static const unsigned int			SERVER_RECONNECT_DELAY		= 1;


	/***********/
	/* Methods */
	/***********/

	// Get reconnection attempt parameters
	virtual void getReconnectParms(unsigned int& maxConnectAttempts,
		unsigned int& reconnectDelay, unsigned int failedAttempts);


	/********/
	/* Data */
	/********/

	// Tunneling parameters to use for each connected client
	std::map<TunnelerID, ServerTunnelParms>	m_tunParms;

	// Socket tunnel server
	TCPSockTunServer*									m_tcpSockTunServer;
	UDPSockTunServer*									m_udpSockTunServer;


private:

	/*************/
	/* Constants */
	/*************/

	// Timeout for client reconnecting a broken tunnel
	static const unsigned int		TUNNEL_RECONNECT_TIMEOUT		= 30;

	// Control when we expire old tunnel parms
	static const int					EXPIRE_OLD_TUN_PARMS_AGE		= OLD_TUN_PARMS_AGE * 5;


	/***********/
	/* Methods */
	/***********/

	// Create a new carried tunnel in response to a received message
	Tunnel* newTunnelConnectionRequest(Connection* conn, TunnelID* tid,
		const HTTPMsgParms* parms);

	// Reconfigure an existing tunnel in response to a reconnection message
	void tunnelReconnectionRequest(Tunnel* tun, Connection* conn,
		Tunnel::TunnelState reportedState);

	// Reconfigure an existing tunnel in response to a new pipe connection
	void newTunnelPipeConnectionRequest(Tunnel* tun, Connection* conn);

	// Send back a tunnel error response to a client
	void sendFatalErrorResponse(Connection* conn, const HTTPMsg* msg);

	// Check if a new test tunnel connection should clear out any old results
	void checkNewTestTunnelConnection(const TunnelerID& tunnelerID);

};


}	// namespace

#endif

