/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP/UDP socket.
 *
 * $Id: Sock.cc,v 1.6 2009-02-27 11:56:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <cstring>
#include <cerrno>
#include <unistd.h>
#include <netdb.h>
#include <poll.h>
#include <fcntl.h>
#include "Sock.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "Utils.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Initialise statics
unsigned int Sock::c_idCount = 1;
Mutex Sock::c_idCountMtx;


// Construction/destruction
Sock::Sock(bool blocking) throw(SockError)
: m_mtx(), m_fd(-1), m_listenAddr(), m_connectedClientAddr(), m_blocking(blocking),
  m_id(0)
{
	/* Construct unconnected */

	// Set socket ID
	m_id = getNextSockID();
}

Sock::Sock(int fd, std::string cliAddr) throw(SockError)
: m_mtx(), m_fd(fd), m_listenAddr(), m_connectedClientAddr(cliAddr), m_blocking(false),
  m_id(0)
{
	/* Construct already connected */

	// Set socket ID
	m_id = getNextSockID();
}

Sock::~Sock()
{
}


/*************/
/* Accessors */
/*************/


// Get address of connected client
std::string Sock::getConnectedClientAddr(void) const
{
	return m_connectedClientAddr;
}


// Get underlying socket descriptor
int Sock::getFD(void) const
{
	return m_fd;
}


/************/
/* Mutators */
/************/


// Release and return our underlying socket descriptor
int Sock::release(void) throw(SockError)
{
	int sockfd = -1;

	MutexVar mv(m_mtx);

	// Check we have a socket to release
	if(m_fd >= 0) {
		sockfd = m_fd;
		m_fd = -1;
	}
	else {
		throw SockError("Sock::release() - No socket to release", 0);
	}

	return sockfd;
}


// Set blocking mode for socket
void Sock::setBlocking(bool blocking) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Set flag
	m_blocking = blocking;

	// Set blocking state
	blocknonblock();
}


/********************/
/* Public Interface */
/********************/


// Bind socket to an address
void Sock::bind(std::string addr, unsigned int& port) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Reuse address
	int reuse = 1;
	setSockOpt(SOL_SOCKET, SO_REUSEADDR, reuse);

	// Set up server socket address structure
	memset(&m_listenAddr, 0, sizeof(m_listenAddr));
	m_listenAddr.sin_family = AF_INET;
	m_listenAddr.sin_port = htons(port);
	if(addr.size() > 0) {
		// Bind to specific interface
		struct hostent hostAddr;
		char auxHostBuff[Utils::AUX_HOST_BUFF_SIZE];
		if(Utils::getHostByName(addr, &hostAddr, auxHostBuff, sizeof(auxHostBuff)) == 0) {
			memcpy(&m_listenAddr.sin_addr, hostAddr.h_addr_list[0],
				sizeof(m_listenAddr.sin_addr));
		}
		else {
			throw SockError("Unable to resolve bind address " + addr, 0);
		}
	}
	else {
		m_listenAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	}

	// Bind the socket
	if(::bind(m_fd, (struct sockaddr*)&m_listenAddr, sizeof(m_listenAddr)) < 0)
		throw SockError("Failed to bind socket", errno);

	// Did we let the OS pick an ephemeral port?
	if(port == 0) {
		// Find out what port we bound to
		port = Utils::getSockPort(m_fd);
	}
}


// Poll socket for reading
bool Sock::pollRead(int timeout, bool& err, bool& hup) throw(SockError)
{
	bool ready = false;
	err = hup = false;

	MutexVar mv(m_mtx);

	// Create poll structure
	struct pollfd pollStruct[] = { { m_fd, POLLIN, 0 } };

	// Do the poll
	int readyfds = ::poll(pollStruct, 1, timeout);
	if(readyfds > 0) {
		// Events pending
		ready = true;

		// Check for error events
		short revents = pollStruct[0].revents;
		if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) {
			err = true;
			if(revents & POLLHUP)
				hup = true;
		}
	}
	else if(readyfds < 0) {
		// Poll error
		throw SockError("Error polling socket", errno);
	}

	return ready;
}


// Poll socket for writing
bool Sock::pollWrite(int timeout, bool& err) throw(SockError)
{
	bool ready = false;
	err = false;

	MutexVar mv(m_mtx);

	// Create poll structure
	struct pollfd pollStruct[] = { { m_fd, POLLOUT, 0 } };

	// Do the poll
	int readyfds = ::poll(pollStruct, 1, timeout);
	if(readyfds > 0) {
		// Events pending
		ready = true;

		// Check for error events
		short revents = pollStruct[0].revents;
		if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL))
			err = true;
	}
	else if(readyfds < 0) {
		// Poll error
		throw SockError("Error polling socket", errno);
	}

	return ready;
}


/***********************/
/* Protected Interface */
/***********************/


// Close socket
void Sock::close(void) throw(SockError)
{
	DPRINTV("Sock::close()");

	MutexVar mv(m_mtx);

	// Got an open socket?
	if(m_fd >= 0) {
		int fd = m_fd;
		m_fd = -1;

again:
		// Close socket descriptor
		if(::close(fd) != 0) {
			if(errno == EINTR) {
				goto again;
			}
			else {
				std::stringstream err;
				err << "Error closing socket id " << m_id << ": " << strerror(errno);
				throw SockError(err.str(), errno);
			}
		}
	}
}


// Setup blocking mode for this socket
void Sock::blocknonblock(void) throw(SockError)
{
	DPRINTV("Sock::blocknonblock()");

	MutexVar mv(m_mtx);

	// Turn on blocking mode for socket
	int sflags = fcntl(m_fd, F_GETFL);
	if(sflags >= 0) {
		if(m_blocking) {
			// Blocking
			sflags = sflags & ~O_NONBLOCK;
			if(fcntl(m_fd, F_SETFL, sflags) < 0) {
				throw SockError("Unable to clear socket nonblocking flag", errno);
			}
		}
		else {
			// Non-blocking
			sflags = sflags | O_NONBLOCK;
			if(fcntl(m_fd, F_SETFL, sflags) < 0) {
				throw SockError("Unable to set socket nonblocking flag", errno);
			}
		}
	}
	else {
		throw SockError("Unable to get socket blocking flags", errno);
	}
}


/*********************/
/* Private Interface */
/*********************/


// Get next socket ID
unsigned int Sock::getNextSockID(void)
{
	unsigned int id = 0;

	MutexVar mv(c_idCountMtx);
	id = c_idCount++;
	if(c_idCount == 0)
		c_idCount = 1;
	mv.unlock();

	return id;
}


}	// namespace

