/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP/UDP socket.
 *
 * $Id: Sock.hh,v 1.5 2009-02-27 11:56:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _SOCK_HH_
#define _SOCK_HH_

#include <string>
#include <sys/socket.h>
#include <netinet/in.h>
#include "Exceptions.hh"
#include "Mutex.hh"

namespace HTTPTunnel {

class Tunneler;

class Sock {
public:
	// Construction/destruction
	Sock(bool blocking) throw(SockError);
	Sock(int fd, std::string cliAddr) throw(SockError);
	virtual ~Sock();

	// Accessors
	std::string getConnectedClientAddr(void) const;
	int getFD(void) const;

	// Mutators
	int release(void) throw(SockError);
	void setBlocking(bool blocking) throw(SockError);

	// Utilities
	virtual void newConnection(Tunneler* tunneler) throw(SockError) = 0;
	virtual void listen(void) throw(SockError) = 0;
	virtual Sock* accept(void) throw(SockError) = 0;
	virtual void connect(std::string addr, unsigned int port) throw(SockError, CommsError) = 0;
	virtual int read(unsigned char* buff, size_t buffSize) throw(SockError, CommsError) = 0;
	virtual int write(const unsigned char* buff, size_t buffSize) throw(SockError, CommsError) = 0;
	virtual void getConnectResult(bool& resultAvailable, bool& connected) throw(SockError) = 0;

	virtual void bind(std::string addr, unsigned int& port) throw(SockError);
	virtual bool pollRead(int timeout, bool& err, bool& hup) throw(SockError);
	virtual bool pollWrite(int timeout, bool& err) throw(SockError);


protected:

	/***********/
	/* Methods */
	/***********/

	// Close our socket descriptor
	virtual void close(void) throw(SockError);

	// Setup blocking/nonblocking mode
	virtual void blocknonblock(void) throw(SockError);

	// Get/set socket options
	template<class T>
	void getSockOpt(int level, int optname, T& optres) throw(SockError);
	template<class T>
	void setSockOpt(int level, int optname, T& optval) throw(SockError);


	/********/
	/* Data */
	/********/

	// Mutex
	Mutex									m_mtx;

	// Socket descriptor
	int									m_fd;

	// Listening socket address structure
	struct sockaddr_in				m_listenAddr;

	// Address of connected client
	std::string							m_connectedClientAddr;

	// Should we be blocking or non-blocking?
	bool									m_blocking;


private:

	// Forbid copying
	Sock(const Sock& that);
	Sock& operator=(const Sock& that);

	/***********/
	/* Methods */
	/***********/

	// Get next socket ID
	static unsigned int getNextSockID(void);


	/********/
	/* Data */
	/********/

	// Unique ID
	static unsigned int				c_idCount;
	static Mutex						c_idCountMtx;
	unsigned int						m_id;

};


}	// namespace


// Include template methods
#include "SockT.cc"


#endif

