/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP socket.
 *
 * $Id: TCPSock.cc,v 1.5 2008-11-21 18:35:53 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cerrno>
#include <cstring>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include "TCPSock.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "Utils.hh"
#include "Tunneler.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
TCPSock::TCPSock(bool blocking) throw(SockError)
: Sock(blocking)
{
	// Create new TCP socket
	int sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if(sockfd < 0) {
		throw SockError("Failed to create new TCP socket descriptor", errno);
	}
	else {
		m_fd = sockfd;
	}

	// Set blocking mode
	blocknonblock();
}

TCPSock::TCPSock(int fd, std::string cliAddr) throw(SockError)
: Sock(fd, cliAddr)
{
}

TCPSock::~TCPSock()
{
	DPRINTV("TCPSock::~TCPSock()");

	MutexVar mv(m_mtx);

	try {
		// Close socket descriptor
		close();
	}
	catch(SockError& e) {
		g_pLogger->info(e.what());
	}
}


/********************/
/* Public Interface */
/********************/


// Wait for a new incoming TCP connection
void TCPSock::newConnection(Tunneler* tunneler) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Socket ready to accept
	Sock* connectedSock = accept();

	try {
		// Tell tunneler
		if(tunneler)
			tunneler->newConnectionArrived(connectedSock);
		else
			throw SockError("No tunneler to handle new TCP connection", 0);
	}
	catch(TunnelError& e) {
		g_pLogger->info(e.what());
		delete connectedSock;
	}
}


// Set socket listening
void TCPSock::listen(void) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Set listening
	if(::listen(m_fd, 5) < 0)
		throw SockError("Failed to activate listening TCP socket", errno);
}


// Wait for a new incoming connection
Sock* TCPSock::accept(void) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Accept connection
	struct sockaddr_in cliaddr;
	socklen_t clilen = sizeof(cliaddr);
	int connfd = -1;
	if((connfd = ::accept(m_fd, (struct sockaddr*)&cliaddr, &clilen)) < 0) {
		throw SockError("Failed to accept TCP connection", errno);
	}

	// Find client address
	std::string clientAddr(inet_ntoa(cliaddr.sin_addr));
	DPRINTI("TCPSock::accept() - Got new connection from " + clientAddr);

	// Create and return new TCPSock to handle this connected socket
	return new TCPSock(connfd, clientAddr);
}


// Establish connection out
void TCPSock::connect(std::string addr, unsigned int port)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Lookup destination address
	struct hostent resremotehost;
	char auxHostBuff[Utils::AUX_HOST_BUFF_SIZE];
	if(Utils::getHostByName(addr, &resremotehost, auxHostBuff, sizeof(auxHostBuff)) != 0) {
		throw ConnectError("Failed to resolve TCP server address " + addr);
	}

	// Set up server address structure
	struct sockaddr_in destaddr;
	memset(&destaddr, 0, sizeof(destaddr));
	destaddr.sin_family = AF_INET;
	destaddr.sin_port = htons(port);
	memcpy(&destaddr.sin_addr, resremotehost.h_addr_list[0],
		sizeof(destaddr.sin_addr));

	// Try the connect
	if(::connect(m_fd, (struct sockaddr*)&destaddr, sizeof(destaddr)) < 0) {
		throw SockError("Failed to connect to TCP server", errno);
	}
}


// Read from socket
int TCPSock::read(unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Read data from socket
again:
	int noread = ::read(m_fd, buff, buffSize);
	if(noread < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Read error from TCP socket", errno);
	}
	else if(noread == 0) {
		// EOF
		throw ConnectionClosed("Other end closed TCP socket");
	}

	return noread;
}


// Write to socket
int TCPSock::write(const unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Write data to socket
again:
	int nowritten = ::write(m_fd, buff, buffSize);
	if(nowritten < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Write error to TCP socket", errno);
	}

	return nowritten;
}


// Get the result of a non-blocking connect attempt
void TCPSock::getConnectResult(bool& resultAvailable, bool& connected)
	throw(SockError)
{
	MutexVar mv(m_mtx);

	// Poll for connect result
	bool pollError = false;
	resultAvailable = pollWrite(0, pollError);
	if(resultAvailable) {
		// Get connect result
		int cres = -1;
		getSockOpt(SOL_SOCKET, SO_ERROR, cres);
		if(cres == 0) {
			// Success
			connected = true;
		}
		else {
			// Failed
			connected = false;
		}
	}
}


}	// namespace

