/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Client side Tunnel for testing.
 *
 * $Id: TestClientTunnel.cc,v 1.5 2008-11-09 18:50:21 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "TestClientTunnel.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "GlobalConfig.hh"
#include "MutexVar.hh"
#include "ClientTunneler.hh"
#include "HTTPMsgParms.hh"
#include "HTTPMsg.hh"
#include "NewTunnelParmsEvent.hh"
#include "ShutdownTunnelEvent.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
TestClientTunnel::TestClientTunnel(TunnelConfig* tcfg, ClientTunneler* tunneler)
: ClientTunnel(tcfg, tunneler), TestTunnel(), m_testType(CTEST_IDLE),
  m_goodURLPayloadSize(0), m_goodBodyPayloadSize(0), m_currURLPayloadSize(0),
  m_currBodyPayloadSize(0)
{
	DPRINTI("TestClientTunnel::TestClientTunnel()");

	// Setup initial test parameters
	ClientTunnelParms tunParms = tunneler->getTunParms();
	m_tunnelTestState = tunParms.getTestState();
	if(m_tunnelTestState == BEGIN_TESTING) {
		m_currURLPayloadSize = MIN_TEST_URL_PAYLOAD;
		m_currBodyPayloadSize = MIN_TEST_BODY_PAYLOAD;
	}
	else {
		m_currURLPayloadSize = tunParms.getURLPayloadSize();
		m_currBodyPayloadSize = tunParms.getBodyPayloadSize();
		m_goodURLPayloadSize = m_currURLPayloadSize;
		m_goodBodyPayloadSize = m_currBodyPayloadSize;
	}

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "TestClientTunnel created in state: " << testStateAsString(getTestState());
	dbg << ", URL payload size: " << m_currURLPayloadSize;
	dbg << ", Body payload size: " << m_currBodyPayloadSize;
	DPRINTV(dbg.str());
	#endif
}

TestClientTunnel::~TestClientTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "TestClientTunnel::~TestClientTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif
}


/*************/
/* Accessors */
/*************/


// Get default tunneling parameters
ClientTunnelParms TestClientTunnel::getDefaultTunParms(void)
{
	return ClientTunnelParms(BEGIN_TESTING, DEFAULT_URL_PAYLOAD,
		DEFAULT_BODY_PAYLOAD);
}


// Get whether we are a test tunnel
bool TestClientTunnel::isTestTunnel(void) const
{
	return true;
}


// Get whether we have completed these tests
bool TestClientTunnel::testsCompleted(void)
{
	bool completed = false;

	MutexVar mv(m_mtx);

	// Only completed once both we and the server side have finished testing
	if(getTestState() == COMPLETED_TESTING) {
		completed = true;
	}

	return completed;
}


/************/
/* Mutators */
/************/


// Record connection error
void TestClientTunnel::recordConnectionError(TunnelErrorType errType)
{
	MutexVar mv(m_mtx);

	// Base class
	ClientTunnel::recordConnectionError(errType);

	// Does this error indicate a problem with our testing?
	switch(errType) {
		case(BAD_RESPONSE_ERROR): {
			// Assume our tests broke something
			nextTest(false);
			break;
		}
		case(GENERAL_COMMS_ERROR): {
			// Are we in the process of testing?
			if(getTestState() == CLIENT_TESTING) {
				// Assume our tests broke something
				nextTest(false);
			}
			break;
		}
		case(UNPROCESSED_DATA_ERROR): {
			// Assume our tests broke something
			nextTest(false);
			break;
		}
		default: {
			break;
		}
	}
}


/******************************/
/* Tunnel Protected Interface */
/******************************/


// Create payload for tunneled msg to send to the destination - Caller holds mutex
void TestClientTunnel::makeTunnelPayload(RawBuff& data, bool& split, unsigned int max)
	throw(MemoryError)
{
	unsigned int payloadLen = 0;

	// Are we currently testing?
	if(getTestState() == CLIENT_TESTING && m_testType != CTEST_IDLE) {
		// What are we currently testing?
		if(m_testType == CTEST_URLMAXSIZE) {
			payloadLen = m_currURLPayloadSize;
		}
		else if(m_testType == CTEST_BODYMAXSIZE) {
			payloadLen = m_currBodyPayloadSize;
		}

		// Create some payload data of the required length
		if(payloadLen > 0) {
			data = RawBuff(payloadLen, true);
		}
	}
}


// Check message parms before resending - Caller holds mutex
void TestClientTunnel::checkResendMsgParms(HTTPMsgParms* parms)
{
	// We don't want to resend any potentially problematic payload
	Payload* payload = parms->getPayloadMutable();
	payload->getPayloadMutable().clear();
}


// Advance our internal state machine - Caller holds mutex
void TestClientTunnel::advanceState(const HTTPMsgParms* parms)
{
	// Let base class update our state
	ClientTunnel::advanceState(parms);

	// Are we waiting to start testing?
	if(getTestState() == BEGIN_TESTING) {
		// Once we become CONNECTED we can begin testing
		if(getState() == CONNECTED) {
			// We're now fully up and ready to begin testing
			setTestState(CLIENT_TESTING);

			// What type of testing should we do?
			if(!m_tunCfg->getUseHeaders()) {
				// Need to test URL payload size
				DPRINTV("TestClientTunnel now testing max URL payload size");
				m_testType = CTEST_URLMAXSIZE;
			}
			else {
				// Need to test msg body payload size
				DPRINTV("TestClientTunnel now testing max body payload size");
				m_testType = CTEST_BODYMAXSIZE;
			}
		}
	}
	else {
		// Does server report later testing state than us?
		TunnelTestState svrState = parms->getTestState();
		if(svrState > getTestState()) {
			// Track server testing state
			setTestState(svrState);

			// Testing completed?
			if(getTestState() == COMPLETED_TESTING) {
				DPRINTV("Both tunnel ends have completed testing");

				// Send final results to controlling tunneler
				sendTestResults();

				// We've finished with this test tunnel for now
				m_tunneler->scheduleEvent(new ShutdownTunnelEvent(0, m_tunneler, m_tid, false));
			}
		}
	}
}


// A sent message has been acknowledged - Caller holds mutex
void TestClientTunnel::msgAcked(SentMsgParms* parms)
{
	// Base class
	ClientTunnel::msgAcked(parms);

	// Are we currently testing?
	if(getTestState() == CLIENT_TESTING) {
		// Get details for acked msg
		HTTPMsgParms* msgParms = parms->m_parms;
		unsigned int payloadLen = msgParms->getPayload()->getPayloadSize();
		if(payloadLen > 0) {
			// What type of testing are we doing?
			if(m_testType == CTEST_URLMAXSIZE) {
				// Record new good baseline
				m_goodURLPayloadSize = payloadLen;
			}
			else if(m_testType == CTEST_BODYMAXSIZE) {
				// Record new good baseline
				m_goodBodyPayloadSize = payloadLen;
			}
		}

		// Consider moving tests on
		nextTest(true);
	}
}


// Record we need to resend a msg - Caller holds mutex
void TestClientTunnel::needToResendMsg(const SeqNo& seqNo, bool msgLost)
{
	// Base class
	ClientTunnel::needToResendMsg(seqNo, msgLost);

	// Is this a lost msg?
	if(msgLost) {
		// Are we currently testing?
		if(getTestState() == CLIENT_TESTING) {
			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "TestClientTunnel::needToResendMsg() - Resend of seq no " << seqNo;
			nfo << " indicates test failed";
			DPRINTI(nfo.str());
			#endif

			// Assume our tests broke something
			nextTest(false);
		}
	}
}


/**********************************/
/* TestTunnel Protected Interface */
/**********************************/


// Move testing on - Caller holds mutex
void TestClientTunnel::nextTest(bool ok)
{
	// Are we actually currently testing?
	if(getTestState() == CLIENT_TESTING) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "TestClientTunnel::nextTest() - Test " << (ok? "passed" : "failed");
		nfo << ", URL payload size: " << m_currURLPayloadSize;
		nfo << ", Body payload size: " << m_currBodyPayloadSize;
		DPRINTV(nfo.str());
		#endif

		// What are we currently testing?
		if(m_testType == CTEST_URLMAXSIZE) {
			// Did test pass?
			if(ok) {
				// Do we want to try another test?
				if(m_currURLPayloadSize < MAX_TEST_URL_PAYLOAD) {
					// Increase size of tested URL payload
					m_currURLPayloadSize = m_currURLPayloadSize + MIN_TEST_URL_PAYLOAD;
				}
				else {
					// We've finished this test, move state on
					DPRINTV("TestClientTunnel finished our testing");
					setTestState(SERVER_TESTING);
				}
			}
			else {
				// Test failed; Use last known good value and move on
				DPRINTV("TestClientTunnel finished our testing");
				setTestState(SERVER_TESTING);
			}
		}

		else if(m_testType == CTEST_BODYMAXSIZE) {
			// Did test pass?
			if(ok) {
				// Do we want to try another test?
				if(m_currBodyPayloadSize < MAX_TEST_BODY_PAYLOAD) {
					// Increase size of tested body payload
					m_currBodyPayloadSize = m_currBodyPayloadSize + MIN_TEST_BODY_PAYLOAD;
				}
				else {
					// We've finished this test, move state on
					DPRINTV("TestClientTunnel finished our testing");
					setTestState(SERVER_TESTING);
				}
			}
			else {
				// Test failed; Use last known good value and move on
				DPRINTV("TestClientTunnel finished our testing");
				setTestState(SERVER_TESTING);
			}
		}

		// Send latest results to controlling tunneler
		sendTestResults();
	}
}


// Inform controlling tunneler of our test results - Caller holds mutex
void TestClientTunnel::sendTestResults(void)
{
	// Collect results and pass onto tunneler
	ClientTunnelParms* parms = new ClientTunnelParms(getTestState(),
		m_goodURLPayloadSize, m_goodBodyPayloadSize);
	m_tunneler->scheduleEvent(new NewTunnelParmsEvent(m_tunneler, parms,
		m_tid.getTunnelerID()));
}


}	// namespace

