/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Server side Tunnel for testing.
 *
 * $Id: TestServerTunnel.cc,v 1.5 2008-11-09 18:50:21 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "TestServerTunnel.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "GlobalConfig.hh"
#include "ServerTunneler.hh"
#include "HTTPMsgParms.hh"
#include "HTTPMsg.hh"
#include "NewTunnelParmsEvent.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
TestServerTunnel::TestServerTunnel(TunnelConfig* tcfg, ServerTunneler* tunneler)
: ServerTunnel(tcfg, tunneler), TestTunnel(), m_testType(STEST_IDLE),
  m_goodPayloadSize(0), m_currPayloadSize(0)
{
	DPRINTI("TestServerTunnel::TestServerTunnel()");

	// Setup initial test parameters
	ServerTunnelParms tunParms = tunneler->getTunParms(m_tid.getTunnelerID());
	m_tunnelTestState = tunParms.getTestState();
	if(m_tunnelTestState == BEGIN_TESTING) {
		m_currPayloadSize = MIN_TEST_PAYLOAD;
	}
	else {
		m_currPayloadSize = tunParms.getMaxPayloadSize(true);
		m_goodPayloadSize = m_currPayloadSize;
	}

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "TestServerTunnel created in state: " << testStateAsString(getTestState());
	dbg << ", Payload size: " << m_currPayloadSize;
	DPRINTV(dbg.str());
	#endif
}

TestServerTunnel::~TestServerTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "TestServerTunnel::~TestServerTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif
}


/*************/
/* Accessors */
/*************/


// Get default tunneling parameters
ServerTunnelParms TestServerTunnel::getDefaultTunParms(void)
{
	return ServerTunnelParms(BEGIN_TESTING, DEFAULT_PAYLOAD);
}


// Get whether we are a test tunnel
bool TestServerTunnel::isTestTunnel(void) const
{
	return true;
}


/************/
/* Mutators */
/************/


// Record connection error
void TestServerTunnel::recordConnectionError(TunnelErrorType errType)
{
	MutexVar mv(m_mtx);

	// Base class
	ServerTunnel::recordConnectionError(errType);

	// Does this error indicate a problem with our testing?
	switch(errType) {
		case(GENERAL_COMMS_ERROR): {
			// Are we in the process of testing?
			if(getTestState() == SERVER_TESTING) {
				// Assume our tests broke something
				nextTest(false);
			}
			break;
		}
		default: {
			break;
		}
	}
}


/******************************/
/* Tunnel Protected Interface */
/******************************/


// Create payload for tunneled msg to send to the destination - Caller holds mutex
void TestServerTunnel::makeTunnelPayload(RawBuff& data, bool& split, unsigned int max)
	throw(MemoryError)
{
	unsigned int payloadLen = 0;

	// Are we currently testing?
	if(getTestState() == SERVER_TESTING) {
		// What are we currently testing?
		if(m_testType == STEST_PAYLOADMAXSIZE) {
			payloadLen = m_currPayloadSize;
		}

		// Create some payload data of the required length
		if(payloadLen > 0) {
			data = RawBuff(payloadLen, true);
		}
	}
}


// Check message parms before resending - Caller holds mutex
void TestServerTunnel::checkResendMsgParms(HTTPMsgParms* parms)
{
	// We don't want to resend any potentially problematic payload
	Payload* payload = parms->getPayloadMutable();
	payload->getPayloadMutable().clear();
}


// Advance our internal state machine - Caller holds mutex
void TestServerTunnel::advanceState(const HTTPMsgParms* parms)
{
	// Let base class update our state
	ServerTunnel::advanceState(parms);

	// Does client report later testing state than us?
	TunnelTestState cliState = parms->getTestState();
	if(cliState > getTestState()) {
		// Track client testing state
		setTestState(cliState);

		// Has client just told us it's our turn to start testing?
		if(cliState == SERVER_TESTING) {
			// What type of testing should we do?
			DPRINTV("TestServerTunnel now testing max payload size");
			m_testType = STEST_PAYLOADMAXSIZE;
		}
	}
}


// A sent message has been acknowledged - Caller holds mutex
void TestServerTunnel::msgAcked(SentMsgParms* parms)
{
	// Base class
	ServerTunnel::msgAcked(parms);

	// Are we currently testing?
	if(getTestState() == SERVER_TESTING) {
		// Get details for acked msg
		HTTPMsgParms* msgParms = parms->m_parms;
		unsigned int payloadLen = msgParms->getPayload()->getPayloadSize();
		if(payloadLen > 0) {
			// What type of testing are we doing?
			if(m_testType == STEST_PAYLOADMAXSIZE) {
				// Record new good baseline
				m_goodPayloadSize = payloadLen;
			}
		}

		// Consider moving tests on
		nextTest(true);
	}
}


// Record we need to resend a msg - Caller holds mutex
void TestServerTunnel::needToResendMsg(const SeqNo& seqNo, bool msgLost)
{
	// Base class
	ServerTunnel::needToResendMsg(seqNo, msgLost);

	// Is this a lost msg?
	if(msgLost) {
		// Are we currently testing?
		if(getTestState() == SERVER_TESTING) {
			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "TestServerTunnel::needToResendMsg() - Resend of seq no " << seqNo;
			nfo << " indicates test failed";
			DPRINTI(nfo.str());
			#endif

			// Assume our tests broke something
			nextTest(false);
		}
	}
}


/**********************************/
/* TestTunnel Protected Interface */
/**********************************/


// Move testing on - Caller holds mutex
void TestServerTunnel::nextTest(bool ok)
{
	// Are we actually currently testing?
	if(getTestState() == SERVER_TESTING) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "TestServerTunnel::nextTest() - Test " << (ok? "passed" : "failed");
		nfo << ", Payload size: " << m_currPayloadSize;
		DPRINTV(nfo.str());
		#endif

		// What are we currently testing?
		if(m_testType == STEST_PAYLOADMAXSIZE) {
			// Did test pass?
			if(ok) {
				// Do we want to try another test?
				if(m_currPayloadSize < MAX_TEST_PAYLOAD) {
					// Increase size of tested payload
					m_currPayloadSize = m_currPayloadSize + MIN_TEST_PAYLOAD;
				}
				else {
					// We've finished this test, move state on
					DPRINTV("TestServerTunnel finished our testing");
					setTestState(COMPLETED_TESTING);
				}
			}
			else {
				// Test failed; Use last known good value and move on
				DPRINTV("TestServerTunnel finished our testing");
				setTestState(COMPLETED_TESTING);
			}
		}

		// Send latest results to controlling tunneler
		sendTestResults();
	}
}


// Inform controlling tunneler of our test results - Caller holds mutex
void TestServerTunnel::sendTestResults(void)
{
	// Collect results and pass onto tunneler
	ServerTunnelParms* parms = new ServerTunnelParms(getTestState(), m_goodPayloadSize);
	m_tunneler->scheduleEvent(new NewTunnelParmsEvent(m_tunneler, parms,
		m_tid.getTunnelerID()));
}


}	// namespace

