/* (C) 2003 Kevin Manning.
 * (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Thread manager class.
 *
 * $Id: ThreadManager.cc,v 1.2 2008-11-09 18:50:21 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <algorithm>
#include <poll.h>
#include <unistd.h>
#include "ThreadManager.hh"
#include "Callback.hh"
#include "Thread.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


/*****************/
/* ThreadDetails */
/*****************/


// Construction/destruction
ThreadDetails::ThreadDetails(Thread* thread)
: m_thread(thread)
{
}

ThreadDetails::~ThreadDetails()
{
}

// Get/set thread
Thread* ThreadDetails::getThread(void) const
{
	return m_thread;
}
void ThreadDetails::setThread(Thread* thread)
{
	m_thread = thread;
}



/*****************/
/* ThreadManager */
/*****************/


// Construction/Destruction
ThreadManager::ThreadManager() throw(MutexError)
: m_active(true), m_threadDetails(), m_mtx(), m_cv_thread_finished()
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::ThreadManager()" << std::endl;
	#endif
}

ThreadManager::~ThreadManager()
{
	#ifdef DEBUG_THREADS
	std::cerr << "Entered ThreadManager::~ThreadManager()" << std::endl;
	#endif

	// Shutdown
	shutdown();

	#ifdef DEBUG_THREADS
	std::cerr << "Leaving ThreadManager::~ThreadManager()" << std::endl;
	#endif
}


// Create a new thread and add it to the pool
Thread* ThreadManager::newThread(Callback& callBack) throw(ThreadError)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::newThread()" << std::endl;
	#endif

	// Take lock
	MutexVar mv(m_mtx);

	// Can we still schedule?
	if(!m_active)
		return 0;

	// Create new thread
	Thread* newThread = new Thread(this, callBack);
	m_threadDetails.push_back(new ThreadDetails(newThread));
	return newThread;
}


// Tidy up and shutdown
void ThreadManager::shutdown(void)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::shutdown()" << std::endl;
	#endif

	try {
		// Take lock
		MutexVar mv(m_mtx);

		// Don't bother shutting down again if we're already inactive
		if(m_active) {
			// Flag we are no longer active
			m_active = false;

			// Wait for all threads to finish
			waitForAll(mv);
		}
	}
	catch(ThreadError& e) {
		#ifdef DEBUG_THREADS
		std::cerr << "ThreadError caught in shutdown(): " << e.what() << std::endl;
		#endif
	}
}


/*********************/
/* Private Interface */
/*********************/


// One of our threads has started
void ThreadManager::started(Thread* thread)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::started()" << std::endl;
	#endif
}


// One of our threads has finished
void ThreadManager::finished(Thread* thread)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::finished()" << std::endl;
	#endif

	// Take mutex
	MutexVar mv(m_mtx);

	// Remove thread from list of active threads
	std::vector<ThreadDetails*>::iterator i = find_if(m_threadDetails.begin(),
		m_threadDetails.end(), PredGetDetails(thread));
	if(i != m_threadDetails.end()) {
		delete *i;
		m_threadDetails.erase(i);
	}

	// Signal thread completion to any waiters
	m_cv_thread_finished.broadcast();
}


// Wait for all currently running threads to finish
void ThreadManager::waitForAll(MutexVar& mv)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::waitForAll()" << std::endl;
	#endif

	// Still waiting for any threads?
	while(m_threadDetails.size() > 0) {
		// Get thread
		Thread* thread = m_threadDetails[0]->getThread();

		#ifdef DEBUG_THREADS
		std::cerr << "ThreadManager::waitForAll() - Waiting for thread " << thread << std::endl;
		#endif

		// Ensure thread is awake
		thread->wake();

		// Wait for signal thread has finished
		m_cv_thread_finished.wait(mv);
	}
}


}	// namespace

