/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Represent a time in seconds & milli-seconds.
 *
 * $Id: Time.cc,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <sys/time.h>
#include "Time.hh"


namespace HTTPTunnel {


// Construction/destruction
Time::Time(time_t secs, unsigned int millis)
: m_timeVal()
{
	// Get current time
	gettimeofday(&m_timeVal, 0);

	// Add interval
	addInterval(secs, millis);
}

Time::Time(time_t secs, unsigned int millis, bool abs)
: m_timeVal()
{
	// Fill out timeval
	m_timeVal.tv_sec = secs;
	m_timeVal.tv_usec = (millis * MILLISCALE);
}

Time::Time(const Time& that)
: m_timeVal(that.m_timeVal)
{
}

Time::~Time()
{
}


// Assignment
Time& Time::operator=(const Time& that)
{
	if(this != &that) {
		m_timeVal = that.m_timeVal;
	}

	return *this;
}


// Subtraction
Time Time::operator-(const Time& that) const
{
	// Start with a copy of ourselves
	Time res(*this);

	// Substract 'that'
	res.subInterval(that.getSecs(), that.getMillis());
	return res;
}

// Addition
Time Time::operator+(const Time& that) const
{
	// Start with a copy of ourselves
	Time res(*this);

	// Add 'that'
	res.addInterval(that.getSecs(), that.getMillis());
	return res;
}


// Equality operator
bool Time::operator==(const Time& that) const
{
	return (m_timeVal.tv_sec == that.m_timeVal.tv_sec &&
		m_timeVal.tv_usec == that.m_timeVal.tv_usec);
}

// Less than operator
bool Time::operator<(const Time& that) const
{
	bool res = false;

	if(m_timeVal.tv_sec < that.m_timeVal.tv_sec) {
		res = true;
	}
	else {
		if(m_timeVal.tv_sec == that.m_timeVal.tv_sec) {
			res = (m_timeVal.tv_usec < that.m_timeVal.tv_usec);
		}
	}

	return res;
}


// Get current time
Time Time::now(void)
{
	// Get current time
	struct timeval now;
	gettimeofday(&now, 0);
	return Time(now.tv_sec, (now.tv_usec / MILLISCALE), true);
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get raw internal time val
const struct timeval& Time::getTimeVal(void) const
{
	return m_timeVal;
}

// Get seconds component
time_t Time::getSecs(void) const
{
	return (m_timeVal.tv_sec);
}

// Get milli-seconds component
unsigned int Time::getMillis(void) const
{
	return (m_timeVal.tv_usec / MILLISCALE);
}

// Get total milli-seconds
unsigned int Time::getTotalMillis(void) const
{
	unsigned int tot = getMillis();
	tot += getSecs() * 1000;
	return tot;
}


/************/
/* Mutators */
/************/


// Subtract some fixed interval
void Time::subInterval(time_t secs, unsigned int millis)
{
	// Subtract seconds
	m_timeVal.tv_sec -= secs;

	// Subtract multiple milli-second seconds
	if(millis >= 1000) {
		m_timeVal.tv_sec -= (millis / 1000);
		millis = millis % 1000;
	}

	// Subtract millis
	unsigned int newMillis = getMillis();
	if(newMillis < millis) {
		// Deal with underflow
		newMillis = 1000 - (millis - newMillis);
		m_timeVal.tv_sec -= 1;
	}
	else {
		newMillis -= millis;
	}
	m_timeVal.tv_usec = newMillis * MILLISCALE;
}


// Add some fixed interval
void Time::addInterval(time_t secs, unsigned int millis)
{
	// Add seconds
	m_timeVal.tv_sec += secs;

	// Add millis
	unsigned int newMillis = getMillis() + millis;
	if(newMillis >= 1000) {
		// Deal with overlow
		m_timeVal.tv_sec += (newMillis / 1000);
		newMillis = newMillis % 1000;
	}
	m_timeVal.tv_usec = newMillis * MILLISCALE;
}


/*************/
/* Utilities */
/*************/


// Has this time elapsed?
bool Time::elapsed(void) const
{
	bool res = true;

	// Get current time
	struct timeval now;
	gettimeofday(&now, 0);

	// Have we passed the current time
	if(m_timeVal.tv_sec > now.tv_sec) {
		// Seconds have not passed
		res = false;
	}
	else {
		if(m_timeVal.tv_sec == now.tv_sec) {
			// Seconds are equal
			if(m_timeVal.tv_usec > now.tv_usec) {
				// Microseconds have not passed
				res = false;
			}
		}
	}

	return res;
}


// Format as a string
std::string Time::toString(void) const
{
	std::stringstream str;
	str << "Secs: " << getSecs() << ", Mili: " << getMillis();
	return str.str();
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const Time& time)
{
	str << time.toString();
	return str;
}


}	// namespace

