/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate tunable parameters for a tunnel.
 *
 * $Id: TuneSettings.cc,v 1.2 2009-02-27 11:56:19 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "TuneSettings.hh"


namespace HTTPTunnel {


// Construction/destruction
TuneSettings::TuneSettings()
: m_tune(""), m_tunedMaxUnackedMsgs(0), m_tunedFastPollTimeout(0),
  m_tunedMediumPollTimeout(0), m_tunedSlowPollTimeout(0), m_tunedMaxPollTimeout(0)
{
	// Setup defaults
	setup();
}

TuneSettings::TuneSettings(std::string tune)
: m_tune(tune), m_tunedMaxUnackedMsgs(0), m_tunedFastPollTimeout(0),
  m_tunedMediumPollTimeout(0), m_tunedSlowPollTimeout(0), m_tunedMaxPollTimeout(0)
{
	// Setup from description
	setup();
}

TuneSettings::TuneSettings(const TuneSettings& that)
: m_tune(that.m_tune), m_tunedMaxUnackedMsgs(that.m_tunedMaxUnackedMsgs),
  m_tunedFastPollTimeout(that.m_tunedFastPollTimeout),
  m_tunedMediumPollTimeout(that.m_tunedMediumPollTimeout),
  m_tunedSlowPollTimeout(that.m_tunedSlowPollTimeout),
  m_tunedMaxPollTimeout(that.m_tunedMaxPollTimeout)
{
}

TuneSettings::~TuneSettings()
{
}


// Assignment
TuneSettings& TuneSettings::operator=(const TuneSettings& that)
{
	// Self assignment?
	if(this != &that) {
		m_tune = that.m_tune;
		m_tunedMaxUnackedMsgs = that.m_tunedMaxUnackedMsgs;
		m_tunedFastPollTimeout = that.m_tunedFastPollTimeout;
		m_tunedMediumPollTimeout = that.m_tunedMediumPollTimeout;
		m_tunedSlowPollTimeout = that.m_tunedSlowPollTimeout;
		m_tunedMaxPollTimeout = that.m_tunedMaxPollTimeout;
	}

	return *this;
}


/*************/
/* Accessors */
/*************/


// Get tuning description
std::string TuneSettings::getTuneDesc(void) const
{
	return m_tune;
}

// Get maximum no. of unacked msgs
unsigned int TuneSettings::getMaxUnackedMsgs(void) const
{
	return m_tunedMaxUnackedMsgs;
}

// Get fast poll timeout
unsigned int TuneSettings::getFastPollTimeout(void) const
{
	return m_tunedFastPollTimeout;
}

// Get medium poll timeout
unsigned int TuneSettings::getMediumPollTimeout(void) const
{
	return m_tunedMediumPollTimeout;
}

// Get slow poll timeout
unsigned int TuneSettings::getSlowPollTimeout(void) const
{
	return m_tunedSlowPollTimeout;
}

// Get maximum poll timeout
unsigned int TuneSettings::getMaxPollTimeout(void) const
{
	return m_tunedMaxPollTimeout;
}


/*********************/
/* Private Interface */
/*********************/


// Set tuning settings based on the tuning description
void TuneSettings::setup(void)
{
	// What tuning is required?
	if(m_tune == "bursty") {
		// Bursty tuning
		m_tunedMaxUnackedMsgs = BRST_MAX_UNACKED_MSGS;
		m_tunedFastPollTimeout = BRST_FAST_POLL_TIMEOUT;
		m_tunedMediumPollTimeout = BRST_MEDIUM_POLL_TIMEOUT;
		m_tunedSlowPollTimeout = BRST_SLOW_POLL_TIMEOUT;
		m_tunedMaxPollTimeout = BRST_MAX_POLL_TIMEOUT;
	}
	else if(m_tune == "realtime") {
		// Realtime tuning
		m_tunedMaxUnackedMsgs = RLTM_MAX_UNACKED_MSGS;
		m_tunedFastPollTimeout = RLTM_FAST_POLL_TIMEOUT;
		m_tunedMediumPollTimeout = RLTM_MEDIUM_POLL_TIMEOUT;
		m_tunedSlowPollTimeout = RLTM_SLOW_POLL_TIMEOUT;
		m_tunedMaxPollTimeout = RLTM_MAX_POLL_TIMEOUT;
	}
	else {
		// Default tuning
		m_tunedMaxUnackedMsgs = DFLT_MAX_UNACKED_MSGS;
		m_tunedFastPollTimeout = DFLT_FAST_POLL_TIMEOUT;
		m_tunedMediumPollTimeout = DFLT_MEDIUM_POLL_TIMEOUT;
		m_tunedSlowPollTimeout = DFLT_SLOW_POLL_TIMEOUT;
		m_tunedMaxPollTimeout = DFLT_MAX_POLL_TIMEOUT;
	}
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const TuneSettings& settings)
{
   str << settings.getTuneDesc();
   return str;
}


}	// namespace

