/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store all information about a unique Tunnel.
 *
 * $Id: Tunnel.hh,v 1.7 2009-05-08 13:44:54 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _TUNNEL_HH_
#define _TUNNEL_HH_

#include <sys/time.h>
#include <vector>
#include <functional>
#include <sstream>
#include <time.h>
#include "Mutex.hh"
#include "Payload.hh"
#include "Exceptions.hh"
#include "TunnelID.hh"
#include "Time.hh"
#include "User.hh"
#include "TuneSettings.hh"


namespace HTTPTunnel {

class TunnelConfig;
class Tunneler;
class RawBuff;
class Connection;
class HTTPConnection;
class HTTPMsg;
class HTTPMsgParms;
class ProxySettings;
class ServerAuthSettings;
class Authenticator;
class TunnelParms;


class Tunnel {
public:

	/*********/
	/* Types */
	/*********/

	// Left and right
	enum Direction { LEFT, RIGHT, UNKNOWNDIR };

	// Tunnel states
	enum TunnelState {
		CONNECT, AUTHREQ, RECONNECT, CLOSING, CLOSED, CONNECTED, UNKNOWNSTATE
	};

	// Tunnel testing states
	enum TunnelTestState {
		NO_TESTING, BEGIN_TESTING, CLIENT_TESTING, SERVER_TESTING, COMPLETED_TESTING
	};

	// Enumerate all communication error type
	enum TunnelErrorType {
		GENERAL_COMMS_ERROR, PROXY_AUTH_ERROR, SERVER_AUTH_ERROR, BAD_RESPONSE_ERROR,
		MSG_RESEND_ERROR, MAX_RESEND_ERROR, CONNECTION_CLOSED_ERROR, UNVIABLE_ERROR,
		QUOTAS_EXCEEDED_ERROR, FATAL_QUOTAS_ERROR, SERVER_CONNECT_ERROR,
		CLIENT_AUTH_FAILED_ERROR, UNPROCESSED_DATA_ERROR
	};


	/*************/
	/* Selectors */
	/*************/

	// Base class
	struct Selector {
		Selector() {}
		virtual ~Selector() {}
		virtual void operator()(std::vector<Connection*>& sel, Tunnel* tun) = 0;
	};

	// Select only left side connections
	struct SelLeftConn : public Selector {
		void operator()(std::vector<Connection*>& sel, Tunnel* tun);
	};

	// Select only right side connections
	struct SelRightConn : public Selector {
		void operator()(std::vector<Connection*>& sel, Tunnel* tun);
	};

	// Sort according to most recently first connected connection
	struct SelRecentConn : public Selector {
		void operator()(std::vector<Connection*>& sel, Tunnel* tun);
	};

	// Exclude the specified connection
	struct SelNotConn : public Selector {
		SelNotConn(Connection* conn) : m_conn(conn) {}
		void operator()(std::vector<Connection*>& sel, Tunnel* tun);
		Connection* m_conn;
	};

	// Select only connected connections
	struct SelConnectedConn : public Selector {
		void operator()(std::vector<Connection*>& sel, Tunnel* tun);
	};


	/**************/
	/* Predicates */
	/**************/

	// Determine whether connection is from left side
	struct PredIsLeft : public std::unary_function<Connection*, bool> {
		PredIsLeft(Tunnel* tun) : m_tun(tun) {}
		bool operator()(const Connection* const& c) {
			return (m_tun->leftConnection(c));
		}
		Tunnel* m_tun;
	};

	// Determine whether connection is from right side
	struct PredIsRight : public std::unary_function<Connection*, bool> {
		PredIsRight(Tunnel* tun) : m_tun(tun) {}
		bool operator()(const Connection* const& c) {
			return (m_tun->rightConnection(c));
		}
		Tunnel* m_tun;
	};


	/***********/
	/* Methods */
	/***********/

	// Construction/destruction
	Tunnel(TunnelConfig* tcfg, Tunneler* tunneler, Direction tunDir, TunnelState state);
	virtual ~Tunnel();


	/*************/
	/* Accessors */
	/*************/

	// Get our config
	const TunnelConfig* getConfig(void) const;
	ProxySettings* getProxySettings(void);
	ServerAuthSettings* getServerAuthSettings(void);

	// Get our tunnel ID
	TunnelID getTid(void) const;
	std::string getTidStr(void) const;

	// Get our owning tunneler
	Tunneler* getTunneler(void);

	// Does the specified connection belong to this Tunnel?
	bool ownsConnection(const Connection* conn);

	// Is the specified connection from the left or right side?
	bool leftConnection(const Connection* conn);
	bool rightConnection(const Connection* conn);

	// Get the number of left/right side connections for this tunnel
	unsigned int noLeftConnections(void);
	unsigned int noRightConnections(void);

	// Get whether we've ever been connected in some direction
	bool everConnected(Direction dir);

	// Get whether we are closing/closed
	bool isClosing(void);
	bool isClosed(void);

	// Get the number of connections matching the provided criteria
	unsigned int noConnFromSel(std::vector<Selector*>& crit);

	// Get connections matching the provided criteria
	Connection* getConnFromSel(std::vector<Selector*>& crit);
	void getConnsFromSel(std::vector<Selector*>& crit, std::vector<Connection*>& res);

	// Get name of owning user
	std::string getUserName(void) const;

	// Get owning user details
	User getUserDetails(void);

	// Get source/destination addresses of connections for this tunnel
	virtual std::string getSrcConnectionAddresses(void);
	virtual std::string getDstConnectionAddresses(void);

	// Get our current states
	TunnelState getState(void) const;
	TunnelTestState getTestState(void) const;

	// Have we not received any data for a suspiciously long time?
	bool rawRecvTimeout(void);
	bool tunRecvTimeout(void);

	// Get whether we are a test tunnel
	virtual bool isTestTunnel(void) const;

	// Get bytes sent/received
	unsigned int getRawBytesIn(void) const;
	unsigned int getRawBytesOut(void) const;
	unsigned int getHTTPBytesIn(void) const;
	unsigned int getHTTPBytesOut(void) const;

	// Get notification msgs
	std::vector<std::string> getSendNotifMsgs(void);
	std::vector<std::string> getRecvNotifMsgs(void);

	// Are we persistent?
	virtual bool isPersistent(void);


	/************/
	/* Mutators */
	/************/

	// Add new connections to this tunnel
	void addLeftConnection(Connection* conn);
	void addRightConnection(Connection* conn);

	// Record connection error
	virtual void recordConnectionError(TunnelErrorType errType);

	// Remove closing connection from this tunnel
	virtual void removeConnection(Connection* conn);

	// A connection that carries us has connected
	virtual void connectionConnected(Connection* conn);

	// An attempt to connect out somewhere failed and has been abandoned
	virtual void connectAttemptFailed(Connection* conn);

	// Decrease leaky bucket error count
	void decreaseErrorCount(void);

	// State change notifications
	void closing(void);
	void reconnecting(void);

	// Add a new notification msg for sending
	void addSendNotifMsg(std::string msg);


	/*************/
	/* Utilities */
	/*************/

	// Process some newly extracted tunnel/raw data from one of our connections
	virtual void processTunnelMsg(HTTPMsg* msg, HTTPConnection* httpConn);
	void processRawData(const RawBuff& raw);

	// Determine whether we have experienced too many errors to continue
	bool errorLevelExceeded(Direction dir);
	bool errorLevelExceeded(void);

	// Send messages in the appropriate direction
	void sendTunnelMsg(bool retry, Connection* conn);
	void respondToTunnelMsg(Connection* conn, const HTTPMsgParms* recvd);
	void sendRawMsgs(void);

	// Check ack timeout on all unacknowledged messages
	bool checkAckTimeout(void);

	// Are we in a state where we can be shutdown?
	bool canBeShutdown(void);


protected:

	/*********/
	/* Types */
	/*********/

	// HTTPMsgParms for a sent message, plus other details
	struct SentMsgParms {
		// Construction/destruction
		SentMsgParms(HTTPMsgParms* parms, unsigned int connID);
		~SentMsgParms();

		// Data
		HTTPMsgParms*			m_parms;
		unsigned int			m_connID;
		Time						m_firstSentTime;
		Time						m_lastSentTime;
		unsigned int			m_ackCount;
		unsigned int			m_resentCount;
	};


	/*************/
	/* Constants */
	/*************/

	// Maximum no. of times we will try resending a msg
	static const unsigned int		MAX_MSG_RESEND					= 3;

	// Raw data buffer fill threshhold
	static const unsigned int		MAX_RAW_DATA_FILL				= 1024 * 1024;
	static const unsigned int		RAW_DATA_UNBLOCK_LEVEL		= 90;


	/**************/
	/* Predicates */
	/**************/

	// Find msg parameter by payload sequence number
	struct CmpSentParmsSeqNo {
		CmpSentParmsSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const SentMsgParms* p);
		SeqNo m_seqNo;
	};

	// Find msg parameter less than payload sequence number
	struct CmpSentParmsLessSeqNo {
		CmpSentParmsLessSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const SentMsgParms* p);
		SeqNo m_seqNo;
	};


	/***********/
	/* Methods */
	/***********/

	// Create a new payload msg for sending down the tunnel
	virtual HTTPMsgParms* makeTunnelMsg(const HTTPMsgParms* recvd, bool retry)
		throw(MemoryError);

	// Create payload for tunneled msg to send to the destination
	virtual void makeTunnelPayload(RawBuff& data, bool& split, unsigned int max)
		throw(MemoryError);

	// Record that we've sent a new tunnel msg
	virtual void sentNewTunnelMsg(HTTPConnection* conn, HTTPMsgParms* msgParms) = 0;
	void sentNewTunnelMsg(HTTPConnection* conn, HTTPMsgParms* msgParms, bool remember);

	// Determine whether an ack implies we should resend a msg
	virtual void checkForResendRequest(const SeqNo& ackNo);
	virtual bool checkForMultiAckResendRequest(SentMsgParms* sentMsgParms) = 0;

	// Check message parms before resending
	virtual void checkResendMsgParms(HTTPMsgParms* parms);

	// Get authenticator
	virtual Authenticator* getAuthenticator(void);

	// Have we got a stable HTTP connection to other end?
	virtual bool httpStable(void) const;

	// A sent message has been acknowledged
	virtual void msgAcked(SentMsgParms* parms);

	// Can we send a new tunnel msg?
	virtual bool clearToSendNewMsg(HTTPConnection* conn) = 0;

	// Check whether any connections need to be blocked/unblocked
	virtual void checkBlockedConnections(void) = 0;

	// Advance our internal state machine
	virtual void advanceState(const HTTPMsgParms* parms) = 0;
	virtual void stateChanged(TunnelState oldState);

	// Get maximum allowable payload size
	virtual unsigned int getMaxPayloadSize(bool useHeader) = 0;

	// Something has been sent down the tunnel
	virtual void tunnelAnyMsgSent(void);

	// Record we need to resend a msg - Caller holds mutex
	virtual void needToResendMsg(const SeqNo& seqNo, bool msgLost);

	// Handle resend of old msgs
	void checkUnackedMgmtMsgs(void);
	bool resendUnackedMsgs(Connection* conn) throw(ResendError);
	bool isMgmtMsg(const HTTPMsgParms* parms) const;

	// Get tunnel state as string
	std::string stateAsString(TunnelState state) const;
	std::string testStateAsString(TunnelTestState state) const;

	// Set our current state
	void setState(TunnelState newState);
	void setTestState(TunnelTestState newState);

	// Get no. of unacked msgs on a connection
	unsigned int noUnackedMsgsOnConn(unsigned int connID);


	/********/
	/* Data */
	/********/

	// A Mutex
	Mutex										m_mtx;

	// Owning tunneler
	Tunneler*								m_tunneler;

	// Lists of connections that comprise this Tunnel
	std::vector<Connection*>			m_leftConn;
	std::vector<Connection*>			m_rightConn;

	// Configuration for this Tunnel
	TunnelConfig*							m_tunCfg;
	TunnelID									m_tid;

	// Min/max and current ack timeout values
	time_t									m_minAckTimeout;
	time_t									m_maxAckTimeout;
	time_t									m_ackTimeout;

	// Round trip ack times
	std::vector<Time>						m_ackRTTs;
	unsigned int							m_avgAckRTT;
	unsigned int							m_biasAvgAckRTT;

	// User of this tunnel
	std::string								m_user;

	// Current tunnel states
	TunnelState								m_tunnelState;
	TunnelTestState						m_tunnelTestState;

	// Queue of parameters for sent but unacknowledged messages
	std::vector<SentMsgParms*>			m_unackedMsgQueue;

	// Time we last sent ANY msg down the tunnel
	Time										m_lastTunnelAnySendTime;

	// Queues of received tunnel/raw data
	std::vector<Payload*>				m_tunnelDataQueue;
	std::vector<RawBuff*>				m_rawDataQueue;
	unsigned int							m_rawDataQueueSize;

	// Maximum no. of outstanding unacked msgs
	unsigned int							m_maxUnackedMsgs;

	// Last received ack sequence no.
	SeqNo										m_lastRecvAckSeqNo;
	bool										m_lastRecvAckResent;

	// Queue of sequence nos. for messages we must resend
	std::vector<SeqNo>					m_resendQueue;


private:

	/*************/
	/* Constants */
	/*************/

	// Amount to decrease leaky bucket error count by each period
	static const unsigned int		ERROR_COUNT_DECREASE			= 1;

	// Level for leaky bucket error count which we consider fatal
	static const unsigned int		FATAL_ERROR_COUNT				= 20;

	// Ack timeout in seconds
	static const time_t				MIN_ACK_TIMEOUT				= 5;
	static const time_t				MAX_ACK_TIMEOUT				= 20;

	// Seconds tunnel can stay completely silent before we call it an error
	static const time_t				MAX_SILENCE_TIME				= 60;

	// Seconds we go with no data received before considering tunnel silent
	static const time_t				MAX_NO_RECV_TIME				= 60;

	// Time after sending last tunnel msg we act on unviable errors (seconds)
	static const unsigned int		UNVIABLE_ERROR_TIME			= 5;

	// Maximum length we allow the old connection list to grow to
	static const unsigned int		MAX_OLD_CONNS_LENGTH			= 100;

	// Number of ack round trip times to keep for averaging
	static const unsigned int		ACKRTT_AVG_LENGTH				= TuneSettings::MAX_UNACKED_MSGS * 3;

	// Maximum no. of recvd notification msgs we keep
	static const unsigned int		MAX_RECV_NOTIFS				= 25;


	/**************/
	/* Predicates */
	/**************/

	// Order message payloads by sequence number
	struct CmpPayloadsSeqNo {
		bool operator()(const Payload* p1, const Payload* p2) {
			return (*p1 < *p2);
		}
	};

	// Find message payload by sequence number
	struct CmpPayloadSeqNo {
		CmpPayloadSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const Payload* p) {
			return (p->getSeqNo() == m_seqNo);
		}
		SeqNo m_seqNo;
	};


	/***********/
	/* Methods */
	/***********/

	// Increment next sent/received packet sequence numbers
	void incSendSeqNo(void);
	void incRecvSeqNo(void);

	// Clear up raw packet queue after sending tunneled msg
	void sentTunnelPayload(HTTPMsgParms* parms);

	// Handle raw data for sending to the destination
	Payload* makeRawPayload(void) throw(MemoryError);
	void sentRawPayload(bool sentOk);

	// Handle some received acknowledgements/non-acknowledgements
	void processAck(std::string& ack);

	// Calculate ack-rtt averages
	void calculateAvgAckRTT(void);

	// Check some new payload isn't from a duplicate message
	bool duplicateTunnelMsg(SeqNo& seqNo);

	// Check & get a suitable connection for sending a tunnel msg out of
	Connection* getTunnelConnection(Connection* conn);

	// Is a connection an old removed one?
	bool oldConnection(Connection* conn);

	// Send a tunnel msg with the given paramaters
	void sendTunnelMsgWithParms(HTTPConnection* conn, HTTPMsgParms* parms)
		throw(TunnelError);

	// Process received notification msgs
   void processRecvNotifs(const HTTPMsgParms* parms);


	/********/
	/* Data */
	/********/

	// Remember IDs of connections we've explicitly dropped
	std::vector<unsigned int>			m_oldConns;

	// Sequence numbers for msgs sent/received
	SeqNo										m_nextSendSeqNo;
	SeqNo										m_nextRecvSeqNo;

	// Leaky bucket error counter
	unsigned int							m_errorCount;

	// Flag to indicate if the last dead connections ever connected
	bool										m_leftEverConnected;
	bool										m_rightEverConnected;

	// In which direction does the tunnel lie?
	Direction								m_tunnelDirection;

	// Flag to indicate whether we have scheduled a shutdown
	bool										m_shutdownScheduled;

	// Things to determine whether we can be shutdown
	time_t									m_lastTunnelSendTime;
	bool										m_sentCloseMsg;

	// Time we last sent a standard CONNECTED msg down the tunnel
	Time										m_lastTunnelConnectedSendTime;

	// Time we last received data
	Time										m_lastRawRecvTime;
	Time										m_lastTunRecvTime;

	// Count of bytes sent/received
	unsigned int							m_rawBytesIn;
	unsigned int							m_rawBytesOut;
	unsigned int							m_httpBytesIn;
	unsigned int							m_httpBytesOut;

	// Lists of notification msgs for sending and been received
	std::vector<std::string>			m_sendNotifMsgs;
	std::vector<std::string>			m_recvNotifMsgs;

};


}	// namespace

#endif

