/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all per tunnel configuration.
 *
 * $Id: TunnelConfig.cc,v 1.3 2008-07-11 11:22:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <string>
#include <sstream>
#include <vector>
#include "TunnelConfig.hh"
#include "TunnelID.hh"
#include "MutexVar.hh"
#include "Utils.hh"
#include "ProxySettings.hh"
#include "ServerAuthSettings.hh"


namespace HTTPTunnel {


// Construction/destruction
TunnelConfig::TunnelConfig(TunnelID* tid, std::string tune, const ProxySettings* proxyCfg,
	const ServerAuthSettings* saCfg, std::string destIP, unsigned int destPort)
: m_mtx(), m_tid(tid), m_httpVer("1.1"), m_destIP(destIP), m_destPort(destPort),
  m_useHeaders(true), m_proxyCfg(0), m_serverAuthCfg(0), m_tuning(tune)
{
	// Copy proxy config
	if(proxyCfg)
		m_proxyCfg = new ProxySettings(*proxyCfg);

	// Copy server auth config
	if(saCfg)
		m_serverAuthCfg = new ServerAuthSettings(*saCfg);
}

TunnelConfig::TunnelConfig(TunnelID* tid, std::string tune, std::string destAddr,
	const ServerAuthSettings* saCfg)
: m_mtx(), m_tid(tid), m_httpVer("1.1"), m_destIP(""), m_destPort(0), m_useHeaders(true),
  m_proxyCfg(0), m_serverAuthCfg(0), m_tuning(tune)
{
	// Set destination IP and port from address string
	setDestAddress(destAddr);

	// Copy server auth config
	if(saCfg) {
		m_serverAuthCfg = new ServerAuthSettings(*saCfg);

		// Create new authentication nonce for this tunnel
		m_serverAuthCfg->newNonce();
	}
}

TunnelConfig::~TunnelConfig()
{
	MutexVar mv(m_mtx);

	if(m_proxyCfg) {
		delete m_proxyCfg;
		m_proxyCfg = 0;
	}
	if(m_serverAuthCfg) {
		delete m_serverAuthCfg;
		m_serverAuthCfg = 0;
	}
	if(m_tid) {
		delete m_tid;
		m_tid = 0;
	}
}


/*************/
/* Accessors */
/*************/

const TunnelID* TunnelConfig::getTunnelID(void) const
{
	return m_tid;
}

std::string TunnelConfig::getHTTPVersion(void) const
{
	return m_httpVer;
}

std::string TunnelConfig::getDestIP(void) const
{
	return m_destIP;
}

unsigned int TunnelConfig::getDestPort(void) const
{
	return m_destPort;
}

Enums::Protocol TunnelConfig::getDestProtocol(void) const
{
	return m_tid->getProtocol();
}

std::string TunnelConfig::getDestAddress(void) const
{
	std::stringstream addr;
	addr << getDestIP() << ":" << getDestPort();
	return addr.str();
}

bool TunnelConfig::getUseHeaders(void) const
{
	return m_useHeaders;
}

ProxySettings* TunnelConfig::getProxySettings(void)
{
	return m_proxyCfg;
}

ServerAuthSettings* TunnelConfig::getServerAuthSettings(void)
{
	return m_serverAuthCfg;
}

const TuneSettings& TunnelConfig::getTuning(void) const
{
	return m_tuning;
}


/************/
/* Mutators */
/************/

void TunnelConfig::setHTTPVersion(std::string ver)
{
	MutexVar mv(m_mtx);
	m_httpVer = ver;
}

void TunnelConfig::setDestIP(std::string destIP)
{
	MutexVar mv(m_mtx);
	m_destIP = destIP;
}

void TunnelConfig::setDestPort(unsigned int destPort)
{
	MutexVar mv(m_mtx);
	m_destPort = destPort;
}

void TunnelConfig::setDestAddress(std::string addr)
{
	// Split into host & port
	std::vector<std::string> parms;
	Utils::SplitString(addr, ":", parms);
	setDestIP(parms[0]);
	setDestPort(atoi(parms[1].c_str()));
}

void TunnelConfig::setUseHeaders(bool uh)
{
	MutexVar mv(m_mtx);
	m_useHeaders = uh;
}


}	// namespace

