/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Uniquely identify an individual tunneled connection.
 *
 * $Id: TunnelID.cc,v 1.3 2008-06-26 09:47:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <string>
#include <sstream>
#include <iostream>
#include "TunnelID.hh"
#include "ClientSideConfig.hh"
#include "GlobalConfig.hh"
#include "Utils.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;


// Initialise static data
unsigned int TunnelID::c_idCount = 0;
Mutex TunnelID::c_idCountMtx;


// Construction/destruction
TunnelID::TunnelID()
: m_id(0), m_cliProcID(0), m_tunnelerID()
{
}

TunnelID::TunnelID(ClientSideConfig* cfg)
: m_id(getNextID()), m_cliProcID(g_pConfig->getPid()), m_tunnelerID(cfg->getTunnelerID())
{
}

TunnelID::TunnelID(std::string id) throw(InvalidTunnelID, InvalidTunnelerID)
: m_id(0), m_cliProcID(0), m_tunnelerID(id)
{
	// Split composite ID
	std::vector<std::string> parts;
	Utils::SplitString(id, ":", parts);

	// Got everything we need?
	if(parts.size() != 6) {
		std::stringstream err;
		err << "Invalid tunnel ID string: " << id;
		throw InvalidTunnelID(err.str());
	}

	// Construct individual ID values
	m_id = strtoul(parts[0].c_str(), 0, 10);
	m_cliProcID = atoi(parts[1].c_str());
}

TunnelID::TunnelID(const TunnelID& that)
: m_id(that.m_id), m_cliProcID(that.m_cliProcID), m_tunnelerID(that.m_tunnelerID)
{
}

TunnelID::~TunnelID()
{
}


// Assignment
TunnelID& TunnelID::operator=(const TunnelID& that)
{
	// Self assignment?
	if(this != &that) {
		m_id = that.m_id;
		m_cliProcID = that.m_cliProcID;
		m_tunnelerID = that.m_tunnelerID;
	}

	return *this;
}


// Operators
bool TunnelID::operator==(const TunnelID& that) const
{
	// Everything must match
	bool res = (that.m_id == m_id);
	res &= (that.m_cliProcID == m_cliProcID);
	res &= (that.m_tunnelerID == m_tunnelerID);
	return res;
}

bool TunnelID::operator<(const TunnelID& that) const
{
	// Just create a string of all members and compare that
	return (toString() < that.toString());
}


// Accessors
unsigned int TunnelID::getID(void) const
{
	return m_id;
}

unsigned int TunnelID::getProcID(void) const
{
	return m_cliProcID;
}

const TunnelerID& TunnelID::getTunnelerID(void) const
{
	return m_tunnelerID;
}

Enums::Protocol TunnelID::getProtocol(void) const
{
	return m_tunnelerID.getProtocol();
}


// Create a representation of ourselves as a string
std::string TunnelID::toString(void) const
{
	std::stringstream val;
	val << m_id << ":" << m_cliProcID << ":" << m_tunnelerID;
	return val.str();
}


/*********************/
/* Private Interface */
/*********************/


// Get next unique ID
unsigned int TunnelID::getNextID(void)
{
	MutexVar mv(c_idCountMtx);
	if(++c_idCount == 0)
		c_idCount = 1;
	return c_idCount;
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const TunnelID& tid)
{
	str << tid.toString();
	return str;
}


}	// namespace

