/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Tunneler.cc,v 1.7 2009-04-24 15:37:33 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <vector>
#include <iostream>
#include <sstream>
#include <algorithm>
#include "Tunneler.hh"
#include "Debug.hh"
#include "TunnelID.hh"
#include "Tunnel.hh"
#include "MutexVar.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "Connection.hh"
#include "HTTPConnection.hh"
#include "Payload.hh"
#include "RawBuff.hh"
#include "HTTPMsg.hh"
#include "ConnectResultEvent.hh"
#include "ConnectionErrorEvent.hh"
#include "LeakTunnelErrorEvent.hh"
#include "ShutdownTunnelEvent.hh"
#include "TunnelDestroyedEvent.hh"
#include "PollTimeoutEvent.hh"
#include "CheckAckEvent.hh"
#include "TunnelStateEvent.hh"
#include "TunnelUnviableEvent.hh"
#include "ConnectionManagementEvent.hh"
#include "TunnelManagementEvent.hh"
#include "AuditEvent.hh"
#include "UDPSilentEvent.hh"
#include "SendMsgEvent.hh"
#include "FlushConnectionEvent.hh"
#include "AuthSuccessEvent.hh"
#include "ConnectionManager.hh"
#include "Server.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"
#include "TunnelStats.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
Tunneler::Tunneler(Server* server) throw(TunnelError)
: EventReceiver(), m_server(server), m_tunnels(), m_tunnelsMtx(), m_closingTunnels(),
  m_closingTunnelsMtx(), m_connections(), m_connectionsMtx(), m_closingConnections(),
  m_closingConnectionsMtx(), m_connectionMgr(0), m_lastTunParmsCheckTime(Time::now())
{
	DPRINTV("Tunneler::Tunneler()");

	// Create connection manager
	m_connectionMgr = new ConnectionManager();

	// Schedule initial events
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
	scheduleEvent(new ConnectionManagementEvent(Time(0,CONN_MANAGE_PERIOD), this));
	scheduleEvent(new TunnelManagementEvent(TUN_MANAGE_PERIOD, this));
	scheduleEvent(new LeakTunnelErrorEvent(LEAK_ERROR_PERIOD, this));

	#ifdef DEBUG_TUN
	scheduleEvent(new AuditEvent(AUDIT_PERIOD, this));
	#endif
}

Tunneler::~Tunneler()
{
	DPRINTV("Tunneler::~Tunneler()");

	// Clear up list of Tunnels
	MutexVar tmv(m_tunnelsMtx);
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		delete *tit;
		*tit = 0;
	}
	m_tunnels.clear();
	tmv.unlock();

	// Destroy connection manager
	MutexVar cmv(m_connectionsMtx);
	MutexVar ccmv(m_closingConnectionsMtx);
	delete m_connectionMgr;
	m_connectionMgr = 0;

	// Clear up list of connections
	std::vector<Connection*>::iterator cit;
	for(cit = m_connections.begin(); cit != m_connections.end(); ++cit) {
		delete *cit;
		*cit = 0;
	}
	m_connections.clear();

	// Clear up list of closing connections
	std::vector<Connection*>::iterator ccit;
	for(ccit = m_closingConnections.begin(); ccit != m_closingConnections.end(); ++ccit) {
		delete *ccit;
		*ccit = 0;
	}
	m_closingConnections.clear();
}


/*************/
/* Accessors */
/*************/


// Get owning server
const Server* Tunneler::getServer(void) const
{
	return m_server;
}

Server* Tunneler::getServerMutable(void)
{
	return m_server;
}


// Get owning server configuration
const ClientSideConfig* Tunneler::getClientSideConfig(void) const
	throw(ConfigError)
{
	// Get basic config and cast it
	const ClientSideConfig* cfg = dynamic_cast<const ClientSideConfig*>
		(m_server->getConfig());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve client side config");
	return cfg;
}

ClientSideConfig* Tunneler::getClientSideConfigMutable(void)
	throw(ConfigError)
{
	// Get basic config and cast it
	ClientSideConfig* cfg = dynamic_cast<ClientSideConfig*>
		(m_server->getConfigMutable());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve client side config");
	return cfg;
}

const ServerSideConfig* Tunneler::getServerSideConfig(void) const
	throw(ConfigError)
{
	// Get basic config and cast it
	const ServerSideConfig* cfg = dynamic_cast<const ServerSideConfig*>
		(m_server->getConfig());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve server side config");
	return cfg;
}

ServerSideConfig* Tunneler::getServerSideConfigMutable(void)
	throw(ConfigError)
{
	// Get basic config and cast it
	ServerSideConfig* cfg = dynamic_cast<ServerSideConfig*>
		(m_server->getConfigMutable());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve server side config");
	return cfg;
}


// Get number of active managed tunnels
unsigned int Tunneler::getNoActiveTunnels(void)
{
	MutexVar tmv(m_tunnelsMtx);
	return m_tunnels.size();
}

// Get IDs of all active managed tunnels
void Tunneler::getActiveTunnelIDs(std::vector<TunnelID>& ids)
{
	// Start with an empty list
	ids.clear();

	MutexVar tmv(m_tunnelsMtx);

	// Iterate over all active tunnels
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		Tunnel* tun = *tit;
		ids.push_back(tun->getTid());
	}
}


// Get connection counts for some tunnel
unsigned int Tunneler::getTunnelConnectionCount(const TunnelID& tid,
	std::vector<Tunnel::Selector*>& crit)
{
	unsigned int res = 0;

	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Get connection count for specified criteria
		res = tun->noConnFromSel(crit);
	}

	return res;
}

// Get stats for some tunnel
TunnelStats* Tunneler::getTunnelStats(const TunnelID& tid)
{
	TunnelStats* stats = 0;

	// Lookup tunnel
	MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Get stats
		bool isTest = tun->isTestTunnel();
		std::string stateStr = "Unknown";
		std::string tstateStr = "Unknown";

		switch(tun->getState()) {
			case(Tunnel::CONNECT):
			case(Tunnel::AUTHREQ):
				stateStr = "Connecting";
				break;
			case(Tunnel::CLOSING):
			case(Tunnel::CLOSED):
				stateStr = "Closing";
				break;
			case(Tunnel::RECONNECT):
				stateStr = "Reconnecting";
				break;
			case(Tunnel::CONNECTED):
				stateStr = "Connected";
				break;
			default:
				break;
		}

		switch(tun->getTestState()) {
			case(Tunnel::NO_TESTING):
				tstateStr = "Not Testing";
				break;
			case(Tunnel::BEGIN_TESTING):
				tstateStr = "Beginning Testing";
				break;
			case(Tunnel::CLIENT_TESTING):
				tstateStr = "Client Testing";
				break;
			case(Tunnel::SERVER_TESTING):
				tstateStr = "Server Testing";
				break;
			case(Tunnel::COMPLETED_TESTING):
				tstateStr = "Testing Complete";
				break;
			default:
				break;
		}

		const TunnelConfig* tunCfg = tun->getConfig();
		std::string destAddr = tunCfg->getDestAddress();
		unsigned int rawBytesIn = tun->getRawBytesIn();
		unsigned int rawBytesOut = tun->getRawBytesOut();
		unsigned int httpBytesIn = tun->getHTTPBytesIn();
		unsigned int httpBytesOut = tun->getHTTPBytesOut();
		User user = tun->getUserDetails();
		std::vector<std::string> notifs = tun->getRecvNotifMsgs();

		// Create stats
		stats = new TunnelStats(tid, isTest, stateStr, tstateStr, destAddr,
			rawBytesIn, rawBytesOut, httpBytesIn, httpBytesOut, user, notifs);
	}

	return stats;
}


/************/
/* Mutators */
/************/


// Initiate a shutdown of the named tunnel
void Tunneler::initiateTunnelShutdown(const TunnelID& tid)
{
	try {
		// Lookup tunnel
		MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Schedule a shutdown for this tunnel
			DPRINTI("Tunneler::initiateTunnelShutdown() - " + tid.toString());
			scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
		}
		else {
			std::stringstream err;
			err << "Tunneler::initiateTunnelShutdown() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error initiating tunnel shutdown: " + e.what());
	}
}

// Initiate a shutdown of all tunnels
void Tunneler::initiateAllTunnelsShutdown(void)
{
	MutexVar tmv(m_tunnelsMtx);

	// Iterate over all tunnels
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Initiate shutdown on this tunnel
		TunnelID tid = (*tit)->getTid();
		initiateTunnelShutdown(tid);
	}
}


// Enable/disable testing based on current config setting
void Tunneler::setServerTesting(void)
{
	// Do nothing in base class
}


// Add connection error to list of scheduled events
void Tunneler::scheduleConnectionError(unsigned int connID, Tunnel::TunnelErrorType type)
{
	// Cancel any existing event
	ConnectionErrorEvent* event = 0;
	event = cancelAndGetEvent<ConnectionErrorEvent, unsigned int>(connID);
	if(event) {
		// Add error to this event
		event->addError(type);
	}
	else {
		// Create new event
		event = new ConnectionErrorEvent(connID, type, this);
	}

	// Schedule new event
	scheduleEvent(event);
}


/**********/
/* Events */
/**********/


// Notification of new HTTP message arrival
void Tunneler::messageArrived(unsigned int connID, HTTPMsg* msg)
	throw(LookupManagementError)
{
	#ifdef DEBUG_HTTP_MSG_LOSS
	// Random HTTP msg loss
	int randLoss = ((float)rand() / RAND_MAX) * 20;
	if(randLoss == 1) {
		std::stringstream los;
		los << "Dropping HTTP msg from connection " << connID;
		DPRINTV(los.str());
		return;
	}
	#endif

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::messageArrived() - On connection " << connID;
	DPRINTV(nfo.str());
	#endif

	// Lookup connection
	MutexVar cmv(m_connectionsMtx);
	MutexVar tmv(m_tunnelsMtx);
	std::vector<Connection*>::iterator cit = getConnection(connID);

	// Let connection know it has received a new HTTP msg
	HTTPConnection* httpConn = dynamic_cast<HTTPConnection*>(*cit);
	if(httpConn) {
		httpConn->msgRecvd();
	}
}


// Notification of some new default tunnel parameters
void Tunneler::newTunnelParms(TunnelParms* parms, const TunnelerID& tid)
{
	// Nothing to do in base class
}


// Notification of a new UDP 'connection'
void Tunneler::newUDPConnectionArrived(Sock* insock) throw(CommsError)
{
	DPRINTI("Tunneler::newUDPConnectionArrived()");

	// Error to receive this in a non-overriden class
	throw CommsError("UDP tunneling not supported in this tunnel type");
}


// Notification that a tunnel needs to be cleared down
void Tunneler::shutdownTunnel(const TunnelID& tid, bool otherEndInitiated)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::shutdownTunnel() - " << tid.toString();
	nfo << " OEI = " << otherEndInitiated;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup tunnel
		MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Check for test tunnel
			checkScheduleServerTest(tun);

			// How was this tunnel shutdown?
			if(otherEndInitiated) {
				// Other end not expecting anything else from us, so just destroy tunnel
				destroyTunnel(tun);
			}
			else {
				// Try to shutdown tunnel cleanly
				tunnelClosing(tun);
			}
		}
		else {
			std::stringstream err;
			err << "Tunneler::shutdownTunnel() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error shutting down tunnel: " + e.what());
	}
}


// Notification that a tunnel has been shutdown (destroyed)
void Tunneler::tunnelDestroyed(const TunnelID& tid)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::tunnelDestroyed() - " << tid.toString();
	DPRINTV(nfo.str());
	#endif
}


// Notification of a poll timeout on a tunnel
void Tunneler::pollTimeout(const TunnelID& tid)
{
	// Not interested in base class
}


// Check UDP datagram timeout
void Tunneler::checkDatagramTimeout(const TunnelID& tid)
{
	DPRINTI("Tunneler::checkDatagramTimeout() - " + tid.toString());

	try {
		// Lookup tunnel
		MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// UDP silence threshold exceeded?
			if(tun->rawRecvTimeout() && tun->tunRecvTimeout()) {
				DPRINTV("UDP silence threshold exceeded for tunnel " + tid.toString());

				// Try to shutdown tunnel cleanly
				tunnelClosing(tun);
			}
			else {
				// Schedule next UDP silent check event
				scheduleEvent(new UDPSilentEvent(this, tid));
			}
		}
		else {
			std::stringstream err;
			err << "Tunneler::checkDatagramTimeout() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error processing UDP datagram timeout: " + e.what());
	}
}


// Notification that a connection has changed state
void Tunneler::connectionStateChange(unsigned int connID,
	Connection::ConnectionState state)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::connectionStateChange () - connection " << connID;
	nfo << " state: " << state;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		MutexVar cmv(m_connectionsMtx);
		MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit;
		cit = getConnection(connID);
		Connection* conn = *cit;

		// Do whatever the state change requires
		if(state == Connection::CONNECTED) {
			conn->tellTunnelsAboutConnect();
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connection state change: " + e.what());
	}
}


// Notification that a tunnel has changed state
void Tunneler::tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state)
{
	// Not interested in base class
}


// Notification of a socket connect failure
void Tunneler::connectResult(ConnectResultEvent* result)
{
	DPRINTI("Tunneler::connectResult()");

	try {
		// Lookup connection
		MutexVar cmv(m_connectionsMtx);
		MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit;
		unsigned int connID = result->getConnectionID();
		cit = getConnection(connID);
		Connection* conn = *cit;

		// Let connection manager clear up after connect attempt
		m_connectionMgr->delProxyConnect(conn);

		// Connect failed?
		ConnectResultEvent::ConnectResultType resType = result->getType();
		if(resType != ConnectResultEvent::SUCCESS) {
			// Record failed connect attempt
			conn->incConnectFailedCount();

			// Was this a test connection?
			Tunnel* tun = getTunFromPred(PredTunConnection(conn));
			if(tun && tun->isTestTunnel()) {
				// Keep trying
				DPRINTV("Test connection failed");
				unsigned int failedCount = conn->getConnectFailedCount();
				unsigned int reconnDelay = MIN_TEST_RECONNECT_DELAY * failedCount;
				if(reconnDelay > MAX_TEST_RECONNECT_DELAY)
					reconnDelay = MAX_TEST_RECONNECT_DELAY;
				m_connectionMgr->addProxyConnect(conn, reconnDelay, result->getHost(),
					result->getPort(), result->getProtocol(), result->getSSL());
			}
			else {
				// Shall we try again?
				bool retry = true;
				unsigned int failedCount = conn->getConnectFailedCount();
				unsigned int maxAttempts = 0, reconnDelay = 0;
				getReconnectParms(maxAttempts, reconnDelay, failedCount);
				if(resType == ConnectResultEvent::FIREWALLED || failedCount >= maxAttempts) {
					retry = false;
				}

				if(retry) {
					// Schedule reconnect attempt
					m_connectionMgr->addProxyConnect(conn, reconnDelay, result->getHost(),
						result->getPort(), result->getProtocol(), result->getSSL());
				}
				else {
					// Tell tunnel connect attempt failed
					if(tun)
						tun->connectAttemptFailed(conn);

					// Clear up connection
					destroyConnection(cit);
				}
			}
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connect result: " + e.what());
	}
}


// Notification of error(s) on a socket connection
void Tunneler::connectionErrors(unsigned int connID,
	std::vector<Tunnel::TunnelErrorType>& errors)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::connectionError() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
		MutexVar cmv(m_connectionsMtx);
		MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);

		// Record errors
		std::vector<Tunnel::TunnelErrorType>::const_iterator eit;
		for(eit = errors.begin(); eit != errors.end(); ++eit) {
			#ifdef DEBUG_TUN
			std::stringstream nfo2;
			nfo2 << "Connection error type " << *eit;
			DPRINTI(nfo2.str());
			#endif

			(*cit)->tellTunnelsAboutError(*eit);
		}

		// Move connection to the closing list
		connectionClosing(cit);
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connection error: " + e.what());
	}
}


// Notification that a connection needs to be cleared down
void Tunneler::shutdownConnection(unsigned int connID)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::shutdownConnection() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		MutexVar cmv(m_connectionsMtx);
		MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);

		// Move connection to the closing list
		connectionClosing(cit);
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error shutting down connection: " + e.what());
	}
}


// Notification we should force a send of a message down a tunnel/connection
void Tunneler::sendTunnelMsg(const TunnelID& tid, unsigned int connID, bool retry)
{
	DPRINTV("Tunneler::sendTunnelMsg() - " + tid.toString());

	try {
		// Lookup connection
		MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		// Lookup tunnel
		MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Force send
			tun->sendTunnelMsg(retry, conn);
		}
		else {
			std::stringstream wrn;
			wrn << "Failed to force send of tunnel msg down unknown tunnel ";
			wrn << tid;
			g_pLogger->warn(wrn.str());
		}
	}
	catch(LookupManagementError& e) {
		std::stringstream wrn;
		wrn << "Error forcing send of tunnel msg down connection " << connID;
		wrn << ": " << e.what();
		g_pLogger->warn(wrn.str());
	}
}


// Notification we should try to flush a connection
void Tunneler::flushConnection(unsigned int connID, bool retry)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::flushConnection() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		try {
			// Flush connection
			conn->flush(retry);
		}
		catch(CommsError& e) {
			std::stringstream wrn;
			wrn << "Error flushing connection: " << e.what();
			g_pLogger->warn(wrn.str());
		}
	}
	catch(LookupManagementError& e) {
		std::stringstream wrn;
		wrn << "Error flushing connection: " << e.what();
		g_pLogger->warn(wrn.str());
	}
}


// Notification to decrease tunnels leaky error counts
void Tunneler::decreaseTunnelsErrorCount(void)
{
	MutexVar tmv(m_tunnelsMtx);

	// Iterate over all tunnels
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Decrease error count
		Tunnel* tun = *tit;
		tun->decreaseErrorCount();
	}
	tmv.unlock();

	// Schedule next error leak event
	scheduleEvent(new LeakTunnelErrorEvent(LEAK_ERROR_PERIOD, this));
}


// Notification that authentication has succeeded for a tunnel
void Tunneler::authenticationSuccess(const TunnelID& tid)
{
	// Nothing to do in base class
}


// Notification that we should perform connection manangement
void Tunneler::performConnectionManagement(void)
{
	MutexVar ccmv(m_closingConnectionsMtx);

	#ifdef DEBUG_TUN
	if(m_closingConnections.size() > 0) {
		std::stringstream dbg;
		dbg << "Tunneler::performConnectionManagement() - Managing ";
		dbg << m_closingConnections.size() << " connections";
		DPRINTV(dbg.str());
	}
	#endif

	// Delete any connections that can be shutdown
	std::transform(m_closingConnections.begin(), m_closingConnections.end(),
		m_closingConnections.begin(), PredConnShutdown(m_connectionMgr));
	std::vector<Connection*>::iterator cit = std::remove_if(m_closingConnections.begin(),
		m_closingConnections.end(), PredPtrNull<Connection>());
	m_closingConnections.erase(cit, m_closingConnections.end());

	// Schedule next event
	scheduleEvent(new ConnectionManagementEvent(Time(0,CONN_MANAGE_PERIOD), this));
}


// Notification that we should perform tunnel manangement
void Tunneler::performTunnelManagement(void)
{
	MutexVar tmv(m_tunnelsMtx);
	MutexVar ctmv(m_closingTunnelsMtx);

	#ifdef DEBUG_TUN
	if(m_closingTunnels.size() > 0) {
		std::stringstream dbg;
		dbg << "Tunneler::performTunnelManagement() - Managing ";
		dbg << m_closingTunnels.size() << " tunnels";
		DPRINTV(dbg.str());
	}
	#endif

	// Find all tunnels from closing list that can be shutdown
	std::vector<Tunnel*> sdTunnels;
	std::vector<Tunnel*>::iterator ctit;
	for(ctit = m_closingTunnels.begin(); ctit != m_closingTunnels.end(); ++ctit) {
		// Can this tunnel be shutdown now?
		Tunnel* tun = *ctit;
		if(tun->canBeShutdown())
			sdTunnels.push_back(tun);
	}

	// Destroy all tunnels that can be shutdown
	for(ctit = sdTunnels.begin(); ctit != sdTunnels.end(); ++ctit) {
		try {
			destroyTunnel(*ctit);
		}
		catch(LookupManagementError& e) {
			g_pLogger->warn("Error destroying tunnel: " + e.what());
		}
	}

	// Schedule next event
	scheduleEvent(new TunnelManagementEvent(TUN_MANAGE_PERIOD, this));
}


// Perform some audit
void Tunneler::audit(void)
{
	DPRINTV("Tunneler::audit()");

	std::stringstream str;

	// Currently active connections
	MutexVar acmv(m_connectionsMtx);
	int acCount = m_connections.size();
	str << "Active connection count: " << acCount;
	DPRINTV(str.str());
	str.str("");
	if(acCount > 0) {
		for(int i = 0; i < acCount; ++i) {
			str << "\tConnection id " << m_connections[i]->getID();
			DPRINTV(str.str());
			str.str("");
		}
	}
	acmv.unlock();

	// Closing connections
	MutexVar ccmv(m_closingConnectionsMtx);
	int ccCount = m_closingConnections.size();
	str << "Closing connection count: " << ccCount;
	DPRINTV(str.str());
	str.str("");
	if(ccCount > 0) {
		for(int i = 0; i < ccCount; ++i) {
			str << "\tConnection id " << m_closingConnections[i]->getID();
			DPRINTV(str.str());
			str.str("");
		}
	}
	ccmv.unlock();

	// Currently active tunnels
	MutexVar tmv(m_tunnelsMtx);
	int tCount = m_tunnels.size();
	str << "Active tunnels count: " << tCount;
	DPRINTV(str.str());
	str.str("");
	if(tCount > 0) {
		for(int i = 0; i < tCount; ++i) {
			str << "\tTunnel id " << m_tunnels[i]->getTidStr();
			DPRINTV(str.str());
			str.str("");
		}
	}
	tmv.unlock();

	// Closing tunnels
	MutexVar ctmv(m_closingTunnelsMtx);
	int ctCount = m_closingTunnels.size();
	str << "Closing tunnels count: " << ctCount;
	DPRINTV(str.str());
	str.str("");
	if(ctCount > 0) {
		for(int i = 0; i < ctCount; ++i) {
			str << "\tTunnel id " << m_closingTunnels[i]->getTidStr();
			DPRINTV(str.str());
			str.str("");
		}
	}
	ctmv.unlock();

	// ConnectionManager audit
	acmv.lock();
	ccmv.lock();
	m_connectionMgr->audit();
	ccmv.unlock();
	acmv.unlock();

	// Schedule next audit
	scheduleEvent(new AuditEvent(AUDIT_PERIOD, this));
}


/***********************/
/* Protected Interface */
/***********************/


/**************/
/* Predicates */
/**************/


// Find connections that can be shutdown and delete them
Connection* Tunneler::PredConnShutdown::operator()(Connection* conn) const
{
	Connection* res = conn;
	bool shutdown = false;

	try {
		// Ask connection
		shutdown = conn->canBeShutdown();
	}
	catch(CommsError& e) {
		// If we can't even check this connection, go ahead & kill it
		g_pLogger->warn("Error checking connection shutdown state: " + e.what());
		shutdown = true;
	}

	if(shutdown) {
		// Let connection manager know
		m_mgr->delConnection(conn);

		// Delete connection
		delete conn;
		res = 0;
	}

	return res;
}


/***********/
/* Methods */
/***********/


// Check whether we should relaunch a server test - Caller holds mutex
void Tunneler::checkScheduleServerTest(Tunnel* tun)
{
	// Nothing to do in base class
}


// Add a new managed connection
void Tunneler::addConnection(Connection* conn)
{
	// Maintain an ordered list
	std::vector<Connection*>::iterator pos;
	MutexVar cmv(m_connectionsMtx);
	pos = std::lower_bound(m_connections.begin(), m_connections.end(),
		conn, Connection::PredID());
	m_connections.insert(pos, conn);

	// Let connection manager know
	m_connectionMgr->addConnection(conn);
}


// Destroy a managed connection - Caller holds mutex
void Tunneler::destroyConnection(std::vector<Connection*>::iterator& cit)
{
	// Get connection
	Connection* conn = *cit;

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunneler::destroyConnection() - Destroying connection id ";
	dbg << conn->getID();
	DPRINTV(dbg.str());
	#endif	

	// Let connection manager know
	m_connectionMgr->delConnection(conn);

	// Remove from list
	m_connections.erase(cit);

	// Cancel any events that might still be pending for this connection
	cancelEvents<ConnectionErrorEvent, unsigned int>(conn->getID());
	cancelEvents<SendMsgEvent, unsigned int>(conn->getID());
	cancelEvents<FlushConnectionEvent, unsigned int>(conn->getID());

	// Now destroy connection itself
	delete conn;
}


// Make a managed connection a candidate for shutting down - Caller holds mutex
void Tunneler::connectionClosing(std::vector<Connection*>::iterator& cit)
{
	// Get connection
	Connection* conn = *cit;

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunneler::connectionClosing() - Moving connection id ";
	dbg << conn->getID() << " to closing list";
	DPRINTV(dbg.str());
	#endif	

	// Remove from list
	m_connections.erase(cit);

	// Cancel any events that might still be pending for this connection
	cancelEvents<ConnectionErrorEvent, unsigned int>(conn->getID());
	cancelEvents<SendMsgEvent, unsigned int>(conn->getID());
	cancelEvents<FlushConnectionEvent, unsigned int>(conn->getID());

	// Let connection know it is now inactive
	conn->setInactive();

	// Place connection on closing list
	MutexVar ccmv(m_closingConnectionsMtx);
	m_closingConnections.push_back(conn);
}


// Locate a managed connection with some ID - Caller holds Mutex
std::vector<Connection*>::iterator Tunneler::getConnection(unsigned int id)
	throw(LookupManagementError)
{
	// Find connection in managed list
	std::vector<Connection*>::iterator first = m_connections.begin();
	std::vector<Connection*>::iterator cit;

	int top = m_connections.size() - 1;
	int bottom = 0;
	bool found = false;

	while(!found && top >= bottom) {
		int mid = (top + bottom) / 2;
		cit = first + mid;
		unsigned int connID = (*cit)->getID();
		if(id == connID) {
			found = true;
		}
		else if(id < connID) {
			top = mid - 1;
		}
		else {
			bottom = mid + 1;
		}
	}

	// Found it?
	if(!found) {
		std::stringstream err;
		err << "Unknown connection " << id;
		throw LookupManagementError(err.str());
	}

	return cit;
}


// Destroy a managed tunnel - Caller holds tunnels mutex
void Tunneler::destroyTunnel(Tunnel* tun) throw(LookupManagementError)
{
	if(tun) {
		// Get tunnel ID
		TunnelID tid = tun->getTid();
		DPRINTV("Destroying tunnel " + tid.toString());

		// Find tunnel in managed list
		std::vector<Tunnel*>::iterator tit;
		tit = std::find(m_tunnels.begin(), m_tunnels.end(), tun);
		if(tit != m_tunnels.end()) {
			// Remove from list
			m_tunnels.erase(tit);
		}
		else {
			// Not found
			std::stringstream err;
			err << "Tunneler::destroyTunnel() - Unknown tunnel " << tid;
			throw LookupManagementError(err.str());
		}

		// Cancel any events that might still be pending for this tunnel
		cancelEvents<PollTimeoutEvent, TunnelID>(tid);
		cancelEvents<TunnelStateEvent, TunnelID>(tid);
		cancelEvents<ShutdownTunnelEvent, TunnelID>(tid);
		cancelEvents<TunnelUnviableEvent, TunnelID>(tid);
		cancelEvents<UDPSilentEvent, TunnelID>(tid);
		cancelEvents<AuthSuccessEvent, TunnelID>(tid);

		// Remove tunnel ID from closing list if it's there
		std::vector<Tunnel*>::iterator ctit;
		MutexVar ctmv(m_closingTunnelsMtx);
		ctit = std::find(m_closingTunnels.begin(), m_closingTunnels.end(), tun);
		if(ctit != m_closingTunnels.end()) {
			m_closingTunnels.erase(ctit);
		}

		// Now destroy tunnel itself
		delete tun;

		// Schedule notification
		scheduleEvent(new TunnelDestroyedEvent(this, tid));
	}
	else {
		throw LookupManagementError("Tunneler::destroyTunnel() - Null tunnel");
	}
}


// Tag a tunnel as closing - Caller holds tunnels mutex
void Tunneler::tunnelClosing(Tunnel* tun) throw(LookupManagementError)
{
	if(tun) {
		DPRINTV("Tunneler::tunnelClosing() - " + tun->getTidStr());

		// Find tunnel in managed list
		std::vector<Tunnel*>::iterator tit;
		tit = std::find(m_tunnels.begin(), m_tunnels.end(), tun);
		if(tit != m_tunnels.end()) {
			// Check we don't already know this tunnel is closing
			std::vector<Tunnel*>::iterator ctit;
			MutexVar ctmv(m_closingTunnelsMtx);
			ctit = std::find(m_closingTunnels.begin(), m_closingTunnels.end(), tun);
			if(ctit == m_closingTunnels.end()) {
				// Copy tunnel to closing list
				m_closingTunnels.push_back(tun);

				// Tell tunnel it is closing
				tun->closing();
			}
		}
		else {
			// Not found
			std::stringstream err;
			err << "Tunneler::tunnelClosing() - Unknown tunnel " << tun->getTidStr();
			throw LookupManagementError(err.str());
		}
	}
	else {
		throw LookupManagementError("Tunneler::tunnelClosing() - Null tunnel");
	}
}


}	// namespace

