/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Tunneler.hh,v 1.7 2009-04-24 15:37:33 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _TUNNELER_HH_
#define _TUNNELER_HH_

#include <vector>
#include "EventReceiver.hh"
#include "TunnelID.hh"
#include "Mutex.hh"
#include "Exceptions.hh"
#include "Tunnel.hh"
#include "Connection.hh"
#include "TunnelConfig.hh"
#include "Time.hh"


namespace HTTPTunnel {

class Server;
class Payload;
class RawBuff;
class HTTPMsg;
class ConnectResultEvent;
class ConnectionErrorEvent;
class ConnectionManager;
class ClientSideConfig;
class ServerSideConfig;
class Sock;
class TunnelStats;
class TunnelerID;


class Tunneler : public EventReceiver {
public:
	// Construction/destruction
	Tunneler(Server* server) throw(TunnelError);
	virtual ~Tunneler();


	/*************/
	/* Accessors */
	/*************/

	// Get owning server
	const Server* getServer(void) const;
	Server* getServerMutable(void);

	// Get owning server configuration
	const ClientSideConfig* getClientSideConfig(void) const throw(ConfigError);
	const ServerSideConfig* getServerSideConfig(void) const throw(ConfigError);
	ClientSideConfig* getClientSideConfigMutable(void) throw(ConfigError);
	ServerSideConfig* getServerSideConfigMutable(void) throw(ConfigError);

	// Get number of active managed tunnels
	unsigned int getNoActiveTunnels(void);

	// Get IDs of all active managed tunnels
	void getActiveTunnelIDs(std::vector<TunnelID>& ids);

	// Get connection counts for some tunnel
	unsigned int getTunnelConnectionCount(const TunnelID& tid,
		std::vector<Tunnel::Selector*>& crit);

	// Get stats for some tunnel
	TunnelStats* getTunnelStats(const TunnelID& tid);


	/************/
	/* Mutators */
	/************/

	// Initiate a shutdown of the named tunnels
	void initiateTunnelShutdown(const TunnelID& tid);
	void initiateAllTunnelsShutdown(void);

	// Enable/disable testing based on current config setting
	virtual void setServerTesting(void);

	// Add connection error to list of scheduled events
	void scheduleConnectionError(unsigned int connID, Tunnel::TunnelErrorType type);


	/**********/
	/* Events */
	/**********/

	// Notification of a new socket connection
	virtual void newConnectionArrived(Sock* insock) throw(CommsError) = 0;

	// Notification that a tunnel has insufficient left/right connections to be viable
	virtual void unviableTunnel(const TunnelID& tid) = 0;

	// Notification of new tunneled/raw data arrival
	virtual void rawDataArrived(unsigned int connID, const TunnelID& tid,
		const RawBuff& raw) throw(LookupManagementError) = 0;

	// Check we have received an ack for some sent message
	virtual void checkAckReceived(void) = 0;


	// Notification of new HTTP message arrival
	virtual void messageArrived(unsigned int connID, HTTPMsg* msg)
		throw(LookupManagementError);

	// Notification of some new default tunnel parameters
	virtual void newTunnelParms(TunnelParms* parms, const TunnelerID& tid);

	// Notification of a new UDP 'connection'
	virtual void newUDPConnectionArrived(Sock* insock) throw(CommsError);

	// Notification that a tunnel needs to be cleared down
	virtual void shutdownTunnel(const TunnelID& tid, bool otherEndInitiated);

	// Notification that a tunnel has been shutdown (destroyed)
	virtual void tunnelDestroyed(const TunnelID& tid);

	// Notification of a poll timeout on a tunnel
	virtual void pollTimeout(const TunnelID& tid);

	// Check UDP datagram timeout
	virtual void checkDatagramTimeout(const TunnelID& tid);

	// Notification that connection has changed state
	virtual void connectionStateChange(unsigned int connID,
		Connection::ConnectionState state);

	// Notification that a tunnel has changed state
	virtual void tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state);

	// Notification of a socket connect result
	virtual void connectResult(ConnectResultEvent* result);

	// Notification of error(s) on a socket connection
	virtual void connectionErrors(unsigned int connID,
		std::vector<Tunnel::TunnelErrorType>& errors);

	// Notification that a connection needs to be cleared down
	virtual void shutdownConnection(unsigned int connID);

	// Notification we should force a send of a message down a tunnel/connection
	virtual void sendTunnelMsg(const TunnelID& tid, unsigned int connID, bool retry);

	// Notification we should try to flush a connection
	virtual void flushConnection(unsigned int connID, bool retry);

	// Notification to decrease tunnels leaky error counts
	virtual void decreaseTunnelsErrorCount(void);

	// Notification that authentication has succeeded for a tunnel
	virtual void authenticationSuccess(const TunnelID& tid);

	// Notification that we should perform connection manangement
	virtual void performConnectionManagement(void);

	// Notification that we should perform tunnel manangement
	virtual void performTunnelManagement(void);

	// Perform some audit
	virtual void audit(void);


protected:

	/*************/
	/* Constants */
	/*************/

	// Time between test reconnection attempts
	static const unsigned int			MIN_TEST_RECONNECT_DELAY	= 2;
	static const unsigned int			MAX_TEST_RECONNECT_DELAY	= 10;

	// Time between acknowledgement checks
	static const unsigned int			ACK_CHECK_PERIOD				= 1;

	// Time between management events
	static const unsigned int			CONN_MANAGE_PERIOD			= 500;
	static const unsigned int			TUN_MANAGE_PERIOD				= 2;

	// Period in seconds between leaky bucket error count decreases
	static const unsigned int			LEAK_ERROR_PERIOD				= 2;

	// Time between audits
	static const unsigned int			AUDIT_PERIOD					= 60 * 10;

	// Control when we clear out old tunnel parms
	#ifdef DEBUG_TUN
	static const int						OLD_TUN_PARMS_AGE				= 60 * 1;
	#else
	static const int						OLD_TUN_PARMS_AGE				= 60 * 60;
	#endif
	static const int						CHECK_TUN_PARMS_PERIOD		= OLD_TUN_PARMS_AGE / 10;


	/***********/
	/* Methods */
	/***********/

	// Get reconnection attempt parameters
	virtual void getReconnectParms(unsigned int& maxConnectAttempts,
		unsigned int& reconnectDelay, unsigned int failedAttempts) = 0;

	// Check whether we should relaunch a server test
	virtual void checkScheduleServerTest(Tunnel* tun);

	// Connection management
	void addConnection(Connection* conn);
	void destroyConnection(std::vector<Connection*>::iterator& cit);
	void connectionClosing(std::vector<Connection*>::iterator& cit);
	std::vector<Connection*>::iterator getConnection(unsigned int id)
		throw(LookupManagementError);

	// Tunnel management
	void destroyTunnel(Tunnel* tun) throw(LookupManagementError);
	void tunnelClosing(Tunnel* tun) throw(LookupManagementError);

	// Locate a tunnel matching some predicate
	template<class P> Tunnel* getTunFromPred(P pred);
	template<class P> void getAllTunsFromPred(P pred, std::vector<Tunnel*>& res);


	/**************/
	/* Predicates */
	/**************/

	// Match tunnel from an ID
	struct PredTunID {
		PredTunID(const TunnelID* tid) : m_tid(tid) {}
		bool operator()(const Tunnel* tun) const {
			return (tun->getTid() == *m_tid);
		}
		const TunnelID* m_tid;
	};

	// Match tunnel from a connection
	struct PredTunConnection {
		PredTunConnection(Connection* conn) : m_conn(conn) {}
		bool operator()(Tunnel* tun) const {
			return (tun->ownsConnection(m_conn));
		}
		Connection* m_conn;
	};

	// Match test tunnels
	struct PredTestTun {
		bool operator()(const Tunnel* tun) const {
			return (tun->isTestTunnel());
		}
	};

	// Match test tunnel from a tunneler ID
	struct PredTestTunID {
		PredTestTunID(const TunnelerID* tid) : m_tid(tid) {}
		bool operator()(const Tunnel* tun) const {
			return (tun->isTestTunnel() && tun->getTid().getTunnelerID() == *m_tid);
		}
		const TunnelerID* m_tid;
	};

	// Match connections that can be shutdown
	struct PredConnShutdown {
		PredConnShutdown(ConnectionManager* mgr) : m_mgr(mgr) {}
		Connection* operator()(Connection* conn) const;
		ConnectionManager* m_mgr;
	};

	// Match null pointers
	template<class T> struct PredPtrNull {
		bool operator()(T* ptr) const {
			return (ptr == 0);
		}
	};


	/********/
	/* Data */
	/********/

	// Our owning server
	Server*											m_server;

	// All the currently active tunnels
	std::vector<Tunnel*>							m_tunnels;
	Mutex												m_tunnelsMtx;

	// Tunnels we are trying to shutdown
	std::vector<Tunnel*>							m_closingTunnels;
	Mutex												m_closingTunnelsMtx;

	// All active connections
	std::vector<Connection*>					m_connections;
	Mutex												m_connectionsMtx;

	// Connections we are trying to shutdown
	std::vector<Connection*>					m_closingConnections;
	Mutex												m_closingConnectionsMtx;

	// A connection manager
	ConnectionManager*							m_connectionMgr;

	// Time of last old tunnel parms check
	Time												m_lastTunParmsCheckTime;


private:


};


}	// namespace


// Include template methods
#include "TunnelerT.cc"


#endif

