/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Uniquely identify an individual tunneled connection.
 *
 * $Id: TunnelerID.cc,v 1.3 2008-06-26 09:47:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cstdlib>
#include <string>
#include <sstream>
#include <iostream>
#include "TunnelerID.hh"
#include "SideConfig.hh"
#include "GlobalConfig.hh"
#include "Utils.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


// Global references
extern GlobalConfig* g_pConfig;


// Initialise static data
unsigned int TunnelerID::c_idCount = 65535;
Mutex TunnelerID::c_idCountMtx;


// Construction/destruction
TunnelerID::TunnelerID()
: m_cliName(), m_cliIP(), m_cliPort(0), m_protocol(Enums::TCP)
{
}

TunnelerID::TunnelerID(SideConfig* cfg)
: m_cliName(g_pConfig->getUniqueHostName()), m_cliIP(cfg->getListenIP()),
  m_cliPort(cfg->getListenPort()), m_protocol(cfg->getProtocol())
{
	// Are we for an ephemeral tunneler?
	if(m_cliIP == "-") {
		// Ensure uniquness by replacing port with unique ID
		m_cliPort = getNextID();
	}
}

TunnelerID::TunnelerID(std::string id) throw(InvalidTunnelerID)
: m_cliName(), m_cliIP(), m_cliPort(0), m_protocol(Enums::TCP)
{
	// Split composite ID
	std::vector<std::string> parts;
	Utils::SplitString(id, ":", parts);

	// Got everything we need?
	if(parts.size() != 6) {
		std::stringstream err;
		err << "Invalid tunneler ID string: " << id;
		throw InvalidTunnelerID(err.str());
	}

	// Construct individual ID values
	m_cliName = parts[2];
	m_cliIP = parts[3];
	m_cliPort = atoi(parts[4].c_str());
	m_protocol = (Enums::Protocol)(atoi(parts[5].c_str()));
}

TunnelerID::TunnelerID(const TunnelerID& that)
: m_cliName(that.m_cliName), m_cliIP(that.m_cliIP), m_cliPort(that.m_cliPort),
  m_protocol(that.m_protocol)
{
}

TunnelerID::~TunnelerID()
{
}


// Assignment
TunnelerID& TunnelerID::operator=(const TunnelerID& that)
{
	// Self assignment?
	if(this != &that) {
		m_cliName = that.m_cliName;
		m_cliIP = that.m_cliIP;
		m_cliPort = that.m_cliPort;
		m_protocol = that.m_protocol;
	}

	return *this;
}


// Operators
bool TunnelerID::operator==(const TunnelerID& that) const
{
	// Everything must match
	bool res = (that.m_cliName == m_cliName);
	res &= (that.m_cliIP == m_cliIP);
	res &= (that.m_cliPort == m_cliPort);
	res &= (that.m_protocol == m_protocol);
	return res;
}

bool TunnelerID::operator<(const TunnelerID& that) const
{
	// Just create a string of all members and compare that
	return (toString() < that.toString());
}


// Accessors
std::string TunnelerID::getClientName(void) const
{
	return m_cliName;
}

std::string TunnelerID::getClientIP(void) const
{
	return m_cliIP;
}

unsigned int TunnelerID::getClientPort(void) const
{
	return m_cliPort;
}

Enums::Protocol TunnelerID::getProtocol(void) const
{
	return m_protocol;
}


// Create a representation of ourselves as a string
std::string TunnelerID::toString(void) const
{
	std::stringstream val;
	val << m_cliName << ":" << m_cliIP << ":" << m_cliPort << ":";
	val << (unsigned int)(m_protocol);
	return val.str();
}


/*********************/
/* Private Interface */
/*********************/


// Get next unique ID
unsigned int TunnelerID::getNextID(void)
{
	MutexVar mv(c_idCountMtx);
	if(++c_idCount == 0)
		c_idCount = 65536;
	return c_idCount;
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const TunnelerID& tid)
{
	str << tid.toString();
	return str;
}


}	// namespace

