/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a UDP socket.
 *
 * $Id: UDPSock.cc,v 1.7 2009-03-20 11:52:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <arpa/inet.h>
#include <netdb.h>
#include <cerrno>
#include <cstring>
#include <strings.h>
#include "UDPSock.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "Tunneler.hh"
#include "Utils.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
UDPSock::UDPSock(bool blocking) throw(SockError)
: Sock(blocking), m_sockAddr(), m_sockAddrLen(sizeof(m_sockAddr)), m_ownSocket(true)
{
	// Create new UDP socket
	int sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if(sockfd < 0) {
		throw SockError("Failed to create new UDP socket descriptor", errno);
	}
	else {
		m_fd = sockfd;
	}

	// Set blocking mode
	blocknonblock();
}

UDPSock::UDPSock(int fd, std::string cliAddr) throw(SockError)
: Sock(fd, cliAddr), m_sockAddr(), m_sockAddrLen(sizeof(m_sockAddr)),
  m_ownSocket(false)
{
}

UDPSock::~UDPSock()
{
	DPRINTV("UDPSock::~UDPSock()");

	MutexVar mv(m_mtx);

	// Should we close?
	if(m_ownSocket) {
		try {
			// Close socket descriptor
			close();
		}
		catch(SockError& e) {
			g_pLogger->info(e.what());
		}
	}
}


/********************/
/* Public Interface */
/********************/


// Wait for a new incoming UDP connection
void UDPSock::newConnection(Tunneler* tunneler) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Is this really a new UDP connection, or simply a continuation?
	if(tunneler->getNoActiveTunnels() == 0) {
		// Accept connection
		Sock* connectedSock = accept();

		try {
			// Tell tunneler
			if(tunneler)
				tunneler->newUDPConnectionArrived(connectedSock);
			else
				throw SockError("No tunneler to handle new UDP connection", 0);
		}
		catch(TunnelError& e) {
			g_pLogger->info(e.what());
			delete connectedSock;
		}
	}
}


// Set socket listening
void UDPSock::listen(void) throw(SockError)
{
	// Do nothing for UDP
}


// Wait for a new incoming connection
Sock* UDPSock::accept(void) throw(SockError)
{
	// Return connected UDP socket
	return new UDPSock(m_fd, "Unknown");
}


// Establish connection out
void UDPSock::connect(std::string addr, unsigned int port)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Lookup destination address
	struct hostent hostAddr;
	char auxHostBuff[Utils::AUX_HOST_BUFF_SIZE];
	if(Utils::getHostByName(addr, &hostAddr, auxHostBuff, sizeof(auxHostBuff)) != 0) {
		throw ConnectError("Failed to resolve UDP server address " + addr);
	}

	// Setup socket address structure
	bzero(&m_sockAddr, sizeof(m_sockAddr));
	m_sockAddr.sin_family = AF_INET;
	m_sockAddr.sin_port = htons(port);
	memcpy(&m_sockAddr.sin_addr, hostAddr.h_addr_list[0],
		sizeof(m_sockAddr.sin_addr));
}


// Read from socket
int UDPSock::read(unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Read data from socket
again:
	ssize_t noread = recvfrom(m_fd, buff, buffSize, 0,
		(struct sockaddr*)&m_sockAddr, &m_sockAddrLen);
	if(noread < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Read error from UDP socket", errno);
	}

	// Get connected client address
	m_connectedClientAddr = inet_ntoa(m_sockAddr.sin_addr);

	return noread;
}


// Write to socket
int UDPSock::write(const unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Write data to socket
again:
	ssize_t nowritten = sendto(m_fd, buff, buffSize, 0,
		(struct sockaddr*)&m_sockAddr, sizeof(m_sockAddr));
	if(nowritten < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Write error to UDP socket", errno);
	}

	return nowritten;
}


// Get the result of a non-blocking connect attempt
void UDPSock::getConnectResult(bool& resultAvailable, bool& connected)
	throw(SockError)
{
	// Do nothing for UDP
	resultAvailable = connected = true;
}


}	// namespace

