/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * A server to accept incoming UDP socket tunnel connections.
 *
 * $Id: UDPSockTunServer.cc,v 1.3 2009-03-20 11:52:26 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include <errno.h>
#include "UDPSockTunServer.hh"
#include "Debug.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "UDPSock.hh"
#include "Time.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
UDPSockTunServer::UDPSockTunServer(const SockTunCallback* userCallback) throw(TunnelError)
: SockTunServer(userCallback), m_threadMgr()
{
	DPRINTV("UDPSockTunServer::UDPSockTunServer()");

	// Create blocking UDP socket
	m_prot = Enums::UDP;
	m_sock = new UDPSock(true);

	// Bind to ephemeral port on the loopback interface
	m_sock->bind(m_listenIP, m_listenPort);

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "UDPSockTunServer bound to " << m_listenIP << ":" << m_listenPort;
	DPRINTV(dbg.str());
	#endif

	// Launch server to process UDP connections
	MutexVar mv(m_mtx);
	ZeroArgClassCallback<UDPSockTunServer> cb(&UDPSockTunServer::startServer, this);
	m_thread = m_threadMgr.newThread(cb);
}

UDPSockTunServer::~UDPSockTunServer()
{
	DPRINTV("UDPSockTunServer::~UDPSockTunServer()");

	// Take mutex
	MutexVar mv(m_mtx);

	// Signal we are no longer running
	m_shutDown = true;
}


/*********************/
/* Private Interface */
/*********************/


// Wait for arriving connections - Thread entry point
void UDPSockTunServer::startServer(void)
{
	DPRINTI("UDPSockTunServer::startServer() - Waiting for connection");

	MutexVar mv(m_mtx);

	bool accepted = false;

	while(!m_shutDown && !accepted) {
		// Poll socket until we think it's worth trying to accept()
		try {
			bool pollError, pollHup;
			if(m_sock->pollRead(0, pollError, pollHup)) {
				// Accept pending connection
				Sock* connectedSock = m_sock->accept();

				// Pass connected socket back to user
				callbackToUser(connectedSock);
				accepted = true;
			}
			else {
				// Poll timeout, wait a while before retrying
				Time pollTime(0, POLL_TIMEOUT);
				m_thread->sleep(mv, pollTime.getTimeVal());
			}
		}
		catch(SockError& e) {
			if(e.m_errno == EINTR) {
				// Interrupted system call
				continue;
			}
			else {
				// General error, just carry on
				continue;
			}
		}
	}

	DPRINTI("UDPSockTunServer closing down");
}


}	// namespace

