/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Class to encode/decode an URL.
 *
 * $Id: URLCode.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <cstring>
#include "URLCode.hh"


// Initialise static data
const char* HTTPTunnel::URLCode::c_unsafeChars = " \"<>%\\^[]`+$,@:;/!#?=&{}|";
const char HTTPTunnel::URLCode::c_hexVals[] = {
	'0','1','2','3','4','5','6','7','8', '9','A','B','C','D','E','F'
};


namespace HTTPTunnel {


// Construction/destruction
URLCode::URLCode()
{
}

URLCode::~URLCode()
{
}


/********************/
/* Public Interface */
/********************/


// Encode an URL
std::string URLCode::encode(const std::string& orig) const
{
	std::stringstream encoded;

	// Iterator over original string
	for(unsigned int pos = 0; pos < orig.size(); pos++) {
		char ch = orig[pos];
		if(!unsafe(ch)) {
			// Safe Character
			encoded << ch;
		}
		else {
			// get Hex Value of the Character
			escape(ch, encoded);
		}
	}

	return encoded.str();
}


// Decode an URL
std::string URLCode::decode(const std::string& orig) const
{
	std::stringstream decoded;

	// Iterator over original string
	for(unsigned int pos = 0; pos < orig.size(); pos++) {
		char ch = orig[pos];
		if(ch == '%') {
			// Get full escape code
			std::string esc(orig, pos + 1, 2);

			// Decode escaped character
			unescape(esc, decoded);

			// Skip remainder of escape code
			pos += 2;
		}
		else {
			decoded << ch;
		}
	}

	return decoded.str();
}


/*********************/
/* Private Interface */
/*********************/


// Is a character unsafe & in need of encoding?
bool URLCode::unsafe(char ch) const
{
	if(ch < 32 || ch >= 127)
		return true;

	return (strchr(c_unsafeChars, ch) != 0);
}


// Escape an unsafe character
void URLCode::escape(char ch, std::stringstream& str) const
{
	str << "%" << c_hexVals[ch / 16] << c_hexVals[ch % 16];
}

// Unescape an encoded char
void URLCode::unescape(std::string& esc, std::stringstream& str) const
{
	// Work out ascii value of char
	char ascVal = 0;
	for(unsigned int i = 0; i < esc.size(); ++i) {
		char ch = esc[i];
		int val = 0;
		if(ch >= 'A' && ch <= 'F') {
			val = ch - 'A' + 10;
		}
		else {
			val = ch - '0';
		}

		int mult = (esc.size() - i - 1);
		if(mult > 0)
			val *= mult * 16;
		ascVal += val;
	}

	// Write out final char
	str << ascVal;
}


}	// namespace

