/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a user of a tunnel.
 *
 * $Id: User.cc,v 1.2 2008-06-26 09:47:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <algorithm>
#include "User.hh"
#include "Debug.hh"
#include "Tunnel.hh"
#include "TunnelConfig.hh"
#include "Tunneler.hh"
#include "ShutdownTunnelEvent.hh"


namespace HTTPTunnel {


// Construction/destruction
User::User()
: m_alwaysAuthenticated(true), m_name(), m_passwd(), m_quotas(), m_sentData(),
  m_recvData(), m_activeTunnels(), m_authenticatedTunnels(), m_keepCached(false)
{
}

User::User(bool auth, std::string name, std::string passwd, const Quotas& quotas,
	bool keepCached)
: m_alwaysAuthenticated(auth), m_name(name), m_passwd(passwd), m_quotas(quotas),
  m_sentData(), m_recvData(), m_activeTunnels(), m_authenticatedTunnels(),
  m_keepCached(keepCached)
{
}

User::User(const User& that)
: m_alwaysAuthenticated(that.m_alwaysAuthenticated), m_name(that.m_name),
  m_passwd(that.m_passwd), m_quotas(that.m_quotas), m_sentData(that.m_sentData),
  m_recvData(that.m_recvData), m_activeTunnels(that.m_activeTunnels),
  m_authenticatedTunnels(that.m_authenticatedTunnels), m_keepCached(that.m_keepCached)
{
}


User::~User()
{
}


// Assignment
User& User::operator=(const User& that)
{
	// Self assignment?
	if(this != &that) {
		m_alwaysAuthenticated = that.m_alwaysAuthenticated;
		m_name = that.m_name;
		m_passwd = that.m_passwd;
		m_quotas = that.m_quotas;
		m_sentData = that.m_sentData;
		m_recvData = that.m_recvData;
		m_activeTunnels = that.m_activeTunnels;
		m_authenticatedTunnels = that.m_authenticatedTunnels;
		m_keepCached = that.m_keepCached;
	}

	return *this;
}

// Equality
bool User::operator==(const User& that)
{
	return (m_alwaysAuthenticated == that.m_alwaysAuthenticated &&
		m_name == that.m_name && m_passwd == that.m_passwd &&
		m_quotas == that.m_quotas && m_keepCached == that.m_keepCached);
}


/*************/
/* Accessors */
/*************/


// Is this user authenticated?
bool User::getAuthenticated(Tunnel* tun) const
{
	bool auth = false;

	// Allways authenticated?
	if(m_alwaysAuthenticated) {
		auth = true;
	}
	else {
		// For a specific tunnel?
		if(tun) {
			// Find tunnel index
			for(unsigned int tit = 0; tit < m_activeTunnels.size(); ++tit) {
				if(m_activeTunnels[tit] == tun) {
					// Cross reference to authenticated tunnels
					auth = m_authenticatedTunnels[tit];
				}
			}
		}
		else {
			// Non-specific tunnel, so are we authenticated for all tunnels?
			auth = true;
			for(unsigned int tit = 0; tit < m_activeTunnels.size(); ++tit) {
				auth &= m_authenticatedTunnels[tit];
			}
		}
	}

	return auth;
}

// Get users login name
std::string User::getName(void) const
{
	return m_name;
}

// Get users passwd
std::string User::getPasswd(void) const
{
	return m_passwd;
}

// Get users quotas
const Quotas& User::getQuotas(void) const
{
	return m_quotas;
}

// Get number of active tunnels for this user
unsigned int User::getNoActiveTunnels(void) const
{
	return m_activeTunnels.size();
}

// Get whether this user is using the specified tunnel
bool User::usingTunnel(Tunnel* tun) const
{
	std::vector<Tunnel*>::const_iterator tit;
	tit = std::find(m_activeTunnels.begin(), m_activeTunnels.end(), tun);
	return (tit != m_activeTunnels.end());
}

// Get whether this user should remain cached
bool User::keepCached(void) const
{
	return (m_keepCached || getNoActiveTunnels() > 0);
}

// Get current average upload bandwidth
unsigned int User::getAvgUpBandwidth(void)
{
	// Get current average bandwidth usage
	unsigned int avgBandwidth = 0;
	unsigned int totBandwidth = 0;
	unsigned int avgMillis = 0;
	getAvgBandwidth(m_sentData, avgBandwidth, totBandwidth, avgMillis);
	return avgBandwidth;
}

// Get current average download bandwidth
unsigned int User::getAvgDownBandwidth(void)
{
	// Get current average bandwidth usage
	unsigned int avgBandwidth = 0;
	unsigned int totBandwidth = 0;
	unsigned int avgMillis = 0;
	getAvgBandwidth(m_recvData, avgBandwidth, totBandwidth, avgMillis);
	return avgBandwidth;
}


/************/
/* Mutators */
/************/


// Set whether this user is authenticated
void User::setAuthenticated(bool auth, Tunnel* tun)
{
	// For a specific tunnel?
	if(tun) {
		// If we're not auth'd for this tunnel we're not always auth'd
		if(!auth)
			m_alwaysAuthenticated = false;

		// Find tunnel index
		for(unsigned int tit = 0; tit < m_activeTunnels.size(); ++tit) {
			if(m_activeTunnels[tit] == tun) {
				// Cross reference to authenticated tunnels
				m_authenticatedTunnels[tit] = auth;
			}
		}
	}
	else {
		// Non-specific tunnel, so set always auth'd
		m_alwaysAuthenticated = auth;
	}
}

// Set users quotas
void User::setQuotas(const Quotas& quotas)
{
	m_quotas = quotas;
}


// Update all our pertinant details
void User::setAll(std::string name, std::string passwd, bool auth,
	const Quotas& quotas, Tunnel* newTun)
{
	m_name = name;
	m_passwd = passwd;
	m_quotas = quotas;
	addTunnel(newTun, auth);
}


// Add a new tunnel to list of those carried
void User::addTunnel(Tunnel* tun, bool auth)
{
	// Check we don't already have this tunnel
	std::vector<Tunnel*>::iterator tit;
	tit = std::find(m_activeTunnels.begin(), m_activeTunnels.end(), tun);
	if(tit == m_activeTunnels.end()) {
		DPRINTV("User::addTunnel() - User " + m_name + " owns tunnel " + tun->getTidStr());
		m_activeTunnels.push_back(tun);
		m_authenticatedTunnels.push_back(auth);
	}
	else {
		DPRINTV("User::addTunnel() - User " + m_name + " already knows about tunnel " +
			tun->getTidStr());
	}

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "User::addTunnel() - User " << m_name << " now owns ";
	nfo << getNoActiveTunnels() << " active tunnels";
	DPRINTI(nfo.str());
	#endif
}

// Remove a tunnel from list of those carried
void User::removeTunnel(Tunnel* tun)
{
	// Check we know about this tunnel
	bool foundIt = false;
	unsigned int tindex = 0;
	for(tindex = 0; tindex < m_activeTunnels.size(); ++tindex) {
		if(m_activeTunnels[tindex] == tun) {
			foundIt = true;
			break;
		}
	}

	if(foundIt) {
		// Remove tunnel details
		std::vector<Tunnel*>::iterator tit = m_activeTunnels.begin() + tindex;
		std::vector<bool>::iterator ait = m_authenticatedTunnels.begin() + tindex;
		m_activeTunnels.erase(tit);
		m_authenticatedTunnels.erase(ait);

		DPRINTV("User::removeTunnel() - User " + m_name + " no longer owns tunnel " +
			tun->getTidStr());
	}
	else {
		DPRINTV("User::removeTunnel() - Unknown tunnel " + tun->getTidStr());
	}

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "User::removeTunnel() - User " << m_name << " now owns ";
	nfo << getNoActiveTunnels() << " active tunnels";
	DPRINTI(nfo.str());
	#endif
}


/********************/
/* Public Interface */
/********************/


// Get the maximum no. of bytes our quotas will let us send
unsigned int User::getBytesCanSend(void)
{
	return getBytesBelowQuota(m_sentData);
}

// We have just sent some no. of bytes
void User::sentBytes(unsigned int bytes)
{
	if(bytes > 0) {
		// Record new bytes sent
		trackBandwidth(bytes, m_sentData);
	}
}


// Do our quotas allow receiving some no. of bytes?
bool User::canRecvBytes(unsigned int bytes)
{
	bool res = true;

	if(bytes > 0) {
		// Record new bytes received
		trackBandwidth(bytes, m_recvData);

		// Does average received bandwidth exceed quota?
		res = bandwidthWithinQuota(m_recvData);
	}

	return res;
}


// Tunnel creation according to our quotas
bool User::canCreateNewTunnel(void)
{
	bool res = true;

	// Get tunnel count limit (if set) for this user
	unsigned int maxTunnels = m_quotas.getNoTunnels();
	if(maxTunnels > 0) {
		// User below limit?
		res = (getNoActiveTunnels() < maxTunnels);
	}

	return res;
}


// Force a shutdown for all our tunnels
void User::closeTunnels(void)
{
	DPRINTV("User::closeTunnels() - User " + m_name);

	std::vector<Tunnel*>::iterator tit;
	for(tit = m_activeTunnels.begin(); tit != m_activeTunnels.end(); ++tit) {
		Tunnel* tun = *tit;
		TunnelID tid = *(tun->getConfig()->getTunnelID());
		DPRINTV("Scheduling close for tunnel " + tid.toString());

		// Schedule a tunnel shutdown
		Tunneler* tunneler = tun->getTunneler();
		tunneler->scheduleEvent(new ShutdownTunnelEvent(0, tunneler, tid, false));
	}
}


/*********************/
/* Private Interface */
/*********************/


// Update bandwidth usage tracking
void User::trackBandwidth(unsigned int bytes, std::vector<DataTrack>& track)
{
	// Drop any old data points
	trimBandwidthTrack(track);

	// Append new track point
	DataTrack tp(Time::now(), bytes);
	track.push_back(tp);
}

// Discard any tracked bandwidth outside our range
void User::trimBandwidthTrack(std::vector<DataTrack>& track)
{
	if(track.size() > 0) {
		// Get cutoff time
		Time cutoff = Time::now();
		cutoff.subInterval(DATA_TRACK_PERIOD, 0);

		// Find any tracked items outside cutoff
		bool keepLooking = true;
		while(keepLooking && track.size() > 0) {
			std::vector<DataTrack>::iterator tit = track.begin();
			if(tit->m_tm < cutoff) {
				track.erase(tit);
			}
			else {
				keepLooking = false;
			}
		}
	}
}


// Get average bandwidth used
void User::getAvgBandwidth(std::vector<DataTrack>& track,
	unsigned int& avg, unsigned int& tot, unsigned int& millis)
{
	avg = tot = millis = 0;

	// Trim tracked data
	trimBandwidthTrack(track);

	// Calculate average
	if(track.size() > 0) {
		// Get time over which to calculate average
		Time now = Time::now();
		Time start = track[0].m_tm;
		millis = (now.getTotalMillis() - start.getTotalMillis());

		// Do we have long enough sample time for estimate?
		if(millis > 0) {
			// Get total bandwidth
			std::vector<DataTrack>::iterator bit;
			for(bit = track.begin(); bit != track.end(); ++bit) {
				tot += bit->m_bytes;
			}

			// Average bandwidth in bytes/s
			double avgd = ((double)tot * 1000) / (double)millis;
			avg = (unsigned int)avgd;
		}
	}
}


// Is the users bandwidth usage within their quota?
bool User::bandwidthWithinQuota(std::vector<DataTrack>& track)
{
	bool res = true;

	// Convert quota bandwidth to bytes/sec
	unsigned int quotaBandwidth = m_quotas.getBandwidth() * 1024;
	if(quotaBandwidth > 0) {
		// Get current average bandwidth usage
		unsigned int avgBandwidth = 0;
		unsigned int totBandwidth = 0;
		unsigned int avgMillis = 0;
		getAvgBandwidth(track, avgBandwidth, totBandwidth, avgMillis);

		// Is usage within tolerated range?
		unsigned int maxBandwidth = quotaBandwidth;
		maxBandwidth += (quotaBandwidth * QUOTA_TOLERENCE) / 100;
		res = (avgBandwidth <= maxBandwidth);

		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "bandwidthWithinQuota: Max = " << maxBandwidth << ", ";
		dbg << "Avg = " << avgBandwidth;
		DPRINTV(dbg.str());
		#endif
	}

	return res;
}


// Get how far we are currently below bandwidth quota
unsigned int User::getBytesBelowQuota(std::vector<DataTrack>& track)
{
	unsigned int res = MAX_BANDWIDTH;

	// Get quota gandwidth in bytes/sec
	unsigned int quotaBandwidth = m_quotas.getBandwidth() * 1024;

	if(quotaBandwidth > 0) {
		res = quotaBandwidth;

		// Get current average rate
		unsigned int avgBandwidth = 0;
		unsigned int totBandwidth = 0;
		unsigned int avgMillis = 0;
		getAvgBandwidth(track, avgBandwidth, totBandwidth, avgMillis);
		if(avgMillis > 0) {
			// Currently below quota?
			if(avgBandwidth < quotaBandwidth) {
				// How much can we send to take us up to quota?
				res = ((quotaBandwidth * avgMillis) / 1000) - totBandwidth;

				// Still apply absolute maximum
				if(res > MAX_BANDWIDTH)
					res = MAX_BANDWIDTH;
			}
			else {
				// Currently at or over quota, so throttle back
				res = 0;
			}
		}

		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "getBytesBelowQuota: Max = " << quotaBandwidth << ", ";
		dbg << "Allowed = " << res;
		DPRINTV(dbg.str());
		#endif
	}

	return res;
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const User& user)
{
	std::string name = user.getName();
	bool authenticated = user.getAuthenticated(0);
	unsigned int bandwidth = user.getQuotas().getBandwidth();

	str << "Name: " << name << ", ";
	str << "Authenticated: " << authenticated << ", ";
	str << "Bandwidth: ";
	if(bandwidth == 0)
		str << "Unlimited";
	else
		str << bandwidth;

	return str;
}


}	// namespace

