/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a user of a tunnel.
 *
 * $Id: User.hh,v 1.1.1.1 2008-03-28 13:06:25 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _USER_HH_
#define _USER_HH_

#include <string>
#include <ostream>
#include <vector>
#include "Quotas.hh"
#include "Time.hh"

namespace HTTPTunnel {

class Tunnel;

class User {
public:
	// Construction/destruction
	User();
	User(bool auth, std::string name, std::string passwd, const Quotas& quotas,
		bool keepCached);
	User(const User& that);
	~User();

	// Operators
	User& operator=(const User& that);
	bool operator==(const User& that);

	// Accessors
	bool getAuthenticated(Tunnel* tun) const;
	std::string getName(void) const;
	std::string getPasswd(void) const;
	const Quotas& getQuotas(void) const;
	unsigned int getNoActiveTunnels(void) const;
	bool usingTunnel(Tunnel* tun) const;
	bool keepCached(void) const;
	unsigned int getAvgUpBandwidth(void);
	unsigned int getAvgDownBandwidth(void);

	// Mutators
	void setAuthenticated(bool auth, Tunnel* tun);
	void setQuotas(const Quotas& quotas);
	void setAll(std::string name, std::string passwd, bool auth,
		const Quotas& quotas, Tunnel* newTun);
	void addTunnel(Tunnel* tun, bool auth);
	void removeTunnel(Tunnel* tun);


	// Sending data according to our quotas
	unsigned int getBytesCanSend(void);
	void sentBytes(unsigned int bytes);

	// Receiving data according to our quotas
	bool canRecvBytes(unsigned int bytes);

	// Tunnel creation accprding to our quotas
	bool canCreateNewTunnel(void);

	// Force a shutdown for all our tunnels
	void closeTunnels(void);


private:

	/*********/
	/* Types */
	/*********/

	// Track amount of data transfered
	struct DataTrack {
		DataTrack(Time tm, unsigned int bytes) : m_tm(tm), m_bytes(bytes) {}
		Time				m_tm;
		unsigned int	m_bytes;
	};


	/*************/
	/* Constants */
	/*************/

	// How long do we track sent/recvd data?
	static const unsigned int		DATA_TRACK_PERIOD		= 10;

	// What percentage above quotas do we tolerate?
	static const unsigned int		QUOTA_TOLERENCE		= 10;

	// Maximum bandwidth usage we will allow
	static const unsigned int		MAX_BANDWIDTH			= 65536;


	/***********/
	/* Methods */
	/***********/

	// Bandwidth quota tracking
	void trackBandwidth(unsigned int bytes, std::vector<DataTrack>& track);
	void trimBandwidthTrack(std::vector<DataTrack>& track);
	void getAvgBandwidth(std::vector<DataTrack>& track, unsigned int& avg,
		unsigned int& tot, unsigned int& millis);
	bool bandwidthWithinQuota(std::vector<DataTrack>& track);
	unsigned int getBytesBelowQuota(std::vector<DataTrack>& track);


	/********/
	/* Data */
	/********/

	// Are we always authenticated?
	bool							m_alwaysAuthenticated;

	// Name/passwd
	std::string					m_name;
	std::string					m_passwd;

	// Quotas
	Quotas						m_quotas;

	// Track bandwidth usage
	std::vector<DataTrack>	m_sentData;
	std::vector<DataTrack>	m_recvData;

	// Our current active tunnels
	std::vector<Tunnel*>		m_activeTunnels;
	std::vector<bool>			m_authenticatedTunnels;

	// Should this user always be kept cached?
	bool							m_keepCached;

};


// Output to a stream
std::ostream& operator<<(std::ostream& str, const User& user);


}	// namespace

#endif

