/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for data writers.
 *
 * $Id: Writer.cc,v 1.3 2008-07-25 12:14:54 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef DEBUG_LOG_DATA
#include <iomanip>
#endif

#include <sstream>
#include <poll.h>
#include <cerrno>
#include <cstring>
#include "Writer.hh"
#include "Debug.hh"
#include "Mutex.hh"
#include "MutexVar.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "Connection.hh"
#include "Tunneler.hh"
#include "RawBuff.hh"
#include "MsgParms.hh"
#include "Sock.hh"
#include "FlushConnectionEvent.hh"


namespace HTTPTunnel {


#ifdef DEBUG_LOG_DATA
// Reference to log file mutex
extern Mutex g_logMtx;
#endif

// Reference to the logger
extern Logger* g_pLogger;


// Construction/destruction
Writer::Writer(Connection* conn)
: m_mtx(), m_conn(conn), m_closing(false), m_buff()
{
}

Writer::~Writer()
{
	#ifdef DEBUG_TUN
	if(m_buff.getLength() > 0) {
		DPRINTW("Writer::~Writer() - Error! Residual write data");
	}
	#endif
}


/********************/
/* Public Interface */
/********************/


// Process and write some message data to our socket
void Writer::write(MsgParms* parms) throw(CommsError)
{
	// Lock data buffer while we process more data into it
	MutexVar mv(m_mtx);

	// Process data into whatever form is required
	processData(parms, m_buff);

	// Flush all buffered data
	flush(parms->getSendRetry());
}

// Write some out-of-band data
void Writer::writeOutOfBand(RawBuff& outOfBand) throw(CommsError)
{
	// Lock data buffer while we process more data into it
	MutexVar mv(m_mtx);

	// Append out of band data to write buffer
	m_buff.append(outOfBand);

	// Just send the data
	rawWrite();
}

// Try to flush all buffered data
bool Writer::flush(bool retry) throw(CommsError)
{
	MutexVar mv(m_mtx);

	// See if we can write any queued data
	rawWrite();

	// Did we manage to flush all data?
	bool flushed = (m_buff.getLength() == 0);
	if(!flushed && retry) {
		// Retry flushing
		DPRINTV("Flush failed, scheduling retry");
		Tunneler* tunneler = m_conn->getTunneler();
		unsigned int connID = m_conn->getID();

		// Check we don't already have a flush pending
		if(!tunneler->isScheduled<FlushConnectionEvent, unsigned int>(connID)) {
			// Schedule flush
			tunneler->scheduleEvent(new FlushConnectionEvent(Time(0,FLUSH_TIME),
				connID, true, tunneler));
		}
	}

	return flushed;
}


/*************/
/* Accessors */
/*************/


// Do we have unprocessed data remaining?
bool Writer::dataRemaining(void) const
{
	return (m_buff.getLength() > 0);
}


/***********************/
/* Protected Interface */
/***********************/


#ifdef DEBUG_LOG_DATA
// Log some newly written data
void Writer::logData(const RawBuff& data, std::ofstream* file,
	unsigned int& lineChar)
{
	// Lock global log file
	MutexVar mv(g_logMtx);

	// Get pointer to data
	const unsigned char* dataPtr = data.begin();

	// Log bytes
	for(unsigned int i = 0; i < data.getLength(); ++i) {
		*file << std::hex << std::setw(2) << std::setfill('0');
		*file << (unsigned int)dataPtr[i];

		// Space or newline?
		if((++lineChar % 20) == 0) {
			lineChar = 0;
			*file << std::endl;
		}
		else
			*file << " ";
	}

	// Flush
	*file << std::flush;
}
#endif


/*********************/
/* Private Interface */
/*********************/


// Write raw data to our socket - Caller holds mutex
void Writer::rawWrite(void) throw(WriteError)
{
	// Any data to send?
	int size = m_buff.getLength();
	if(size > 0) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "Writer::rawWrite() - Buff contains " << size << " bytes";
		DPRINTV(nfo.str());
		#endif

		// Somewhere to record progress
		std::stringstream err;
		bool ok = true;
		int nowritten = 0;
		const unsigned char* rawBuf = m_buff.begin();
		Sock* sock = m_conn->getSocket();

		// Try to write as much as we can
		while(nowritten < size) {
			try {
				// Check write will not block
				bool pollError = false;
				if(sock->pollWrite(0, pollError)) {
					// Check for error condition on socket
					if(pollError) {
						// Error on socket
						ok = false;
						err << "Poll returned error for write on socket for connection ";
						err << m_conn->getID();
						break;
					}

					try {
						// Write should not block
						int writeres = sock->write(rawBuf + nowritten, size - nowritten);
						nowritten += writeres;
					}
					catch(SockError& e) {
						if(e.m_errno == EAGAIN) {
							// Write would block
							std::stringstream nfo;
							nfo << "Socket write says it would block for connection ";
							nfo << m_conn->getID();
							DPRINTV(nfo.str());
							break;
						}
						else {
							// Rethrow
							throw;
						}
					}
				}
				else {
					// Poll timeout
					std::stringstream nfo;
					nfo << "Poll says write would block for connection " << m_conn->getID();
					DPRINTV(nfo.str());
					break;
				}
			}
			catch(SockError& e) {
				ok = false;
				err << "SockError writing for connection " << m_conn->getID();
				err << ": " << e.what() << "(" << e.m_errno << " ";
				err << strerror(e.m_errno) << ")";
				break;
			}
			catch(CommsError& e) {
				ok = false;
				err << "CommsError writing for connection " << m_conn->getID();
				err << ": " << e.what();
				break;
			}
		}

		// Erase written data from buffer
		if(nowritten > 0) {
			m_buff.erase(nowritten);
		}

		// Any problems?
		if(!ok) {
			throw WriteError(err.str());
		}
	}
}


}	// namespace

