#include <signal.h>
#include <poll.h>
#include <unistd.h>
#include <string.h>
#include <cstdlib>
#include <iostream>
#include "HTTPTunneler.hh"

using namespace HTTPTunnel;


/*
 * Read and write between stdin and HTTP socket.
 *
 * Reading and writing of tunnel sockets is done using the 'read()' and
 * 'write()' methods from the 'Sock' socket class returned by the HTTP
 * tunneling library. This allows us to not care whether we are actually
 * reading/writing from an underlying TCP or UDP socket.
 *
 * If you know the type of the underlying socket however and you wish to use
 * the standard library IO functions (read, write, poll, etc), you may
 * retrieve the underlying socket descriptor by calling the 'Sock::getFD()'
 * method.
 */
void ReadWrite(Sock* tunsock)
{
	bool keepRunning = true;
	unsigned char buff[1024];
	size_t noread = 0;

	while(keepRunning) {
		// Poll stdin and socket
		struct pollfd pollStruct[] = {
			{ STDIN_FILENO, POLLIN, 0 },
			{ tunsock->getFD(), POLLIN, 0 }
		};
		int readyfds = poll(pollStruct, 2, -1);
		if(readyfds > 0) {
			// Check events
			for(int i = 0; i < 2; ++i) {
				short revents = pollStruct[i].revents;
				if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) {
					if(revents & POLLERR)
						std::cerr << "Got ERR on fd ";
					else if(revents & POLLHUP)
						std::cerr << "Got HUP on fd ";
					else
						std::cerr << "Got NVAL on fd ";
					std::cerr << i << " (" << pollStruct[i].fd << ")" << std::endl;

					keepRunning = false;
					break;
				}
				else if(revents & POLLIN) {
					// Got data to read
					try {
						if(pollStruct[i].fd == STDIN_FILENO) {
							// Read from stdin
							noread = read(STDIN_FILENO, buff, sizeof(buff));
							if(noread == 0) {
								// Stdin closed
								std::cerr << "Got EOF from STDIN - shutting down" << std::endl;
								keepRunning = false;
							}
							else {
								// Write to socket
								tunsock->write(buff, noread);
							}
						}
						else {
							// Read from socket, write to stdout
							noread = tunsock->read(buff, sizeof(buff));
							write(STDOUT_FILENO, buff, noread);
						}
					}
					catch(ConnectionClosed& e) {
						// Socket closed
						std::cerr << "Got EOF from server - shutting down" << std::endl;
						keepRunning = false;
					}
					catch(TunnelError& e) {
						// Error on tunnel connection
						std::cerr << "Got error on tunnel connection: " << e.what() << std::endl;
						keepRunning = false;
					}
				}
			}
		}
		else if(readyfds < 0) {
			// Poll error
			std::cerr << "Error while polling" << std::endl;
			keepRunning = false;
		}
	}
}


// Main
int main(int argc, char** argv)
{
	// Are we going to use TCP or UDP?
	Enums::Protocol protocol = Enums::TCP;
	if(argc == 2) {
		if(strcmp(argv[1], "-u") == 0)
			protocol = Enums::UDP;
	}

	try {
		// Configure global settings
		HTTPTunneler* tunneler = HTTPTunneler::Get();
		GlobalConfigBlob globalBlob(QUIET);
		tunneler->ConfigureGlobal(globalBlob);

		// Start the tunneling library
		tunneler->Start();
		std::cerr << "Tunneling library started" << std::endl;

		/* Configure client tunnel socket. Client will connect to server
		 * at IP address "localhost", port 8080.
		 */
		SockTunClientConfigBlob cliBlob(protocol, false, "localhost", 8080);
		ClientSockTunHandle cliHandle = tunneler->LaunchClientSocket(cliBlob);
		TunnelerID cliID = cliHandle.getTunID();
		std::cerr << "Client tunneler ID: " << cliID << std::endl;

		// Do some simple I/O
		ReadWrite(cliHandle.getSock());

		// If our tunneler isn't already shutting down, shut it down now
		tunneler->ShutdownTunneler(cliID, true);

		// Remove configured tunnels
		tunneler->RemoveTunneler(cliID);

		// Stop the tunneling library
		tunneler->Stop();
		std::cerr << "Tunneling library stopped" << std::endl;
	}
	catch(TunnelError& e) {
		std::cerr << "Error running tunneler: " << e.what() << std::endl;
	}

	// Tidy up and exit
	try {
		HTTPTunneler::Release();
	}
	catch(TunnelError& e) {
		std::cerr << "Error releasing tunneler: " << e.what() << std::endl;
	}

	exit(0);
}


