#include <signal.h>
#include <poll.h>
#include <unistd.h>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <algorithm>
#include "HTTPTunneler.hh"
#include "Mutex.hh"
#include "MutexVar.hh"
#include "ConditionVar.hh"
#include "ThreadManager.hh"
#include "Time.hh"

using namespace HTTPTunnel;

// List of connected client socket descriptors
std::vector<Sock*> ClientSocks;
Mutex SocksMtx;
ConditionVar SocksCond;
ThreadManager ThreadMgr;
Thread* StdinThread = 0;
bool KeepGoingSocks = true, KeepGoingStdin = true;


// Thread for reading from stdin
void ReadStdin(void)
{
	MutexVar mv(SocksMtx);

	// Setup poll structure for stdin
	struct pollfd pollStruct[1];
	pollStruct[0].fd = STDIN_FILENO;
	pollStruct[0].events = POLLIN;
	pollStruct[0].revents = 0;
	unsigned char buff[1024];

	// Start polling
	while(KeepGoingStdin) {
		if(KeepGoingSocks) {
			// Poll stdin
			int readyfds = poll(pollStruct, 1, 0);
			if(readyfds > 0) {
				if(pollStruct[0].revents & POLLIN) {
					ssize_t noread = read(STDIN_FILENO, buff, sizeof(buff));
					// Check for EOF
					if(noread == 0) {
						std::cerr << "Got EOF from STDIN - shutting down" << std::endl;
						KeepGoingSocks = false;
					}
					else {
						// Write data from stdin to all connected clients
						std::vector<Sock*>::iterator sit;
						for(sit = ClientSocks.begin(); sit != ClientSocks.end(); ++sit) {
							try {
								Sock* sock = *sit;
								sock->write(buff, noread);
							}
							catch(TunnelError& e) {
								std::cerr << "Error writing to connected client: " << e.what() << std::endl;
								KeepGoingSocks = false;
							}
						}
					}
				}
			}
			else if(readyfds < 0) {
				// Poll error
				std::cerr << "Error while polling stdin" << std::endl;
				KeepGoingSocks = false;
			}
		}

		// Sleep before polling again
		Time sleepTime(0,10);
		StdinThread->sleep(mv, sleepTime.getTimeVal());
	}
}


/*
 * Read and write between stdin and HTTP client sockets.
 *
 * Reading and writing of tunnel sockets is done using the 'read()' and
 * 'write()' methods from the 'Sock' socket class returned by the HTTP
 * tunneling library. This allows us to not care whether we are actually
 * reading/writing from an underlying TCP or UDP socket.
 *
 * If you know the type of the underlying socket however and you wish to use
 * the standard library IO functions (read, write, poll, etc), you may
 * retrieve the underlying socket descriptor by calling the 'Sock::getFD()'
 * method.
 */
void ReadWrite(void)
{
	MutexVar mv(SocksMtx);

	// Launch thread to read stdin
	ZeroArgGlobalCallback stdincb(&ReadStdin);
	StdinThread = ThreadMgr.newThread(stdincb);

	unsigned char buff[1024];

	// Read from connected clients
	while(KeepGoingSocks) {
		// Work out how many descriptors we're dealing with
		unsigned int nofds = ClientSocks.size();

		// Setup poll struct for STDIN and the connected client sockets
		struct pollfd pollStruct[nofds];
		for(unsigned int i = 0; i < ClientSocks.size(); ++i) {
			pollStruct[i].fd = ClientSocks[i]->getFD();
			pollStruct[i].events = POLLIN;
			pollStruct[i].revents = 0;
		}

		// Do the poll
		int readyfds = poll(pollStruct, nofds, 0);
		if(readyfds > 0) {
			// Check events
			for(unsigned int i = 0; i < nofds; ++i) {
				short revents = pollStruct[i].revents;
				if((revents & POLLERR) || (revents & POLLHUP) || (revents & POLLNVAL)) {
					if(revents & POLLERR)
						std::cerr << "Got ERR on fd ";
					else if(revents & POLLHUP)
						std::cerr << "Got HUP on fd ";
					else
						std::cerr << "Got NVAL on fd ";
					std::cerr << i << " (" << pollStruct[i].fd << ")" << std::endl;

					KeepGoingSocks = false;
					break;
				}
				else if(revents & POLLIN) {
					// Lookup socket
					Sock* sock = 0;
					std::vector<Sock*>::iterator sit;
					for(sit = ClientSocks.begin(); sit != ClientSocks.end(); ++sit) {
						if((*sit)->getFD() == pollStruct[i].fd) {
							sock = *sit;
							break;
						}
					}

					if(sock) {
						// Got data to read
						try {
							// Read from socket, write to stdout
							int noread = sock->read(buff, sizeof(buff));
							write(STDOUT_FILENO, buff, noread);
						}
						catch(ConnectionClosed& e) {
							// Socket closed
							std::cerr << "Got EOF from connected client ";
							std::cerr << i << " (" << sock->getFD() << ")" << std::endl;

							// Tidy up closed client connection
							delete sock;
							ClientSocks.erase(sit);
						}
						catch(TunnelError& e) {
							// Error on tunnel connection
							std::cerr << "Error reading from connected client: " << e.what() << std::endl;
							KeepGoingSocks = false;
						}
					}
				}
			}
		}
		else if(readyfds < 0) {
			// Poll error
			std::cerr << "Error while polling" << std::endl;
			KeepGoingSocks = false;
		}
		else {
			// Poll timeout - wait a while before trying again
			Time pollTime(0,10);
			SocksCond.timedwait(mv, pollTime.getTimeVal());
		}
	}

	// Close down any remaining connections
	while(ClientSocks.size() > 0) {
		delete ClientSocks[0];
		ClientSocks.erase(ClientSocks.begin());
	}
	
	// Shutdown stdin reading thread
	KeepGoingStdin = false;
	mv.unlock();
	ThreadMgr.shutdown();
	mv.lock();
}


// Callback to notify us of a new connected client
void NewClientConnection(Sock* sock)
{
	std::cerr << "Got new connection from client" << std::endl;

	// Add client socket to list of connected clients
	MutexVar mv(SocksMtx);
	ClientSocks.push_back(sock);

	// Signal I/O thread it has a new socket to process
	SocksCond.signal();
}


// Main
int main(int argc, char** argv)
{
	try {
		// Configure global settings
		HTTPTunneler* tunneler = HTTPTunneler::Get();
		GlobalConfigBlob globalBlob(QUIET);
		tunneler->ConfigureGlobal(globalBlob);

		// Start the tunneling library
		tunneler->Start();
		std::cerr << "Tunneling library started" << std::endl;

		/* Configure server tunnel socket. Server will listen for connections
		 * from clients on IP address "localhost", port 8080.
		 */
		SockTunServerConfigBlob svrBlob("localhost", 8080, false,
			new GlobalSockTunCallback(&NewClientConnection));
		ServerSockTunHandle svrHandle = tunneler->LaunchServerSocket(svrBlob);
		TunnelerID svrID = svrHandle.getTunID();
		std::cerr << "Server tunneler ID: " << svrID << std::endl;

		// Do I/O
		ReadWrite();

		// If our tunneler isn't already shutting down, shut it down now
		tunneler->ShutdownTunneler(svrID, true);

		// Remove configured tunnels
		tunneler->RemoveTunneler(svrID);

		// Stop the tunneling library
		tunneler->Stop();
		std::cerr << "Tunneling library stopped" << std::endl;
	}
	catch(TunnelError& e) {
		std::cerr << "Error running tunneler: " << e.what() << std::endl;
	}

	// Tidy up and exit
	try {
		HTTPTunneler::Release();
	}
	catch(TunnelError& e) {
		std::cerr << "Error releasing tunneler: " << e.what() << std::endl;
	}

	exit(0);
}


