/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Wrap a socket connection in HTTP and tunnel it to some destination.
 *
 * $Id: main.cc,v 1.50 2008-11-14 19:06:27 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <signal.h>
#include <unistd.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include "HTTPTunneler.hh"


/********/
/* Data */
/********/


// Config file name
std::string g_configFile = "";

// Global flag to indicate we should keep running
bool g_bRunning = true;


/*************/
/* Functions */
/*************/


// Re-read config and restart
void rereadConfig(void)
{
	try {
		// Try to reconfigure from the config file
		HTTPTunnel::HTTPTunneler::Get()->ReconfigureFromFile(g_configFile);
	}
	catch(HTTPTunnel::ConfigError& e) {
		// Error in reconfiguration config, so exit
		std::cerr << "Configuration error rereading config: " << e.what() << std::endl;
		g_bRunning = false;
	}
	catch(HTTPTunnel::TunnelError& e) {
		// General error while reconfiguring, so exit
		std::cerr << "General error rereading config: " << e.what() << std::endl;
		g_bRunning = false;
	}
}


// Display usage.
void usage(char* exe)
{
	std::cerr << STRNAME << " version " << VERSION << " (using ";
	std::cerr << HTTPTunnel::HTTPTunneler::GetLibBuildName() << ")" << std::endl;
	std::cerr << "Usage: " << exe << " <config.file>" << std::endl;
}


// Parse the command line arguments and setup the global configuration
void parseArgs(int argc, char** argv, bool& argerr, bool& cfgerr)
{
	argerr = cfgerr = false;

	try {
		if(argc == 2) {
			// Specified config file to read
			g_configFile = argv[1];
			HTTPTunnel::HTTPTunneler::Get()->ConfigureFromFile(g_configFile);
		}
		else if(argc == 3) {
			if(strcmp(argv[1], "-create") == 0) {
				// Create and use default config file
				g_configFile = argv[2];
				HTTPTunnel::HTTPTunneler::Get()->RegenerateCfgFile(g_configFile);
				HTTPTunnel::HTTPTunneler::Get()->ConfigureFromFile(g_configFile);
			}
			else {
				// Bad option
				argerr = true;
			}
		}
		else {
			// Bad number of args
			argerr = true;
		}
	}
	catch(HTTPTunnel::ConfigError& e) {
		std::cerr << "Configuration error processing config: " << e.what() << std::endl;
		cfgerr = true;
	}
	catch(HTTPTunnel::TunnelError& e) {
		std::cerr << "General error processing config: " << e.what() << std::endl;
		cfgerr = true;
	}
}


// Handler for SIGTERM/SIGINT
void signalHandler(void)
{
	// We will be waiting for SIGTERM/SIGINT
	sigset_t sigSet;

	#if defined(_POSIX_REALTIME_SIGNALS) && (_POSIX_REALTIME_SIGNALS != -1)
	sigemptyset(&sigSet);
	sigaddset(&sigSet, SIGTERM);
	sigaddset(&sigSet, SIGINT);
	sigaddset(&sigSet, SIGHUP);
	struct timespec sigTime = { 1, 0 };
	#endif

	// Just spin
	while(g_bRunning) {
		#if defined(_POSIX_REALTIME_SIGNALS) && (_POSIX_REALTIME_SIGNALS != -1)
		int sigNo = sigtimedwait(&sigSet, 0, &sigTime);
		if(sigNo == SIGTERM || sigNo == SIGINT) {
			// Stop running
			g_bRunning = false;
		}
		else if(sigNo == SIGHUP) {
			// Reload config
			rereadConfig();
		}

		#else
		int sigNo = sigpending(&sigSet);
		if(sigNo == -1) {
			std::cerr << "sigpending() returned error" << std::endl;
			g_bRunning = false;
		}
		else if(sigismember(&sigSet, SIGTERM) || sigismember(&sigSet, SIGINT)) {
			// Stop running
			g_bRunning = false;
		}
		else if(sigismember(&sigSet, SIGHUP)) {
			// Reload config
			rereadConfig();
		}
		else {
			sleep(1);
		}
		#endif
	}
}


// Initialise general process functions
bool initProcess(int argc, char** argv)
{
	bool ok = true;

	// Set logging name to use
	HTTPTunnel::HTTPTunneler::SetName(STRNAME);

	// Parse args and setup config appropriately
	bool argerr = false, cfgerr = false;
	parseArgs(argc, argv, argerr, cfgerr);
	if(argerr) {
		usage(argv[0]);
		ok = false;
	}
	else if(cfgerr) {
		ok = false;
	}

	// Block all signals
	sigset_t newmask;
	sigfillset(&newmask);
	if(pthread_sigmask(SIG_BLOCK, &newmask, 0) == -1) {
		std::cerr << "Fatal error setting default process signal mask" << std::endl;
		ok = false;
	}

	return ok;
}


// Main
int main(int argc, char** argv)
{
	// Initialise
	if(initProcess(argc, argv)) {
		try {
			// Start the tunneling
			HTTPTunnel::HTTPTunneler::Get()->Start();

			// Run signal catcher in main thread
			signalHandler();

			// Stop the tunneling
			HTTPTunnel::HTTPTunneler::Get()->Stop();
		}
		catch(HTTPTunnel::TunnelError& e) {
			std::cerr << "Error running tunneler: " << e.what() << std::endl;
		}
	}

	// Tidy up and exit
	try {
		HTTPTunnel::HTTPTunneler::Release();
	}
	catch(HTTPTunnel::TunnelError& e) {
		std::cerr << "Error releasing tunneler: " << e.what() << std::endl;
	}

	exit(0);
}


